/***************************************************************************************
 * Systems Tester specific external function implementation (t3nontri_external.c)
 *
 * 2010 IBM Corporation
 *
 * This example shows how to implement external functions using direct function implementation
 * that is supported in Systems Tester TTCN-3 compiler.
 *
 * It implements getCurrentDateTime external function as defined below:
 *
 * external function getCurrentDateTime(in charstring format, in integer max_length) return charstring
 *
 * This function accepts two parameters:
 *   - format specification defined as character string
 *   - maximum length of the resulting string
 * It returns current date and time formatted according to the provided format specification.
 *
 * Format specification should be done according to the ANSI C "strftime" function.
 *
 * NOTE:
 * Ensure that generated code is compiled with /DEXT_FUNC_BYPASS_TRI_ExternalFuncExample_getCurrentDateTime
 *
 ***************************************************************************************/

#include "t3rts.h"
#include "t3tri.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#ifndef EXT_FUNC_BYPASS_TRI_ExternalFuncExample_getCurrentDateTime
#error "Check that EXT_FUNC_BYPASS_TRI_ExternalFuncExample_getCurrentDateTime is defined"
#endif

t3rt_value_t ExternalFuncExample_getCurrentDateTime(unsigned long argc, t3rt_value_t argv[], t3rt_context_t ctx)
{
    time_t ltime;
    struct tm *today;
    char *time_buf;
    const char *format;
    int max_length;

    /* check parameter list length (we have 2 parameters) */
    if (argc != 2)
    {
      t3rt_report_error(__LINE__, __FILE__, 
                        t3rt_wide_string_construct_from_ascii("Unexpected parameter length", t3rt_temp_alloc_c, ctx),
                        ctx);
    }
    
    /* STEP 1: PROCESS IN AND INOUT PARAMETERS */
    
    format = t3rt_value_get_string(argv[0], ctx);
    max_length = t3rt_value_get_integer(argv[1], ctx);
      
    /* STEP 2: PERFORM NECESSARY CALCULATIONS */
   
    /* call standard ANSI C routines to obtain current date/time */  
    time(&ltime);
    today = localtime(&ltime);

    /* format date/time value according to provided format string */
    time_buf = malloc(max_length + 1);
    strftime(time_buf, max_length, format, today);
    
    /* STEP 2: PREPARE INOUT, OUT PARAMETERS AND RETURN VALUE */

    /* return value is passed through argv[argc] array member */    
    t3rt_value_set_string(argv[argc], time_buf, ctx);

    /* free data */
    free(time_buf);
     
    return t3rt_no_value_c; /* return value is not used */
}

/**************************************************************************
 *
 * TE -> SA
 *
 * Function to be implemented by the System Adaptor
 *
 */

TriStatus 
triSAReset() 
{ 
  return TRI_OK; 
}

TriStatus 
triExecuteTestcase (const TriTestCaseId *testCaseId,  
                    const TriPortIdList *tsiPortList) 
{
  return TRI_OK; 
}

TriStatus 
triEndTestcase () 
{
  return TRI_OK; 
}

TriStatus 
triMap  (const TriPortId *compPortId,  
         const TriPortId *tsiPortId) 
{ 
  return TRI_Error; 
}


TriStatus 
triUnmap (const TriPortId *compPortId, 
          const TriPortId *tsiPortId) 
{ 
  return TRI_Error; 
}


TriStatus 
triSend (const TriComponentId *componentId, 
         const TriPortId      *tsiPortId, 
         const TriAddress     *sutAddress, 
         const TriMessage     *sendMessage) 
{
  return TRI_Error;
}


TriStatus 
triSendMC (const TriComponentId *componentId,
           const TriPortId      *tsiPortId,
           const TriAddressList *sutAddresses,
           const TriMessage     *sendMessage)
{
  return TRI_Error;
}           


TriStatus
triSendBC (const TriComponentId *componentId,
           const TriPortId      *tsiPortId,
           const TriMessage     *sendMessage)
{
  return TRI_Error;
}           


TriStatus 
triCall (const TriComponentId   *componentId, 
         const TriPortId        *tsiPortId,  
         const TriAddress       *sutAddress,  
         const TriSignatureId   *signatureId,  
         const TriParameterList *parameterList) 
{
  return TRI_Error; 
}


TriStatus
triCallMC (const TriComponentId   *componentId,
           const TriPortId        *tsiPortId, 
           const TriAddressList   *sutAddresses, 
           const TriSignatureId   *signatureId, 
           const TriParameterList *parameterList)
{
  return TRI_Error; 
}           


TriStatus
triCallBC (const TriComponentId   *componentId,
           const TriPortId        *tsiPortId,
           const TriSignatureId   *signatureId,
           const TriParameterList *parameterList)
{
  return TRI_Error; 
}           


TriStatus 
triReply (const TriComponentId    *componentId, 
          const TriPortId         *tsiPortId,  
          const TriAddress        *sutAddress,  
          const TriSignatureId    *signatureId,  
          const TriParameterList  *parameterList,  
          const TriParameter      *returnValue) 
{ 
  return TRI_Error; 
}


TriStatus 
triReplyMC (const TriComponentId   *componentId,
            const TriPortId        *tsiPortId,
            const TriAddressList   *sutAddresses,
            const TriSignatureId   *signatureId,
            const TriParameterList *parameterList,
            const TriParameter     *returnValue)
{
  return TRI_Error; 
}
            

TriStatus
triReplyBC (const TriComponentId   *componentId,
            const TriPortId        *tsiPortId,
            const TriSignatureId   *signatureId,
            const TriParameterList *parameterList,
            const TriParameter     *returnValue)
{
  return TRI_Error; 
}
            

TriStatus
triRaise (const TriComponentId  *componentId,
          const TriPortId       *tsiPortId,
          const TriAddress      *sutAddress,
          const TriSignatureId  *signatureId,
          const TriException    *exception) 
{ 
  return TRI_Error; 
}


TriStatus
triRaiseMC (const TriComponentId *componentId,
            const TriPortId      *tsiPortId,
            const TriAddressList *sutAddresses,
            const TriSignatureId *signatureId,
            const TriException   *exception)
{
  return TRI_Error; 
}
            

TriStatus
triRaiseBC (const TriComponentId *componentId,
            const TriPortId      *tsiPortId,
            const TriSignatureId *signatureId,
            const TriException   *exception)
{
  return TRI_Error; 
}            

            
TriStatus 
triSUTActionInformal (char* description) 
{ 
  return TRI_Error; 
}


TriStatus 
triSUTActionTemplate (const TriActionTemplate* templateValue) 
{ 
  return TRI_Error; 
}


/**************************************************************************
 *
 * TE -> PA
 *
 * Function to be implemented by the Platform Adaptor
 *
 */

TriStatus
triPAReset() 
{ 
  return TRI_OK; 
}


TriStatus
triStartTimer (const TriTimerId *timerId,
               TriTimerDuration  timerDuration) 
{ 
  return TRI_Error; 
}


TriStatus
triStopTimer (const TriTimerId *timerId) 
{ 
  return TRI_Error; 
}


TriStatus 
triReadTimer (const TriTimerId *timerId,
              TriTimerDuration *elapsedTime) 
{ 
  return TRI_Error; 
}


TriStatus
triTimerRunning (const TriTimerId *timerId,
                 unsigned char    *running) 
{ 
  return TRI_Error; 
}

TriStatus 
triExternalFunction (const TriFunctionId *functionId,		     
                     TriParameterList    *parameterList,
                     TriParameter        *returnValue) 
{ 
  return TRI_Error; 
}

