#include <stdio.h>

#ifdef _WIN32
#include <windows.h>
#include <process.h>
#else
#include <sys/types.h>
#include <pthread.h>
#endif

#include <time.h>

#include <t3rts.h>
#include "t3tri.h"

#ifndef _WIN32
#include <sys/poll.h>

static void Sleep(unsigned long milliseconds)
{
  poll(NULL, 0, milliseconds);
}
#endif

/**************************************************************************
 *
 * TE -> PA
 *
 * This timer implementation uses a new thread for each started timer. 
 * All the thread does is to sleep for the timers duration and then issues 
 * a time out event (using triTimeout). 
 */

/* Timer representation */

typedef struct tut_timer_struct_t 
{
  unsigned long start_time;
  double        duration;
#ifdef _WIN32  
  HANDLE        thread;
#else
  pthread_t     thread;
#endif  
  TriTimerId    id;
} *tut_timer_t;


#define MAX_TIMERS_IN_TESTSUITE 5

struct timer_list_struct_t
{ 
  tut_timer_t timer_array[MAX_TIMERS_IN_TESTSUITE];
  unsigned int timer_in_list;
};

static struct timer_list_struct_t timer_list;

/***********************************************************
 *
 * tut_compare_binary
 *
 * Compares two TRI-defined binary strings. 
 * Zero (0) is returned if they are equal, non-zero if they 
 * are not.
 *
 ***********************************************************/

int 
tut_compare_binary(const BinaryString *bs1, const BinaryString *bs2)
{
  int bytes, i;

  if (bs1->bits != bs2->bits)
    return -1;

  bytes = bs1->bits / 8;
  for (i = 0; i < bytes; ++i) {
    if (bs1->data[i] != bs2->data[i])
      return -1;
  }

  if ((bs1->bits % 8) != 0)
    printf("Warning! The binary string comparision only support full bytes.\n");

  return 0;
}

/***********************************************************
 *
 * tut_copy_binary
 *
 * Copies a TRI binary string from 'src' into 'dest'. 
 *
 ***********************************************************/

int 
tut_copy_binary(BinaryString *dest, const BinaryString *src)
{
  if (dest == NULL || src == NULL) {
    printf("BinaryString copy failed.\n");
    return -1;
  }

  if (src->bits > 0) { 
    long bytes = (src->bits)/8;
    if (src->bits % 8 != 0)
      bytes += 1;

    dest->bits = src->bits;
    dest->data = (unsigned char*)malloc(sizeof(unsigned char) * bytes);
    memcpy(dest->data, src->data, bytes);
  }

  return 0;
}

/***********************************************************
 *
 * triPAReset
 *
 * This function is called one time only, before the control 
 * part is executed.
 * The timer list is cleared.
 *
 ***********************************************************/

TriStatus
triPAReset() 
{ 
  unsigned int i;

  timer_list.timer_in_list = 0;
  for (i = 0; i < MAX_TIMERS_IN_TESTSUITE; ++i)
    timer_list.timer_array[i] = NULL;

  return TRI_OK;
}


/***********************************************************
 *
 * run_timer
 *
 * This function sleeps for the timer duration and generates
 * a timeout when it wakes up.
 *
 * The time to sleep is adjusted with the time it took for 
 * the thread to get to this point from when it was started.
 * This is the sec_delay variable. 
 *
 ***********************************************************/
#ifdef _WIN32
unsigned run_timer(void *t)
#else
void * run_timer(void *t)
#endif
{
  tut_timer_t timer = (tut_timer_t)t;
  double sec_delay  = (clock() - timer->start_time) / (double)CLOCKS_PER_SEC;
  double sleep_time = timer->duration - sec_delay;

#ifndef _WIN32
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
#endif

#ifdef DEBUG
  printf(TRI_PREFIX "Sleeping for %.4lf seconds\n", sleep_time);
#endif

  if (sleep_time > 0.0)
    Sleep((unsigned long)(sleep_time * 1000));

  triTimeout(&(timer->id));

  timer->duration = 0.0;
  timer->thread = 0;

#ifdef _WIN32
  return 0;
#else
  return NULL;
#endif
}

/***********************************************************
 *
 * find_timer
 *
 * Lookup function for timers.
 *
 ***********************************************************/

int
find_timer(struct timer_list_struct_t *list, 
           const TriTimerId *timerId)
{
  unsigned int i;

  for (i = 0; i < list->timer_in_list; ++i) {
    if (list->timer_array[i] != NULL && list->timer_array[i]->id.data != NULL && 
        tut_compare_binary(timerId, &(list->timer_array[i]->id)) == 0) 
      return i;
  }

  return -1;
}


/***********************************************************
 *
 * triStartTimer
 *
 * Creates a new timer instance, initializes it and creates 
 * the timeout thread for it.
 *
 ***********************************************************/

TriStatus
triStartTimer (const TriTimerId *timerId,
               TriTimerDuration  timerDuration) 
{ 
  tut_timer_t timer;
  unsigned long stime = clock();

  if (timer_list.timer_in_list == MAX_TIMERS_IN_TESTSUITE) {
    printf("Error in triStartTimer. No more room in timer list (maximum size is '%d' elements).\n", MAX_TIMERS_IN_TESTSUITE);
    return TRI_Error;
  }
  
  timer = (tut_timer_t)malloc(sizeof(struct tut_timer_struct_t));
  
  timer->start_time = stime;
  timer->duration   = timerDuration;
  tut_copy_binary(&(timer->id), timerId);
  timer_list.timer_array[timer_list.timer_in_list++] = timer;

#ifdef _WIN32
  timer->thread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)run_timer, (void*)timer, 0, NULL);
#else
  pthread_create(&timer->thread, NULL, run_timer, (void*)timer);
#endif
  
  return TRI_OK; 
}


/***********************************************************
 *
 * triStopTimer
 *
 * Looks up the timer, if ir is running, stop it and clear its
 * data.
 *
 ***********************************************************/

TriStatus
triStopTimer (const TriTimerId *timerId) 
{ 
  int i = find_timer(&timer_list, timerId);

  if (i >= 0) {
    tut_timer_t t = timer_list.timer_array[i];
    if (t->thread != 0) {
#ifdef _WIN32
      TerminateThread((HANDLE)t->thread, 1UL); 
#else
      pthread_cancel(t->thread);
#endif

      t->duration = 0.0;
      free(t->id.data);
      t->id.data = NULL;
      t->start_time = 0;
      t->thread = 0;
    }
  }

  return TRI_OK;
}


/***********************************************************
 *
 * triReadTimer
 *
 * Set elapsed_time to a proper value. If no timer is found 
 * or it has been stopped it is set to 0.0.
 *
 ***********************************************************/

TriStatus 
triReadTimer (const TriTimerId *timerId,
              TriTimerDuration *elapsedTime) 
{
  int i = find_timer(&timer_list, timerId);

  *elapsedTime = 0.0;

  if (i >= 0) {
    tut_timer_t t = timer_list.timer_array[i];
    if (t->thread != 0)
      *elapsedTime = (clock() - t->start_time) / (double)CLOCKS_PER_SEC;
  }

  return TRI_OK; 
}


/***********************************************************
 *
 * triTimerRunning
 *
 * The timer is running if the thread handle is non-null,
 * otherwie it is not. 
 *
 ***********************************************************/

TriStatus
triTimerRunning (const TriTimerId *timerId,
                 unsigned char    *running) 
{ 
  int i = find_timer(&timer_list, timerId);

  *running = 0;

  if (i >= 0) {
    tut_timer_t t = timer_list.timer_array[i];
    if (t->thread != 0)
      *running = 1;
  } 

  return TRI_OK; 
}


/***********************************************************
 *
 * triExternalFunction
 *
 * Not implemented in this tutorial.
 *
 ***********************************************************/

TriStatus 
triExternalFunction (const TriFunctionId *functionId,		     
                     TriParameterList    *parameterList,
                     TriParameter        *returnValue) 
{ 
  printf("Calling external functions is not implemented.\n");

  return TRI_Error; 
}

/* End of File */
