#ifdef _WIN32
#include <windows.h>
#include <process.h>
#else
#include <sys/types.h>
#include <pthread.h>
#endif

#include <stdlib.h>
#include <string.h>

#include "t3tri.h"
#include "t3rts.h"

static const TriPortId *triTsiport;
static const TriComponentId *triTsiportTC;

static const TriPortId *triAsnport;
static const TriComponentId *triAsnportTC;

static const TriPortId *triSccpport;
static const TriComponentId *triSccpportTC;


#ifndef _WIN32
#include <sys/poll.h>

static void Sleep(unsigned long milliseconds)
{
  poll(NULL, 0, milliseconds);
}
#endif

/**************************************************************************
 *
 * Implementation of Client - Server Paradigm
 * A thread processes the function calls of the Remote Procedure Functions
 * such that TTCN can process code, while the function call is being performed.
 */

class Workload
{
public:
	Workload() : m_next(0) { }
	virtual ~Workload() { }
	virtual void perform_work() = 0;

	Workload *m_next;
};

static Workload *work = 0;
static Workload **workadd = &work;
static void* cs_workload;
static bool worker_started = 0;

#ifdef _WIN32
static DWORD WINAPI PerformWorkload(void* pv)
#else
static void* PerformWorkload(void* pv)
#endif
{
	for (;;)
	{
		t3pl_sem_wait(cs_workload, t3rt_no_context_c);

		if (work == 0)
		{	// Wait some time for new work
			for (int i = 0; i < 10; i++)
			{
				t3pl_sem_post(cs_workload, t3rt_no_context_c);
				Sleep(100);
				t3pl_sem_wait(cs_workload, t3rt_no_context_c);
				if (work) break;
			}

			if (work == 0)	// No New work, stop worker thread.
				break;
		}

		Workload *my_work = work;
		work = my_work->m_next;
		if (workadd == &(my_work->m_next))
			workadd = &work;
		t3pl_sem_post(cs_workload, t3rt_no_context_c);

		my_work->perform_work();
		delete my_work;
		
		Sleep(1);
	}

	worker_started = false;
	t3pl_sem_post(cs_workload, t3rt_no_context_c);

	return 0;
}

static void plan(Workload *new_work)
{
	t3pl_sem_wait(cs_workload, t3rt_no_context_c);
	*workadd = new_work;
	workadd = &(new_work->m_next);
	if (!worker_started)
	{
		worker_started = true;

#ifdef _WIN32
	::CreateThread(NULL, 0, PerformWorkload, NULL, 0, NULL);
#else
	pthread_t thread_id;
	pthread_create(&thread_id, NULL, PerformWorkload, NULL);
#endif
	}
	t3pl_sem_post(cs_workload, t3rt_no_context_c);
}	

void InitializeWorker()		// Must be called once at startup
{
	if (!cs_workload)
		cs_workload = t3pl_sem_create(1, t3rt_no_context_c);
}


class SendBinaryStringWorkload : public Workload 
{
	unsigned char *message;
	unsigned long length;
	char *portname;

public:
	SendBinaryStringWorkload (char *port, unsigned char *msg, unsigned long len): 
		Workload (),
		message ((unsigned char *)malloc(len)),
		length (len),
		portname(port)
	{ 
		memcpy (message, msg, len);
	}

	virtual ~SendBinaryStringWorkload() 
	{ 
		if (message) free(message);
	}

	virtual void perform_work()
	{
		TriMessage msg;
		msg.data = message;
		msg.bits = length * 8;
		msg.aux = 0;

		if (strcmp (portname, "asnport") == 0) {
			triEnqueueMsg(triAsnport, 0, triAsnportTC, &msg);
		} else if (strcmp (portname, "sccpport") == 0) {
			triEnqueueMsg(triSccpport, 0, triSccpportTC, &msg);
		} else {
			triEnqueueMsg(triTsiport, 0, triTsiportTC, &msg);
		}
	}

};


/**************************************************************************
 *
 * TE -> SA
 *
 * Function to be implemented by the System Adaptor
 *
 */

TriStatus 
triSAReset() 
{ 
	InitializeWorker();
  	return TRI_OK; 
}


TriStatus 
triExecuteTestcase (const TriTestCaseId *testCaseId,  
                    const TriPortIdList *tsiPortList) 
{
  	return TRI_OK; 
}

TriStatus 
triEndTestcase () 
{
  	return TRI_OK; 
}

TriStatus 
triMap  (const TriPortId *compPortId,  
         const TriPortId *tsiPortId) 
{ 
	if (strcmp(tsiPortId->portName, "tsiport") == 0) {
		triTsiport = tsiPortId;
		triTsiportTC = &(compPortId->compInst);
	}

	if (strcmp(tsiPortId->portName, "asnport") == 0) {
		triAsnport = tsiPortId;
		triAsnportTC = &(compPortId->compInst);
	}

	if (strcmp(tsiPortId->portName, "sccpport") == 0) {
		triSccpport = tsiPortId;
		triSccpportTC = &(compPortId->compInst);
	}

  	return TRI_OK; 
}


TriStatus 
triUnmap (const TriPortId *compPortId, 
          const TriPortId *tsiPortId) 
{ 
  	return TRI_OK; 
}


TriStatus 
triSend (const TriComponentId *componentId, 
         const TriPortId      *tsiPortId, 
         const TriAddress     *sutAddress, 
         const TriMessage     *sendMessage) 
{
	static char *asnport = "asnport";
	static char *sccpport = "sccpport";
	static char *tsiport = "tsiport";

	/* This is part of the workaround */
	if (strcmp(tsiPortId->portName, asnport) == 0) {
		// Message from asncomponent to sccpcomponent. These are theoritically connected together.
		plan (new SendBinaryStringWorkload (sccpport, sendMessage->data, sendMessage->bits / 8)); 
	}

	if (strcmp(tsiPortId->portName, sccpport) == 0) {
		// Message from sccpcomponent to asncomponent. These are theoritically connected together.
		plan (new SendBinaryStringWorkload (asnport, sendMessage->data, sendMessage->bits / 8)); 
	}
	/* Until here */

	if (strcmp (tsiPortId->portName, tsiport) == 0) {
		// This is a message to the system under test, which will be processed there, and then an
		// answer is sent back. For simplicity, here the same message is sent back.
		plan (new SendBinaryStringWorkload (tsiport, sendMessage->data, sendMessage->bits / 8)); 
	}

  	return TRI_OK;
}


TriStatus 
triSendMC (const TriComponentId *componentId,
           const TriPortId      *tsiPortId,
           const TriAddressList *sutAddresses,
           const TriMessage     *sendMessage)
{
  return TRI_Error;
}           


TriStatus
triSendBC (const TriComponentId *componentId,
           const TriPortId      *tsiPortId,
           const TriMessage     *sendMessage)
{
  return TRI_Error;
}           


TriStatus 
triCall (const TriComponentId   *componentId, 
         const TriPortId        *tsiPortId,  
         const TriAddress       *sutAddress,  
         const TriSignatureId   *signatureId,  
         const TriParameterList *parameterList) 
{
  return TRI_Error; 
}


TriStatus
triCallMC (const TriComponentId   *componentId,
           const TriPortId        *tsiPortId, 
           const TriAddressList   *sutAddresses, 
           const TriSignatureId   *signatureId, 
           const TriParameterList *parameterList)
{
  return TRI_Error; 
}           


TriStatus
triCallBC (const TriComponentId   *componentId,
           const TriPortId        *tsiPortId,
           const TriSignatureId   *signatureId,
           const TriParameterList *parameterList)
{
  return TRI_Error; 
}           


TriStatus 
triReply (const TriComponentId    *componentId, 
          const TriPortId         *tsiPortId,  
          const TriAddress        *sutAddress,  
          const TriSignatureId    *signatureId,  
          const TriParameterList  *parameterList,  
          const TriParameter      *returnValue) 
{ 
  return TRI_Error; 
}


TriStatus 
triReplyMC (const TriComponentId   *componentId,
            const TriPortId        *tsiPortId,
            const TriAddressList   *sutAddresses,
            const TriSignatureId   *signatureId,
            const TriParameterList *parameterList,
            const TriParameter     *returnValue)
{
  return TRI_Error; 
}
            

TriStatus
triReplyBC (const TriComponentId   *componentId,
            const TriPortId        *tsiPortId,
            const TriSignatureId   *signatureId,
            const TriParameterList *parameterList,
            const TriParameter     *returnValue)
{
  return TRI_Error; 
}
            

TriStatus
triRaise (const TriComponentId  *componentId,
          const TriPortId       *tsiPortId,
          const TriAddress      *sutAddress,
          const TriSignatureId  *signatureId,
          const TriException    *exception) 
{ 
  return TRI_Error; 
}


TriStatus
triRaiseMC (const TriComponentId *componentId,
            const TriPortId      *tsiPortId,
            const TriAddressList *sutAddresses,
            const TriSignatureId *signatureId,
            const TriException   *exception)
{
  return TRI_Error; 
}
            

TriStatus
triRaiseBC (const TriComponentId *componentId,
            const TriPortId      *tsiPortId,
            const TriSignatureId *signatureId,
            const TriException   *exception)
{
  return TRI_Error; 
}            

            
TriStatus 
triSUTActionInformal (char* description) 
{ 
  return TRI_Error; 
}


TriStatus 
triSUTActionTemplate (const TriActionTemplate* templateValue) 
{ 
  return TRI_Error; 
}


/* End of File */

