/*************************************************************************
 * TCI Codec binary string routines
 *
 * 2010 IBM Corporation
 *
 * This file contains the functions used for processing binary strings:
 * 1) Packing 1,2,4,N bytes into the binary string
 * 2) Unpacking 1,2,4,N bytes from the binary string
 *
 * Note: This implementation supports only byte-aligned binary strings
 *
 *************************************************************************/

#include <stdlib.h>
#include "tci.h"
#include "binary_string.h"

/* Assertin macros */
#define bstring_assert(cond, msg) if (!(cond)) {tciErrorReq(msg); return;}
#define bstring_assert_wret(cond, msg, retval) if (!(cond)) {tciErrorReq(msg); return (retval);}

/* Appends byte to the end of binary string
 * It's assumed that binary string has enough free space
 */
static void append_byte (MyBinaryString *string, unsigned char data)
{
  unsigned long free_pos;

  free_pos = string->string.bits/8;
  string->string.data[free_pos] = data;
  string->string.bits += 8;
}

/* Expands binary string to the required size  */
static void ensure_unused_size(MyBinaryString *string, unsigned long bit_size)
{
  if ((string->string.bits + bit_size)/8 > string->allocated) {
    /* expand string (doubling required size) */
    unsigned long new_size = (string->string.bits + bit_size)/8 * 2;
    unsigned char *buf = tciMemoryAllocate(new_size);

    bstring_assert(buf != 0, "Out of memory");

    if (string->allocated != 0)
      memcpy(buf, string->string.data, string->string.bits/8);

    string->string.data = buf;
    string->allocated = new_size;
  }
}

/* Appends one byte to the end of binary string */
void binary_string_append_char (MyBinaryString *string, unsigned char data)
{
  bstring_assert(string->string.bits % 8 == 0, "This implementation supports only byte-aligned binary strings");

  ensure_unused_size(string, 8);
  append_byte(string, data);
}


/* Appends two bytes to the end of binary string */
void binary_string_append_short (MyBinaryString *string, unsigned short data)
{
  unsigned char one_byte;

  bstring_assert(string->string.bits % 8 == 0, "This implementation supports only byte-aligned binary strings");
  
  ensure_unused_size(string, 16);

  one_byte = (unsigned char)(data >> 8);
  append_byte (string, one_byte);

  one_byte = (unsigned char)data;
  append_byte (string, one_byte);
}

/* Appends four bytes to the end of binary string */
void binary_string_append_long (MyBinaryString *string, unsigned long data)
{
  unsigned char one_byte;
  char bits = 32;

  bstring_assert(string->string.bits % 8 == 0, "This implementation supports only byte-aligned binary strings");
  
  ensure_unused_size(string, 32);

  while (bits > 0) {
    one_byte = (unsigned char)(data >> (bits-8)); 
    append_byte (string, one_byte);
    bits -= 8;
  }
}

/* Appends N bytes to the end of binary string */
void binary_string_append_bytes (MyBinaryString *string, const unsigned char *data, unsigned long bytesize)
{
  unsigned long i;
  unsigned long free_pos;

  bstring_assert(string->string.bits % 8 == 0, "This implementation supports only byte-aligned binary strings");
  
  ensure_unused_size(string, bytesize*8);

  free_pos = string->string.bits/8;
  memcpy(&string->string.data[free_pos], data, bytesize);
  string->string.bits += bytesize*8;
}

/* Extracts next byte from the binary string
 * Iterator is moved forward
 */
static unsigned char get_byte (BinaryStringIter *iter)
{
  unsigned char byte;

  byte = iter->string.data[iter->bit_index/8];
  iter->bit_index += 8;

  return byte;
}

/* Extracts next byte from the binary string
 * Iterator is moved forward
 */
unsigned char binary_string_get_char (BinaryStringIter *iter)
{
  bstring_assert_wret(iter->string.bits % 8 == 0, "This implementation supports only byte-aligned binary strings", 0);
  bstring_assert_wret(iter->bit_index % 8 == 0, "Brokem binary string iterator", 0);
  bstring_assert_wret(iter->string.bits - iter->bit_index >= 8, "Broken binary string", 0);

  return get_byte(iter);
}


/* Extracts next two bytes from the binary string
 * Iterator is moved forward
 */
unsigned short binary_string_get_short (BinaryStringIter *iter)
{
  unsigned short left = 0, right = 0;

  bstring_assert_wret(iter->string.bits % 8 == 0, "This implementation supports only byte-aligned binary strings", 0);
  bstring_assert_wret(iter->bit_index % 8 == 0, "Brokem binary string iterator", 0);
  bstring_assert_wret(iter->string.bits - iter->bit_index >= 16, "Broken binary string", 0);

  left  = ((unsigned short) get_byte(iter)) << 8;
  right = (unsigned short) get_byte(iter);

  return (left | right);
}


/* Extracts next four bytes from the binary string
 * Iterator is moved forward
 */
unsigned long binary_string_get_long (BinaryStringIter *iter)
{
  unsigned long first = 0, second = 0, third = 0, fourth= 0;

  bstring_assert_wret(iter->string.bits % 8 == 0, "This implementation supports only byte-aligned binary strings", 0);
  bstring_assert_wret(iter->bit_index % 8 == 0, "Brokem binary string iterator", 0);
  bstring_assert_wret(iter->string.bits - iter->bit_index >= 32, "Broken binary string", 0);
  
  first  = ((unsigned long) get_byte(iter)) << 24;
  second = ((unsigned long) get_byte(iter)) << 16;
  third  = ((unsigned long) get_byte(iter)) <<  8;
  fourth = (unsigned long) get_byte(iter);

  return (first | second | third | fourth);
}


/* Extracts next N bytes from the binary string
 * Iterator is moved forward
 */
void binary_string_get_bytes (BinaryStringIter *iter, unsigned long n, unsigned char *data)
{
  unsigned long i;

  bstring_assert(iter->string.bits % 8 == 0, "This implementation supports only byte-aligned binary strings");
  bstring_assert(iter->bit_index % 8 == 0, "Brokem binary string iterator");
  bstring_assert(iter->string.bits - iter->bit_index >= n*8, "Broken binary string");

  for(i = 0; i < n; i++)
    data[i] = get_byte(iter);
}

/* End of File */
