/***************************************************************************************
 * TCI Codec encoding/decoding routines
 *
 * 2010 IBM Corporation
 *
 * This file contains encoding and decoding routines
 * that use TCI CD interface to encode and decode values.
 * 
 *
 * TCI public functions implemented:
 *  tciEncode 
 *  tciDecode
 *
 * Every type is encoded by its:
 *   - qualified type name
 *   - "omitted" flag
 *   - type dependent data
 *    
 * Structured types are encoded by recursive encoding of all their fields (elements)
 *    
 * Two special encoding functions encode_string and encode_double are used
 * to encode ASCII strings (including textual representations of bitstrings, 
 * hexstrings, octetstrings) and TTCN-3 float values.
 *
 ***************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "tci.h"
#include "binary_string.h"

#define tci_assert(cond, msg) if (!(cond)) {tciErrorReq(msg); return;}
#define tci_assert_wret(cond, msg, retval) if (!(cond)) {tciErrorReq(msg); return (retval);}

#include <float.h>
#include <limits.h>

#ifdef __linux__

#if !defined __USE_ISOC99 && !defined __USE_ISOC9X
#define __USE_ISOC99 1
#define __USE_ISOC9X 1
#include <math.h>
#undef __USE_ISOC99
#undef __USE_ISOC9X
#else
#include <math.h>
#endif

#else

#include <math.h>

#ifndef _WIN32
#include <ieeefp.h>
#endif

#endif

#define FP_NORM     0
#define FP_BAD      1
#define FP_NINF     2
#define FP_PINF     3
#define FP_NZERO    4
#define FP_PZERO    5


#if defined(_MSC_VER)  

static unsigned short get_fpclass (double x)
{
  switch (_fpclass (x)) {
  case _FPCLASS_SNAN:
  case _FPCLASS_QNAN:
      return FP_BAD;
  case _FPCLASS_NINF:
      return FP_NINF;
  case _FPCLASS_PINF:
      return FP_PINF;
  case _FPCLASS_PZ:
      return FP_NZERO;
  case _FPCLASS_NZ:
      return FP_PZERO;
  default:
      return FP_NORM;
  }
}

#else

#if defined(__linux__) || defined(_WIN32) 

static unsigned short get_fpclass (double x)
{
  switch (fpclassify (x)) {
  case FP_NAN:
      return FP_BAD;
  case FP_INFINITE:
      return (signbit (x) == 0) ? FP_PINF : FP_NINF;
  case FP_ZERO:
      return (signbit (x) == 0) ? FP_PZERO : FP_NZERO;
  default:
      return FP_NORM;
  }
}

#else 

static unsigned short get_fpclass (double x)
{
  switch (fpclass (x)) {
  case FP_SNAN:
  case FP_QNAN:
      return FP_BAD;
  case FP_NINF:
      return FP_NINF;
  case FP_PINF:
      return FP_PINF;
  case FP_PZERO:
      return FP_NZERO;
  case FP_NZERO:
      return FP_PZERO;
  default:
      return FP_NORM;
  }
}

#endif

#endif

#define EXP_MINUS_MASK       (1 << 11)
#define MANTISSA_MINUS_MASK  (1 << 12)

#define MUL_BASE    ((unsigned long)INT_MAX + 1)

/*
 * encode_double
 *
 * Special encoding function used to encode float values
*/

static void encode_double (MyBinaryString *message, double number)
{
  double          abs_mantissa;
  double          int_val;
  int             exponent;
  unsigned short  header;
  unsigned long   chunk;

  header = get_fpclass (number);
  header <<= 13;

  if (header != 0) { /* Special value */
    binary_string_append_short(message, header);
  } else { /* Ordinary value */
    abs_mantissa = frexp (fabs (number), &exponent);

    tci_assert ((((unsigned short)abs(exponent)) >> 11) == 0, "Illegal double");

    header |= (unsigned short) abs (exponent);
    if (exponent < 0) header |= EXP_MINUS_MASK;
    if (number < 0) header |= MANTISSA_MINUS_MASK;

    binary_string_append_short(message, header);

    abs_mantissa = modf (abs_mantissa * MUL_BASE, &int_val);
    chunk = (unsigned long)int_val;

    binary_string_append_long(message, chunk);

    chunk = (unsigned long)(abs_mantissa * MUL_BASE);

    binary_string_append_long(message, chunk);
  }
}

/*
 * decode_double
 *
 * Special decoding function used to decode float values
*/

static double decode_double(BinaryStringIter *iter)
{
  extern const double     pZero;
  extern const double     nZero;
  const double     NaN = pZero/pZero;
  const double     pINF = +1.0/pZero;
  const double     nINF = -1.0/pZero;

  double            mantissa;
  char              mantissa_minus_sign = 0;
  int               exponent;
  unsigned short    header;
  unsigned long     chunk1, chunk2;
  
  header = binary_string_get_short (iter);

  switch (header >> 13) {
  case FP_BAD:
      return NaN;
  case FP_NINF:
      return nINF;
  case FP_PINF:
      return pINF;
  case FP_NZERO:
      return nZero;
  case FP_PZERO:
      return pZero;
  case FP_NORM:
      break;
  default:
      tci_assert_wret (0, "Illegal float value", 0.0); /* Should never happen! */
  }

  if ((header & MANTISSA_MINUS_MASK) != 0) {
    header &= ~(MANTISSA_MINUS_MASK);
    mantissa_minus_sign = 1;
  }
  if ((header & EXP_MINUS_MASK) != 0) {
    header &= ~(EXP_MINUS_MASK);
    exponent = -((int)header);
  } else {
    exponent = (int)header;
  }

  chunk1 = binary_string_get_long (iter);
  chunk2 = binary_string_get_long (iter);

  mantissa = (((double)chunk2 / MUL_BASE) + (double)chunk1) / MUL_BASE;
  if (mantissa_minus_sign)  mantissa = -mantissa;

  return ldexp (mantissa, exponent);
}

/*
 * encode_string
 * 
 * This is a encoding function that is used for encoding of all kinds of ASCII strings.
 * It's used for charstrings, bitstrings, hexstrings, octetstrings as well as for encoding
 * service data passed within BinaryString (eg. type names)
*/

static void encode_string(MyBinaryString *msg, String str)
{
  unsigned long length;
  
  if (str)
    length = strlen(str);
  else
    length = 0;

  binary_string_append_long(msg, length);
  binary_string_append_bytes(msg, str, length);
}

/*
 * decode_string
 * 
 * This is a decoding function that is used for decoding of all kinds of ASCII strings.
 * It's used for charstrings, bitstrings, hexstrings, octetstrings as well as for decoding
 * service data passed within BinaryString (eg. type names)
*/

static String decode_string(BinaryStringIter *iter, unsigned long *length, String *str)
{
  unsigned long ulong;
  
  ulong = binary_string_get_long(iter);
  if (length)
    *length = ulong;
  
  *str = tciMemoryAllocate(ulong + 1);
  binary_string_get_bytes(iter, ulong, *str);
  (*str)[ulong] = '\0';

  return *str;
}

/*
 * encode_type
 * 
 * Encodes type by the type name.
 * User defined types are presented wityh fully qualified type names,
 * while built-in types are not prefixed with module names.
*/

static void encode_type(MyBinaryString *msg, TciType type) {

  TciModuleIdType module_id = tciGetDefiningModule(type);
  String module_name = module_id.moduleName;
  String type_name = tciGetName(type);

  tci_assert(type_name != NULL, "Illegal type");

  if (module_name) {
    /* user-defined type */

    binary_string_append_long(msg, strlen(module_name) + 1 + strlen(type_name)); /* additional space for '.' */ 
    binary_string_append_bytes(msg, module_name, strlen(module_name));
    binary_string_append_char(msg, '.');
    binary_string_append_bytes(msg, type_name, strlen(type_name));
  }
  else {
    /* builtin type */
    binary_string_append_long(msg, strlen(type_name));
    binary_string_append_bytes(msg, type_name, strlen(type_name));
  }
}

/*
 * decode_type
 * 
 * Decodes type.
*/

static void decode_type(BinaryStringIter *iter, TciType *type) {

  String type_name;

  /* get type name */
  decode_string(iter, NULL, &type_name);

  *type = tciGetTypeForName(type_name);
  tci_assert(*type != NULL, "Incorrect type name");
}

/*
 * encode
 * 
 * This is the encoder "worker" function.
 * It contains encoder implementation for all TTCN-3 types
 * Structured types are processed recursively.
 * Encoded value is returned as an "out" parameter.
*/

static void encode(MyBinaryString *msg, TciValue value)
{
  TciType type;
  TciTypeClassType kind;

  tci_assert(value != NULL, "Illegal value");

  type = tciGetType(value);
  tci_assert(type != NULL, "Illegal type");

  kind = tciGetTypeClass(type);
    
  switch(kind) { 
    case TCI_ADDRESS_TYPE:
      {
        TciValue underlying_value;
        
        /* Encode underlying address value*/
        underlying_value = tciGetAddressValue(value);
        tci_assert(underlying_value != NULL, "Illegal address value");
        encode(msg, underlying_value);
      }
      break;
      
    case TCI_BITSTRING_TYPE:
      {
        String bstring;
      
        /* Encode bit string */
        bstring = tciGetBStringValue(value);
        tci_assert(bstring != NULL, "Illegal bitstring value");
        encode_string(msg, bstring);
      }
      break;
      
    case TCI_BOOLEAN_TYPE:
      {
        unsigned char bool;
        
        /* Encode boolean */
        bool = tciGetBooleanValue(value);
        binary_string_append_char(msg, bool);
      }
      break;
    
    case TCI_CHAR_TYPE:
      {
        unsigned char character;
        
        /* Encode character */
        character = tciGetCharValue(value);
        binary_string_append_char(msg, character);
      }
      break;
      
    case TCI_CHARSTRING_TYPE:
      {
        TciCharStringValue cstr_value;

        /* encode charstring value */
        cstr_value = tciGetCStringValue(value);
        tci_assert(cstr_value != NULL, "Illegal charstring value");
        binary_string_append_long(msg, cstr_value->length);
        binary_string_append_bytes(msg, cstr_value->string, cstr_value->length);
      }
      break;
      
    case TCI_ENUMERATED_TYPE:
      {
        String enum_choice;
        
        /* encode enumeration choice */
        enum_choice = tciGetEnumValue(value);
        tci_assert(enum_choice != NULL, "Illegal enumerated value");
        encode_string(msg, enum_choice);
      }
      break;
      
    case TCI_FLOAT_TYPE:
      {
        double dbl;
      
        /* encode double value */
        dbl = tciGetFloatValue(value);
        encode_double(msg, dbl); /* decomposition into abs, mantissa, etc*/
      }
      break;
      
    case TCI_HEXSTRING_TYPE:
      {
        String hstring;
        
        /* encode hex string */
        hstring = tciGetHStringValue(value);
        tci_assert(hstring != NULL, "Illegal hexstring value")
        encode_string(msg, hstring);
      }
      break;
      
    case TCI_INTEGER_TYPE:
      {
        unsigned char sign;
        String abs_str;
        
        /* encode sign */
        sign = tciGetIntSign(value);
        binary_string_append_char(msg, sign);

        /* encode string representing absolute value */
        abs_str = tciGetIntAbs(value);
        tci_assert(abs_str != NULL, "Illegal integer value");
        encode_string(msg, abs_str);
      }
      break;
      
    case TCI_OBJID_TYPE:
      {
        TciObjidValue objid_val;
        unsigned long i;
        
        /* get list of objid elements */
        objid_val = tciGetTciObjidValue(value);
        /* encode objid length */
        binary_string_append_long(msg, objid_val->length); 
      
        for (i = 0; i < objid_val->length; i++) {
          /* encode objid elements one by one */
          binary_string_append_long(msg, objid_val->elements[i].elem_as_number);
        }
      }
      break;
      
    case TCI_OCTETSTRING_TYPE:
      {
        String ostring;
        
        ostring = tciGetOStringValue(value);
        tci_assert(ostring, "Illegal octet string value");
        
        /* encode octet string */
        encode_string(msg, ostring);
      }
      break;
      
    case TCI_RECORD_TYPE:
    case TCI_SET_TYPE:
      {
        String *field_names;
        unsigned long i;
      
        /* get list of record/set firld names */
        field_names = tciGetRecFieldNames(value);
        for (i = 0; field_names[i] != NULL; i++) {
          TciValue elem;
          
          /* get record/set field value */
          elem = tciGetRecFieldValue(value, field_names[i]);
          tci_assert(elem, "Ilegal record/set field value");
        
          if (tciNotPresent(elem)) {
            binary_string_append_char(msg, 1);
          } else {
            binary_string_append_char(msg, 0);

            /* encode record/set field value */
            encode(msg, elem);
          }
  
  
        }
	  }
      break;
      
    case TCI_RECORD_OF_TYPE:
    case TCI_SET_OF_TYPE:
      {
        unsigned long length;
        unsigned long i;

        /* encode length of record of/set of*/
        length = tciGetRecOfLength(value);
        binary_string_append_long(msg, length); 
      
        for (i = 0; i < length; i++) {
          TciValue elem;
          
          /* get record of/set of element */
          elem = tciGetRecOfFieldValue(value, i);
          tci_assert(elem, "Illegal record of/set of element value");

          /* encode record of/set of element */
          encode(msg, elem);
        }
      }
      break;
         
    case TCI_ANYTYPE_TYPE:
    case TCI_UNION_TYPE:
      {
        String variant;
        TciValue elem;
        
        variant = tciGetUnionPresentVariantName(value);
        tci_assert(variant, "Union variant is not set");
        
        /* encode variant name */
        encode_string(msg, variant);

        /* get variant value */
        elem = tciGetUnionVariant(value, variant);
        tci_assert(elem, "Illegal union variant");
        
        /* encode variant value */
        encode(msg, elem);
      }
      break;
      
    case TCI_UNIVERSAL_CHAR_TYPE:
      {
        TciUCReturnValue wchar;
        
        /* get widechar value */
        wchar = tciGetUniversalCharValue(value);
        /* encode widechar */
        binary_string_append_bytes(msg, wchar, sizeof(TciUCValue));
      }
      break;
      
    case TCI_UNIVERSAL_CHARSTRING_TYPE:
      {
        unsigned long length;  
        unsigned long i;
        TciUCReturnValue wchar;
      
        length = tciGetUCStringLength(value);
        
        /* encode string length */
        binary_string_append_long(msg, length); 
      
        for (i=0; i < length; i++) {
          /* encode widestring elements one by one */
          wchar = tciGetUCStringCharValue(value, i);
          binary_string_append_bytes(msg, wchar, sizeof(TciUCValue));
        }
      }
      break;
      
    case TCI_VERDICT_TYPE:
      {
        unsigned long verdict;

        /* encode verdict */ 
        verdict = tciGetVerdictValue(value);
        binary_string_append_long(msg, verdict);
      }
      break;
      
    default:
      tci_assert(0, "Unsupported type kind in Encoder");
  }
}

/*
 * decode
 * 
 * This is the decoder "worker" function.
 * It contains decoder implementation for all TTCN-3 types
 * Structured types are processed recursively.
 * Decoded value is returned as an "out" parameter.
*/

static void decode(BinaryStringIter *iter, TciValue value)
{
  TciType type;
  TciTypeClassType kind;

  tci_assert(value != NULL, "Illegal value");

  type = tciGetType(value);
  tci_assert(type != NULL, "Illegal type");

  kind = tciGetTypeClass(type);

  switch(kind) {
    case TCI_ADDRESS_TYPE:
      {
        TciValue address_value;

        /* Instantiate underlying address value*/
        address_value = tciGetAddressValue(value);
        
        /* decode underlying representation of the address value */
        decode(iter, address_value);

        tciSetAddressValue(value, address_value);
      }
      break;
      
    case TCI_BITSTRING_TYPE:
      {
        unsigned long length;
        String bstring;
         
        /* decode bit string */
        decode_string(iter, NULL, &bstring);
        tciSetBStringValue(value, bstring);
      }
      break;
      
    case TCI_BOOLEAN_TYPE:
      {
        unsigned char bool;
        
        /* decode boolean */ 
        bool = binary_string_get_char(iter);
        tciSetBooleanValue(value, bool);
      }
      break;
    
    case TCI_CHAR_TYPE:
      {
        char character;

        /* decode char */ 
        character = binary_string_get_char(iter);
        tciSetCharValue(value, character);
      }
      break;
      
    case TCI_CHARSTRING_TYPE:
      {
        unsigned long length;
        String cstring;
        TciCharStringValue cstr_value;
        
        /* decode character string */
        decode_string(iter, &length, &cstring);
        
        /* allocate memory for strucuture representing character string in TCI */
        cstr_value = tciMemoryAllocate(sizeof(struct _TciCharStringValue));
        cstr_value->length = length;
        cstr_value->string = cstring;
        tciSetCStringValue(value, cstr_value);
      }
      break;
      
    case TCI_ENUMERATED_TYPE:
      {
        String choice;
        
        /* decode string representing enumerated choice */
        decode_string(iter, NULL, &choice);
        tciSetEnumValue(value, choice);
      }
      break;
      
    case TCI_FLOAT_TYPE:
      {
        double dbl;
        
        /* decode float value */
        dbl = decode_double(iter);
        tciSetFloatValue(value, dbl);
      }
      break;
      
    case TCI_HEXSTRING_TYPE:
      {
        String hstring;
        
        /* decode hex string */ 
	    decode_string(iter, NULL, &hstring);
	    tciSetHStringValue(value, hstring);
	  }
      break;
	  
    case TCI_INTEGER_TYPE:
      {
        unsigned char sign;
        String abs_str;
        
        /* decode sign of integer (1('+') or 0('-')) */ 
        sign = binary_string_get_char(iter);
        tciSetIntSign(value, sign);
        
        /* decode string representing absolute inetegr value  */ 
        decode_string(iter, NULL, &abs_str);
        tciSetIntAbs(value, abs_str);
      }
      break;
      
    case TCI_OBJID_TYPE:
      {
        TciObjidValue objid_val;
        unsigned long i;
        
        /* allocate memory for strucuture representing object identifiers in TCI */
        objid_val = tciMemoryAllocate(sizeof(struct _TciObjidValue));

        /* decode number of elements on object identifier */
        objid_val->length = binary_string_get_long(iter);
        objid_val->elements = tciMemoryAllocate(sizeof(TciObjidElemValue) * objid_val->length);

        for (i = 0; i < objid_val->length; i++) {
          /* decode objid elements one by one */
          objid_val->elements[i].elem_as_number = binary_string_get_long(iter);
        }
        tciSetObjidValue(value, objid_val);
      }
      break;
      
    case TCI_OCTETSTRING_TYPE:
      {
        String ostring;

        /* decode octet string */ 
	    decode_string(iter, NULL, &ostring);
	    tciSetOStringValue(value, ostring);
	  }
      break;
      
    case TCI_RECORD_TYPE:
    case TCI_SET_TYPE:
      {
        String *field_names;
        unsigned long i;
        TciValue elem;
        
        /* get list of record/set field names */
        field_names = tciGetRecFieldNames(value);

        /* iterate until NULL element - it marks end of list */
        for (i = 0; field_names[i] != NULL; i++) {
          /* decode structure elements one by one */

          /* check whether value is omitted */
          unsigned char omitted = binary_string_get_char(iter);

          if (omitted) {
            tciSetFieldOmitted(value, field_names[i]);
          } else {
            /* instantiate field value */
            elem = tciGetRecFieldValue(value, field_names[i]);

            decode(iter, elem);

            tciSetRecFieldValue(value, field_names[i], elem);
          }
        }
      }	
      break;
      
    case TCI_RECORD_OF_TYPE:
    case TCI_SET_OF_TYPE:
      {
        unsigned long length;
        unsigned long i;
        TciValue elem;
        TciType elem_type;

        /* get element type of record/set */
        elem_type = tciGetRecOfElementType(value);
        
        /* decode length of record of/set of */
        length = binary_string_get_long(iter);

        tciSetRecOfLength(value, length);

        for (i = 0; i < length; i++) {
          /* decode record of/set of elements one by one */

          elem = tciNewInstance(elem_type);

          decode(iter, elem);

          tciSetRecOfFieldValue(value, i, elem);
        }
      }
      break;
         
    case TCI_ANYTYPE_TYPE:
    case TCI_UNION_TYPE:
      {
        String variant;
        TciValue elem;
        
        /* decode variant name */
        decode_string(iter, NULL, &variant);

        /* instantiate variant value */
        elem = tciGetUnionVariant(value, variant);
	    
        /* decode variant value */
        decode(iter, elem);

        tciSetUnionVariant(value, variant, elem);
      }
      break;
      
    case TCI_UNIVERSAL_CHAR_TYPE:
      {
        TciUCValue wchar;
        
        /* universal char is represented in TCI as unsigned char[4] array */
        binary_string_get_bytes(iter, sizeof(TciUCValue), wchar);
        tciSetUniversalCharValue(value, wchar);
      }
      break;
      
    case TCI_UNIVERSAL_CHARSTRING_TYPE:
      {
        unsigned long length;
        unsigned long i;
        TciUCValue wchar;
        
        /* decode string length */
        length = binary_string_get_long(iter);
        tciSetUCStringLength(value, length);
      
        for(i = 0; i < length; i++) {
          /*decode universal charstring elements one by one */
          binary_string_get_bytes(iter, sizeof(TciUCValue), wchar);
          tciSetUCStringCharValue(value, i, wchar);
        }
      }
      break;
      
    case TCI_VERDICT_TYPE:
      {
        unsigned long verdict;
        
        /* verdict is represented in TCI as integer */
        verdict = binary_string_get_long(iter);
        tciSetVerdictValue(value, verdict);
      }
      break;
      
    default:
      tci_assert(0, "Unsupported type kind in Decoder");
  }
}

/*
 * tciEncode
 * 
 * This is the main encoding function that is called 
 * every time there is a need to encode TciValue.
 * 
 * BinaryString iterator allocates data in permanent memory
 * therefore this function creates copy of encoded data
 * in temporary memory and returns it. Temporary memory
 * will be freed lately by the runtime system.
 *
*/

BinaryString tciEncode(TciValue value)
{
  MyBinaryString msg;
  TciType type;

  msg.allocated = 0;
  msg.string.data = NULL;
  msg.string.bits = 0;
  msg.string.aux = NULL;

  type = tciGetType(value);
  tci_assert_wret(type != NULL, "Illegal type", msg.string);
  encode_type(&msg, type);
  
  encode(&msg, value);

  return msg.string;
}

/*
 * tciDecode
 * 
 * This is the main decoding function that is called 
 * every time there is a need to decode BinaryString.
 * It returns NULL in case when BinaryString cannot be decoded.
 *
*/

TciValue tciDecode(BinaryString tri_bstr, TciType decHypothesis)
{
  BinaryString msg;
  BinaryStringIter iter;
  TciValue value;
  TciType type;

  iter.string = tri_bstr;
  iter.bit_index = 0;

  decode_type(&iter, &type);
  tci_assert_wret(type != NULL, "Illegal type", NULL);

  value = tciNewInstance(type);
  
  decode(&iter, value);

  tci_assert_wret(iter.bit_index == iter.string.bits, "Decoder failed to fully decode binary string", NULL);

  /* check strict type compatibility */ 
  if (decHypothesis && value) {
    TciModuleIdType hyp_module_id = tciGetDefiningModule(decHypothesis);
    String hyp_module_name = hyp_module_id.moduleName;
    String hyp_type_name = tciGetName(decHypothesis);

    TciType type = tciGetType(value);
    TciModuleIdType module_id = tciGetDefiningModule(type);
    String module_name = module_id.moduleName;
    String type_name = tciGetName(type);
    
    if (!((hyp_module_name && module_name) || (!hyp_module_name && !module_name)))
      return NULL;

    if (strcmp(hyp_type_name, type_name) != 0)
      return NULL;
  }
    
  return value;
}  

/* End of File */

