/***************************************************************************************
 * TCI Test Logging example implementation
 *
 * 2010 IBM Corporation
 *
 * This is an example implementation of TCI Test Logging (TL) interface
 *
 * This implementation tries to be close to default builtin log mechanism
 *
 * By default all events are written to "tli_log.txt" file 
 * (to avoid clashes with builtin log messages).
 * If you compile your project with /DT3RT_NO_BUILTIN_LOG then default builtin log will
 * be switched off and this implementation will write events to stdout stream (to screen).
 *
 * You may find useful the implementation of the function that prints contents 
 * of arbitrary TciValue object.
 *
 ***************************************************************************************/
 
#ifdef __cplusplus
extern "C" {
#endif

#include "tci.h"
#include <stdio.h>
#include <stdlib.h>

/* common platform independent name for snprintf function */
#if defined(_WIN32) && ! defined(WIN32_GCC)   /* Win32 platform */
#define str_printf _snprintf
#else
#define str_printf snprintf
#endif

/* default output file name for TLI events */
#ifndef TLI_LOG_FILE_NAME
#define TLI_LOG_FILE_NAME "tli_log.txt"
#endif

/* Change or remove this definition if you would like whole event printed (it may be quite large) */
#define MAX_VALUE_LENGTH 100

/* Macros to print TCI and TRI lists. */

#define TCI_LOG_LIST(buf, list, member, type) {                                                       \
                                                  int i;                                              \
                                                  if ((list).length != 0)                             \
                                                  {                                                   \
                                                      int i;                                          \
                                                      for (i = 0; i < (list).length; i++)             \
                                                      {                                               \
                                                          if (i != 0) bufAppendString(buf, ", ");     \
                                                          log##type(buf, (list).member[i]);           \
                                                      }                                               \
                                                  }                                                   \
                                              }

#define TRI_LOG_LIST(buf, list, member, type) {                                                       \
                                                  int i;                                              \
                                                  if ((list).length != 0)                             \
                                                  {                                                   \
                                                      int i;                                          \
                                                      for (i = 0; i < (list).length; i++)             \
													  {                                               \
                                                          if (i != 0) bufAppendString(buf, ", ");     \
                                                          if ((list).member[i])                       \
                                                              log##type(buf, *((list).member[i]));    \
                                                          else                                        \
                                                              bufAppendString(buf, "(null)");         \
                                                      }                                               \
                                                  }                                                   \
                                              }

#define DECLARE_BUFFER log_print_buffer_t buf = {NULL, 0, 0}
#define PBUF &buf
#define WRITE_BUFFER bufWrite(&buf)

#define APPEND_STR(str) bufAppendString(&buf, str)
#define APPEND_STR2(str1, str2) {bufAppendString(&buf, str1); bufAppendString(&buf, str2);}
#define APPEND_STR3(str1, str2, str3) {bufAppendString(&buf, str1); bufAppendString(&buf, str2); bufAppendString(&buf, str3);}
#define APPEND_SEP bufAppendString(&buf, ", ")

#define STATUS_OK(status) (status == TRI_OK)
#define GET_REASON(encoderStatus, trasmissionStatus) ((!STATUS_OK(encoderStatus)) ? "encoder error" : "transmission error")
/* Log buffer declaration and management routines */

typedef struct _log_print_buffer_t {
    char *buf;
    unsigned long length;
    unsigned long allocated;
} log_print_buffer_t;

/* Prints TciValue into ASCII string and appends it to buffer */
static void to_string(TciValue value, log_print_buffer_t *buf, int indent);

static void bufAppendString(log_print_buffer_t *buf, const char *str)
{
    unsigned long len = str ? strlen(str) : 0;

    if (!len) return;
    
    if (buf->allocated - buf->length < len + 1) 
    {
        /* 1 for '\0', 100 - reserved */
        unsigned long new_size = buf->length + len + 1 + 100;
        if (buf->allocated)
          buf->buf = realloc(buf->buf, new_size);
        else
        {
          buf->buf = malloc(new_size);
          buf->buf[0] = '\0';
        }

        buf->allocated = new_size;
    }
    strcat(buf->buf, str);
    buf->length += len;
}

static void bufWrite(log_print_buffer_t *buf)
{
	static FILE* TLfile = NULL; 

	if (!buf->buf) return;
	
#ifdef T3RT_NO_BUILTIN_LOG	
    TLfile = stdout;
#else
    if (TLfile == NULL) 
        TLfile = fopen(TLI_LOG_FILE_NAME, "wt");
#endif

    fprintf(TLfile, "%s\n", buf->buf);
    fflush(TLfile);
    
    free(buf->buf);
}                                            
                                            
/* Operations to log basic types */
static void logString(log_print_buffer_t *buf, String str) 
{
    bufAppendString(buf, str);
}

static void logBinaryString(log_print_buffer_t *buf, BinaryString binary_string) 
{
    int i;
    char str[100];
    
    bufAppendString(buf, "'");
    if ((binary_string.bits%8) == 0) 
    {
        for (i = 0; i < binary_string.bits / 8; i++) 
        {
            sprintf(str, "%02x", binary_string.data[i]);
            bufAppendString(buf, str);
        }
        bufAppendString(buf, "'H");
    }
    else 
    {
        for (i = 0; i < binary_string.bits; i++) 
        {
            sprintf(str, "%d", ((binary_string.data[i/8] & ((unsigned char)(1) << (7-i%8))) == 0) ? 0 : 1);
            bufAppendString(buf, str);
        }
        bufAppendString(buf, "'B");
    }
}

static void logQualifiedName(log_print_buffer_t *buf, QualifiedName qualified_name)
{
    if (qualified_name.moduleName) 
    {
        bufAppendString(buf, qualified_name.moduleName);
        bufAppendString(buf, ".");
    }
    bufAppendString(buf, qualified_name.objectName);
}

static void logTriComponentId(log_print_buffer_t *buf, TriComponentId component_id)
{
    logString(buf, component_id.compName);
}

static void logTciBehaviourIdType(log_print_buffer_t *buf, TciBehaviourIdType behaviour_id_type) 
{
    logQualifiedName(buf, behaviour_id_type);
}

static void logTciValue(log_print_buffer_t *buf, TciValue value)
{
    log_print_buffer_t tmp_buf = {NULL, 0, 0};
    to_string(value, &tmp_buf, 0);

	if (tmp_buf.length > MAX_VALUE_LENGTH)
	{
		tmp_buf.buf[MAX_VALUE_LENGTH] = '\0';
       	bufAppendString(buf, tmp_buf.buf);
       	bufAppendString(buf, "<...(truncated)>");

	}
	else
	{
       	bufAppendString(buf, tmp_buf.buf);
    }
    
    if (tmp_buf.buf)
      free(tmp_buf.buf);
}

static void logTciParameterType(log_print_buffer_t *buf, TciParameterType parameter)
{
/*
    switch (parameter.parPassMode) 
    {
        case TCI_IN_PAR    : bufAppendString(buf, "in");
                             break;
        case TCI_INOUT_PAR : bufAppendString(buf, "inout");
                             break;
        case TCI_OUT_PAR   : bufAppendString(buf, "out");
                             break;
    }
    bufAppendString(buf, " ");
*/
	if (parameter.parValue)
		logTciValue(buf, parameter.parValue);
	else
		bufAppendString(buf, "-");
}

static void logTciParameterListType(log_print_buffer_t *buf, TciParameterListType parameter_list_type, char *open_br, char *close_br)
{
	bufAppendString(buf, open_br);
    if (parameter_list_type != NULL) 
    {
        TCI_LOG_LIST(buf, *parameter_list_type, parList, TciParameterType);
    }    
	else 
	{
		bufAppendString(buf, "<not available>");
    }
	bufAppendString(buf, close_br);
}

static void logTciTestCaseIdType(log_print_buffer_t *buf, TciTestCaseIdType test_case_id_type)
{
    logQualifiedName(buf, test_case_id_type);
}

static void logTciValueTemplate(log_print_buffer_t *buf, TciValueTemplate value_template)
{
    if (!value_template) 
    {
        bufAppendString(buf, "(null)");
    }
    else 
    {
		bufAppendString(buf, tciGetValueTemplateDef(value_template));
    }
}

static void logTciNonValueTemplate(log_print_buffer_t *buf, TciNonValueTemplate non_value_template)
{
    if (!non_value_template) 
    {
        bufAppendString(buf, "(null)");
    }
    else 
    {
        bufAppendString(buf, tciGetNonValueTemplateDef(non_value_template));
    }
}

static void logTciValueDifference(log_print_buffer_t *buf, TciValueDifference value_difference)
{
    bufAppendString(buf, "'"); 
	bufAppendString(buf, value_difference.desc);
    bufAppendString(buf, "'");
	
	bufAppendString(buf, "(expected: ");
	logTciValueTemplate(buf, value_difference.tmpl);
	bufAppendString(buf, ", got: ");
	logTciValue(buf, value_difference.val);
	bufAppendString(buf, ")");
}

static void logTciValueDifferenceList(log_print_buffer_t *buf, TciValueDifferenceList value_difference_list)
{
	if (value_difference_list != NULL) 
	{
		if (value_difference_list->length > 1)
			bufAppendString(buf, "(");

		TCI_LOG_LIST(buf, *value_difference_list, diffList, TciValueDifference);

		if (value_difference_list->length > 1)
			bufAppendString(buf, ")");
	}
	else 
	{
		bufAppendString(buf, "<unknown reason>");
	}
}

static void logTciValueList(log_print_buffer_t *buf, TciValueList value_list)
{
	bufAppendString(buf, "(");
	if (value_list != NULL) 
	{
    	TCI_LOG_LIST(buf, *value_list, valueList, TciValue);
    }
	else 
	{
		bufAppendString(buf, "<not available>");
	}
	bufAppendString(buf, ")");
}

static void logTciVerdictValue(log_print_buffer_t *buf, TciValue value)
{
	TciVerdictValue verdict_value = tciGetVerdictValue(value);
	
    switch (verdict_value) 
    {
        case TCI_VERDICT_NONE_STATIC   : bufAppendString(buf, "none");
                                         break;
        case TCI_VERDICT_PASS_STATIC   : bufAppendString(buf, "pass");
                                         break;
        case TCI_VERDICT_INCONC_STATIC : bufAppendString(buf, "inconc");
                                         break;
        case TCI_VERDICT_FAIL_STATIC   : bufAppendString(buf, "fail");
                                         break;
        case TCI_VERDICT_ERROR_STATIC  : bufAppendString(buf, "error");
                                         break;
    }
}

static void logTriAddress(log_print_buffer_t *buf, TriAddress address)
{
    logBinaryString(buf, address);
}

static void logTriAddressList(log_print_buffer_t *buf, TriAddressList address_list)
{
	bufAppendString(buf, "(");
	if (address_list.length) 
	{
	    TRI_LOG_LIST(buf, address_list, addrList, TriAddress);
	}
	else
	{
		bufAppendString(buf, "<not available>");
	}
	bufAppendString(buf, ")");
}

static void logTriComponentIdList(log_print_buffer_t *buf, TriComponentIdList component_id_list)
{
	bufAppendString(buf, "(");
	if (component_id_list.length) 
	{
	    TRI_LOG_LIST(buf, component_id_list, compIdList, TriComponentId);
	}
	else 
	{
		bufAppendString(buf, "<not available>");
	}
	bufAppendString(buf, ")");
}

static void logCDestination(log_print_buffer_t *buf, TriComponentId comp_id)
{
    logString(buf, "'");
    logTriComponentId(buf, comp_id);
    logString(buf, "'");
}

static void logPDestination(log_print_buffer_t *buf, TriPortId port_id)
{
    logCDestination(buf, port_id.compInst);
}

static void logPDestinationList(log_print_buffer_t *buf, TriPortIdList port_id_list)
{
	bufAppendString(buf, "(");
	if (port_id_list.length) 
	{
	    TRI_LOG_LIST(buf, port_id_list, portIdList, PDestination);
	}
	else
	{
		bufAppendString(buf, "<not available>");
	}
	bufAppendString(buf, ")");
}

static void logTriException(log_print_buffer_t *buf, TriException exception)
{
    logBinaryString(buf, exception);
}

static void logTriMessage(log_print_buffer_t *buf, TriMessage message)
{
    logBinaryString(buf, message);
}

static void logTriParameter(log_print_buffer_t *buf, TriParameter parameter)
{
    switch (parameter.mode) 
    {
        case TRI_IN    : bufAppendString(buf, "in");
                         break;
        case TRI_INOUT : bufAppendString(buf, "inout");
                         break;
        case TRI_OUT   : bufAppendString(buf, "out");
                         break;
    }
    bufAppendString(buf, " ");
    logBinaryString(buf, parameter.par);
}

static void logTriParameterList(log_print_buffer_t *buf, TriParameterList parameter_list)
{
	bufAppendString(buf, "(");
	if (parameter_list.length) 
	{
		TRI_LOG_LIST(buf, parameter_list, parList, TriParameter);
	}
	else 
	{
		bufAppendString(buf, "<not available>");
	}
	bufAppendString(buf, ")");
}

static void logTriPortId(log_print_buffer_t *buf, TriPortId port_id)
{
    char str[100];

    bufAppendString(buf, port_id.portName);
	if (port_id.portIndex != -1) 
	{
	    char str[100];

	    bufAppendString(buf, "[");
		sprintf(str, "%d", port_id.portIndex);
		bufAppendString(buf, str);
	    bufAppendString(buf, "]");
	}
}

static void logTriSignatureId(log_print_buffer_t *buf, TriSignatureId signature_id)
{
    logQualifiedName(buf, signature_id);
}

static void logTriStatus(log_print_buffer_t *buf, TriStatus status)
{
    switch (status) 
    {
        case TRI_Error : bufAppendString(buf, "'TRI Error'");
                         break;
        case TRI_OK    : bufAppendString(buf, "'TRI OK'");
                         break;
    }
}

static void logTriTimerDuration(log_print_buffer_t *buf, TriTimerDuration timer_duration)
{
    char str[100];
    
    sprintf(str, "%f", timer_duration);
    bufAppendString(buf, str);
}

static void logTriTimerId(log_print_buffer_t *buf, TriTimerId timer_id)
{
    logBinaryString(buf, timer_id);
}

static void logTimerStatus(log_print_buffer_t *buf, TimerStatus status)
{
    switch(status)
    {
      case runningT:
        bufAppendString(buf, "RUNNING");
        break;
      case inactiveT:
        bufAppendString(buf, "INACTIVE");
        break;
      case expiredT:
        bufAppendString(buf, "EXPIRED");
        break;
    }
}

static void logComponentStatus(log_print_buffer_t *buf, ComponentStatus status)
{
    switch(status)
    {
      case inactiveC:
        bufAppendString(buf, "INACTIVE");
        break;
      case runningC:
        bufAppendString(buf, "RUNNING");
        break;
      case stoppedC:
        bufAppendString(buf, "STOPPED");
        break;
      case killedC:
        bufAppendString(buf, "KILLED");
        break;
    }
}

/*
 * Timestamp is in milliseconds elapsed since midnight
 */
static void
logTimeStamp(log_print_buffer_t *buf, int ts)
{
    int hour, min, sec, msec;
    char timebuffer[100];

    hour = ts / (3600 * 1000);
    min  = (ts - hour * 3600 * 1000) / (60 * 1000);
    sec  = (ts - (hour * 3600 + min * 60) * 1000) / 1000;
    msec = ts - (hour * 3600 + min * 60 + sec) * 1000;
 
    sprintf(timebuffer, "[%.2d:%.2d:%.2d.%d] ", hour, min, sec, msec);
    bufAppendString(buf, timebuffer);
}

/* Operation executed for all events */

static void logTliEventInformation(log_print_buffer_t *buf, String event, String am, long int ts, String src, long int line, TriComponentId c)
{
    char str[100];

    if (src && line)
    {
	    bufAppendString(buf, src);
	    sprintf(str, " (%d): ", line);
	    bufAppendString(buf, str);
	}

    logTimeStamp(buf, ts);
	
    bufAppendString(buf, "[");
	logTriComponentId(buf, c);
    bufAppendString(buf, "] ");
    
    bufAppendString(buf, event);
    bufAppendString(buf, ":\t");
}


/* TCI-TL operations provided to TE */

/* tliTcExecute
 * 
 * Shall be called by TE to log the execute test case request.
*/
                                              
void tliTcExecute (String am,
                   long int ts,
                   String src,
                   long int line,
                   TriComponentId c,
                   TciTestCaseIdType tcId,
                   TciParameterListType pars,
                   TriTimerDuration dur)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "testcase_executed", am, ts, src, line, c);
    /* log parameters */
    logTciTestCaseIdType(PBUF, tcId); logTciParameterListType(PBUF, pars, "(", ")");

    WRITE_BUFFER;
}

/* tliTcStart
 * 
 * Shall be called by TE to log the start of a testcase. This event occurs before the testcase is started.
*/
                          
void tliTcStart (String am,
                 long int ts,
                 String src,
                 long int line,
                 TriComponentId c,
                 TciTestCaseIdType tcId,
                 TciParameterListType pars,
                 TriTimerDuration dur)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "testcase_starting", am, ts, src, line, c);
    /* log parameters */
    logTciTestCaseIdType(PBUF, tcId); logTciParameterListType(PBUF, pars, "(", ")");

    WRITE_BUFFER;
}


/* tliTcStop
 * 
 * Shall be called by TE to log the stop of a testcase.
*/
                        
void tliTcStop (String am,
                long int ts,
                String src,
                long int line,
                TriComponentId c)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "testcase_stopped", am, ts, src, line, c);
    /* no parameters */

    WRITE_BUFFER;
}

/* tliTcStarted
 * 
 * Shall be called by TM to log the start of a testcase. This event occurs after the testcase was started.
*/
                       
void tliTcStarted (String am,
                   long int ts,
                   String src,
                   long int line,
                   TriComponentId c,
                   TciTestCaseIdType tcId,
                   TciParameterListType pars,
                   TriTimerDuration dur)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "testcase_started", am, ts, src, line, c);
    /* log parameters */
    logTciTestCaseIdType(PBUF, tcId); logTciParameterListType(PBUF, pars, "(", ")");

    WRITE_BUFFER;
}


/* tliTcTerminated
 * 
 * Shall be called by TM to log the termination of a testcase. This event occurs after the testcase
 * terminated.
*/
                          
void tliTcTerminated (String am,
                      long int ts,
                      String src,
                      long int line,
                      TriComponentId c,
                      TciTestCaseIdType tcId,
                      TciParameterListType pars,
                      TciValue outcome)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "testcase_terminated", am, ts, src, line, c);
    /* log parameters */
	logTciVerdictValue(PBUF, outcome);
    APPEND_STR(" <= "); 
	logTciTestCaseIdType(PBUF, tcId); logTciParameterListType(PBUF, pars, "(", ")");

    WRITE_BUFFER;
}

/* tliCtrlStart
 * 
 * Shall be called by TE to log the start of the control part. This event occurs before the control is
 * started. If the control is not represented by a TRI component, 'c' is null.
*/
                             
void tliCtrlStart (String am,
                   long int ts,
                   String src,
                   long int line,
                   TriComponentId c)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "control_part_started", am, ts, src, line, c);
    /* no parameters */

    WRITE_BUFFER;
}

/* tliCtrlStop
 * 
 * Shall be called by TE to log the stop of the control part. This event occurs before the control is
 * stopped. If the control is not represented by a TRI component, 'c' is null.
*/
                          
void tliCtrlStop (String am,
                  long int ts,
                  String src,
                  long int line,
                  TriComponentId c)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "control_part_stopped", am, ts, src, line, c);
    /* no parameters */

    WRITE_BUFFER;
}

/* tliCtrlTerminated
 * 
 * Shall be called by TM to log the termination of the control part. This event occurs after the control
 * has terminated. If the control is not represented by a TRI component, 'c' is null.
*/
 
void tliCtrlTerminated (String am,
                        long int ts,
                        String src,
                        long int line,
                        TriComponentId c)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "control_part_terminated", am, ts, src, line, c);
    /* no parameters */

    WRITE_BUFFER;
}

/* tliMSend_m
 * 
 * Shall be called by SA to log a unicast send operation. This event occurs after sending. This event is
 * used for logging the communication with the SUT.
*/
                               
void tliMSend_m (String am,
                 long int ts,
                 String src,
                 long int line,
                 TriComponentId c,
                 TriPortId atPort,
                 TriPortId toPort,
                 TciValue msgValue,
                 TriAddress address,
                 TriStatus encoderStatus,
                 TriMessage msg,
                 TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(encoderStatus) && STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "sut_message_sent";
	else
		eventName = "sut_message_sending_failed";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */
    logTriPortId(PBUF, atPort); 
	APPEND_STR(".send("); logTciValue(PBUF, msgValue); APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(encoderStatus, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliMSend_m_BC
 * 
 * Shall be called by SA to log a broadcast send operation. This event occurs after sending. This event is
 * used for logging the communication with the SUT.
*/

void tliMSend_m_BC (String am,
                    long int ts,
                    String src,
                    long int line,
                    TriComponentId c,
                    TriPortId atPort,
                    TriPortId toPort,
                    TciValue msgValue,
                    TriStatus encoderStatus,
                    TriMessage msg,
                    TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(encoderStatus) && STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "sut_message_sent_bc";
	else
		eventName = "sut_message_sending_failed_bc";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */
    logTriPortId(PBUF, atPort); 
	APPEND_STR(".send("); logTciValue(PBUF, msgValue); APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(encoderStatus, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliMSend_m_MC
 * 
 * Shall be called by SA to log a multicast send operation. This event occurs after sending. This event is
 * used for logging the communication with the SUT.
*/

void tliMSend_m_MC (String am,
                    long int ts,
                    String src,
                    long int line,
                    TriComponentId c,
                    TriPortId atPort,
                    TriPortId toPort,
                    TciValue msgValue,
                    TriAddressList addresses,
                    TriStatus encoderStatus,
                    TriMessage msg,
                    TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(encoderStatus) && STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "sut_message_sent_mc";
	else
		eventName = "sut_message_sending_failed_mc";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */
    logTriPortId(PBUF, atPort); 
	APPEND_STR(".send("); logTciValue(PBUF, msgValue); APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(encoderStatus, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliMSend_c
 * 
 * Shall be called by CH to log a unicast send operation. This event occurs after sending. This event is
 * used for logging the intercomponent communication.
*/
                           
void tliMSend_c (String am,
                 long int ts,
                 String src,
                 long int line,
                 TriComponentId c,
                 TriPortId atPort,
                 TriPortId toPort,
                 TciValue msgValue,
                 TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "message_sent";
	else
		eventName = "message_sending_failed";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

    APPEND_STR("to "); logPDestination(PBUF, toPort); APPEND_STR(" ");
	logTriPortId(PBUF, atPort);
	APPEND_STR(".send("); logTciValue(PBUF, msgValue); APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(TRI_OK, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliMSend_c_BC
 * 
 * Shall be called by CH to log a broadcast send operation. This event occurs after sending. This event is
 * used for logging the intercomponent communication.
*/

void tliMSend_c_BC (String am,
                    long int ts,
                    String src,
                    long int line,
                    TriComponentId c,
                    TriPortId atPort,
                    TriPortIdList toPortList,
                    TciValue msgValue,
                    TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "message_sent_bc";
	else
		eventName = "message_sending_failed_bc";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

    APPEND_STR("to "); logPDestinationList(PBUF, toPortList); APPEND_STR(" ");
	logTriPortId(PBUF, atPort);
	APPEND_STR(".send("); logTciValue(PBUF, msgValue); APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(TRI_OK, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliMSend_c_MC
 * 
 * Shall be called by CH to log a multicast send operation. This event occurs after sending. This event is
 * used for logging the intercomponent communication.
*/

void tliMSend_c_MC (String am,
                    long int ts,
                    String src,
                    long int line,
                    TriComponentId c,
                    TriPortId atPort,
                    TriPortIdList toPortList,
                    TciValue msgValue,
                    TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "message_sent_mc";
	else
		eventName = "message_sending_failed_mc";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

    APPEND_STR("to "); logPDestinationList(PBUF, toPortList); APPEND_STR(" ");
	logTriPortId(PBUF, atPort);
	APPEND_STR(".send("); logTciValue(PBUF, msgValue); APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(TRI_OK, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliMDetected_m
 * 
 * Shall be called by SA to log the enqueuing of a message. This event occurs after the message is
 * enqueued. This event is used for logging the communication with the SUT.
*/
                           
void tliMDetected_m (String am,
                     long int ts,
                     String src,
                     long int line,
                     TriComponentId c,
                     TriPortId atPort,
                     TriPortId fromPort,
                     TriMessage msg,
                     TriAddress address)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "sut_message_detected", am, ts, src, line, c);
    /* log parameters */
	APPEND_STR("at port "); 
    logTriPortId(PBUF, atPort);
	APPEND_STR(" detected message"); 

    WRITE_BUFFER;
}

/* tliMDetected_c
 * 
 * Shall be called by CH to log the enqueuing of a message. This event occurs after the message is
 * enqueued. This event is used for logging the intercomponent communication.
*/
                            
void tliMDetected_c (String am,
                     long int ts,
                     String src,
                     long int line,
                     TriComponentId c,
                     TriPortId atPort,
                     TriPortId fromPort,
                     TciValue msgValue)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "message_detected", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("from "); logPDestination(PBUF, fromPort);
	APPEND_STR(" at port "); 
    logTriPortId(PBUF, atPort); 
	APPEND_STR(" detected message"); 

    WRITE_BUFFER;
}

/* tliMMismatch_m
 * 
 * Shall be called by TE to log the mismatch of a template. This event occurs after checking a
 * template match. This event is used for logging the communication with the SUT.
*/
                            
void tliMMismatch_m (String am,
                     long int ts,
                     String src,
                     long int line,
                     TriComponentId c,
                     TriPortId port,
                     TciValue msgValue,
                     TciValueTemplate msgTmpl,
                     TciValueDifferenceList diffs,
                     TriAddress address,
                     TciValueTemplate addressTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "sut_message_mismatch", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("at port "); logTriPortId(PBUF, port); APPEND_STR(" ");
	logTciValueDifferenceList(PBUF, diffs);

    WRITE_BUFFER;
}

/* tliMMismatch_c
 * 
 * Shall be called by TE to log the mismatch of a template. This event occurs after checking a
 * template match. This event is used for logging the intercomponent communication.
*/

void tliMMismatch_c (String am,
                     long int ts,
                     String src,
                     long int line,
                     TriComponentId c,
                     TriPortId port,
                     TciValue msgValue,
                     TciValueTemplate msgTmpl,
                     TciValueDifferenceList diffs,
                     TriComponentId from,
                     TciNonValueTemplate fromTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "message_mismatch", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("from "); logCDestination(PBUF, from);
    APPEND_STR(" at port "); logTriPortId(PBUF, port); APPEND_STR(" ");
	logTciValueDifferenceList(PBUF, diffs);

    WRITE_BUFFER;
}

/* tliMReceive_m
 * 
 * Shall be called by TE to log the receive of a message. This event occurs after checking a template
 * match. This event is used for logging the communication with SUT.
*/

void tliMReceive_m (String am,
                    long int ts,
                    String src,
                    long int line,
                    TriComponentId c,
                    TriPortId port,
                    TciValue msgValue,
                    TciValueTemplate msgTmpl,
                    TriAddress address,
                    TciValueTemplate addressTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "sut_message_received", am, ts, src, line, c);
    /* log parameters */

    logTriPortId(PBUF, port);
	APPEND_STR(".receive");
	if (msgValue) 
	{
		APPEND_STR("(");
		logTciValue(PBUF, msgValue);
		APPEND_STR(")");
	}
	APPEND_STR(" received");

	WRITE_BUFFER;
}

/* tliMReceive_c
 * 
 * Shall be called by TE to log the receive of a message. This event occurs after checking a template
 * match. This event is used for logging the intercomponent communication.
*/
                           
void tliMReceive_c (String am,
                    long int ts,
                    String src,
                    long int line,
                    TriComponentId c,
                    TriPortId port,
                    TciValue msgValue,
                    TciValueTemplate msgTmpl,
                    TriComponentId from,
                    TciNonValueTemplate fromTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "message_received", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("from "); logCDestination(PBUF, from); APPEND_STR(" ");
    logTriPortId(PBUF, port);
	APPEND_STR(".receive");
	if (msgValue) 
	{
		APPEND_STR("(");
		logTciValue(PBUF, msgValue);
		APPEND_STR(")");
	}
	APPEND_STR(" received");

    WRITE_BUFFER;
}

/* tliPrCall_m
 * 
 * Shall be called by SA to log a unicast call operation. This event occurs after call execution. This
 * event is used for logging the communication with the SUT.
*/
                           
void tliPrCall_m (String am,
                  long int ts,
                  String src,
                  long int line,
                  TriComponentId c,
                  TriPortId atPort,
                  TriPortId toPort,
                  TriSignatureId signature,
                  TciParameterListType parsValue,
                  TriAddress address,
                  TriStatus encoderStatus,
                  TriParameterList pars,
                  TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(encoderStatus) && STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "sut_call_initiated";
	else
		eventName = "sut_call_failed";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

    logTriPortId(PBUF, atPort);
	APPEND_STR(".call("); 
	logTriSignatureId(PBUF, signature); APPEND_STR(":"); logTciParameterListType(PBUF, parsValue, "{", "}"); 
	APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(encoderStatus, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrCall_m_BC
 * 
 * Shall be called by SA to log a broadcast call operation. This event occurs after call execution. This
 * event is used for logging the communication with the SUT.
*/

void tliPrCall_m_BC (String am,
                     long int ts,
                     String src,
                     long int line,
                     TriComponentId c,
                     TriPortId atPort,
                     TriPortId toPort,
                     TriSignatureId signature,
                     TciParameterListType parsValue,
                     TriStatus encoderStatus,
                     TriParameterList pars,
                     TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(encoderStatus) && STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "sut_call_initiated_bc";
	else
		eventName = "sut_call_failed_bc";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

    logTriPortId(PBUF, atPort);
	APPEND_STR(".call("); 
	logTriSignatureId(PBUF, signature); APPEND_STR(":"); logTciParameterListType(PBUF, parsValue, "{", "}"); 
	APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(encoderStatus, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrCall_m_MC
 * 
 * Shall be called by SA to log a multicast call operation. This event occurs after call execution. This
 * event is used for logging the communication with the SUT.
*/

void tliPrCall_m_MC (String am,
                     long int ts,
                     String src,
                     long int line,
                     TriComponentId c,
                     TriPortId atPort,
                     TriPortId toPort,
                     TriSignatureId signature,
                     TciParameterListType parsValue,
                     TriAddressList addresses,
                     TriStatus encoderStatus,
                     TriParameterList pars,
                     TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(encoderStatus) && STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "sut_call_initiated_mc";
	else
		eventName = "sut_call_failed_mc";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

    logTriPortId(PBUF, atPort);
	APPEND_STR(".call("); 
	logTriSignatureId(PBUF, signature); APPEND_STR(":"); logTciParameterListType(PBUF, parsValue, "{", "}"); 
	APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(encoderStatus, transmissionStatus));
	}

    WRITE_BUFFER;
}

          
/* tliPrCall_c
 * 
 * Shall be called by CH to log a unicast call operation. This event occurs after call execution. This
 * event is used for logging the intercomponent communication.
*/
                            
void tliPrCall_c (String am,
                  long int ts,
                  String src,
                  long int line,
                  TriComponentId c,
                  TriPortId atPort,
                  TriPortId toPort,
                  TriSignatureId signature,
                  TciParameterListType parsValue,
                  TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "sut_call_initiated";
	else
		eventName = "sut_call_failed";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

	APPEND_STR("to "); logPDestination(PBUF, toPort); APPEND_STR(" ");
	logTriPortId(PBUF, atPort);
	APPEND_STR(".call("); 
	logTriSignatureId(PBUF, signature); APPEND_STR(":"); logTciParameterListType(PBUF, parsValue, "{", "}"); 
	APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(TRI_OK, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrCall_c_BC
 * 
 * Shall be called by CH to log a broadcast call operation. This event occurs after call execution. This
 * event is used for logging the intercomponent communication.
*/

void tliPrCall_c_BC (String am,
                     long int ts,
                     String src,
                     long int line,
                     TriComponentId c,
                     TriPortId atPort,
                     TriPortIdList toPortList,
                     TriSignatureId signature,
                     TciParameterListType parsValue,
                     TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "sut_call_initiated_bc";
	else
		eventName = "sut_call_failed_bc";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

	APPEND_STR("to "); logPDestinationList(PBUF, toPortList); APPEND_STR(" ");
	logTriPortId(PBUF, atPort);
	APPEND_STR(".call("); 
	logTriSignatureId(PBUF, signature); APPEND_STR(":"); logTciParameterListType(PBUF, parsValue, "{", "}"); 
	APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(TRI_OK, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrCall_c_MC
 * 
 * Shall be called by CH to log a multicast call operation. This event occurs after call execution. This
 * event is used for logging the intercomponent communication.
*/

void tliPrCall_c_MC (String am,
                     long int ts,
                     String src,
                     long int line,
                     TriComponentId c,
                     TriPortId atPort,
                     TriPortIdList toPortList,
                     TriSignatureId signature,
                     TciParameterListType parsValue,
                     TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "sut_call_initiated_mc";
	else
		eventName = "sut_call_failed_mc";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

	APPEND_STR("to "); logPDestinationList(PBUF, toPortList); APPEND_STR(" ");
	logTriPortId(PBUF, atPort);
	APPEND_STR(".call("); 
	logTriSignatureId(PBUF, signature); APPEND_STR(":"); logTciParameterListType(PBUF, parsValue, "{", "}"); 
	APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(TRI_OK, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrGetCallDetected_m
 * 
 * Shall be called by SA to log the getcall enqueue operation. This event occurs after call is enqueued.
 * This event is used for logging the communication with the SUT.
*/
                            
void tliPrGetCallDetected_m (String am,
                             long int ts,
                             String src,
                             long int line,
                             TriComponentId c,
                             TriPortId atPort,
                             TriPortId fromPort,
                             TriSignatureId signature,
                             TriParameterList pars,
                             TriAddress address)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "sut_call_detected", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("at port "); logTriPortId(PBUF, atPort);
    APPEND_STR(" detected request to call "); logTriSignatureId(PBUF, signature);;

    WRITE_BUFFER;
}

/* tliPrGetCallDetected_c
 * 
 * Shall be called by CH to log the getcall enqueue operation. This event occurs after call is enqueued.
 * This event is used for logging the intercomponent communication.
*/
                                    
void tliPrGetCallDetected_c (String am,
                             long int ts,
                             String src,
                             long int line,
                             TriComponentId c,
                             TriPortId atPort,
                             TriPortId fromPort,
                             TriSignatureId signature,
                             TciParameterListType parsValue)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "call_detected", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("from "); logPDestination(PBUF, fromPort);
    APPEND_STR(" at port "); logTriPortId(PBUF, atPort);
    APPEND_STR(" detected request to call "); logTriSignatureId(PBUF, signature);

    WRITE_BUFFER;
}

/* tliPrGetCallMismatch_m
 * 
 * Shall be called by TE to log the mismatch of a getcall. This event occurs after getcall is checked
 * against a template. This event is used for logging the communication with the SUT.
*/
                                    
void tliPrGetCallMismatch_m (String am,
                             long int ts,
                             String src,
                             long int line,
                             TriComponentId c,
                             TriPortId port,
                             TriSignatureId signature,
                             TciParameterListType parsValue,
                             TciValueTemplate parsTmpl,
                             TciValueDifferenceList diffs,
                             TriAddress address,
                             TciValueTemplate addressTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "sut_call_mismatch", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("at port "); logTriPortId(PBUF, port); APPEND_STR(" "); 
	logTciValueDifferenceList(PBUF, diffs);

    WRITE_BUFFER;
}

/* tliPrGetCallMismatch_c
 * 
 * Shall be called by TE to log the mismatch of a getcall. This event occurs after getcall is checked
 * against a template. This event is used for logging the intercomponent communication.
*/
                                    
void tliPrGetCallMismatch_c (String am,
                             long int ts,
                             String src,
                             long int line,
                             TriComponentId c,
                             TriPortId port,
                             TriSignatureId signature,
                             TciParameterListType parsValue,
                             TciValueTemplate parsTmpl,
                             TciValueDifferenceList diffs,
                             TriComponentId from,
                             TciNonValueTemplate fromTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "call_mismatch", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("from "); logCDestination(PBUF, from);
    APPEND_STR(" at port "); logTriPortId(PBUF, port); APPEND_STR(" ");
	logTciValueDifferenceList(PBUF, diffs);

    WRITE_BUFFER;
}

/* tliPrGetCall_m
 * 
 * Shall be called by TE to log getting a call. This event occurs after getcall has matched against a
 * template. This event is used for logging the communication with the SUT.
*/
                                    
void tliPrGetCall_m (String am,
                     long int ts,
                     String src,
                     long int line,
                     TriComponentId c,
                     TriPortId port,
                     TriSignatureId signature,
                     TciParameterListType parsValue,
                     TciValueTemplate parsTmpl,
                     TriAddress address,
                     TciValueTemplate addressTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "sut_call_received", am, ts, src, line, c);
    /* log parameters */
    logTriPortId(PBUF, port);
    APPEND_STR(".getcall("); 
	logTriSignatureId(PBUF, signature); APPEND_STR(":"); logTciParameterListType(PBUF, parsValue, "{", "}");
	APPEND_STR(") received");

    WRITE_BUFFER;
}

/* tliPrGetCall_c
 * 
 * Shall be called by TE to log getting a call. This event occurs after getcall has matched against a
 * template. This event is used for logging the intercomponent communication.
*/
                            
void tliPrGetCall_c (String am,
                     long int ts,
                     String src,
                     long int line,
                     TriComponentId c,
                     TriPortId port,
                     TriSignatureId signature,
                     TciParameterListType parsValue,
                     TciValueTemplate parsTmpl,
                     TriComponentId from,
                     TciNonValueTemplate fromTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "call_received", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("from "); logCDestination(PBUF, from); APPEND_STR(" ");
    logTriPortId(PBUF, port);
    APPEND_STR(".getcall("); 
	logTriSignatureId(PBUF, signature); APPEND_STR(":"); logTciParameterListType(PBUF, parsValue, "{", "}");
	APPEND_STR(") received");

    WRITE_BUFFER;
}

/* tliPrReply_m
 * 
 * Shall be called by SA to log a unicast reply operation. This event occurs after reply execution. This
 * event is used for logging the communication with the SUT.
*/

void tliPrReply_m (String am,
                   long int ts,
                   String src,
                   long int line,
                   TriComponentId c,
                   TriPortId atPort,
                   TriPortId toPort,
                   TriSignatureId signature,
                   TciParameterListType parsValue,
                   TciValue replValue,
                   TriAddress address,
                   TriStatus encoderStatus,
                   TriParameterList pars,
                   TriParameter repl,
                   TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(encoderStatus) && STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "sut_reply_sent";
	else
		eventName = "sut_reply_failed";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

	logTriPortId(PBUF, atPort); 
	APPEND_STR(".reply("); 
	logTriSignatureId(PBUF, signature); APPEND_STR(":"); logTciParameterListType(PBUF, parsValue, "{", "}");
	APPEND_STR(")");
	if (replValue) 
	{
		APPEND_STR(" return "); logTciValue(PBUF, replValue);
	}

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(encoderStatus, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrReply_m_BC
 * 
 * Shall be called by SA to log a broadcast reply operation. This event occurs after reply execution. This
 * event is used for logging the communication with the SUT.
*/

void tliPrReply_m_BC (String am,
                      long int ts,
                      String src,
                      long int line,
                      TriComponentId c,
                      TriPortId atPort,
                      TriPortId toPort,
                      TriSignatureId signature,
                      TciParameterListType parsValue,
                      TciValue replValue,
                      TriStatus encoderStatus,
                      TriParameterList pars,
                      TriParameter repl,
                      TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(encoderStatus) && STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "sut_reply_sent_bc";
	else
		eventName = "sut_reply_failed_bc";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

	logTriPortId(PBUF, atPort); 
	APPEND_STR(".reply("); 
	logTriSignatureId(PBUF, signature); APPEND_STR(":"); logTciParameterListType(PBUF, parsValue, "{", "}");
	APPEND_STR(")");
	if (replValue) 
	{
		APPEND_STR(" return "); logTciValue(PBUF, replValue);
	}

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(encoderStatus, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrReply_m_MC
 * 
 * Shall be called by SA to log a multicast reply operation. This event occurs after reply execution. This
 * event is used for logging the communication with the SUT.
*/

void tliPrReply_m_MC (String am,
                      long int ts,
                      String src,
                      long int line,
                      TriComponentId c,
                      TriPortId atPort,
                      TriPortId toPort,
                      TriSignatureId signature,
                      TciParameterListType parsValue,
                      TciValue replValue,
                      TriAddressList addresses,
                      TriStatus encoderStatus,
                      TriParameterList pars,
                      TriParameter repl,
                      TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(encoderStatus) && STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "sut_reply_sent_mc";
	else
		eventName = "sut_reply_failed_mc";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

	logTriPortId(PBUF, atPort); 
	APPEND_STR(".reply("); 
	logTriSignatureId(PBUF, signature); APPEND_STR(":"); logTciParameterListType(PBUF, parsValue, "{", "}");
	APPEND_STR(")");
	if (replValue) 
	{
		APPEND_STR(" return "); logTciValue(PBUF, replValue);
	}

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(encoderStatus, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrReply_c
 * 
 * Shall be called by CH to log a unicast reply operation. This event occurs after reply execution. This
 * event is used for logging the intercomponent communication.
*/
 
void tliPrReply_c (String am,
                   long int ts,
                   String src,
                   long int line,
                   TriComponentId c,
                   TriPortId atPort,
                   TriPortId toPort,
                   TriSignatureId signature,
                   TciParameterListType parsValue,
                   TciValue replValue,
                   TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "reply_sent";
	else
		eventName = "reply_failed";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

    APPEND_STR("to "); logPDestination(PBUF, toPort); APPEND_STR(" ");
	logTriPortId(PBUF, atPort); 
	APPEND_STR(".reply("); 
	logTriSignatureId(PBUF, signature); APPEND_STR(":"); logTciParameterListType(PBUF, parsValue, "{", "}");
	APPEND_STR(")");
	if (replValue) 
	{
		APPEND_STR(" return "); logTciValue(PBUF, replValue);
	}

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(TRI_OK, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrReply_c_BC
 * 
 * Shall be called by CH to log a broadcast reply operation. This event occurs after reply execution. This
 * event is used for logging the intercomponent communication.
*/

void tliPrReply_c_BC (String am,
                      long int ts,
                      String src,
                      long int line,
                      TriComponentId c,
                      TriPortId atPort,
                      TriPortIdList toPortList,
                      TriSignatureId signature,
                      TciParameterListType parsValue,
                      TciValue replValue,
                      TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "reply_sent_bc";
	else
		eventName = "reply_failed_bc";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

    APPEND_STR("to "); logPDestinationList(PBUF, toPortList); APPEND_STR(" ");
	logTriPortId(PBUF, atPort); 
	APPEND_STR(".reply("); 
	logTriSignatureId(PBUF, signature); APPEND_STR(":"); logTciParameterListType(PBUF, parsValue, "{", "}");
	APPEND_STR(")");
	if (replValue) 
	{
		APPEND_STR(" return "); logTciValue(PBUF, replValue);
	}

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(TRI_OK, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrReply_c_MC
 * 
 * Shall be called by CH to log a multicast reply operation. This event occurs after reply execution. This
 * event is used for logging the intercomponent communication.
*/

void tliPrReply_c_MC (String am,
                      long int ts,
                      String src,
                      long int line,
                      TriComponentId c,
                      TriPortId atPort,
                      TriPortIdList toPortList,
                      TriSignatureId signature,
                      TciParameterListType parsValue,
                      TciValue replValue,
                      TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "reply_sent_mc";
	else
		eventName = "reply_failed_mc";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

    APPEND_STR("to "); logPDestinationList(PBUF, toPortList); APPEND_STR(" ");
	logTriPortId(PBUF, atPort); 
	APPEND_STR(".reply("); 
	logTriSignatureId(PBUF, signature); APPEND_STR(":"); logTciParameterListType(PBUF, parsValue, "{", "}");
	APPEND_STR(")");
	if (replValue) 
	{
		APPEND_STR(" return "); logTciValue(PBUF, replValue);
	}

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(TRI_OK, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrGetReplyDetected_m
 * 
 * Shall be called by SA to log the getreply enqueue operation. This event occurs after getreply is
 * enqueued. This event is used for logging the communication with the SUT.
*/
                             
void tliPrGetReplyDetected_m (String am,
                              long int ts,
                              String src,
                              long int line,
                              TriComponentId c,
                              TriPortId atPort,
                              TriPortId fromPort,
                              TriSignatureId signature,
                              TriParameterList pars,
                              TriParameter repl,
                              TriAddress address)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "sut_reply_detected", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("at port "); logTriPortId(PBUF, atPort);
    APPEND_STR(" detected reply to "); logTriSignatureId(PBUF, signature);

    WRITE_BUFFER;
}

/* tliPrGetReplyDetected_c
 * 
 * Shall be called by CH to log the getreply enqueue operation. This event occurs after getreply is
 * enqueued. This event is used for logging the intercomponent communication.
*/
                                     
void tliPrGetReplyDetected_c (String am,
                              long int ts,
                              String src,
                              long int line,
                              TriComponentId c,
                              TriPortId atPort,
                              TriPortId fromPort,
                              TriSignatureId signature,
                              TciParameterListType parsValue,
                              TciValue replValue)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "sut_reply_detected", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("from "); logPDestination(PBUF, fromPort);
    APPEND_STR(" at port "); logTriPortId(PBUF, atPort);
    APPEND_STR(" detected reply to "); logTriSignatureId(PBUF, signature);

    WRITE_BUFFER;
}

/* tliPrGetReplyMismatch_m
 * 
 * Shall be called by TE to log the mismatch of a getreply operation. This event occurs after getreply is
 * checked against a template. This event is used for logging the communication with SUT.
*/
                                     
void tliPrGetReplyMismatch_m (String am,
                              long int ts,
                              String src,
                              long int line,
                              TriComponentId c,
                              TriPortId port,
                              TriSignatureId signature,
                              TciParameterListType parsValue,
                              TciValueTemplate parsTmpl,
                              TciValue replValue,
                              TciValueTemplate replyTmpl,
                              TciValueDifferenceList diffs,
                              TriAddress address,
                              TciValueTemplate addressTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "sut_reply_mismatch", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("at port "); logTriPortId(PBUF, port); APPEND_STR(" ");
	logTciValueDifferenceList(PBUF, diffs);

    WRITE_BUFFER;
}

/* tliPrGetReplyMismatch_c
 * 
 * Shall be called by TE to log the mismatch of a getreply operation. This event occurs after getreply is
 * checked against a template. This event is used for logging the intercomponent communication.
*/

void tliPrGetReplyMismatch_c (String am,
                              long int ts,
                              String src,
                              long int line,
                              TriComponentId c,
                              TriPortId port,
                              TriSignatureId signature,
                              TciParameterListType parsValue,
                              TciValueTemplate parsTmpl,
                              TciValue replValue,
                              TciValueTemplate replyTmpl,
                              TciValueDifferenceList diffs,
                              TriComponentId from,
                              TciNonValueTemplate fromTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "sut_reply_mismatch", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("from "); logCDestination(PBUF, from);
    APPEND_STR(" at port "); logTriPortId(PBUF, port); APPEND_STR(" ");
	logTciValueDifferenceList(PBUF, diffs);

    WRITE_BUFFER;
}

/* tliPrGetReply_m
 * 
 * Shall be called by TE to log getting a reply. This event occurs after getreply is checked against a
 * template. This event is used for logging the communication with SUT.
*/
                                     
void tliPrGetReply_m (String am,
                      long int ts,
                      String src,
                      long int line,
                      TriComponentId c,
                      TriPortId port,
                      TriSignatureId signature,
                      TciParameterListType parsValue,
                      TciValueTemplate parsTmpl,
                      TciValue replValue,
                      TciValueTemplate replyTmpl,
                      TriAddress address,
                      TciValueTemplate addressTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "sut_reply_received", am, ts, src, line, c);
    /* log parameters */
    logTriPortId(PBUF, port);
    APPEND_STR(".getreply("); 
	logTriSignatureId(PBUF, signature); APPEND_STR(":"); logTciParameterListType(PBUF, parsValue, "{", "}");
	APPEND_STR(")");
	if (replValue)
	{
		APPEND_STR(" return "); logTciValue(PBUF, replValue);
	}
	APPEND_STR(" received");

    WRITE_BUFFER;
}

/* tliPrGetReply_c
 * 
 * Shall be called by TE to log getting a reply. This event occurs after getreply is checked against a
 * template. This event is used for logging the intercomponent communication.
*/
                             
void tliPrGetReply_c (String am,
                      long int ts,
                      String src,
                      long int line,
                      TriComponentId c,
                      TriPortId port,
                      TriSignatureId signature,
                      TciParameterListType parsValue,
                      TciValueTemplate parsTmpl,
                      TciValue replValue,
                      TciValueTemplate replyTmpl,
                      TriComponentId from,
                      TciNonValueTemplate fromTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "sut_reply_received", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("from "); logCDestination(PBUF, from); APPEND_STR(" ");
    logTriPortId(PBUF, port);
    APPEND_STR(".getreply("); 
	logTriSignatureId(PBUF, signature); APPEND_STR(":"); logTciParameterListType(PBUF, parsValue, "{", "}");
	APPEND_STR(")");
	if (replValue)
	{
		APPEND_STR(" return "); logTciValue(PBUF, replValue);
	}
	APPEND_STR(" received");

    WRITE_BUFFER;
}

/* tliPrRaise_m
 * 
 * Shall be called by SA to log a unicast raise operation. This event occurs after reply execution. This
 * event is used for logging the communication with the SUT.
*/
                             
void tliPrRaise_m (String am,
                   long int ts,
                   String src,
                   long int line,
                   TriComponentId c,
                   TriPortId atPort,
                   TriPortId toPort,
                   TriSignatureId signature,
                   TciParameterListType parsValue,
                   TciValue excValue,
                   TriAddress address,
                   TriStatus encoderStatus,
                   TriException exc,
                   TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(encoderStatus) && STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "sut_exception_raised";
	else
		eventName = "sut_raise_failed";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

	logTriPortId(PBUF, atPort);
    APPEND_STR(".raise("); logTriSignatureId(PBUF, signature); 
	APPEND_STR(", "); logTciValue(PBUF, excValue); APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(encoderStatus, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrRaise_m_BC
 * 
 * Shall be called by SA to log a broadcast raise operation. This event occurs after reply execution. This
 * event is used for logging the communication with the SUT.
*/

void tliPrRaise_m_BC (String am,
                      long int ts,
                      String src,
                      long int line,
                      TriComponentId c,
                      TriPortId atPort,
                      TriPortId toPort,
                      TriSignatureId signature,
                      TciParameterListType parsValue,
                      TciValue excValue,
                      TriStatus encoderStatus,
                      TriException exc,
                      TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(encoderStatus) && STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "sut_exception_raised_bc";
	else
		eventName = "sut_raise_failed_bc";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

	logTriPortId(PBUF, atPort);
    APPEND_STR(".raise("); logTriSignatureId(PBUF, signature); 
	APPEND_STR(", "); logTciValue(PBUF, excValue); APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(encoderStatus, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrRaise_m_MC
 * 
 * Shall be called by SA to log a multicast raise operation. This event occurs after reply execution. This
 * event is used for logging the communication with the SUT.
*/

void tliPrRaise_m_MC (String am,
                      long int ts,
                      String src,
                      long int line,
                      TriComponentId c,
                      TriPortId atPort,
                      TriPortId toPort,
                      TriSignatureId signature,
                      TciParameterListType parsValue,
                      TciValue excValue,
                      TriAddressList addresses,
                      TriStatus encoderStatus,
                      TriException exc,
                      TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(encoderStatus) && STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "sut_exception_raised_mc";
	else
		eventName = "sut_raise_failed_mc";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */

	logTriPortId(PBUF, atPort);
    APPEND_STR(".raise("); logTriSignatureId(PBUF, signature); 
	APPEND_STR(", "); logTciValue(PBUF, excValue); APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(encoderStatus, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrRaise_c
 * 
 * Shall be called by CH to log a unicast raise operation. This event occurs after reply execution. This
 * event is used for logging the intercomponent communication.
*/
                             
void tliPrRaise_c (String am,
                   long int ts,
                   String src,
                   long int line,
                   TriComponentId c,
                   TriPortId atPort,
                   TriPortId toPort,
                   TriSignatureId signature,
                   TciParameterListType parsValue,
                   TciValue excValue,
                   TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "exception_raised";
	else
		eventName = "raise_failed";

    logTliEventInformation(&buf, eventName, am, ts, src, line, c);
    /* log parameters */

    APPEND_STR("to "); logPDestination(PBUF, toPort); APPEND_STR(" ");
	logTriPortId(PBUF, atPort);
    APPEND_STR(".raise("); logTriSignatureId(PBUF, signature); 
	APPEND_STR(", "); logTciValue(PBUF, excValue); APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(TRI_OK, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrRaise_c_BC
 * 
 * Shall be called by CH to log a broadcast raise operation. This event occurs after reply execution. This
 * event is used for logging the intercomponent communication.
*/

void tliPrRaise_c_BC (String am,
                      long int ts,
                      String src,
                      long int line,
                      TriComponentId c,
                      TriPortId atPort,
                      TriPortIdList toPortList,
                      TriSignatureId signature,
                      TciParameterListType parsValue,
                      TciValue excValue,
                      TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "exception_raised_bc";
	else
		eventName = "raise_failed_bc";

    logTliEventInformation(&buf, eventName, am, ts, src, line, c);
    /* log parameters */

    APPEND_STR("to "); logPDestinationList(PBUF, toPortList); APPEND_STR(" ");
	logTriPortId(PBUF, atPort);
    APPEND_STR(".raise("); logTriSignatureId(PBUF, signature); 
	APPEND_STR(", "); logTciValue(PBUF, excValue); APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(TRI_OK, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrRaise_c_MC
 * 
 * Shall be called by CH to log a multicast raise operation. This event occurs after reply execution. This
 * event is used for logging the intercomponent communication.
*/
                             
void tliPrRaise_c_MC (String am,
                      long int ts,
                      String src,
                      long int line,
                      TriComponentId c,
                      TriPortId atPort,
                      TriPortIdList toPortList,
                      TriSignatureId signature,
                      TciParameterListType parsValue,
                      TciValue excValue,
                      TriStatus transmissionStatus)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(transmissionStatus);
	char *eventName;

	if (status)
		eventName = "exception_raised_mc";
	else
		eventName = "raise_failed_mc";

    logTliEventInformation(&buf, eventName, am, ts, src, line, c);
    /* log parameters */

    APPEND_STR("to "); logPDestinationList(PBUF, toPortList); APPEND_STR(" ");
	logTriPortId(PBUF, atPort);
    APPEND_STR(".raise("); logTriSignatureId(PBUF, signature); 
	APPEND_STR(", "); logTciValue(PBUF, excValue); APPEND_STR(")");

	if (!status)
	{
		APPEND_STR2(" failed due to ", GET_REASON(TRI_OK, transmissionStatus));
	}

    WRITE_BUFFER;
}

/* tliPrCatchDetected_m
 * 
 * Shall be called by SA to log the catch enqueue operation. This event occurs after catch is
 * enqueued. This event is used for logging the communication with the SUT.
*/
                             
void tliPrCatchDetected_m (String am,
                           long int ts,
                           String src,
                           long int line,
                           TriComponentId c,
                           TriPortId atPort,
                           TriPortId fromPort,
                           TriSignatureId signature,
                           TriException exc,
                           TriAddress address)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "sut_exception_detected", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("at port "); logTriPortId(PBUF, atPort); 
	APPEND_STR(" detected exception in "); logTriSignatureId(PBUF, signature);

	WRITE_BUFFER;
}

/* tliPrCatchDetected_c
 * 
 * Shall be called by CH to log the catch enqueue operation. This event occurs after catch is enqueued.
 * This event is used for logging the intercomponent communication.
*/
                                  
void tliPrCatchDetected_c (String am,
                           long int ts,
                           String src,
                           long int line,
                           TriComponentId c,
                           TriPortId atPort,
                           TriPortId fromPort,
                           TriSignatureId signature,
                           TciValue excValue)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "exception_detected", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("from "); logPDestination(PBUF, fromPort); 
    APPEND_STR(" at port "); logTriPortId(PBUF, atPort); 
	APPEND_STR(" detected exception in "); logTriSignatureId(PBUF, signature);

	WRITE_BUFFER;
}

/* tliPrCatchMismatch_m
 * 
 * Shall be called by TE to log the mismatch of a catch operation. This event occurs after catch is
 * checked against a template. This event is used for logging the communication with SUT.
*/
                                  
void tliPrCatchMismatch_m (String am,
                           long int ts,
                           String src,
                           long int line,
                           TriComponentId c,
                           TriPortId port,
                           TriSignatureId signature,
                           TciValue excValue,
                           TciValueTemplate excTmpl,
                           TciValueDifferenceList diffs,
                           TriAddress address,
                           TciValueTemplate addressTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "sut_exception_mismatch", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("at port "); logTriPortId(PBUF, port); APPEND_STR(" ");
	logTciValueDifferenceList(PBUF, diffs);

	WRITE_BUFFER;
}

/* tliPrCatchMismatch_c
 * 
 * Shall be called by TE to log the mismatch of a catch operation. This event occurs after catch is
 * checked against a template. This event is used for logging the intercomponent communication.
*/
                                  
void tliPrCatchMismatch_c (String am,
                           long int ts,
                           String src,
                           long int line,
                           TriComponentId c,
                           TriPortId port,
                           TriSignatureId signature,
                           TciValue excValue,
                           TciValueTemplate excTmpl,
                           TciValueDifferenceList diffs,
                           TriComponentId from,
                           TciNonValueTemplate fromTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "sut_exception_mismatch", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("from "); logCDestination(PBUF, from);
    APPEND_STR(" at port "); logTriPortId(PBUF, port); APPEND_STR(" ");
	logTciValueDifferenceList(PBUF, diffs);

	WRITE_BUFFER;
}

/* tliPrCatch_m
 * 
 * Shall be called by SA to log catching an exception. This event occurs after catch is checked against
 * a template. This event is used for logging the communication with SUT.
*/
                                  
void tliPrCatch_m (String am,
                   long int ts,
                   String src,
                   long int line,
                   TriComponentId c,
                   TriPortId port,
                   TriSignatureId signature,
                   TciValue excValue,
                   TciValueTemplate excTmpl,
                   TriAddress address,
                   TciValueTemplate addressTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "sut_exception_caught", am, ts, src, line, c);
    /* log parameters */
    logTriPortId(PBUF, port);
    APPEND_STR(".catch("); logTriSignatureId(PBUF, signature); 
	APPEND_STR(", "); logTciValue(PBUF, excValue); APPEND_STR(") caught");

    WRITE_BUFFER;
}

/* tliPrCatch_c
 * 
 * Shall be called by CH to log catching an exception. This event occurs after catch is checked against
 * a template. This event is used for logging the intercomponent communication.
*/
                          
void tliPrCatch_c (String am,
                   long int ts,
                   String src,
                   long int line,
                   TriComponentId c,
                   TriPortId port,
                   TriSignatureId signature,
                   TciValue excValue,
                   TciValueTemplate excTmpl,
                   TriComponentId from,
                   TciNonValueTemplate fromTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "exception_caught", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("from "); logCDestination(PBUF, from); APPEND_STR(" ");
    logTriPortId(PBUF, port);
    APPEND_STR(".catch("); logTriSignatureId(PBUF, signature); 
	APPEND_STR(", "); logTciValue(PBUF, excValue); APPEND_STR(") caught");

    WRITE_BUFFER;
}

/* tliPrCatchTimeoutDetected
 * 
 * Shall be called by PA to log the detection of a catch timeout. This event occurs after the timeout is
 * enqueued.
*/
                          
void tliPrCatchTimeoutDetected (String am,
                                long int ts,
                                String src,
                                long int line,
                                TriComponentId c,
                                TriPortId port,
                                TriSignatureId signature)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "timeout_exception_detected", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("at port "); logTriPortId(PBUF, port);
    APPEND_STR(" detected timeout exception");

	WRITE_BUFFER;
}

/* tliPrCatchTimeout
 * 
 * Shall be called by TE to log catching a timeout. This event occurs after the catch timeout has been
 * performed.
*/
                                       
void tliPrCatchTimeout (String am,
                        long int ts,
                        String src,
                        long int line,
                        TriComponentId c,
                        TriPortId port,
                        TriSignatureId signature)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "timeout_exception_caught", am, ts, src, line, c);
    /* log parameters */
    logTriPortId(PBUF, port);
    APPEND_STR(".catch(timeout) caught");

	WRITE_BUFFER;
}

/* tliCCreate
 * 
 * Shall be called by TE to log the create component operation. This event occurs after component
 * creation.
*/
                               
void tliCCreate (String am,
                 long int ts,
                 String src,
                 long int line,
                 TriComponentId c,
                 TriComponentId comp,
                 String name,
                 unsigned char isAlive)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "component_created", am, ts, src, line, c);
    /* log parameters */
    logTriComponentId(PBUF, comp); 
    APPEND_STR(".create(\""); 
    APPEND_STR(name);
    APPEND_STR("\")");
    if (isAlive) APPEND_STR(" alive");

    WRITE_BUFFER;
}

/* tliCStart
 * 
 * Shall be called by TE to log the start component operation. This event occurs after component start.
*/
                        
void tliCStart (String am,
                long int ts,
                String src,
                long int line,
                TriComponentId c,
                TriComponentId comp,
                TciBehaviourIdType name,
                TciParameterListType parsValue)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "component_started", am, ts, src, line, c);
    /* log parameters */
    logTriComponentId(PBUF, comp);
	APPEND_STR(".start("); logTciBehaviourIdType(PBUF, name); logTciParameterListType(PBUF, parsValue, "(", ")"); APPEND_STR(")");

    WRITE_BUFFER;
}

/* tliCRunning
 * 
 * Shall be called by TE to log the running component operation. This event occurs after component
 * running.
*/

void tliCRunning (String am,
                  long int ts,
                  String src,
                  long int line,
                  TriComponentId c,
                  TriComponentId comp,
                  ComponentStatus status)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "component_is_running", am, ts, src, line, c);
    /* log parameters */
    logTriComponentId(PBUF, comp); 
	APPEND_STR(".status == ");
	logComponentStatus(PBUF, status);

    WRITE_BUFFER;
}

/* tliCAlive
 * 
 * Shall be called by TE to log the alive component operation. This event occurs after component
 * running.
*/

void tliCAlive (String am,
                long int ts,
                String src,
                long int line,
                TriComponentId c,
                TriComponentId comp,
                ComponentStatus status)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "component_is_alive", am, ts, src, line, c);
    /* log parameters */
    logTriComponentId(PBUF, comp); 
	APPEND_STR(".status == ");
	logComponentStatus(PBUF, status);

    WRITE_BUFFER;
}

/* tliCStop
 * 
 * Shall be called by TE to log the stop component operation. This event occurs after component stop.
*/
                       
void tliCStop (String am,
               long int ts,
               String src,
               long int line,
               TriComponentId c,
               TriComponentId comp)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "component_stopped", am, ts, src, line, c);
    /* log parameters */

    WRITE_BUFFER;
}

/* tliCKill
 * 
 * Shall be called by TE to log the kill component operation. This event occurs after component stop.
*/

void tliCKill (String am,
               long int ts,
               String src,
               long int line,
               TriComponentId c,
               TriComponentId comp)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "component_killed", am, ts, src, line, c);
    /* log parameters */

    WRITE_BUFFER;
}

/* tliCDoneMismatch
 * 
 * Shall be called by TE to log the mismatch of a done component operation. This event occurs after
 * done is checked against a template.
*/
                      
void tliCDoneMismatch (String am,
                       long int ts,
                       String src,
                       long int line,
                       TriComponentId c,
                       TriComponentId comp,
                       TciNonValueTemplate compTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "component_done_mismatch", am, ts, src, line, c);
    /* log parameters */
	if (tciIsAnyNonValueTemplate(compTmpl))
		APPEND_STR("any component");
	else if (tciIsAllNonValueTemplate(compTmpl))
		APPEND_STR("all component");
	else
		logTciNonValueTemplate(PBUF, compTmpl);

	APPEND_STR(".done mismatched");

    WRITE_BUFFER;
}

/* tliCDone
 * 
 * Shall be called by TE to log the done component operation. This event occurs after the done
 * operation.
*/
                              
void tliCDone (String am,
               long int ts,
               String src,
               long int line,
               TriComponentId c,
               TriComponentId comp,
               TciNonValueTemplate compTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "component_done", am, ts, src, line, c);
    /* log parameters */
	if (tciIsAnyNonValueTemplate(compTmpl))
		APPEND_STR("any component");
	else if (tciIsAllNonValueTemplate(compTmpl))
		APPEND_STR("all component");
	else
		logTciNonValueTemplate(PBUF, compTmpl);

	APPEND_STR(".done received");

    WRITE_BUFFER;
}

/* tliCKilledMismatch
 * 
 * Shall be called by TE to log the mismatch of a killed component operation. This event occurs after
 * done is checked against a template.
*/

void tliCKilledMismatch (String am,
                         long int ts,
                         String src,
                         long int line,
                         TriComponentId c,
                         TriComponentId comp,
                         TciNonValueTemplate compTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "component_killed_mismatch", am, ts, src, line, c);
    /* log parameters */
	if (tciIsAnyNonValueTemplate(compTmpl))
		APPEND_STR("any component");
	else if (tciIsAllNonValueTemplate(compTmpl))
		APPEND_STR("all component");
	else
		logTciNonValueTemplate(PBUF, compTmpl);

	APPEND_STR(".killed mismatched");

    WRITE_BUFFER;
}

/* tliCKilled
 * 
 * Shall be called by TE to log the killed component operation. This event occurs after
 * done is checked against a template.
*/

void tliCKilled (String am,
                 long int ts,
                 String src,
                 long int line,
                 TriComponentId c,
                 TriComponentId comp,
                 TciNonValueTemplate compTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "component_killed", am, ts, src, line, c);
    /* log parameters */
	if (tciIsAnyNonValueTemplate(compTmpl))
		APPEND_STR("any component");
	else if (tciIsAllNonValueTemplate(compTmpl))
		APPEND_STR("all component");
	else
		logTciNonValueTemplate(PBUF, compTmpl);

	APPEND_STR(".killed received");

    WRITE_BUFFER;
}

/* tliCTerminated
 * 
 * Shall be called by TE to log the termination of a component. This event occurs after the termination
 * of the component.
*/
                        
void tliCTerminated (String am,
                     long int ts,
                     String src,
                     long int line,
                     TriComponentId c,
                     TciValue verdict)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "component_terminated", am, ts, src, line, c);
    /* log parameters */
	APPEND_STR("=> "); 
    logTciVerdictValue(PBUF, verdict);

    WRITE_BUFFER;
}

/* tliPConnect
 * 
 * Shall be called by CH to log the connect operation. This event occurs after the connect operation.
*/
                            
void tliPConnect (String am,
                  long int ts,
                  String src,
                  long int line,
                  TriComponentId c,
                  TriPortId port1,
                  TriPortId port2)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "port_connected", am, ts, src, line, c);
    /* log parameters */
	APPEND_STR("connect(");
	logTriComponentId(PBUF, port1.compInst); APPEND_STR(":"); logTriPortId(PBUF, port1);
	APPEND_STR(", ");
	logTriComponentId(PBUF, port2.compInst); APPEND_STR(":"); logTriPortId(PBUF, port2);
	APPEND_STR(")");

    WRITE_BUFFER;
}

/* tliPDisconnect
 * 
 * Shall be called by CH to log the disconnect operation. This event occurs after the disconnect 
 * operation.
*/
                         
void tliPDisconnect (String am,
                     long int ts,
                     String src,
                     long int line,
                     TriComponentId c,
                     TriPortId port1,
                     TriPortId port2)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "port_disconnected", am, ts, src, line, c);
    /* log parameters */
	APPEND_STR("disconnect(");
	logTriComponentId(PBUF, port1.compInst); APPEND_STR(":"); logTriPortId(PBUF, port1);
	APPEND_STR(", ");
	logTriComponentId(PBUF, port2.compInst); APPEND_STR(":"); logTriPortId(PBUF, port2);
	APPEND_STR(")");

    WRITE_BUFFER;
}

/* tliPMap
 * 
 * Shall be called by SA to log the map operation. This event occurs after the map operation.
*/
                            
void tliPMap (String am,
              long int ts,
              String src,
              long int line,
              TriComponentId c,
              TriPortId port1,
              TriPortId port2)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "port_mapped", am, ts, src, line, c);
    /* log parameters */
	APPEND_STR("map(");
	logTriComponentId(PBUF, port1.compInst); APPEND_STR(":"); logTriPortId(PBUF, port1);
	APPEND_STR(", ");
	logTriComponentId(PBUF, port2.compInst); APPEND_STR(":"); logTriPortId(PBUF, port2);
	APPEND_STR(")");

    WRITE_BUFFER;
}

/* tliPUnmap
 * 
 * Shall be called by SA to log the unmap operation. This event occurs after the unmap operation.
*/
                     
void tliPUnmap (String am,
                long int ts,
                String src,
                long int line,
                TriComponentId c,
                TriPortId port1,
                TriPortId port2)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "port_unmapped", am, ts, src, line, c);
    /* log parameters */
	APPEND_STR("unmap(");
	logTriComponentId(PBUF, port1.compInst); APPEND_STR(":"); logTriPortId(PBUF, port1);
	APPEND_STR(", ");
	logTriComponentId(PBUF, port2.compInst); APPEND_STR(":"); logTriPortId(PBUF, port2);
	APPEND_STR(")");

    WRITE_BUFFER;
}

/* tliPClear
 * 
 * Shall be called by TE to log the port clear operation. This event occurs after the port clear
 * operation.
*/
                       
void tliPClear (String am,
                long int ts,
                String src,
                long int line,
                TriComponentId c,
                TriPortId port)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "port_cleared", am, ts, src, line, c);
    /* log parameters */
    logTriPortId(PBUF, port);
	APPEND_STR(".clear"); 

    WRITE_BUFFER;
}

/* tliPStart
 * 
 * Shall be called by TE to log the port start operation. This event occurs after the port start operation.
*/
                       
void tliPStart (String am,
                long int ts,
                String src,
                long int line,
                TriComponentId c,
                TriPortId port)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "port_started", am, ts, src, line, c);
    /* log parameters */
    logTriPortId(PBUF, port);
	APPEND_STR(".start"); 

    WRITE_BUFFER;
}

/* tliPStop
 * 
 * Shall be called by TE to log the port stop operation. This event occurs after the port stop operation.
*/
                       
void tliPStop (String am,
               long int ts,
               String src,
               long int line,
               TriComponentId c,
               TriPortId port)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "port_stopped", am, ts, src, line, c);
    /* log parameters */
    logTriPortId(PBUF, port);
	APPEND_STR(".stop"); 

    WRITE_BUFFER;
}

/* tliPHalt
 * 
 * Shall be called by TE to log the port halt operation. This event occurs after the port halt operation.
*/

void tliPHalt (String am,
               long int ts,
               String src,
               long int line,
               TriComponentId c,
               TriPortId port)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "port_halted", am, ts, src, line, c);
    /* log parameters */
    logTriPortId(PBUF, port);
	APPEND_STR(".halt"); 

    WRITE_BUFFER;
}


/* tliEncode
 * 
 * Shall be called by CD to log the encode operation.
*/
                      
void tliEncode (String am,
                long int ts,
                String src,
                long int line,
                TriComponentId c,
                TciValue val,
                TriStatus encoderStatus,
                TriMessage msg,
                String codec)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(encoderStatus);
	char *eventName;

	if (status)
		eventName = "value_encoded";
	else
		eventName = "encoder_failed";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */
    logTciValue(PBUF, val); 
	APPEND_STR(" => ");
    logTriMessage(PBUF, msg);

    WRITE_BUFFER;
}

/* tliDecode
 * 
 * Shall be called by CD to log the decode operation.
*/
                       
void tliDecode (String am,
                long int ts,
                String src,
                long int line,
                TriComponentId c,
                TriMessage msg,
                TriStatus decoderStatus,
                TciValue val,
                String codec)
{
    DECLARE_BUFFER;
	int status = STATUS_OK(decoderStatus);
	char *eventName;

	if (status)
		eventName = "data_decoded";
	else
		eventName = "decoder_failed";

    logTliEventInformation(PBUF, eventName, am, ts, src, line, c);
    /* log parameters */
	logTciValue(PBUF, val);
	APPEND_STR(" <= ");
    logTriMessage(PBUF, msg);

	WRITE_BUFFER;
}

/* tliTTimeoutDetected
 * 
 * Shall be called by PA to log the detection of a timeout. This event occurs after timeout is
 * enqueued.
*/
                       
void tliTTimeoutDetected (String am,
                          long int ts,
                          String src,
                          long int line,
                          TriComponentId c,
                          TriTimerId timer)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "timeout_detected", am, ts, src, line, c);
    /* log parameters */
    logTriTimerId(PBUF, timer);
	APPEND_STR(".timeout detected");

    WRITE_BUFFER;
}

/* tliTTimeoutMismatch
 * 
 * Shall be called by TE to log a timeout mismatch. This event occurs after a timeout match failed.
*/
                                 
void tliTTimeoutMismatch (String am,
                          long int ts,
                          String src,
                          long int line,
                          TriComponentId c,
                          TriTimerId timerId,
                          TciNonValueTemplate timerTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "timeout_mismatch", am, ts, src, line, c);
    /* log parameters */
	if (tciIsAnyNonValueTemplate(timerTmpl))
		APPEND_STR("any timer");
	else if (tciIsAllNonValueTemplate(timerTmpl))
		APPEND_STR("all timer");
	else
		logTciNonValueTemplate(PBUF, timerTmpl);

	APPEND_STR(".timeout mismatched");

    WRITE_BUFFER;
}

/* tliTTimeout
 * 
 * Shall be called by TE to log a timeout match. This event occurs after a timeout matched.
*/
                                 
void tliTTimeout (String am,
                  long int ts,
                  String src,
                  long int line,
                  TriComponentId c,
                  TriTimerId timerId,
                  TciNonValueTemplate timerTmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "timer_timed_out", am, ts, src, line, c);
    /* log parameters */
	if (tciIsAnyNonValueTemplate(timerTmpl))
		APPEND_STR("any timer");
	else if (tciIsAllNonValueTemplate(timerTmpl))
		APPEND_STR("all timer");
	else
		logTciNonValueTemplate(PBUF, timerTmpl);

	APPEND_STR(".timeout received");

    WRITE_BUFFER;
}

/* tliTStart
 * 
 * Shall be called by PA to log the start of a timer. This event occurs after the start timer
 * operation.
*/
                         
void tliTStart (String am,
                long int ts,
                String src,
                long int line,
                TriComponentId c,
                TriTimerId timer,
                TriTimerDuration dur)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "timer_started", am, ts, src, line, c);
    /* log parameters */
    logTriTimerId(PBUF, timer);
	APPEND_STR(".start("); logTriTimerDuration(PBUF, dur); APPEND_STR(")"); 

    WRITE_BUFFER;
}

/* tliTStop
 * 
 * Shall be called by PA to log the stop of a timer. This event occurs after the stop timer operation.
*/
                       
void tliTStop (String am,
               long int ts,
               String src,
               long int line,
               TriComponentId c,
               TriTimerId timer,
               TriTimerDuration dur)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "timer_stopped", am, ts, src, line, c);
    /* log parameters */
    logTriTimerId(PBUF, timer);
	APPEND_STR(".stop"); 

    WRITE_BUFFER;
}

/* tliTRead
 * 
 * Shall be called by PA to log the reading of a timer. This event occurs after the read timer
 * operation.
*/
                      
void tliTRead (String am,
               long int ts,
               String src,
               long int line,
               TriComponentId c,
               TriTimerId timer,
               TriTimerDuration elapsed)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "timer_read", am, ts, src, line, c);
    /* log parameters */
    logTriTimerId(PBUF, timer);
	APPEND_STR(".read == "); logTriTimerDuration(PBUF, elapsed);

    WRITE_BUFFER;
}

/* tliTRunning
 *
 * Shall be called by PA to log the running timer operation. This event occurs after the running timer
 * operation.
*/
                      
void tliTRunning (String am,
                  long int ts,
                  String src,
                  long int line,
                  TriComponentId c,
                  TriTimerId timer,
                  TimerStatus status)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "timer_is_running", am, ts, src, line, c);
    /* no parameters */
	logTriTimerId(PBUF, timer);
	APPEND_STR(".status == ");
	logTimerStatus(PBUF, status);

    WRITE_BUFFER;
}

/* tliSEnter
 *
 * Shall be called by TE to log the entering of a scope. This event occurs after the scoped has been
 * entered.
*/
                         
void tliSEnter (String am,
                long int ts,
                String src,
                long int line,
                TriComponentId c,
                QualifiedName name,
                TciParameterListType parsValue,
                String kind)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "scope_entered", am, ts, src, line, c);
    /* log parameters */
    logQualifiedName(PBUF, name); logTciParameterListType(PBUF, parsValue, "(", ")");

    WRITE_BUFFER;
}

/* tliSLeave
 *
 * Shall be called by TE to log the leaving of a scope. This event occurs after the scoped has been
 * leaved.
*/
                       
void tliSLeave (String am,
                long int ts,
                String src,
                long int line,
                TriComponentId c,
                QualifiedName name,
                TciParameterListType parsValue,
                TciValue returnValue,
                String kind)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "scope_left", am, ts, src, line, c);
    /* log parameters */
    logQualifiedName(PBUF, name); logTciParameterListType(PBUF, parsValue, "(", ")");
	if (returnValue)
	{
		APPEND_STR(" return "); logTciValue(PBUF, returnValue);
	}

    WRITE_BUFFER;
}

/* tliVar
 *
 * Shall be called by TE to log the modification of the value of a variable. This event occurs after the
 * values has been changed.
*/
                       
void tliVar (String am,
             long int ts,
             String src,
             long int line,
             TriComponentId c,
             QualifiedName name,
             TciValue varValue)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "variable_modified", am, ts, src, line, c);
    /* log parameters */
	logQualifiedName(PBUF, name); APPEND_STR(" := "); logTciValue(PBUF, varValue);

    WRITE_BUFFER;
}

/* tliModulePar
 *
 * Shall be called by TE to log the value of a module parameter. This event occurs after the
 * access to the value of a module parameter.
*/
                    
void tliModulePar (String am,
                   long int ts,
                   String src,
                   long int line,
                   TriComponentId c,
                   QualifiedName name,
                   TciValue parValue)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "module_parameter_set", am, ts, src, line, c);
    /* log parameters */
	logQualifiedName(PBUF, name); APPEND_STR(" := "); logTciValue(PBUF, parValue);

    WRITE_BUFFER;
}

/* tliGetVerdict
 *
 * Shall be called by TE to log the getverdict operation. This event occurs after the getverdict
 * operation.
*/
                          
void tliGetVerdict (String am,
                    long int ts,
                    String src,
                    long int line,
                    TriComponentId c,
                    TciValue verdict)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "verdict_queried", am, ts, src, line, c);
    /* log parameters */
    APPEND_STR("getverdict() == "); logTciVerdictValue(PBUF, verdict);

    WRITE_BUFFER;
}

/* tliSetVerdict
 *
 * Shall be called by TE to log the setverdict operation. This event occurs after the setverdict
 * operation.
*/
                           
void tliSetVerdict (String am,
                    long int ts,
                    String src,
                    long int line,
                    TriComponentId c,
                    TciValue verdict)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "verdict_set", am, ts, src, line, c);
    /* log parameters */
	APPEND_STR("setverdict("); logTciVerdictValue(PBUF, verdict); APPEND_STR(")");

    WRITE_BUFFER;
}

/* tliLog
 *
 * Shall be called by TM to log the TTCN-3 statement log. This event occurs after the TTCN-3 log
 * operation.
*/

void tliLog (String am,
             long int ts,
             String src,
             long int line,
             TriComponentId c,
             String log)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "values_logged", am, ts, src, line, c);
    /* log parameters */
	APPEND_STR("log"); logString(PBUF, log);

    WRITE_BUFFER;
}

/* tliAEnter
 *
 * Shall be called by TE to log entering an alt. This event occurs after an alt has been entered.
*/
                    
void tliAEnter (String am,
                long int ts,
                String src,
                long int line,
                TriComponentId c)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "alt_block_entered", am, ts, src, line, c);
    /* no parameters */

    WRITE_BUFFER;
}

/* tliALeave
 *
 * Shall be called by TE to log leaving an alt. This event occurs after the alt has been leaved.
*/
                      
void tliALeave (String am,
                long int ts,
                String src,
                long int line,
                TriComponentId c)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "alt_block_left", am, ts, src, line, c);
    /* no parameters */

    WRITE_BUFFER;
}

/* tliADefaults
 *
 * Shall be called by TE to log entering the default section. This event occurs after the default
 * section has been entered.
*/
                       
void tliADefaults (String am,
                   long int ts,
                   String src,
                   long int line,
                   TriComponentId c)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "defaults_entered", am, ts, src, line, c);
    /* no parameters */

    WRITE_BUFFER;
}

/* tliAActivate
 *
 * Shall be called by TE to log the activation of a default. This event occurs after the default activation.
*/
                          
void tliAActivate (String am,
                   long int ts,
                   String src,
                   long int line,
                   TriComponentId c,
                   QualifiedName name,
                   TciParameterListType pars,
                   TciValue ref)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "default_activated", am, ts, src, line, c);
    /* log parameters */

	if (ref)
	{
		logTciValue(PBUF, ref); APPEND_STR(" := "); 
	}
	APPEND_STR("activate(");
    logQualifiedName(PBUF, name); logTciParameterListType(PBUF, pars, "(", ")");
	APPEND_STR(")");

    WRITE_BUFFER;
}

/* tliADeactivate
 *
 * Shall be called by TE to log the deactivation of a default. This event occurs after the default
 * deactivation.
*/
                          
void tliADeactivate (String am,
                     long int ts,
                     String src,
                     long int line,
                     TriComponentId c,
                     TciValue ref)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "default_deactivated", am, ts, src, line, c);
    /* log parameters */
	APPEND_STR("deactivate");
	if (ref)
	{
		APPEND_STR("(");
		logTciValue(PBUF, ref);
		APPEND_STR(")");
	}

    WRITE_BUFFER;
}

/* tliANomatch
 *
 * Shall be called by TE to log the nomatch of an alt. This event occurs after the alt has not matched.
*/
                            
void tliANomatch (String am,
                  long int ts,
                  String src,
                  long int line,
                  TriComponentId c)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "alternative_mismatch", am, ts, src, line, c);
    /* no parameters */

    WRITE_BUFFER;
}
                         
/* tliARepeat
 *
 * Shall be called by TE to log repeating an alt. This event occurs when the alt is been repeated.
*/
                         
void tliARepeat (String am,
                 long int ts,
                 String src,
                 long int line,
                 TriComponentId c)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "alt_block_repeated", am, ts, src, line, c);
    /* no parameters */

    WRITE_BUFFER;
}

/* tliAWait
 *
 * Shall be called by TE to log that the component awaits events for a new snapshot.
*/
                        
void tliAWait (String am,
               long int ts,
               String src,
               long int line,
               TriComponentId c)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "alt_block_wait_new_events", am, ts, src, line, c);
    /* no parameters */

    WRITE_BUFFER;
}

/* tliAction
 *
 * Shall be called by TE to log that the component executed an SUT action.
*/

void tliAction (String am,
                long int ts,
                String src,
                long int line,
                TriComponentId c,
                String action)
{
    DECLARE_BUFFER;

    logTliEventInformation(&buf, "sut_action", am, ts, src, line, c);
    /* log parameters */
    logString(&buf, action);

    WRITE_BUFFER;
}

/* tliMatch
 *
 * Shall be called by TE to log that the component successfully executed a match operation.
*/
                        
void tliMatch(String am,
              long int ts,
              String src,
              long int line,
              TriComponentId c,
              TciValue expr,
              TciValueTemplate tmpl)
{
    DECLARE_BUFFER;

    logTliEventInformation(&buf, "match_succeeded", am, ts, src, line, c);
    /* log parameters */
    logTciValue(&buf, expr);
    logString(&buf, " against ");
    logTciValueTemplate(&buf, tmpl);

    WRITE_BUFFER;
}

/* tliMatchMismatch
 *
 * Shall be called by TE to log that the component unsuccessfully executed a match operation - 
 * a mismatch occurred.
*/
                        
void tliMatchMismatch(String am,
                      long int ts,
                      String src,
                      long int line,
                      TriComponentId c,
                      TciValue expr,
                      TciValueTemplate tmpl,
                      TciValueDifferenceList diffs)
{
    DECLARE_BUFFER;

    logTliEventInformation(&buf, "match_failed", am, ts, src, line, c);
    /* log parameters */
	logTciValueDifferenceList(PBUF, diffs);

    WRITE_BUFFER;
}

/* tliInfo
 *
 * Can be called by TE to log additional information during test execution. 
 * This generation of this event is tool dependent as well as the usage of the parameters level and info.
*/
                        
void tliInfo(String am,
             long int ts,
             String src,
             long int line,
             TriComponentId c,
             long int level,
             String info)
{
    DECLARE_BUFFER;

    logTliEventInformation(PBUF, "information", am, ts, src, line, c);
    /* log parameters */
	logString(PBUF, info);

    WRITE_BUFFER;
}

/***********************************************************************************
 * TciValue to String routines
 ***********************************************************************************/

static void make_indent(log_print_buffer_t *buf, int indent)
{
  int i;
  
  bufAppendString(buf, "\n");
  for (i = 0; i < indent; i++)
  {
    bufAppendString(buf, "  ");
  }
}

/* forward declaration */
static char* wchar2string (TciUCValue wchar, char *buf);

static void to_string(TciValue value, log_print_buffer_t *buf, int indent)
{
  TciTypeClassType kind;

  if (!value)
  {
    bufAppendString(buf, "<NULL>");
    return;
  }
  
  if (tciNotPresent(value))
  {
    bufAppendString(buf, "-");
    return;
  }
  
  kind = tciGetTypeClass(tciGetType(value));
    
  switch(kind) 
  { 
    case TCI_ADDRESS_TYPE:
      {
        bufAppendString(buf, "<SUT address>");
        
        /*
          You may use tciGetAddressValue function only if builtin 'address' type is redefined in test suite.
          Otherwise you will receive runtime error here as long as binary strings have no representation in TCI.
          
          TciValue address_value = tciGetAddressValue(value);
          to_string(address_value, buf, indent);
        */
      }
      break;
      
    case TCI_BITSTRING_TYPE:
      {
        String bstring = tciGetBStringValue(value);
        bufAppendString(buf, bstring);
      }
      break;
      
    case TCI_BOOLEAN_TYPE:
      {
        unsigned char bool = tciGetBooleanValue(value);
        bufAppendString(buf, bool ? "true" : "false");
      }
      break;
    
    case TCI_CHAR_TYPE:
      {
        char c = tciGetCharValue(value);
        char str[20];
        sprintf(str, "\"%c\"", c);
        bufAppendString(buf, str);
      }
      break;
      
    case TCI_CHARSTRING_TYPE:
      {
        TciCharStringValue cstr_value = tciGetCStringValue(value);
        char *str = (char*)malloc(cstr_value->length + 1);
        bufAppendString(buf, "\"");
        if (str)
        {
          memcpy(str, cstr_value->string, cstr_value->length);
          str[cstr_value->length] = '\0';
          bufAppendString(buf, str);
          free(str);
        }
        bufAppendString(buf, "\"");
      }
      break;
      
    case TCI_ENUMERATED_TYPE:
      {
        String enum_choice = tciGetEnumValue(value);
        bufAppendString(buf, enum_choice);
      }
      break;
      
    case TCI_FLOAT_TYPE:
      {
        double dbl = tciGetFloatValue(value);
        char str[100];
        str_printf(str, 100, "%f", dbl);
        bufAppendString(buf, str);
      }
      break;
      
    case TCI_HEXSTRING_TYPE:
      {
        String bstring = tciGetHStringValue(value);
        bufAppendString(buf, bstring);
      }
      break;
      
    case TCI_INTEGER_TYPE:
      {
        unsigned char sign = tciGetIntSign(value);
        String abs_str = tciGetIntAbs(value);
        
        bufAppendString(buf, sign ? "" : "-");
        bufAppendString(buf, abs_str);
      }
      break;
      
    case TCI_OBJID_TYPE:
      {
        TciObjidValue objid_val = tciGetTciObjidValue(value);
        unsigned long i;

        bufAppendString(buf, "objid {"); 
        for (i = 0; i < objid_val->length; i++) 
        {
          char str[100];
          str_printf(str, 100, "%d", objid_val->elements[i].elem_as_number);
          bufAppendString(buf, str);
          if (i < objid_val->length - 1)
            bufAppendString(buf, " ");
        }
        bufAppendString(buf, "}"); 
      }
      break;
      
    case TCI_OCTETSTRING_TYPE:
      {
        String bstring = tciGetOStringValue(value);
        bufAppendString(buf, bstring);
      }
      break;
      
    case TCI_RECORD_TYPE:
    case TCI_SET_TYPE:
      {
        String *field_names = tciGetRecFieldNames(value);
        unsigned long i;
      
        make_indent(buf, indent);
        bufAppendString(buf, "{");
        for (i = 0; field_names && field_names[i]; i++) 
        {
          TciValue elem = tciGetRecFieldValue(value, field_names[i]);
          make_indent(buf, indent+1);
          bufAppendString(buf, field_names[i]);
          bufAppendString(buf, " := ");
          if (tciNotPresent(elem))
            bufAppendString(buf, "omit");
          else  
            to_string(elem, buf, indent+1);
          if (field_names[i+1])
            bufAppendString(buf, ",");
        }
        make_indent(buf, indent);
        bufAppendString(buf, "}");
	  }
      break;
      
    case TCI_RECORD_OF_TYPE:
    case TCI_SET_OF_TYPE:
      {
        unsigned long length = tciGetRecOfLength(value);
        unsigned long i;

        make_indent(buf, indent);
        bufAppendString(buf, "{");
        for (i = 0; i < length; i++) 
        {
          char str[100];
          TciValue elem = tciGetRecOfFieldValue(value, i);
          str_printf(str, 100, "[%d]", i);
          make_indent(buf, indent+1);
          bufAppendString(buf, str);
          bufAppendString(buf, " := ");
          if (tciNotPresent(elem))
            bufAppendString(buf, "-");
          else  
            to_string(elem, buf, indent+1);
          if (i < length - 1)
            bufAppendString(buf, ",");
        }
        make_indent(buf, indent);
        bufAppendString(buf, "}");
      }
      break;
         
    case TCI_ANYTYPE_TYPE:
    case TCI_UNION_TYPE:
      {
        String variant = tciGetUnionPresentVariantName(value);
        TciValue elem = tciGetUnionVariant(value, variant);

        bufAppendString(buf, "{");
        bufAppendString(buf, variant);
        bufAppendString(buf, " := ");
        to_string(elem, buf, indent);
        bufAppendString(buf, "}");
      }
      break;
      
    case TCI_UNIVERSAL_CHAR_TYPE:
      {
        TciUCReturnValue wchar = tciGetUniversalCharValue(value);
        char str[100];
        
        wchar2string(wchar, str);
        
        bufAppendString(buf, "\"");
        bufAppendString(buf, str);
        bufAppendString(buf, "\"");
      }
      break;
      
    case TCI_UNIVERSAL_CHARSTRING_TYPE:
      {
        unsigned long length= tciGetUCStringLength(value);
        unsigned long i;

        bufAppendString(buf, "\"");
        
        for (i=0; i < length; i++) 
        {
          TciUCReturnValue wchar = tciGetUCStringCharValue(value, i);
          char str[100];

          wchar2string(wchar, str);
          bufAppendString(buf, str);
        }
        bufAppendString(buf, "\"W");
      }
      break;
      
    case TCI_VERDICT_TYPE:
      {
        unsigned long verdict = tciGetVerdictValue(value);
        if (verdict == TCI_VERDICT_NONE)
          bufAppendString(buf, "none");
        else if (verdict == TCI_VERDICT_PASS) 
          bufAppendString(buf, "pass");
        else if (verdict == TCI_VERDICT_INCONC) 
          bufAppendString(buf, "inconc");
        else if (verdict == TCI_VERDICT_FAIL) 
          bufAppendString(buf, "fail");
        else if (verdict == TCI_VERDICT_ERROR) 
          bufAppendString(buf, "error");
      }    
      break;

    case TCI_COMPONENT_TYPE:
      {
        bufAppendString(buf, "<component instance>");
      }
      break;
    case TCI_DEFAULT_TYPE:
      {
        bufAppendString(buf, "<default instance>");
      }
      break;
    case TCI_PORT_TYPE:
      {
        bufAppendString(buf, "<port instance>");
      }
      break;
    case TCI_TIMER_TYPE:
      {
        bufAppendString(buf, "<timer instance>");
      }
      break;
    case TCI_TEMPLATE_TYPE:
      {
        bufAppendString(buf, "<template instance>");
      }
      break;
      
    default:
      {
        bufAppendString(buf, "<unknown object>");
      }
  }
}


static unsigned long wchar2int (TciUCValue wchar)
{
  unsigned long result;
  
  result =  wchar[0]; /* group */
  result <<= 8;
  result |= wchar[1]; /* plane */
  result <<= 8;
  result |= wchar[2]; /* row */
  result <<= 8;
  result |= wchar[3]; /* cell */
  
  return result;
}

static char* wchar2string (TciUCValue wchar, char *buf)
{
  /* Converts a wide char into ASCII string
   * Chars that are mappable to ASCII are printed as such,
   * other chars are printed as a <#> combination.
   * We assume that 'buf' is wide enough
   */

  unsigned long char_code = wchar2int(wchar);

  if ( (char_code >= ' ' && char_code < 0xFF) ||
       (char_code == '\n') || 
       (char_code == '\t')) 
    sprintf (buf, "%c", (char) char_code);
  else 
    sprintf (buf, "<0x%08lX>", char_code);
    
  return buf;
}

/*********************************************************************/

#ifndef DISABLE_TCI_CD_TM_STUBS

TciValue tciDecode(BinaryString message, 
                   TciType decHypothesis) {

  return 0;                   
}

BinaryString tciEncode(TciValue value) {

  BinaryString result;
 
  result.bits = 0;
  return result;
}

void tciTestCaseStarted(TciTestCaseIdType testCaseId, 
                        TciParameterListType parameterList, 
                        double timer) {}

void tciTestCaseTerminated(TciValue verdict, 
                           TciParameterListType parameterlist) {}

/*
 * 'tciControlTerminated' is generated by the compiler
 *
 * void tciControlTerminated () {}
 */

TciValue tciGetModulePar(TciModuleParameterIdType parameterId) {
  
  return 0;
}

void tciError(String message) {}

#endif /* DISABLE_TCI_CD_TM_STUBS */


#ifdef __cplusplus
}
#endif
