using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Xml;
using System.Net;

namespace TestExecutor
{
    public partial class MainForm : Form
    {
        delegate void ConnectedEvent(string HostAddress, string state);
        delegate void DisconnectedEvent();

        event ConnectedEvent DoConnected;
        event DisconnectedEvent DoDisconnected;

        TciSoapAccessService TestExecutable;

        TciSoapAccessService TestEventListener; // this objects runs on separate thread
        Thread providedThread; // a thread using TestEventListener

        public MainForm()
        {
            InitializeComponent();

            DoConnected += delegate(string HostAddress, string state) 
            { 
                // Update status text
                if (HostAddress != null)
                    StatusLabel.Text = "Connected to " + HostAddress;
                if (state != null)
                    StateLabel.Text = state;
            };

            DoDisconnected += delegate() 
            {
                // Update status text
                StatusLabel.Text = "Not connected";
                StateLabel.Text = "";
            };
        }

        private void OnConnect(object sender, EventArgs e)
        {
            ConnectForm frm = new ConnectForm();
            if (frm.ShowDialog(this) == DialogResult.OK)
            {
                if (TestEventListener != null)
                    TestEventListener.Abort();

                TestExecutable = new TciSoapAccessService();
                TestEventListener = new TciSoapAccessService();

                // set address parameters (boths objects are initialized equally)
                if (frm.HostAddress.StartsWith("http://"))
                {
                    TestExecutable.Url = frm.HostAddress;
                    TestEventListener.Url = frm.HostAddress;
                }
                else
                {
                    TestExecutable.Url = "http://" + frm.HostAddress;
                    TestEventListener.Url = "http://" + frm.HostAddress;
                }

                TestExecutable.Timeout = Timeout.Infinite;
                TestEventListener.Timeout = Timeout.Infinite;

                // check connection and request test executable state
                try
                {
                    LoadTree();

                    // clear all items, list will be fully reload
                    TestHistoryListBox.Items.Clear();

                    XmlQualifiedName state = TestExecutable.GetTestExecutableState();
                    lock (this)
                    {
                        DoConnected(frm.HostAddress, state.Name);
                    }

                    // start receiving thread
                    providedThread = new Thread(ProvidedHandler);
                    providedThread.Start();
                }
                catch (Exception exc)
                {
                    MessageBox.Show(exc.ToString(), "Connection error");
                }
            }

        }

        // form initialization
        private void MainForm_Load(object sender, EventArgs e)
        {
            lock (this)
            {
                DoDisconnected(); // initial behavior
            }

            connectToolMenuItem.PerformClick();
        }

        // load tree of modules and testcases
        private void LoadTree()
        {
            TestCaseTree.Nodes.Clear();

            string[] modules = TestExecutable.GetModules();

            // iterate through all modules
            foreach (string module in modules)
            {
                ModuleNode mNode = new ModuleNode(module);
                TestCaseTree.Nodes.Add(mNode);

                // get all testcases defined on the module
                QualifiedName[] testCases = TestExecutable.GetTestCases(module, true);
                foreach (QualifiedName testCase in testCases)
                {
                    // filter-out imported testcases
                    if (testCase.moduleName == module)
                    {
                        TestCaseNode testCaseNode = new TestCaseNode(module, testCase.objectName);
                        mNode.Nodes.Add(testCaseNode);
                    }
                }
            }

            TestCaseTree.ExpandAll();
        }

        private void OnExit(object sender, EventArgs e)
        {
            Application.Exit();
        }

        // Thread listening for events in the test suite
        private void ProvidedHandler()
        {
            int eventId = 1;

            restart:

            try
            {
                while(true)
                {
                    EventInfo tcEvent = TestEventListener.ListenEvent(eventId, true);
                    if (tcEvent != null)
                    {
                        ++eventId; // increase event id only after successful fetch

                        // update 'state' label
                        XmlQualifiedName state = TestEventListener.GetTestExecutableState();

                        lock (this)
                        {
                            DoConnected(null, state.Name);
                        }

                        // add entries to list box
                        #region Switch
                        switch (tcEvent.Item.GetType().Name)
                        {
                            case "TestCaseStartedInfo":
                                TestCaseStartedInfo tcStarted = (TestCaseStartedInfo)tcEvent.Item;

                                TestHistoryListBox.Items.Add("Executing " + tcStarted.testCaseId.objectName + "() ...");
                                break;

                            case "TestCaseTerminatedInfo":
                                TestCaseTerminatedInfo tcTerminated = (TestCaseTerminatedInfo)tcEvent.Item;
                                switch (tcTerminated.verdict.Name)
                                {
                                    case "VERDICT-NONE":
                                        TestHistoryListBox.Items.Add("... none");
                                        break;

                                    case "VERDICT-PASS":
                                        TestHistoryListBox.Items.Add("... passed");
                                        break;

                                    case "VERDICT-FAIL":
                                        TestHistoryListBox.Items.Add("... failed");
                                        break;

                                    case "VERDICT-INCONC":
                                        TestHistoryListBox.Items.Add("... inconclusive");
                                        break;

                                    case "VERDICT-ERROR":
                                        TestHistoryListBox.Items.Add("... test system error");
                                        break;
                                }
                                break;

                            case "ControlTerminatedInfo":
                                TestHistoryListBox.Items.Add("=== control part terminated ===");
                                break;

                            case "ErrorInfo":
                                ErrorInfo errorInfo = (ErrorInfo)tcEvent.Item;
                                TestHistoryListBox.Items.Add("ERROR: " + errorInfo.errorMessage);
                                break;
                        }
                        #endregion
                    }
                }
            }
            catch (WebException exc)
            {
                if (exc.Status != WebExceptionStatus.RequestCanceled)
                    MessageBox.Show(exc.ToString(), "Connection has been lost...");
            }

            catch (ThreadAbortException exc)
            {
                // do nothing
            }

            catch (InvalidOperationException exc)
            {
                // ignore
                goto restart;
            }
 
            catch (Exception exc)
            {
                MessageBox.Show(exc.ToString());
            }

            lock (this)
            {
                DoDisconnected();
            }
        }

        private void OnStart(object sender, EventArgs e)
        {
            TreeNode startedNode = TestCaseTree.SelectedNode;

            if (startedNode != null)
            {
                // selection on module node
                if (startedNode is ModuleNode)
                {
                    ModuleNode mNode = (ModuleNode)startedNode;
                    TestExecutable.StartControl(mNode.ModuleName);
                }
                // selection on testcase node
                else
                {
                    TestCaseNode tcNode = (TestCaseNode)startedNode;

                    QualifiedName qName = new QualifiedName();
                    qName.moduleName = tcNode.ModuleName;
                    qName.objectName = tcNode.TestCaseName;

                    ParameterType[] parameters = TestExecutable.GetTestCaseParameters(qName);

                    TestCaseParameters parForm = new TestCaseParameters(parameters);
                    if (parForm.ShowDialog(this) == DialogResult.OK)
                        TestExecutable.StartTestCase(qName, parForm.ParValues);
                }
            }
        }

        private void OnStop(object sender, EventArgs e)
        {
            TreeNode stoppedNode = TestCaseTree.SelectedNode;

            if (stoppedNode != null)
            {
                // selection on module node
                if (stoppedNode is ModuleNode)
                {
                    TestExecutable.StopControl();
                }
                // selection on testcase node
                else
                {
                    TestExecutable.StopTestCase();
                }
            }
        }

        private void OnClosing(object sender, FormClosingEventArgs e)
        {
            if (TestEventListener != null)
                TestEventListener.Abort();
        }

    }
}