#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "tci.h"

#ifdef DEBUG
#define tci_assert(cond, msg) if (!(cond)) {printf(">>>>> VALUE->STRING: " msg); bufAppendString(buf,"-"); return;}
#else
#define tci_assert(cond, msg) if (!(cond)) {bufAppendString(buf,"-"); return;}
#endif

/* Log buffer declaration and management routines */

typedef struct _log_print_buffer_t {
    char *buf;
    unsigned long length;
    unsigned long allocated;
} log_print_buffer_t;
                                              
static void bufAppendStringN(log_print_buffer_t *buf, const char *str, unsigned long length)
{
    if (length == 0) return;

    // expand buffer if necesary
    if (buf->allocated - buf->length < length + 1) 
    {
        /* 1 for '\0', 100 - reserved */
        unsigned long new_size = buf->length + length + 1 + 100;
    
        char *tmp = NULL;
        if (buf->allocated)
            tmp = (char*)realloc(tmp, new_size);
        else
            tmp = (char*)malloc(new_size);

        if (buf->buf) strcpy(tmp, buf->buf);
        buf->buf = tmp;
        buf->allocated = new_size;
    }

    strncpy(buf->buf + buf->length, str, length);
    buf->length += length;
    buf->buf[buf->length] = '\0';
}

static void bufAppendString(log_print_buffer_t *buf, const char *str)
{
    bufAppendStringN(buf, str, strlen(str));
}

static void value2string(log_print_buffer_t *buf, TciValue value)
{
  TciType type;
  TciTypeClassType kind;
  
  type = tciGetType(value);
  tci_assert(type != NULL, "Illegal type");

  if (tciNotPresent(value))
  {
    tci_assert(0, "Illegal value");
    return;
  }
  
  kind = tciGetTypeClass(type);
    
  switch(kind) { 
    case TCI_ADDRESS_TYPE:
      {
        TciValue underlying_value;
        
        /* print underlying address value */
        underlying_value = tciGetAddressValue(value);
        if (NULL == underlying_value)
          bufAppendString(buf, "null");
        else
          value2string(buf, underlying_value);
      }
      break;
      
    case TCI_BITSTRING_TYPE:
      {
        String bstring;
      
        /* print bit string */
        bstring = tciGetBStringValue(value);
        tci_assert(bstring != NULL, "Illegal bitstring value");
        bufAppendString(buf, bstring);
      }
      break;
      
    case TCI_BOOLEAN_TYPE:
      {
        unsigned char bval;
        
        /* print boolean */
        bval = tciGetBooleanValue(value);
        bufAppendString(buf, bval ? "true" : "false");
      }
      break;
    
    case TCI_CHAR_TYPE:
      {
        unsigned char character;
        char schar[10];
        
        /* print character */
        character = tciGetCharValue(value);
        sprintf(schar, "\"%c\"", character);
        bufAppendString(buf, schar);
      }
      break;
      
    case TCI_CHARSTRING_TYPE:
      {
        TciCharStringValue cstr_value;

        /* print charstring value */
        cstr_value = tciGetCStringValue(value);
        tci_assert(cstr_value != NULL, "Illegal charstring value");
        bufAppendString(buf, "\"");
        bufAppendStringN(buf, cstr_value->string, cstr_value->length);
        bufAppendString(buf, "\"");
      }
      break;
      
    case TCI_ENUMERATED_TYPE:
      {
        String enum_choice;
        
        /* encode enumeration choice */
        enum_choice = tciGetEnumValue(value);
        tci_assert(enum_choice != NULL, "Illegal enumerated value");
        bufAppendString(buf, enum_choice);
      }
      break;
      
    case TCI_FLOAT_TYPE:
      {
        double dbl;
        char sdbl[100];
      
        /* encode double value */
        dbl = tciGetFloatValue(value);
        sprintf(sdbl, "%e", dbl);
        bufAppendString(buf, sdbl);
      }
      break;
      
    case TCI_HEXSTRING_TYPE:
      {
        String hstring;
        
        /* encode hex string */
        hstring = tciGetHStringValue(value);
        tci_assert(hstring != NULL, "Illegal hexstring value")
        bufAppendString(buf, hstring);
      }
      break;
      
    case TCI_INTEGER_TYPE:
      {
        unsigned char sign;
        String abs_str;
        
        /* encode sign */
        sign = tciGetIntSign(value);
        bufAppendString(buf, sign ? "" : "-");

        /* encode string representing absolute value */
        abs_str = tciGetIntAbs(value);
        tci_assert(abs_str != NULL, "Illegal integer value");
        bufAppendString(buf, abs_str);
      }
      break;
      
    case TCI_OBJID_TYPE:
      {
        TciObjidValue objid_val;
        unsigned long i;
        char si[100];

        bufAppendString(buf, "objid {");
        
        /* get list of objid elements */
        objid_val = tciGetTciObjidValue(value);
        
        
        for (i = 0; i < objid_val->length; i++) {
          if (i > 0)
            bufAppendString(buf, " ");
        
          /* print objid elements one by one */
          sprintf(si, "%ld", objid_val->elements[i].elem_as_number);
          bufAppendString(buf, si);
        }

        bufAppendString(buf, "}");
      }
      break;
      
    case TCI_OCTETSTRING_TYPE:
      {
        String ostring;
        
        ostring = tciGetOStringValue(value);
        tci_assert(ostring, "Illegal octet string value");
        
        /* print octet string */
        bufAppendString(buf, ostring);
      }
      break;
      
    case TCI_RECORD_TYPE:
    case TCI_SET_TYPE:
      {
        String *field_names;
        unsigned long i;

        bufAppendString(buf, "{");
        
        /* get list of record/set firld names */
        field_names = tciGetRecFieldNames(value);
        for (i = 0; field_names[i] != NULL; i++) {
          TciValue elem;

          if (i > 0)
            bufAppendString(buf, ", ");
          
          /* get record/set field value */
          elem = tciGetRecFieldValue(value, field_names[i]);
          tci_assert(elem, "Ilegal record/set field value");

          /* print record/set field value */
          if (tciNotPresent(elem))
            bufAppendString(buf, "omit");
          else  
            value2string(buf, elem);
        }

        bufAppendString(buf, "}");
	  }
      break;
      
    case TCI_RECORD_OF_TYPE:
    case TCI_SET_OF_TYPE:
      {
        unsigned long length;
        unsigned long i;

        /* encode length of record of/set of*/
        length = tciGetRecOfLength(value);

        bufAppendString(buf, "{");
        
        for (i = 0; i < length; i++) {
          TciValue elem;

          if (i > 0)
            bufAppendString(buf, ", ");
          
          /* get record of/set of element */
          elem = tciGetRecOfFieldValue(value, i);
          tci_assert(elem, "Illegal record of/set of element value");

          /* print record of/set of element */
          if (tciNotPresent(elem))
            bufAppendString(buf, "-");
          else  
            value2string(buf, elem);
        }

        bufAppendString(buf, "}");
      }
      break;
         
    case TCI_ANYTYPE_TYPE:
    case TCI_UNION_TYPE:
      {
        String variant;
        TciValue elem;
        
        variant = tciGetUnionPresentVariantName(value);
        tci_assert(variant, "Union variant is not set");

        bufAppendString(buf, "{");
        
        /* print variant name */
        bufAppendString(buf, variant);
        bufAppendString(buf, " := ");

        /* get variant value */
        elem = tciGetUnionVariant(value, variant);
        tci_assert(elem, "Illegal union variant");
        
        /* print variant value */
        value2string(buf, elem);
      }
      break;
      
    case TCI_UNIVERSAL_CHAR_TYPE:
      {
        TciUCReturnValue wchar;
        
        /* get widechar value */
        wchar = tciGetUniversalCharValue(value);
        /* print widechar */
        // ??? binary_string_append_bytes(msg, wchar, sizeof(TciUCValue));
        bufAppendString(buf, "<n/a>");
      }
      break;
      
    case TCI_UNIVERSAL_CHARSTRING_TYPE:
      {
        unsigned long length;  
        unsigned long i;
        TciUCReturnValue wchar;
      
        length = tciGetUCStringLength(value);
        
        for (i=0; i < length; i++) {
          /* encode widestring elements one by one */
          wchar = tciGetUCStringCharValue(value, i);
          // ??? binary_string_append_bytes(msg, wchar, sizeof(TciUCValue));
          bufAppendString(buf, "<n/a>");
        }
      }
      break;
      
    case TCI_VERDICT_TYPE:
      {
        unsigned long verdict;
        const char *vname;

        /* encode verdict */ 
        verdict = tciGetVerdictValue(value);

        if (verdict == TCI_VERDICT_NONE)
          vname = "none";
        else if (verdict == TCI_VERDICT_PASS)
          vname = "pass";
        else if (verdict == TCI_VERDICT_INCONC)
          vname = "inconc";
        else if (verdict == TCI_VERDICT_FAIL)
          vname = "fail";
        else if (verdict == TCI_VERDICT_ERROR)
          vname = "error";
        else
          tci_assert(0, "Illegal verdict");
        
        bufAppendString(buf, vname);
      }
      break;
      
    default:
      bufAppendString(buf, "-");
  }
}

String tciValueToString(TciValue value)
{
    log_print_buffer_t buf = {NULL, 0, 0};

    value2string(&buf, value);

    return buf.buf;
}

/* End of File */

