#include "soapH.h"
#include "tci.h"

#include "t3rts.h"

// internal RTS function
extern TciValue tciConstructValueFromString(TciType type, String inst);

// calling function is responsible for releasing memory
extern String tciValueToString(TciValue value);

#define SMALLOC(cnt,type) ((type *)soap_malloc(soap, (cnt) * sizeof(type)))
#define SFREE(ptr) if (ptr) soap_dealloc(soap, ptr);
#define SSTRDUP(str) ((str) ? soap_strdup(soap, str) : NULL)

// global variable, stores state returned by GetTestExecutableState
static enum t3ws__TestExecutableState gTestExecutableState = t3ws__IDLE;

////////////////////////////////////////////////////////////////
// Data conversion functions
////////////////////////////////////////////////////////////////

/*
 * This implementation is made on top of standard TCI implementation
 * that requires setting root module with tciRootModule to be able to 
 * call various TCI TM functions. Calling tciRootModule while test case
 * is running results in testcase error.
 * That's why to make it possible to provide test structure information
 * at arbitrary point of time (and during testcase execution too) we
 * implement caching of necessary static information (like list of modules,
 * list of testcases, etc)
 */

typedef struct t3ws__ModuleEntry
{
    TciModuleIdType moduleId;

    struct {
        int size;
        TciModuleIdType *moduleId;
    } importedModules;

    struct {
        int size;
        TciTestCaseIdType *testCaseId;
    } testCases;
} t3ws__ModuleEntry;

// globally cached test suite structure
static struct t3ws__ModuleEntry *gTestSuiteStructure = NULL;
static int gTotalModules = 0; // size of gTestSuiteStructure list

void t3ws__FillTestSuiteCache(struct soap *soap)
{
    TciModuleIdListType moduleList = tciGetModules();

    if (moduleList && moduleList->length > 0) {
        int i;

        gTotalModules = moduleList->length;
        gTestSuiteStructure = SMALLOC(moduleList->length, t3ws__ModuleEntry);

        for (i = 0; i < moduleList->length; ++i)
	{
	    TciTestCaseIdListType testCases;
	    TciModuleIdListType importedModules;

	    tciRootModule(moduleList->idList[i].moduleName);

            // set module name
            gTestSuiteStructure[i].moduleId.moduleName = SSTRDUP(moduleList->idList[i].moduleName);
            gTestSuiteStructure[i].moduleId.objectName = SSTRDUP(moduleList->idList[i].objectName);

	    gTestSuiteStructure[i].importedModules.size = 0;
	    gTestSuiteStructure[i].importedModules.moduleId = NULL;
	    gTestSuiteStructure[i].testCases.size = 0;
	    gTestSuiteStructure[i].testCases.testCaseId = NULL;

            // set testcases 
            testCases = tciGetTestCases();
            if (testCases && testCases->length > 0) 
	    {
	        int j;

	        gTestSuiteStructure[i].testCases.size = testCases->length;
	        gTestSuiteStructure[i].testCases.testCaseId = SMALLOC(testCases->length, TciTestCaseIdType);

                for (j = 0; j < testCases->length; ++j) {
                    gTestSuiteStructure[i].testCases.testCaseId[j].moduleName = SSTRDUP(testCases->idList[j].moduleName);
                    gTestSuiteStructure[i].testCases.testCaseId[j].objectName = SSTRDUP(testCases->idList[j].objectName);
                }
	    }

	    // set imported modules
            importedModules = tciGetImportedModules();
            if (importedModules && importedModules->length > 0)
	    {
	        int j;

	        gTestSuiteStructure[i].importedModules.size = importedModules->length;
	        gTestSuiteStructure[i].importedModules.moduleId = SMALLOC(importedModules->length + 1, TciModuleIdType);

		for (j = 0; j < importedModules->length; ++j) {
                    gTestSuiteStructure[i].importedModules.moduleId[j].moduleName = SSTRDUP(importedModules->idList[j].moduleName);
                    gTestSuiteStructure[i].importedModules.moduleId[j].objectName = SSTRDUP(importedModules->idList[j].objectName);
		}
	    }
	}
    }
}

static int LookupModuleIdx(const char *moduleId)
{
    int i;

    for (i = 0; gTestSuiteStructure && gTestSuiteStructure[i].moduleId.moduleName; ++i)
        if (strcmp(gTestSuiteStructure[i].moduleId.moduleName, moduleId) == 0)
	    return i;

    return -1;
}

////////////////////////////////////////////////////////////////
// Data conversion functions
////////////////////////////////////////////////////////////////

//
// ws2tci_ParameterList
//
// converts parameter list from WSDL representation to TCI one
//

static void
ws2tci_ParameterList(struct soap* soap, 
                     TciTestCaseIdType tcId,
                     TciParameterListType dstParList, 
                     t3ws__ParameterList *srcParList)
{
    TciParameterTypeListType tcParams = tciGetTestCaseParameters(tcId);

    dstParList->length = srcParList->__sizeparameter;
    dstParList->parList = NULL;

    if (dstParList->length > 0) {
        int i;

        dstParList->parList = SMALLOC(dstParList->length, TciParameterType);
        for (i = 0; i < dstParList->length; ++i) {
            dstParList->parList[i].parName = SSTRDUP(srcParList->parameter[i].parameterName);

            switch(srcParList->parameter[i].passingMode) {
            case TCI_IN_PAR:
                dstParList->parList[i].parPassMode = t3ws__IN_PAR; 
		dstParList->parList[i].parValue = tciConstructValueFromString(tcParams->parList[i].parType,
                                                                              srcParList->parameter[i].parameterValue);
                break;
            case TCI_INOUT_PAR:
                dstParList->parList[i].parPassMode = t3ws__INOUT_PAR; 
		dstParList->parList[i].parValue = tciConstructValueFromString(tcParams->parList[i].parType,
                                                                              srcParList->parameter[i].parameterValue);
                break;
            case TCI_OUT_PAR:
                dstParList->parList[i].parPassMode = t3ws__OUT_PAR; 
		dstParList->parList[i].parValue = NULL;
                break;
            }

        }
    }
}

//
// tci2ws_ParameterList
//
// converts parameter list from TCI representation to WSDL one
//

static void
tci2ws_ParameterList(struct soap* soap,
                     t3ws__ParameterList *dstParList,
                     TciParameterListType srcParList)
{
    dstParList->__sizeparameter = srcParList->length;
    dstParList->parameter = NULL;

    if (srcParList->length > 0) {
        int i;

        dstParList->parameter = SMALLOC(srcParList->length, t3ws__Parameter);
        for (i = 0; i < srcParList->length; ++i) {
            switch(srcParList->parList[i].parPassMode) {
            case TCI_IN_PAR:
                dstParList->parameter[i].passingMode = t3ws__IN_PAR;
                break;
            case TCI_INOUT_PAR:
                dstParList->parameter[i].passingMode = t3ws__INOUT_PAR;
                break;
            case TCI_OUT_PAR:
                dstParList->parameter[i].passingMode = t3ws__OUT_PAR;
                break;
            }

            if (srcParList->parList[i].parValue) {
                char *v2s = tciValueToString(srcParList->parList[i].parValue);
                dstParList->parameter[i].parameterValue = SSTRDUP(v2s);
                free(v2s);

                dstParList->parameter[i].parameterName = SSTRDUP(srcParList->parList[i].parName);
            } else {
                dstParList->parameter[i].parameterName = NULL;
                dstParList->parameter[i].parameterValue = NULL;
            }
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
// WSDL TM required operations
///////////////////////////////////////////////////////////////////////////////

//
// GetModules
//

SOAP_FMAC5 int SOAP_FMAC6 
__t3sa__GetModules(struct soap* soap, 
                   struct _t3ws__GetModules *dummy, 
                   struct _t3ws__GetModulesResponse *retVal)
{
    int nModules = gTotalModules;

    retVal->__sizemoduleId = nModules;
    retVal->moduleId = NULL;

    if (nModules > 0) {
        int i;

        retVal->moduleId = SMALLOC(nModules, char*);
        for (i = 0; i < nModules; ++i)
            retVal->moduleId[i] = SSTRDUP(gTestSuiteStructure[i].moduleId.moduleName);
    }

    return SOAP_OK;
}

//
// GetImportedModules
//

SOAP_FMAC5 int SOAP_FMAC6 
__t3sa__GetImportedModules(struct soap* soap, 
                           struct _t3ws__GetImportedModules *params, 
                           struct _t3ws__GetImportedModulesResponse *retVal)
{
    int moduleIdx = LookupModuleIdx(params->moduleId);
    int nImportedModules;

    if (moduleIdx == -1) return SOAP_USER_ERROR;

    retVal->__sizemoduleId = 0;
    retVal->moduleId = NULL;

    nImportedModules = gTestSuiteStructure[moduleIdx].importedModules.size;
    if (nImportedModules > 0) {
	int i;

        retVal->__sizemoduleId = nImportedModules;
        retVal->moduleId = SMALLOC(nImportedModules, char*);

        for (i = 0; i < gTestSuiteStructure[moduleIdx].importedModules.size; ++i)
            retVal->moduleId[i] = SSTRDUP(gTestSuiteStructure[moduleIdx].importedModules.moduleId[i].moduleName);
    }

    return SOAP_OK;
}

//
// GetModuleParameters
//

SOAP_FMAC5 int SOAP_FMAC6 
__t3sa__GetModuleParameters(struct soap* soap, 
                            struct _t3ws__GetModuleParameters *params, 
                            struct _t3ws__GetModuleParametersResponse *retVal)
{
    TciModuleParameterListType modParList;
    TciTestCaseIdType modId;

    modId.moduleName = modId.objectName = SSTRDUP(params->moduleId);
    modId.aux = 0;

    modParList = tciGetModuleParameters(modId);
    if (!modParList) return SOAP_USER_ERROR;

    retVal->__sizemodulePar = modParList->length;
    retVal->modulePar= NULL;

    if (modParList->length > 0) {
        int i;

        retVal->modulePar = SMALLOC(modParList->length, t3ws__ModuleParameter);
        for(i = 0; i < modParList->length; ++i) {
            TciType parType = tciGetModuleParameterType(modParList->modParList[i].parName);

            retVal->modulePar[i].moduleParId->moduleName = SSTRDUP(modParList->modParList[i].parName.moduleName);
            retVal->modulePar[i].moduleParId->objectName = SSTRDUP(modParList->modParList[i].parName.objectName);

            retVal->modulePar[i].moduleParType->moduleName = SSTRDUP(tciGetDefiningModule(parType).moduleName);
            retVal->modulePar[i].moduleParType->objectName = SSTRDUP(tciGetName(parType));

            if (modParList->modParList[i].defaultValue) {
                char *v2s = tciValueToString(modParList->modParList[i].defaultValue);
                retVal->modulePar[i].defaultValue = SSTRDUP(v2s);
                free(v2s);
            }
            else
                retVal->modulePar[i].defaultValue = NULL;
        }
    }

    return SOAP_OK;
}

//
// GetTestCases
//
// NOTE: I do not return imported testcases here. May be return them as specified in TCI?
//       Test Executor handles both cases well.


SOAP_FMAC5 int SOAP_FMAC6 
__t3sa__GetTestCases(struct soap* soap, 
                     struct _t3ws__GetTestCases *params, 
                     struct _t3ws__GetTestCasesResponse *retVal)
{
    int moduleIdx = LookupModuleIdx(params->moduleId);
    int nTestCases;

    if (moduleIdx == -1) return SOAP_USER_ERROR;

    retVal->__sizetestCaseId = 0;
    retVal->testCaseId = NULL;

    nTestCases = gTestSuiteStructure[moduleIdx].testCases.size;
    if (nTestCases > 0) {
        int i;
	int tcIdx = 0;

        retVal->__sizetestCaseId = nTestCases;
        retVal->testCaseId = SMALLOC(nTestCases, t3ws__QualifiedName);

	for (i = 0; i < nTestCases; ++i)
	    if (params->includeImported 
	           ||
	        strcmp(params->moduleId, gTestSuiteStructure[moduleIdx].testCases.testCaseId[i].moduleName) == 0)
            {
                retVal->testCaseId[tcIdx].moduleName = SSTRDUP(gTestSuiteStructure[moduleIdx].testCases.testCaseId[i].moduleName);
                retVal->testCaseId[tcIdx].objectName = SSTRDUP(gTestSuiteStructure[moduleIdx].testCases.testCaseId[i].objectName);
                ++tcIdx;
	    }

        retVal->__sizetestCaseId = tcIdx; // make correction if do not include imported testcases
    }

    return SOAP_OK;
}

//
// GetTestCaseParameters
//

SOAP_FMAC5 int SOAP_FMAC6 
__t3sa__GetTestCaseParameters(struct soap* soap, 
                              struct _t3ws__GetTestCaseParameters *params, 
                              struct _t3ws__GetTestCaseParametersResponse *retVal)
{
    TciParameterTypeListType tcParams;
    TciTestCaseParameterIdListType tcParNames;
    TciTestCaseIdType tcId;

    tcId.moduleName = SSTRDUP(params->testCaseId->moduleName);
    tcId.objectName = SSTRDUP(params->testCaseId->objectName);
    tcId.aux = 0;

    tcParams = tciGetTestCaseParameters(tcId);
    tcParNames = tciGetTestCaseParametersNames(tcId);
    if ( !(tcParams && tcParNames && tcParams->length == tcParNames->length) ) return SOAP_USER_ERROR;

    retVal->__sizeparameter = tcParams->length;
    retVal->parameter = NULL;

    if (tcParams->length > 0) {
        int i;

        retVal->parameter = SMALLOC(tcParams->length, t3ws__ParameterType);
        for(i = 0; i < tcParams->length; ++i) {
            TciType paramType = tcParams->parList[i].parType;

            retVal->parameter[i].parameterName = SSTRDUP(tcParNames->idList[i]);
            retVal->parameter[i].parameterType = SMALLOC(1, t3ws__QualifiedName);
            retVal->parameter[i].parameterType->moduleName = SSTRDUP(tciGetDefiningModule(paramType).moduleName);
            retVal->parameter[i].parameterType->objectName = SSTRDUP(tciGetName(paramType));

            switch(tcParams->parList[i].parPassMode) {
            case TCI_IN_PAR:
                retVal->parameter[i].passingMode = t3ws__IN_PAR;
                break;
            case TCI_INOUT_PAR:
                retVal->parameter[i].passingMode = t3ws__INOUT_PAR;
                break;
            case TCI_OUT_PAR:
                retVal->parameter[i].passingMode = t3ws__OUT_PAR;
                break;
            }
        }
    }

    return SOAP_OK;
}

//
// StartTestCase
//

SOAP_FMAC5 int SOAP_FMAC6 
__t3sa__StartTestCase(struct soap* soap, 
                      struct _t3ws__StartTestCase *params, 
                      struct _t3ws__StartTestCaseResponse *retVal)
{
    TciTestCaseIdType tcId;
    struct _TciParameterListType paramList;

    // change root module
    tciRootModule(params->testCaseId->moduleName);

    tcId.moduleName = SSTRDUP(params->testCaseId->moduleName);
    tcId.objectName = SSTRDUP(params->testCaseId->objectName);
    tcId.aux = 0;

    ws2tci_ParameterList(soap, tcId, &paramList, params->parameterList);

    tciStartTestCase(tcId, &paramList);

    gTestExecutableState = t3ws__EXECUTING_TESTCASE;

    return SOAP_OK;
}

//
// StopTestCase
//

SOAP_FMAC5 int SOAP_FMAC6 
__t3sa__StopTestCase(struct soap* soap, 
                     struct _t3ws__StopTestCase *dummy, 
                     struct _t3ws__StopTestCaseResponse *dummyRetVal)
{
    tciStopTestCase();

    return SOAP_OK;
}

//
// StartControl
//

SOAP_FMAC5 int SOAP_FMAC6 
__t3sa__StartControl(struct soap* soap, 
                     struct _t3ws__StartControl *params, 
                     struct _t3ws__StartControlResponse *dummyRetVal)
{
    // change root module
    tciRootModule(params->moduleId);

    tciStartControl();

    gTestExecutableState = t3ws__EXECUTING_CONTROL;

    return SOAP_OK;
}

//
// StopControl
//

SOAP_FMAC5 int SOAP_FMAC6 
__t3sa__StopControl(struct soap* soap, 
                    struct _t3ws__StopControl *dummy, 
                    struct _t3ws__StopControlResponse *dummyRetVal)
{
    tciStopControl();

    return SOAP_OK;
}

//
// GetTestExecutableState
//

SOAP_FMAC5 int SOAP_FMAC6 
__t3sa__GetTestExecutableState(struct soap* soap, 
                               struct _t3ws__GetTestExecutableState *dummy, 
			       struct _t3ws__GetTestExecutableStateResponse *retVal)
{
    retVal->state = gTestExecutableState;

    return SOAP_OK;
}

///////////////////////////////////////////////////////////////
// WSDL TM provided operations
//////////////////////////////////////////////////////////////

static void * gProtectEvents;
static void * gNewEvent;

#define WAIT(sem) t3pl_sem_wait(sem, NULL)
#define RELEASE(sem) t3pl_sem_post(sem, NULL)

// list of events
struct EventStorage {
    t3ws__EventInfo eventInfo;
    struct soap *soap;
    int eventId;
    struct EventStorage *next;
    struct EventStorage *prev;
};

static struct EventStorage *gEventStorageHead; // pointer to head
static struct EventStorage *gEventStorageTail; // pointer to tail

//
// AddEvent (add event to the list head)
//

static void AddEvent(struct soap *soap, t3ws__EventInfo *e)
{
    struct EventStorage *node; 
    int eventId;

    WAIT(gProtectEvents);

    if (!gEventStorageHead) {
        node = gEventStorageHead = gEventStorageTail = malloc(sizeof(struct EventStorage));

	node->eventId = 1;

        node->next = NULL;
        node->next = NULL;

    } else {
        node = malloc(sizeof(struct EventStorage));

	node->eventId = gEventStorageTail->eventId + 1;

        node->next = NULL;
        node->prev = gEventStorageTail;
        gEventStorageTail->next = node;
        gEventStorageTail = node;
    }

    node->soap = soap;
    node->eventInfo = *e;

    RELEASE(gProtectEvents);

    RELEASE(gNewEvent);
}

//
// GetEvent (removes event from the list tail)
//

static t3ws__EventInfo* GetEvent(struct soap *soap, int eventId) 
{
    struct EventStorage *node = NULL; 

    if (eventId < 1) return NULL;

    WAIT(gProtectEvents);

    node = gEventStorageTail;
    while(node && node->eventId > eventId)
        node = node->prev;

    RELEASE(gProtectEvents);

    if (node && node->eventId == eventId)
        return &node->eventInfo;
    else
        return NULL;
}

//
// TciWsInit
//
// should be called during test suite initialization (eg. from main)
//

void TciWsInit()
{
    if (!gProtectEvents)
        gProtectEvents = t3pl_sem_create(1, NULL);
    if (!gNewEvent)
        gNewEvent = t3pl_sem_create(0, NULL);
}

//
// ListenEvent
//

SOAP_FMAC5 int SOAP_FMAC6 
__t3sa__ListenEvent(struct soap* soap, 
                    struct _t3ws__ListenEvent *params, 
                    struct _t3ws__ListenEventResponse *retVal)
{
    retVal->eventInfo = GetEvent(soap, params->eventId);

    if (retVal->eventInfo == NULL && params->waitEvent == true_)
        while(retVal->eventInfo == NULL)
        {
            WAIT(gNewEvent);
            retVal->eventInfo = GetEvent(soap, params->eventId);
        }

    return SOAP_OK;
}

//
// Provided server functions
//

static bool tcRunning; // flags 

//
// tciError
//

void tciError(String message)
{
    struct soap *soap = soap_new();
    t3ws__EventInfo e;

    e.__union_EventInfo = SOAP_UNION__t3ws__union_EventInfo_errorInfo;
    e.union_EventInfo.errorInfo = SMALLOC(1, t3ws__ErrorInfo);

    e.union_EventInfo.errorInfo->errorMessage = SSTRDUP(message);

    AddEvent(soap, &e);

    if (tcRunning)
        tciStopTestCase();
    else 
        tciStopControl();
}

//
// tciTestCaseStarted
//

void tciTestCaseStarted(TciTestCaseIdType testCaseId, 
                        TciParameterListType parameterList, 
                        double timer)
{
    struct soap *soap = soap_new();
    t3ws__EventInfo e;

    e.__union_EventInfo = SOAP_UNION__t3ws__union_EventInfo_tcStartedInfo;
    e.union_EventInfo.tcStartedInfo = SMALLOC(1, t3ws__TestCaseStartedInfo);

    // testcase id
    e.union_EventInfo.tcStartedInfo->testCaseId = SMALLOC(1, t3ws__QualifiedName);
    e.union_EventInfo.tcStartedInfo->testCaseId->moduleName = SSTRDUP(testCaseId.moduleName);
    e.union_EventInfo.tcStartedInfo->testCaseId->objectName = SSTRDUP(testCaseId.objectName);

    //parameter list
    e.union_EventInfo.tcStartedInfo->parameterList = SMALLOC(1, t3ws__ParameterList);
    tci2ws_ParameterList(soap, e.union_EventInfo.tcStartedInfo->parameterList, parameterList);
     
    // timeout
    e.union_EventInfo.tcStartedInfo->timeout = NULL;
    if (timer > 0) {
        e.union_EventInfo.tcStartedInfo->timeout = SMALLOC(1, double);
        *e.union_EventInfo.tcStartedInfo->timeout = timer;
    }

    AddEvent(soap, &e);

    tcRunning = true;
}

//
// tciTestCaseTerminated
//

void tciTestCaseTerminated(TciValue verdict,
                           TciParameterListType parameterList)
{
    struct soap *soap = soap_new();
    t3ws__EventInfo e;

    e.__union_EventInfo = SOAP_UNION__t3ws__union_EventInfo_tcTerminatedInfo;
    e.union_EventInfo.tcTerminatedInfo = SMALLOC(1, t3ws__TestCaseTerminatedInfo);

    switch(tciGetVerdictValue(verdict)) {
    case TCI_VERDICT_NONE_STATIC:
        e.union_EventInfo.tcTerminatedInfo->verdict = t3ws__VERDICT_NONE;
        break;
    case TCI_VERDICT_PASS_STATIC:
        e.union_EventInfo.tcTerminatedInfo->verdict = t3ws__VERDICT_PASS;
        break;
    case TCI_VERDICT_INCONC_STATIC:
        e.union_EventInfo.tcTerminatedInfo->verdict = t3ws__VERDICT_INCONC;
        break;
    case TCI_VERDICT_FAIL_STATIC:
        e.union_EventInfo.tcTerminatedInfo->verdict = t3ws__VERDICT_FAIL;
        break;
    case TCI_VERDICT_ERROR_STATIC:
        e.union_EventInfo.tcTerminatedInfo->verdict = t3ws__VERDICT_ERROR;
        break;
    }

    //parameter list
    e.union_EventInfo.tcTerminatedInfo->parameterList = SMALLOC(1, t3ws__ParameterList);
    tci2ws_ParameterList(soap, e.union_EventInfo.tcTerminatedInfo->parameterList, parameterList);

    AddEvent(soap, &e);

    tcRunning = false;

    if (gTestExecutableState == t3ws__EXECUTING_TESTCASE)
        gTestExecutableState = t3ws__IDLE;
}

//
// tciControlTerminated
//

void tciControlTerminated()
{
    struct soap *soap = soap_new();
    t3ws__EventInfo e;

    e.__union_EventInfo = SOAP_UNION__t3ws__union_EventInfo_ctrlTerminatedInfo;
    e.union_EventInfo.ctrlTerminatedInfo = SMALLOC(1, t3ws__ControlTerminatedInfo);

    AddEvent(soap, &e);

    if (gTestExecutableState == t3ws__EXECUTING_CONTROL)
        gTestExecutableState = t3ws__IDLE;
}

//
// tciGetModulePar
//

TciValue tciGetModulePar(TciModuleParameterIdType parameterId)
{
    // we assume that all module parameters should be defined using conf files

    return NULL;
}

