/***************************************************************************************
 * TRI External function implementation (t3tri_external.c)
 *
 * 2010 IBM Corporation
 *
 * This example shows how to implement external functions using TRI and TCI interfaces.
 * It implements getCurrentDateTime external function as defined below:
 *
 * external function getCurrentDateTime(in charstring format, in integer max_length) return charstring
 *
 * This function accepts two parameters:
 *   - format specification defined as character string
 *   - maximum length of the resulting string
 * It returns current date and time formatted according to the provided format specification.
 *
 * Format specification should be done according to the ANSI C "strftime" function.
 *
 ***************************************************************************************/

#include "tci.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>


#ifdef EXT_FUNC_BYPASS_TRI_ExternalFuncExample_getCurrentDateTime
#error "EXT_FUNC_BYPASS_TRI_ExternalFuncExample_getCurrentDateTime is defined"
#endif

TriStatus 
triExternalFunction (const TriFunctionId *functionId,		     
                     TriParameterList    *parameterList,
                     TriParameter        *returnValue) 
{ 
  /*
   * Add implementations of other external functions here
  */ 

  if (strcmp(functionId->objectName, "getCurrentDateTime") == 0)
  {
    time_t ltime;
    struct tm *today;
    char *time_buf;
    char *format;
    int max_length;
    char *endptr;
    TciCharStringValue formatValue;
    struct _TciCharStringValue resultValue;
    TciValue par1, par2, ret;

    /* check parameter list length (we have 2 parameters) */
    if (parameterList->length != 2)
      return TRI_Error;
    
    /* STEP 1: PROCESS(DECODE) IN AND INOUT PARAMETERS */
    
    /* decode first parameter - date/time format specification string */
    par1 = tciDecode(parameterList->parList[0]->par, tciGetTciCharstringType());
    if (!par1)
      return TRI_Error; /* decode failed */

    /* extract TciCharStringValue object (TCI charstring representation) from abstract TciValue */  
    formatValue = tciGetCStringValue(par1);
    if (!formatValue)
      return TRI_Error; /* value is not correct */

    /* prepare format string for strftime() function ('\0' terminated string) */
    format = malloc(formatValue->length + 1);
    memcpy(format, formatValue->string, formatValue->length);
    format[formatValue->length] = '\0';
      
    /* decode second parameter - maximum length of the resulting value */
    par2 = tciDecode(parameterList->parList[1]->par, tciGetIntegerType());
    if (!par2)
      return TRI_Error; /* decode failed */
    
    /* check that value is positive and not zero length */
    if (tciGetIntSign(par2) == 0 || tciGetIntNumberOfDigits(par2) == 0) 
      return TRI_Error;

    /* convert string into integer (integers in TCI are representaed as strings of digits) */  
    max_length = strtol(tciGetIntAbs(par2), &endptr, 10);
    if (*endptr != '\0')
      return TRI_Error; /* overflow */
      
    /* STEP 2: PERFORM NECESSARY CALCULATIONS */

    /* call standard ANSI C routines to obtain current date/time */  
    time(&ltime);
    today = localtime(&ltime);

    /* format date/time value according to provided format string */
    time_buf = malloc(max_length + 1);
    strftime(time_buf, max_length, format, today);
    
    /* STEP 2: PREPARE(ENCODE) INOUT, OUT PARAMETERS AND RETURN VALUE */
    
    /* create fresh charstring TCI value */
    ret = tciNewInstance(tciGetTciCharstringType());
    if (!ret) 
      return TRI_Error; /* failed creating value */

    /* prepare return value of TciCharStringValue type */
    resultValue.length = strlen(time_buf);
    resultValue.string = malloc(resultValue.length);
    memcpy(resultValue.string, time_buf, resultValue.length);
      
    /* fill TciValue object with TciCharStringValue data */  
    tciSetCStringValue(ret, &resultValue);
    
    /* encode return value */
    returnValue->mode = TRI_OUT;
    returnValue->par = tciEncode(ret);

    /* free data */
    free(format);
    free(time_buf);
    free(resultValue.string);
     
    return TRI_OK;
  }

  return TRI_Error; 
}

/**************************************************************************
 *
 * TE -> SA
 *
 * Function to be implemented by the System Adaptor
 *
 */

TriStatus 
triSAReset() 
{ 
  return TRI_OK; 
}

TriStatus 
triExecuteTestcase (const TriTestCaseId *testCaseId,  
                    const TriPortIdList *tsiPortList) 
{
  return TRI_OK; 
}

TriStatus 
triEndTestcase () 
{
  return TRI_OK; 
}

TriStatus 
triMap  (const TriPortId *compPortId,  
         const TriPortId *tsiPortId) 
{ 
  return TRI_Error; 
}


TriStatus 
triUnmap (const TriPortId *compPortId, 
          const TriPortId *tsiPortId) 
{ 
  return TRI_Error; 
}


TriStatus 
triSend (const TriComponentId *componentId, 
         const TriPortId      *tsiPortId, 
         const TriAddress     *sutAddress, 
         const TriMessage     *sendMessage) 
{
  return TRI_Error;
}


TriStatus 
triSendMC (const TriComponentId *componentId,
           const TriPortId      *tsiPortId,
           const TriAddressList *sutAddresses,
           const TriMessage     *sendMessage)
{
  return TRI_Error;
}           


TriStatus
triSendBC (const TriComponentId *componentId,
           const TriPortId      *tsiPortId,
           const TriMessage     *sendMessage)
{
  return TRI_Error;
}           


TriStatus 
triCall (const TriComponentId   *componentId, 
         const TriPortId        *tsiPortId,  
         const TriAddress       *sutAddress,  
         const TriSignatureId   *signatureId,  
         const TriParameterList *parameterList) 
{
  return TRI_Error; 
}


TriStatus
triCallMC (const TriComponentId   *componentId,
           const TriPortId        *tsiPortId, 
           const TriAddressList   *sutAddresses, 
           const TriSignatureId   *signatureId, 
           const TriParameterList *parameterList)
{
  return TRI_Error; 
}           


TriStatus
triCallBC (const TriComponentId   *componentId,
           const TriPortId        *tsiPortId,
           const TriSignatureId   *signatureId,
           const TriParameterList *parameterList)
{
  return TRI_Error; 
}           


TriStatus 
triReply (const TriComponentId    *componentId, 
          const TriPortId         *tsiPortId,  
          const TriAddress        *sutAddress,  
          const TriSignatureId    *signatureId,  
          const TriParameterList  *parameterList,  
          const TriParameter      *returnValue) 
{ 
  return TRI_Error; 
}


TriStatus 
triReplyMC (const TriComponentId   *componentId,
            const TriPortId        *tsiPortId,
            const TriAddressList   *sutAddresses,
            const TriSignatureId   *signatureId,
            const TriParameterList *parameterList,
            const TriParameter     *returnValue)
{
  return TRI_Error; 
}
            

TriStatus
triReplyBC (const TriComponentId   *componentId,
            const TriPortId        *tsiPortId,
            const TriSignatureId   *signatureId,
            const TriParameterList *parameterList,
            const TriParameter     *returnValue)
{
  return TRI_Error; 
}
            

TriStatus
triRaise (const TriComponentId  *componentId,
          const TriPortId       *tsiPortId,
          const TriAddress      *sutAddress,
          const TriSignatureId  *signatureId,
          const TriException    *exception) 
{ 
  return TRI_Error; 
}


TriStatus
triRaiseMC (const TriComponentId *componentId,
            const TriPortId      *tsiPortId,
            const TriAddressList *sutAddresses,
            const TriSignatureId *signatureId,
            const TriException   *exception)
{
  return TRI_Error; 
}
            

TriStatus
triRaiseBC (const TriComponentId *componentId,
            const TriPortId      *tsiPortId,
            const TriSignatureId *signatureId,
            const TriException   *exception)
{
  return TRI_Error; 
}            

            
TriStatus 
triSUTActionInformal (char* description) 
{ 
  return TRI_Error; 
}


TriStatus 
triSUTActionTemplate (const TriActionTemplate* templateValue) 
{ 
  return TRI_Error; 
}


/**************************************************************************
 *
 * TE -> PA
 *
 * Function to be implemented by the Platform Adaptor
 *
 */

TriStatus
triPAReset() 
{ 
  return TRI_OK; 
}


TriStatus
triStartTimer (const TriTimerId *timerId,
               TriTimerDuration  timerDuration) 
{ 
  return TRI_Error; 
}


TriStatus
triStopTimer (const TriTimerId *timerId) 
{ 
  return TRI_Error; 
}


TriStatus 
triReadTimer (const TriTimerId *timerId,
              TriTimerDuration *elapsedTime) 
{ 
  return TRI_Error; 
}


TriStatus
triTimerRunning (const TriTimerId *timerId,
                 unsigned char    *running) 
{ 
  return TRI_Error; 
}

