/***************************************************************************************
 * TRI Unix native (real-time) timer implementation (triUnixNativeTimers.c)
 *
 * 2010 IBM Corporation
 *
 * This is an example implementation of TRI Platform Adapter (except triExternalFunction)
 * for Linux and Solaris platforms.
 *
 * This file implements TRI timer routines basing on POSIX 1003.b real-time timer functions.
 *
 * When building this example consider several notes:
 * 1) Test suite executable has to be linked with 'librt.a' library (use -lrt switch)
 *
 * 2) CLOCK_ID (by default defined as CLOCK_REALTIME) denotes the clock used by hardware timer routines.
 *    On some systems CLOCK_REALTIME may have resolution only 10 ms.
 *    You may decide to use another clock (e.g. CLOCK_HIGHRES on Solaris or CLOCK_REALTIME_HR on Linux)
 *    as long as it's supported on your system.
 *
 *    Note that on Solaris using of high resolution timers may be allowed only for "root"
 *    (unless the tunable "clock_highres_nonroot" is set to a non-zero value in /etc/system).
 *
 * 3) This implementation uses SIGUSR1 signal for a timeout signal. If you are using this signal
 *    for ther purposes then change TIMER_SIGNAL_ID definition
 *
 * 4) The amount of real-time timers that may be active simultaneously is platform dependent.
 *    If you receive TRI_Error from triStartTimer then check that timer limit is not exceeded.
 *
 ***************************************************************************************/

#include <unistd.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/signal.h>
#include <sys/errno.h>
#include <semaphore.h>

#include "t3tri.h"

#define CLOCK_ID CLOCK_REALTIME

#define TIMER_SIGNAL_ID SIGUSR1

#define NANOSECONDS_IN_SEC 1000000000

#define unt_assert(cond, msg) if (!(cond)) {fprintf(stderr, "%s\n", msg); fflush(stderr); exit(-1);}

typedef struct _UNTTimerData
{
	TriTimerId        id;          /* TRI timer id */
	timer_t           handle;      /* OS timer handle */
	struct timespec   start;       /* start time of the timer (used in triReadTimer) */
	int               is_running;  /* timer state */
} UNTTimerData;

typedef struct _UNTTimerQueue
{
    UNTTimerData       **timers;         /* timer queue */
    unsigned long        size;           /* number of timers in the queue */
    unsigned long        allocated;      /* number of allocated elements */
    sem_t               *mutex;          /* mutex that protects access to timer queue */
} UNTTimerQueue;

static UNTTimerQueue gloTimerQueue; /* global queue of timers */


sem_t *gmutex; /* points to queue->mutex */

/* forward declarations */
static int stopTimer(UNTTimerQueue *queue, UNTTimerData *pTimer);

/*
 * Stops all timers and deallocates resources
 */
 
static void 
freeTimerQueue(UNTTimerQueue *queue)
{
    if (queue->mutex)
    {
      sem_destroy(queue->mutex);
      free(queue->mutex);
      queue->mutex = NULL;
    }
    
    if (queue->timers)
    {
      unsigned int i;
      for (i = 0; i < queue->size; i++)
      {
        stopTimer(queue, queue->timers[i]);
        free(queue->timers[i]->id.data);
        free(queue->timers[i]);
      }
      free(queue->timers);
      queue->timers = NULL;
      queue->size = 0;
      queue->allocated = 0;
    }   
}

/*
 * Initializes timer queue
 */

static int 
initTimerQueue(UNTTimerQueue *queue)
{
    int status;
    
    unt_assert(queue, "initTimerQueue: Unexpected NULL value");
    
    queue->allocated = 100;
    queue->size = 0;
    queue->timers = malloc(queue->allocated * sizeof(UNTTimerData*));
    unt_assert(queue->timers, "initTimerQueue: Memory allocation error");

    queue->mutex = malloc(sizeof(sem_t));
    unt_assert(queue->mutex, "initTimerQueue: Memory allocation error");
    status = sem_init(queue->mutex, 0, 1);

    gmutex = queue->mutex;


    return status == 0;
}

/*
 * Acquires mutex
 */

static int
acquireMutex(sem_t *mutex)
{
    return sem_wait(mutex) == 0;
}

/*
 * Releases mutex
 */

static int
releaseMutex(sem_t *mutex)
{
    return sem_post(mutex) == 0;
}

/* Right mask has zeroes to the right "00000000", "10000000", 
   "11000000", etc. Index is same as "number of ones (1) to the left. */

static const unsigned char right_mask[9] = { 0, 128, 192, 224, 240, 248, 252, 254, 255 };

/*
 * Compares two timers
 */

static int
timersEqual(const TriTimerId *id1,
            const TriTimerId *id2)
{
    unsigned int bytesize, bitsize;
    unsigned char c1, c2;

    unt_assert(id1, "timersEqual: Unexpected NULL value");
    unt_assert(id2, "timersEqual: Unexpected NULL value");
    
    if (id1->bits != id2->bits) 
      return 0;

    bytesize = (unsigned int)id1->bits / 8;
    if (memcmp(id1->data, id2->data, bytesize) != 0)
      return 0;
  
    bitsize = (unsigned int)id1->bits % 8;
    if (!bitsize)
      return 1;
    c1 = id1->data[bytesize] & right_mask[bitsize];
    c2 = id2->data[bytesize] & right_mask[bitsize];
    return c1 == c2;
}

/*
 * Copies timer id
 */
 
static void 
copyTimerId(TriTimerId *dest, const TriTimerId *src)
{
    unsigned int bytesize;
    
    bytesize = (unsigned int)src->bits / 8;
    if ((unsigned int)src->bits % 8 > 0)
      bytesize++;
      
    dest->bits = src->bits;
    dest->data = malloc(bytesize);
    unt_assert(dest->data, "Memory allocation error");
    memcpy(dest->data, src->data, bytesize);
}

/*
 * Locates timer in the queue using its id
 */

static UNTTimerData*
findTimer(const UNTTimerQueue *queue, const TriTimerId *timerId)
{
    int i;
    int bResult;
    UNTTimerData *pTimer = NULL;
    
    bResult = acquireMutex(queue->mutex);
    unt_assert(bResult, "Failed to acquire timer queue mutex")

    for (i = 0; i < queue->size && !pTimer; i++)
    {
      if (timersEqual(&queue->timers[i]->id, timerId))
        pTimer = queue->timers[i];
    }
    
    bResult = releaseMutex(queue->mutex);
    unt_assert(bResult, "Failed to release timer queue mutex")
    
	return pTimer;
}

/*
 * Creates new timer object
 */

static UNTTimerData*
newTimer(UNTTimerQueue *queue, const TriTimerId *timerId)
{
    int bResult;
    UNTTimerData *pTimer = NULL;

    pTimer = malloc(sizeof(UNTTimerData));
    unt_assert(pTimer, "Memory allocation error");
    
    memset(pTimer, 0, sizeof(UNTTimerData));
    copyTimerId(&pTimer->id, timerId);
	
    bResult = acquireMutex(queue->mutex);
    unt_assert(bResult, "Failed to acquire timer queue mutex")

    if (queue->size == queue->allocated)
    {
      queue->allocated *= 2;
      queue->timers = realloc(queue->timers, sizeof(UNTTimerData*) * queue->allocated);
    }
    
    queue->timers[queue->size++] = pTimer;
    
    bResult = releaseMutex(queue->mutex);
    unt_assert(bResult, "Failed to release timer queue mutex")
	
	return pTimer;
}

/*
 * Stops timer
 */

static int
stopTimer(UNTTimerQueue *queue, UNTTimerData *pTimer)
{
    int status = 0;

    acquireMutex(queue->mutex);
      
    if(pTimer->is_running)
      status = timer_delete(pTimer->handle);

    pTimer->is_running = 0;
  
    releaseMutex(queue->mutex);
  	
	return status == 0;
}

/*
 * Timer callback routine
 */

static void
TIMER_SIGNAL_HANDLER(int signo, siginfo_t *info, void *extra)
{
    UNTTimerData *pTimer = (UNTTimerData*)info->si_value.sival_ptr;

    unt_assert(signo == TIMER_SIGNAL_ID, "TIMER_SIGNAL_HANDLER: Unexpected signal id");
    unt_assert(pTimer,                   "TIMER_SIGNAL_HANDLER: Unexpected NULL value");
	
    acquireMutex(gmutex);

    if(pTimer->is_running)
    {
      triTimeout(&pTimer->id);
      timer_delete(pTimer->handle);
    }

    pTimer->is_running = 0;

    releaseMutex(gmutex);

}


/* 
 * Platform Adapter reset function
 */ 

TriStatus
triPAReset() 
{ 
    struct sigaction sig_act;
    
    freeTimerQueue(&gloTimerQueue);
    
    if (!initTimerQueue(&gloTimerQueue))
      return TRI_Error;

    sig_act.sa_flags = SA_SIGINFO;
    sig_act.sa_sigaction = TIMER_SIGNAL_HANDLER;
    sigemptyset(&sig_act.sa_mask);
    sigaction(TIMER_SIGNAL_ID, &sig_act, 0);
      
	return TRI_OK;
}

/* 
 * TRI timer start function
 */ 

TriStatus
triStartTimer (const TriTimerId *timerId,
               TriTimerDuration  timerDuration) 
{ 
    int status;
    struct sigevent evp;
    struct itimerspec timeout;
    UNTTimerData *pTimer = findTimer(&gloTimerQueue, timerId);

    if(pTimer) 
    {
      /* timer exists => restart */
      if(!stopTimer(&gloTimerQueue, pTimer))
        return TRI_Error;
    }
    else 
	{
      /* create new timer */
      pTimer = newTimer(&gloTimerQueue, timerId);
	}

	if (timerDuration == 0.0)
	{
      /* passing zeros to timer_settime simply disables timer so force timeout manually */
      triTimeout(timerId);
      return TRI_OK;
	}
	
    evp.sigev_value.sival_ptr = pTimer;
    evp.sigev_signo = TIMER_SIGNAL_ID;
    evp.sigev_notify = SIGEV_SIGNAL;
    
    status = timer_create(CLOCK_ID, &evp, &pTimer->handle);
    if (status == -1)
      return TRI_Error;

    timeout.it_value.tv_sec = (unsigned long)timerDuration;
    timeout.it_value.tv_nsec = (unsigned long)((timerDuration - timeout.it_value.tv_sec) * NANOSECONDS_IN_SEC);
    timeout.it_interval.tv_sec = 0;
    timeout.it_interval.tv_nsec = 0;

    pTimer->is_running = 1;
      
    status = timer_settime(pTimer->handle, 0 /*relatime timer*/, &timeout, NULL);
    if (status == -1)
      return TRI_Error;
      
    status = clock_gettime(CLOCK_ID, &pTimer->start);
    if (status == -1)
      return TRI_Error;

	return TRI_OK; 
}

/* 
 * TRI timer stop function
 */ 

TriStatus
triStopTimer (const TriTimerId *timerId) 
{ 
    UNTTimerData *pTimer = findTimer(&gloTimerQueue, timerId);

    if(pTimer)
      return stopTimer(&gloTimerQueue, pTimer) ? TRI_OK : TRI_Error;

    /* It's OK to stop not running timer */
    return TRI_OK;
}

/* 
 * TRI timer read function
 */ 

TriStatus 
triReadTimer (const TriTimerId *timerId,
              TriTimerDuration *elapsedTime) 
{ 
    UNTTimerData *pTimer = findTimer(&gloTimerQueue, timerId);

    *elapsedTime = 0.0;

    /* It's OK to query not running timer, return zero in this case */
    if(pTimer && pTimer->is_running) 
    {
      int status;
      struct timespec now;
      
      status = clock_gettime(CLOCK_ID, &now);
      if (status == -1)
        return TRI_Error;

      *elapsedTime = (TriTimerDuration)(now.tv_sec - pTimer->start.tv_sec) + 
                     (TriTimerDuration)(now.tv_nsec - pTimer->start.tv_nsec) / NANOSECONDS_IN_SEC;
	} 

	return TRI_OK;
}

/* 
 * TRI is timer running function
 */ 

TriStatus
triTimerRunning (const TriTimerId *timerId,
                 unsigned char    *running) 
{ 
    UNTTimerData *pTimer = findTimer(&gloTimerQueue, timerId);

	*running = pTimer && pTimer->is_running;

    /* It's OK to query not running timer */
    return TRI_OK; 
}

/* EOF */
