/***************************************************************************************
 * TRI Windows native (real-time) timer implementation (triWindowsNativeTimers.c)
 *
 * 2010 IBM Corporation
 *
 * This is an example implementation of TRI Platform Adapter (except triExternalFunction)
 * for Win32 platform.
 *
 * This file implements TRI timer routines basing on Win32 Platform SDK provided timer API
 *
 ***************************************************************************************/

#ifndef _WIN32_WINNT
/* This is necessary to include Platform SDK headers */
#define _WIN32_WINNT 0x0501
#endif

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "t3tri.h"

#define wnt_assert(cond, msg) if (!(cond)) {fprintf(stderr, "%s\n", msg); fflush(stderr); exit(-1);}

#define MSEC_IN_SEC 1000

typedef struct _WNTTimerData
{
	TriTimerId       id;          /* TRI timer id */
	HANDLE           handle;      /* OS timer handle */
	LARGE_INTEGER    start;       /* start time of the timer (used in triReadTimer) */
	BOOL             is_running;  /* timer state */
} WNTTimerData;

typedef struct _WNTTimerQueue
{
    WNTTimerData       **timers;         /* timer queue */
    unsigned long        size;           /* number of timers in the queue */
    unsigned long        allocated;      /* number of allocated elements */
    HANDLE               handle;         /* handle of the timer queue */
    LARGE_INTEGER        frequency;      /* the frequency of the high-resolution performance counter */
    HANDLE               mutex;          /* mutex that protects access to timer queue */
} WNTTimerQueue;

static WNTTimerQueue gloTimerQueue; /* global queue of timers */

/*
 * Stops all timers and deallocates resources
 */
 
static void 
freeTimerQueue(WNTTimerQueue *queue)
{
    if (queue->handle)
    {
      DeleteTimerQueueEx(queue->handle, INVALID_HANDLE_VALUE);
      queue->handle = NULL;
    }
    if (queue->mutex)
    {
      CloseHandle(queue->mutex);
      queue->mutex = NULL;
    }
    if (queue->timers)
    {
      unsigned int i;
      for (i = 0; i < queue->size; i++)
      {
        free(queue->timers[i]->id.data);
        free(queue->timers[i]);
      }
      free(queue->timers);
      queue->timers = NULL;
      queue->size = 0;
      queue->allocated = 0;
    }   
}

/*
 * Initializes timer queue
 */

static BOOL 
initTimerQueue(WNTTimerQueue *queue)
{
    BOOL bResult;
    
    wnt_assert(queue, "initTimerQueue: Unexpected NULL value");
    
    queue->allocated = 100;
    queue->size = 0;
    queue->timers = malloc(queue->allocated * sizeof(WNTTimerData*));
    wnt_assert(queue->timers, "initTimerQueue: Memory allocation error");
   
    queue->mutex = CreateMutex(NULL, FALSE, NULL);
    queue->handle = CreateTimerQueue();
    bResult = QueryPerformanceFrequency(&queue->frequency);
    
    return queue->mutex && queue->handle && bResult;
}

/*
 * Acquires mutex
 */

static BOOL 
acquireMutex(HANDLE mutex, DWORD timeout)
{
    return WaitForSingleObject(mutex, timeout) == WAIT_OBJECT_0;
}

/*
 * Releases mutex
 */

static BOOL
releaseMutex(HANDLE mutex)
{
    return ReleaseMutex(mutex);
}

/* Right mask has zeroes to the right "00000000", "10000000", 
   "11000000", etc. Index is same as "number of ones (1) to the left. */

static const unsigned char right_mask[9] = { 0, 128, 192, 224, 240, 248, 252, 254, 255 };

/*
 * Compares two timers
 */

static BOOL
timersEqual(const TriTimerId *id1,
            const TriTimerId *id2)
{
    unsigned int bytesize, bitsize;
    unsigned char c1, c2;

    wnt_assert(id1, "timersEqual: Unexpected NULL value");
    wnt_assert(id2, "timersEqual: Unexpected NULL value");
    
    if (id1->bits != id2->bits) 
      return FALSE;

    bytesize = (unsigned int)id1->bits / 8;
    if (memcmp(id1->data, id2->data, bytesize) != 0)
      return FALSE;
  
    bitsize = (unsigned int)id1->bits % 8;
    if (!bitsize)
      return TRUE;
    c1 = id1->data[bytesize] & right_mask[bitsize];
    c2 = id2->data[bytesize] & right_mask[bitsize];
    return c1 == c2;
}

/*
 * Copies timer id
 */
 
static void 
copyTimerId(TriTimerId *dest, const TriTimerId *src)
{
    unsigned int bytesize;
    
    bytesize = (unsigned int)src->bits / 8;
    if ((unsigned int)src->bits % 8 > 0)
      bytesize++;
      
    dest->bits = src->bits;
    dest->data = malloc(bytesize);
    wnt_assert(dest->data, "Memory allocation error");
    memcpy(dest->data, src->data, bytesize);
}

/*
 * Locates timer in the queue using its id
 */

static WNTTimerData*
findTimer(const WNTTimerQueue *queue, const TriTimerId *timerId)
{
    int i;
    BOOL bResult;
    WNTTimerData *pTimer = NULL;
    
    bResult = acquireMutex(queue->mutex, 5000);
    wnt_assert(bResult, "Failed to acquire timer queue mutex")

    for (i = 0; i < queue->size && !pTimer; i++)
    {
      if (timersEqual(&queue->timers[i]->id, timerId))
        pTimer = queue->timers[i];
    }
    
    bResult = releaseMutex(queue->mutex);
    wnt_assert(bResult, "Failed to release timer queue mutex")
    
	return pTimer;
}

/*
 * Creates new timer object
 */

static WNTTimerData*
newTimer(WNTTimerQueue *queue, const TriTimerId *timerId)
{
    BOOL bResult;
    WNTTimerData *pTimer = NULL;

    pTimer = malloc(sizeof(WNTTimerData));
    wnt_assert(pTimer, "Memory allocation error");
    
    memset(pTimer, 0, sizeof(WNTTimerData));
    copyTimerId(&pTimer->id, timerId);
	
    bResult = acquireMutex(queue->mutex, 5000);
    wnt_assert(bResult, "Failed to acquire timer queue mutex")

    if (queue->size == queue->allocated)
    {
      queue->allocated *= 2;
      queue->timers = realloc(queue->timers, sizeof(WNTTimerData*) * queue->allocated);
    }
    
    queue->timers[queue->size++] = pTimer;
    
    bResult = releaseMutex(queue->mutex);
    wnt_assert(bResult, "Failed to release timer queue mutex")
	
	return pTimer;
}

/*
 * Stops timer
 */

static BOOL
stopTimer(WNTTimerQueue *queue, WNTTimerData *pTimer)
{
    BOOL bResult = TRUE;
      
    if(pTimer->is_running)
      bResult = DeleteTimerQueueTimer(queue, pTimer->handle, INVALID_HANDLE_VALUE);

    pTimer->is_running = FALSE;
  	
	return bResult;
}

/*
 * Timer callback routine
 */

static VOID CALLBACK 
TIMER_CALLBACK_FUNC(PVOID lpParam, BOOLEAN TimerOrWaitFired)
{
    WNTTimerData *pTimer = (WNTTimerData*)lpParam;

    wnt_assert(pTimer, "TIMER_CALLBACK_FUNC: Unexpected NULL value");
	
    if(pTimer->is_running) 
      triTimeout(&pTimer->id);

    pTimer->is_running = FALSE;
}


/* 
 * Platform Adapter reset function
 */ 

TriStatus
triPAReset() 
{ 
    freeTimerQueue(&gloTimerQueue);
    
    if (!initTimerQueue(&gloTimerQueue))
      return TRI_Error;

	return TRI_OK;
}

/* 
 * TRI timer start function
 */ 

TriStatus
triStartTimer (const TriTimerId *timerId,
               TriTimerDuration  timerDuration) 
{ 
    BOOL bResult;
    WNTTimerData *pTimer = findTimer(&gloTimerQueue, timerId);

    if(pTimer) 
    {
      /* timer exists => restart */
      if(!stopTimer(&gloTimerQueue, pTimer))
        return TRI_Error;
    }
    else 
	{
      /* create new timer */
      pTimer = newTimer(&gloTimerQueue, timerId);
	}

    pTimer->is_running = TRUE;
    
    bResult = CreateTimerQueueTimer(&pTimer->handle, 
                                    gloTimerQueue.handle, 
                                    TIMER_CALLBACK_FUNC, 
                                    (PVOID)pTimer,
                                    (DWORD)(ceil(timerDuration * MSEC_IN_SEC)), 
                                    0,
                                    WT_EXECUTEONLYONCE);
    
	bResult |= QueryPerformanceCounter(&pTimer->start);
	
	return bResult ? TRI_OK : TRI_Error; 
}

/* 
 * TRI timer stop function
 */ 

TriStatus
triStopTimer (const TriTimerId *timerId) 
{ 
    WNTTimerData *pTimer = findTimer(&gloTimerQueue, timerId);

    if(pTimer)
      return stopTimer(&gloTimerQueue, pTimer) ? TRI_OK : TRI_Error;

    /* It's OK to stop not running timer */
    return TRI_OK;
}

/* 
 * TRI timer read function
 */ 

TriStatus 
triReadTimer (const TriTimerId *timerId,
              TriTimerDuration *elapsedTime) 
{ 
    BOOL bResult = TRUE;
    WNTTimerData *pTimer = findTimer(&gloTimerQueue, timerId);

    *elapsedTime = 0.0;

    /* It's OK to query not running timer, return zero in this case */
    if(pTimer && pTimer->is_running) 
    {
      TriTimerDuration fnow;
      TriTimerDuration fstart;
      LARGE_INTEGER now;

      bResult = QueryPerformanceCounter(&now);

      fnow = (TriTimerDuration) now.QuadPart / gloTimerQueue.frequency.QuadPart;
      fstart = (TriTimerDuration) pTimer->start.QuadPart / gloTimerQueue.frequency.QuadPart;

      /* 
       * CreateTimerQueueTimer resolution is 1 msec while performance counter has much higher
       * resolution. To avoid strange values in "read" operation we round them to 1 msec.
       */ 
      *elapsedTime = (ceil(fnow * MSEC_IN_SEC) - floor(fstart * MSEC_IN_SEC)) / MSEC_IN_SEC;
	} 

	return bResult ? TRI_OK : TRI_Error;
}

/* 
 * TRI is timer running function
 */ 

TriStatus
triTimerRunning (const TriTimerId *timerId,
                 unsigned char    *running) 
{ 
    WNTTimerData *pTimer = findTimer(&gloTimerQueue, timerId);

	*running = pTimer && pTimer->is_running;

    /* It's OK to query not running timer */
    return TRI_OK; 
}

/* EOF */
