/*************************************************************************
 * Tutorial TRI Integration (testertutorial_tri.c)
 *
 * 2010 IBM Corporation
 *
 * This is the TRI integration provided as the solution to the Tau Tester 
 * Tutorial. 
 *
 *************************************************************************/

#include <stdio.h>

#ifdef _WIN32
#include <process.h>
#include <winsock2.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <pthread.h>
#endif

#include <time.h>

#include <t3rts.h>
#include "t3tri.h"

#include "tut_comm.h"
#include "tut_ports.h"

#define DEFAULT_SUT_PORT 30000
#define SUT_PORT_KEYNAME "t3rt.tutorial.sut_port"


/* This variable is accessible at certain times, in the libt3tri.lib.
   It is needed in this implementation to access values stored in the 
   RTconf configuration. See documentation for further information.
*/
extern t3rt_context_t current_ctx;


/* The list of ports. */
#ifdef _WIN32
static struct port_list_struct_t portlist = { NULL, (HANDLE)0 };
#else
static struct port_list_struct_t portlist = { NULL, PTHREAD_MUTEX_INITIALIZER};
#endif

/* The socket for SUT communication. */
static SOCKET         sut_socket = INVALID_SOCKET; 


/**************************************************************************
 *
 * TE -> SA
 *
 * Functions to be implemented by the System Adaptor
 *
 */

/***********************************************************
 *
 * triSAReset
 *
 * This function is called one time, in the initialization 
 * phase before the execution of the control part.
 *
 * It initializes the socket package and tries to connect to
 * the SUT via the (default or given) port number. The host 
 * is hard-coded to be the local machine. If the SUT is 
 * running on a separate machine this have to be changed.
 *
 * The port mapping representation is initialized and the 
 * socket reader thread is also created here.
 *
 ***********************************************************/

TriStatus 
triSAReset() 
{ 
  unsigned short sut_port = DEFAULT_SUT_PORT;
#ifdef _WIN32
  WSADATA wsaData;
#endif
  struct sockaddr_in sut_addr;
  struct hostent* localHost;
  char* localIP;
  int iResult;
#ifdef _WIN32  
  static HANDLE thread_id = 0;
#else
  static pthread_t thread_id = 0;
#endif  
  char this_host[256];
  reader_struct_t r_struct = (reader_struct_t)malloc(sizeof(struct reader_struct_t));

  if (sut_socket == INVALID_SOCKET) {
    /* Initialize Winsock. */
#ifdef _WIN32
    iResult = WSAStartup(MAKEWORD(2,2), &wsaData);
    if (iResult != NO_ERROR) {
      printf(TRI_PREFIX "Error at WSAStartup()\n");
      return TRI_Error;
    }
#endif
    /* Retreive SUT port from configuration. */
    if (current_ctx != NULL) {
      t3rt_value_t val = t3rt_rtconf_get_param(SUT_PORT_KEYNAME, current_ctx);
      if (val != t3rt_illegal_value_c)
        if (val != t3rt_no_value_c)
          sut_port = (unsigned short) t3rt_value_get_integer(val, current_ctx);
    }

    /* Get the local host information */
    gethostname(this_host, 255);
    localHost = gethostbyname(this_host);
    localIP = inet_ntoa (*(struct in_addr *)*localHost->h_addr_list);
  
    /* */
    sut_addr.sin_family = AF_INET;
    sut_addr.sin_addr.s_addr = inet_addr(localIP);
    sut_addr.sin_port = htons(sut_port);

    /* Create new sockets */ 
    sut_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  
    if (sut_socket == INVALID_SOCKET) {
#ifdef _WIN32
      printf(TRI_PREFIX "Error at socket(): %ld\n", WSAGetLastError() );
      WSACleanup();
#else
      printf(TRI_PREFIX "Error at socket()\n");
#endif
      return TRI_Error;
    }

    printf(TRI_PREFIX "Connecting ETS to SUT socket.\n");

    if (connect(sut_socket, (struct sockaddr *) &sut_addr, sizeof(sut_addr)) == SOCKET_ERROR) {
      printf(TRI_PREFIX "Failed to connect. Ensure that the SUT is running correctly.\n");
#ifdef _WIN32
      WSACleanup();
#endif
      return TRI_Error;
    }
  }

#ifdef _WIN32
  if (portlist.mutex == (HANDLE)NULL) {
    portlist.mutex = CreateMutex(NULL, FALSE, "portsMutex");
  }
#endif

  if (thread_id == 0) {
    r_struct->portlist = &portlist;
    r_struct->socket = sut_socket;
#ifdef _WIN32
    thread_id = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)tut_socket_reader, (void*)r_struct, 0, NULL);
#else
    pthread_create(&thread_id, NULL, tut_socket_reader, (void*)r_struct);
#endif
  }

  return TRI_OK; 
}


/***********************************************************
 *
 * triExecuteTestcase
 *
 * This function is called, not surprisingly, before each
 * test case is executed. In this implementation it clears 
 * all the port mappings left from previous test cases. 
 * They reside here if no proper unmap statament was made.
 * A warning is issused if any mappings are found.
 *
 ***********************************************************/

TriStatus 
triExecuteTestcase (const TriTestCaseId *testCaseId,  
                    const TriPortIdList *tsiPortList) 
{
  if (portlist.list != NULL) {
    printf(TRI_PREFIX 
          "Warning: Mapped ports present when starting test case %s.\n",
          testCaseId->objectName);

    tut_clear_port_mappings(&portlist);
  }

  return TRI_OK;
}

/***********************************************************
 *
 * triEndTestcase
 *
 * This function is called, not surprisingly, after each
 * test case is executed. In this implementation it clears 
 * all the port mappings. They reside here if no proper 
 * unmap statament was made.
 *
 ***********************************************************/

TriStatus 
triEndTestcase () 
{
  tut_clear_port_mappings(&portlist);

  return TRI_OK;
}

/***********************************************************
 *
 * triMap
 *
 * Creates a new port mapping and store it. 
 *
 ***********************************************************/

TriStatus 
triMap  (const TriPortId *compPortId,  
         const TriPortId *tsiPortId) 
{ 
  port_mapping_t new_port = NULL;

  new_port = (port_mapping_t)malloc(sizeof(struct port_mapping_struct_t));

  set_port_mapping_comp_port(new_port, tut_copy_TriPortId(compPortId));
  
  if (new_port->comp_port == NULL) {
    printf(TRI_PREFIX "Error copying component port.\n");
#ifdef _WIN32
    WSACleanup();
#endif
    return TRI_Error;
  }

  set_port_mapping_tsi_port(new_port, tut_copy_TriPortId(tsiPortId));
  
  if (new_port->tsi_port == NULL) {
    printf(TRI_PREFIX "Error copying TSI port.\n");
#ifdef _WIN32
    WSACleanup();
#endif
    return TRI_Error;
  }

  tut_add_port_mapping(new_port, &portlist);

  return TRI_OK; 
}


/***********************************************************
 *
 * triUnmap
 *
 * Remove mapping from storage.
 *
 ***********************************************************/

TriStatus 
triUnmap (const TriPortId *compPortId, 
          const TriPortId *tsiPortId) 
{ 
  tut_delete_port_mapping(&portlist, compPortId, tsiPortId);

  return TRI_OK; 
}



/***********************************************************
 *
 * triSend
 *
 * Uses the help function to pack and send the 'send' 
 * statement to the SUT.
 *
 ***********************************************************/

TriStatus 
triSend (const TriComponentId *componentId, 
         const TriPortId      *tsiPortId, 
         const TriAddress     *sutAddress, 
         const TriMessage     *sendMessage) 
{
  if (sut_socket == INVALID_SOCKET) {
    printf(TRI_PREFIX "Socket invalid!\n");
#ifdef _WIN32
    WSACleanup();
#endif
    return TRI_Error;
  }

  tut_send_message(sut_socket, componentId, tsiPortId, sutAddress, sendMessage);

  return TRI_OK;
}

/***********************************************************
 *
 * triSendMC
 *
 * Not implemented in this tutorial.
 *
 ***********************************************************/

TriStatus 
triSendMC (const TriComponentId *componentId,
           const TriPortId      *tsiPortId,
           const TriAddressList *sutAddresses,
           const TriMessage     *sendMessage)
{
  printf(TRI_PREFIX "Multicast 'send' operation is not implemented.\n");

  return TRI_Error;
}


/***********************************************************
 *
 * triSendBC
 *
 * Not implemented in this tutorial.
 *
 ***********************************************************/

TriStatus
triSendBC (const TriComponentId *componentId,
           const TriPortId      *tsiPortId,
           const TriMessage     *sendMessage)
{
  printf(TRI_PREFIX "Broadcast 'send' operation is not implemented.\n");

  return TRI_Error;
}

/***********************************************************
 *
 * triCall
 *
 * Uses the help function to pack and send the 'call' 
 * statement to the SUT.
 *
 ***********************************************************/

TriStatus 
triCall (const TriComponentId   *componentId, 
         const TriPortId        *tsiPortId,  
         const TriAddress       *sutAddress,  
         const TriSignatureId   *signatureId,  
         const TriParameterList *parameterList) 
{
  if (sut_socket == INVALID_SOCKET) {
    printf(TRI_PREFIX "Socket invalid!\n");
#ifdef _WIN32
    WSACleanup();
#endif
    return TRI_Error;
  }

  tut_send_call(sut_socket, componentId, tsiPortId, sutAddress, signatureId, parameterList);

  return TRI_OK;
}

/***********************************************************
 *
 * triCallMC
 *
 * Not implemented in this tutorial.
 *
 ***********************************************************/

TriStatus
triCallMC (const TriComponentId   *componentId,
           const TriPortId        *tsiPortId, 
           const TriAddressList   *sutAddresses, 
           const TriSignatureId   *signatureId, 
           const TriParameterList *parameterList)
{
  printf(TRI_PREFIX "Multicast 'call' operation is not implemented.\n");

  return TRI_Error;
}

/***********************************************************
 *
 * triCallBC
 *
 * Not implemented in this tutorial.
 *
 ***********************************************************/

TriStatus
triCallBC (const TriComponentId   *componentId,
           const TriPortId        *tsiPortId,
           const TriSignatureId   *signatureId,
           const TriParameterList *parameterList)
{
  printf(TRI_PREFIX "Broadcast 'call' operation is not implemented.\n");

  return TRI_Error;
}

/***********************************************************
 *
 * triReply
 *
 * Uses the help function to pack and send the 'reply' 
 * statement to the SUT.
 *
 ***********************************************************/

TriStatus 
triReply (const TriComponentId    *componentId, 
          const TriPortId         *tsiPortId,  
          const TriAddress        *sutAddress,  
          const TriSignatureId    *signatureId,  
          const TriParameterList  *parameterList,  
          const TriParameter      *returnValue) 
{ 
  if (sut_socket == INVALID_SOCKET) {
    printf(TRI_PREFIX "Socket invalid!\n");
#ifdef _WIN32
    WSACleanup();
#endif
    return TRI_Error;
  }

  tut_send_reply(sut_socket, componentId, tsiPortId, sutAddress, 
                 signatureId, parameterList, returnValue);

  return TRI_OK;
}

/***********************************************************
 *
 * triReplyMC
 *
 * Not implemented in this tutorial.
 *
 ***********************************************************/

TriStatus 
triReplyMC (const TriComponentId   *componentId,
            const TriPortId        *tsiPortId,
            const TriAddressList   *sutAddresses,
            const TriSignatureId   *signatureId,
            const TriParameterList *parameterList,
            const TriParameter     *returnValue)
{
  printf(TRI_PREFIX "Multicast 'reply' operation is not implemented.\n");

  return TRI_Error;
}

/***********************************************************
 *
 * triReplyBC
 *
 * Not implemented in this tutorial.
 *
 ***********************************************************/

TriStatus
triReplyBC (const TriComponentId   *componentId,
            const TriPortId        *tsiPortId,
            const TriSignatureId   *signatureId,
            const TriParameterList *parameterList,
            const TriParameter     *returnValue)
{
  printf(TRI_PREFIX "Broadcast 'reply' operation is not implemented.\n");

  return TRI_Error;
}

/***********************************************************
 *
 * triRaise
 *
 * Not implemented in this tutorial. 
 *
 ***********************************************************/

TriStatus
triRaise (const TriComponentId  *componentId,
          const TriPortId       *tsiPortId,
          const TriAddress      *sutAddress,
          const TriSignatureId  *signatureId,
          const TriException    *exception) 
{ 
  printf(TRI_PREFIX "Unicast 'raise' operation is not implemented.\n");

  return TRI_Error; 
}

/***********************************************************
 *
 * triRaiseMC
 *
 * Not implemented in this tutorial.
 *
 ***********************************************************/

TriStatus
triRaiseMC (const TriComponentId *componentId,
            const TriPortId      *tsiPortId,
            const TriAddressList *sutAddresses,
            const TriSignatureId *signatureId,
            const TriException   *exception)
{
  printf(TRI_PREFIX "Multicast 'raise' operation is not implemented.\n");

  return TRI_Error;
}

/***********************************************************
 *
 * triRaiseBC
 *
 * Not implemented in this tutorial.
 *
 ***********************************************************/

TriStatus
triRaiseBC (const TriComponentId *componentId,
            const TriPortId      *tsiPortId,
            const TriSignatureId *signatureId,
            const TriException   *exception)
{
  printf(TRI_PREFIX "Broadcast 'raise' operation is not implemented.\n");

  return TRI_Error;
}

/***********************************************************
 *
 * triSUTActionInformal, triSUTActionTemplate
 *
 * Not implemented in this tutorial. 
 *
 ***********************************************************/

TriStatus 
triSUTActionInformal (char* description) 
{ 
  printf(TRI_PREFIX "Informal sut 'action' operation is not implemented.\n");

  return TRI_Error; 
}


TriStatus 
triSUTActionTemplate (const TriActionTemplate* templateValue) 
{ 
  printf(TRI_PREFIX "Template sut 'action' operation is not implemented.\n");

  return TRI_Error; 
}


/**************************************************************************
 *
 * TE -> PA
 *
 * This timer implementation uses a new thread for each started timer. 
 * All the thread does is to sleep for the timers duration and then issues 
 * a time out event (using triTimeout). 
 */

/* Timer representation */

typedef struct tut_timer_struct_t 
{
  unsigned long start_time;
  double        duration;
#ifdef _WIN32  
  HANDLE        thread;
#else
  pthread_t     thread;
#endif  
  TriTimerId    id;
} *tut_timer_t;


#define MAX_TIMERS_IN_TESTSUITE 5

struct timer_list_struct_t
{ 
  tut_timer_t timer_array[MAX_TIMERS_IN_TESTSUITE];
  unsigned int timer_in_list;
};

static struct timer_list_struct_t timer_list;

/***********************************************************
 *
 * triPAReset
 *
 * This function is called one time only, before the control 
 * part is executed.
 * The timer list is cleared.
 *
 ***********************************************************/

TriStatus
triPAReset() 
{ 
  unsigned int i;

  timer_list.timer_in_list = 0;
  for (i = 0; i < MAX_TIMERS_IN_TESTSUITE; ++i)
    timer_list.timer_array[i] = NULL;

  return TRI_OK;
}


/***********************************************************
 *
 * run_timer
 *
 * This function sleeps for the timer duration and generates
 * a timeout when it wakes up.
 *
 * The time to sleep is adjusted with the time it took for 
 * the thread to get to this point from when it was started.
 * This is the sec_delay variable. 
 *
 ***********************************************************/
#ifdef _WIN32
unsigned run_timer(void *t)
#else
void * run_timer(void *t)
#endif
{
  tut_timer_t timer = (tut_timer_t)t;
  double sec_delay  = (clock() - timer->start_time) / (double)CLOCKS_PER_SEC;
  double sleep_time = timer->duration - sec_delay;

#ifndef _WIN32
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
#endif

#ifdef DEBUG
  printf(TRI_PREFIX "Sleeping for %.4lf seconds\n", sleep_time);
#endif

  if (sleep_time > 0.0)
    Sleep((unsigned long)(sleep_time * 1000));

  triTimeout(&(timer->id));

  timer->duration = 0.0;
  timer->thread = 0;

#ifdef _WIN32
  return 0;
#else
  return NULL;
#endif
}

/***********************************************************
 *
 * find_timer
 *
 * Lookup function for timers.
 *
 ***********************************************************/

int
find_timer(struct timer_list_struct_t *list, 
           const TriTimerId *timerId)
{
  unsigned int i;

  for (i = 0; i < list->timer_in_list; ++i) {
    if (list->timer_array[i] != NULL && list->timer_array[i]->id.data != NULL && 
        tut_compare_binary(timerId, &(list->timer_array[i]->id)) == 0) 
      return i;
  }

  return -1;
}


/***********************************************************
 *
 * triStartTimer
 *
 * Creates a new timer instance, initializes it and creates 
 * the timeout thread for it.
 *
 ***********************************************************/

TriStatus
triStartTimer (const TriTimerId *timerId,
               TriTimerDuration  timerDuration) 
{ 
  tut_timer_t timer;
  unsigned long stime = clock();

  if (timer_list.timer_in_list == MAX_TIMERS_IN_TESTSUITE) {
    printf(TRI_PREFIX "Error in triStartTimer. No more room in timer list (maximum size is '%d' elements).\n", MAX_TIMERS_IN_TESTSUITE);
    return TRI_Error;
  }
  
  timer = (tut_timer_t)malloc(sizeof(struct tut_timer_struct_t));
  
  timer->start_time = stime;
  timer->duration   = timerDuration;
  tut_copy_binary(&(timer->id), timerId);
  timer_list.timer_array[timer_list.timer_in_list++] = timer;

#ifdef _WIN32
  timer->thread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)run_timer, (void*)timer, 0, NULL);
#else
  pthread_create(&timer->thread, NULL, run_timer, (void*)timer);
#endif
  
  return TRI_OK; 
}


/***********************************************************
 *
 * triStopTimer
 *
 * Looks up the timer, if ir is running, stop it and clear its
 * data.
 *
 ***********************************************************/

TriStatus
triStopTimer (const TriTimerId *timerId) 
{ 
  int i = find_timer(&timer_list, timerId);

  if (i >= 0) {
    tut_timer_t t = timer_list.timer_array[i];
    if (t->thread != 0) {
#ifdef _WIN32
      TerminateThread((HANDLE)t->thread, 1UL); 
#else
      pthread_cancel(t->thread);
#endif

      t->duration = 0.0;
      free(t->id.data);
      t->id.data = NULL;
      t->start_time = 0;
      t->thread = 0;
    }
  }

  return TRI_OK;
}


/***********************************************************
 *
 * triReadTimer
 *
 * Set elapsed_time to a proper value. If no timer is found 
 * or it has been stopped it is set to 0.0.
 *
 ***********************************************************/

TriStatus 
triReadTimer (const TriTimerId *timerId,
              TriTimerDuration *elapsedTime) 
{
  int i = find_timer(&timer_list, timerId);

  *elapsedTime = 0.0;

  if (i >= 0) {
    tut_timer_t t = timer_list.timer_array[i];
    if (t->thread != 0)
      *elapsedTime = (clock() - t->start_time) / (double)CLOCKS_PER_SEC;
  }

  return TRI_OK; 
}


/***********************************************************
 *
 * triTimerRunning
 *
 * The timer is running if the thread handle is non-null,
 * otherwie it is not. 
 *
 ***********************************************************/

TriStatus
triTimerRunning (const TriTimerId *timerId,
                 unsigned char    *running) 
{ 
  int i = find_timer(&timer_list, timerId);

  *running = 0;

  if (i >= 0) {
    tut_timer_t t = timer_list.timer_array[i];
    if (t->thread != 0)
      *running = 1;
  } 

  return TRI_OK; 
}


/***********************************************************
 *
 * triExternalFunction
 *
 * Not implemented in this tutorial.
 *
 ***********************************************************/

TriStatus 
triExternalFunction (const TriFunctionId *functionId,		     
                     TriParameterList    *parameterList,
                     TriParameter        *returnValue) 
{ 
  printf(TRI_PREFIX "Calling external functions is not implemented.\n");

  return TRI_Error; 
}

/* End of File */
