/*************************************************************************
 * Tutorial Codecs (tut_codecs.c)
 *
 * 2010 IBM Corporation
 *
 * This contains the functions necessary for encoding and decoding the 
 * non-ASN.1 defined types that are used in the tutorial test suite. 
 * The file also contain the functions necessary for registering the 
 * codecs module and the codecs functions.
 *
 * Types handled are:
 *  boolean
 *  integer
 *  bitstring 
 *  hexstring
 *
 *************************************************************************/

#include <string.h>
#include <stdlib.h>

#include "t3rts.h"


#define BOOLEAN_IDENTIFIER    "boolean"
#define INTEGER_IDENTIFIER    "integer"
#define BITSTRING_IDENTIFIER  "bitstring"
#define HEXSTRING_IDENTIFIER  "hexstring"

/***********************************************************
 *
 * encode_boolean, decode_boolean
 *
 * Encoding format: "boolean" 1|0 (1 byte) total 8 bytes.
 *
 ***********************************************************/

t3rt_codecs_result_t 
encode_boolean(t3rt_value_t value,
               t3rt_binary_string_t buf,
               t3rt_context_t ctx)
{
  t3rt_binary_string_append_nbytes(buf, BOOLEAN_IDENTIFIER, 
                                   strlen(BOOLEAN_IDENTIFIER), ctx);
  t3rt_binary_string_append_1byte(
    buf, 
    (unsigned char)(t3rt_value_get_boolean(value, ctx) ? 1 : 0),
    ctx);

  return t3rt_codecs_result_succeeded_c;
}

t3rt_codecs_result_t
decode_boolean(t3rt_binary_string_iter_t * iter,
               t3rt_type_t type,
               t3rt_alloc_strategy_t strategy,
               t3rt_value_t * decoded_value,
               t3rt_context_t ctx)
{
  char idbuf[32];
  unsigned char v;
  
  if (t3rt_bstring_iter_remaining_room(iter, ctx) < 
      (strlen(BOOLEAN_IDENTIFIER) * 8 + 1))
    return t3rt_codecs_result_not_applicable_c;

  t3rt_bstring_iter_get_nbytes(iter, strlen(BOOLEAN_IDENTIFIER), idbuf, ctx);
  if (strncmp(BOOLEAN_IDENTIFIER, idbuf, strlen(BOOLEAN_IDENTIFIER)) != 0)
    return t3rt_codecs_result_not_applicable_c;

  v = t3rt_bstring_iter_get_1byte(iter, ctx);

  *decoded_value = t3rt_type_instantiate_value(type, 
                                               strategy, ctx);

  t3rt_value_set_boolean(*decoded_value, (v == 1) ? true : false, ctx);

  return t3rt_codecs_result_succeeded_c;
}


/***********************************************************
 *
 * encode_integer, decode_integer
 *
 * Encoding format: <int> (4 bytes) total 4 bytes.
 *
 ***********************************************************/

t3rt_codecs_result_t 
encode_integer(t3rt_value_t value,
               t3rt_binary_string_t buf,
               t3rt_context_t ctx)
{
  t3rt_binary_string_append_4bytes(buf, 
                                   (unsigned long)t3rt_value_get_integer(value, ctx),
                                   ctx);

  return t3rt_codecs_result_succeeded_c;
}

t3rt_codecs_result_t
decode_integer(t3rt_binary_string_iter_t * iter,
               t3rt_type_t type,
               t3rt_alloc_strategy_t strategy,
               t3rt_value_t * decoded_value,
               t3rt_context_t ctx)
{
  long v = t3rt_bstring_iter_get_4bytes(iter, ctx);

  *decoded_value = t3rt_type_instantiate_value(type, 
                                               strategy, ctx);
  t3rt_value_set_integer(*decoded_value, v, ctx);

  return t3rt_codecs_result_succeeded_c;
}


/***********************************************************
 *
 * encode_bitstring, decode_bitstring
 *
 * Encoding format: <len> (4 bytes) <bits> (n bytes)  
 * total 4+n bytes.
 *
 ***********************************************************/

t3rt_codecs_result_t 
encode_bitstring(t3rt_value_t value,
                 t3rt_binary_string_t buf,
                 t3rt_context_t ctx)
{
  const char* str = t3rt_value_get_string(value, ctx);

  t3rt_binary_string_append_4bytes(buf, (unsigned long)strlen(str), ctx);
  t3rt_binary_string_append_nbytes(buf, str, strlen(str), ctx);

  return t3rt_codecs_result_succeeded_c;
}

t3rt_codecs_result_t
decode_bitstring(t3rt_binary_string_iter_t * iter,
                 t3rt_type_t type,
                 t3rt_alloc_strategy_t strategy,
                 t3rt_value_t * decoded_value,
                 t3rt_context_t ctx)
{
  long len = t3rt_bstring_iter_get_4bytes(iter, ctx);
  unsigned char *str;

  *decoded_value = t3rt_type_instantiate_value(type, 
                                               strategy, ctx);

  str = (unsigned char*)malloc(sizeof(unsigned char) * (len + 1));
  t3rt_bstring_iter_get_nbytes(iter, len, str, ctx);
  str[len] = '\0';

  t3rt_value_set_string(*decoded_value, str, ctx);
  free(str);

  return t3rt_codecs_result_succeeded_c;
}


/***********************************************************
 *
 * encode_hexstring, decode_hexstring
 *
 * Encoding format: <len> (4 bytes) <bits> (n bytes)  
 * total 4+n bytes.
 *
 ***********************************************************/

t3rt_codecs_result_t 
encode_hexstring(t3rt_value_t value,
                 t3rt_binary_string_t buf,
                 t3rt_context_t ctx)
{
  const char* str = t3rt_value_get_string(value, ctx);

  t3rt_binary_string_append_4bytes(buf, (unsigned long)strlen(str), ctx);
  t3rt_binary_string_append_nbytes(buf, str, strlen(str), ctx);

  return t3rt_codecs_result_succeeded_c;
}

t3rt_codecs_result_t
decode_hexstring(t3rt_binary_string_iter_t * iter,
                 t3rt_type_t type,
                 t3rt_alloc_strategy_t strategy,
                 t3rt_value_t * decoded_value,
                 t3rt_context_t ctx)
{
  long len = t3rt_bstring_iter_get_4bytes(iter, ctx);
  unsigned char *str;

  *decoded_value = t3rt_type_instantiate_value(type, 
                                               strategy, ctx);

  str = (unsigned char*)malloc(sizeof(unsigned char) * (len + 1));
  t3rt_bstring_iter_get_nbytes(iter, len, str, ctx);
  str[len] = '\0';

  t3rt_value_set_string(*decoded_value, str, ctx);
  free(str);

  return t3rt_codecs_result_succeeded_c;
}


/***********************************************************
 *
 * codecs_init
 *
 * Nothing is required to be initialized.
 *
 ***********************************************************/

void
codecs_init(t3rt_context_t ctx) 
{
  /* empty */ 
}


/***********************************************************
 *
 * codecs_setup
 *
 * This function is called for each type in the test suite. 
 * Codecs functions are registered for the selected builtin 
 * types only.
 *
 ***********************************************************/

void
codecs_setup(t3rt_type_t type, t3rt_context_t ctx)
{
  if (strcmp(t3rt_type_definition_name(type, ctx), "boolean") == 0) {
    t3rt_type_set_encoder(type, encode_boolean, ctx);
    t3rt_type_set_decoder(type, decode_boolean, ctx);
    return;      
  }

  if (strcmp(t3rt_type_definition_name(type, ctx), "integer") == 0) {
    t3rt_type_set_encoder(type, encode_integer, ctx);
    t3rt_type_set_decoder(type, decode_integer, ctx);
    return;      
  }

  if (strcmp(t3rt_type_definition_name(type, ctx), "bitstring") == 0) {
    t3rt_type_set_encoder(type, encode_bitstring, ctx);
    t3rt_type_set_decoder(type, decode_bitstring, ctx);
    return;      
  }

  if (strcmp(t3rt_type_definition_name(type, ctx), "hexstring") == 0) {
    t3rt_type_set_encoder(type, encode_hexstring, ctx);
    t3rt_type_set_decoder(type, decode_hexstring, ctx);
    return;      
  }
}


/***********************************************************
 *
 * t3ud_register_codecs
 *
 * Register the codecs "package" in the RTS. This function is
 * called once, in the beginning of the ETS execution, before 
 * any test behavior has been executed.
 *
 ***********************************************************/

void
t3ud_register_codecs (int argc, char * argv [], t3rt_context_t ctx)
{
  t3rt_codecs_register(codecs_init, codecs_setup, ctx);
}

