/*************************************************************************
 * Tutorial communication (tut_comm.c)
 *
 * 2010 IBM Corporation
 *
 * This contains the functions used for packaging and unpackaging the data 
 * to be sent over the SUT socket. It also handles the sending and reading.
 *
 * In the end of the file there is a socket reader function 
 * (tut_socket_reader) that is executed in its own thread, dispatching 
 * message to the appropriate handler function.
 *
 *************************************************************************/

#include "tut_comm.h"

#include <stdio.h>
#include <stdlib.h>

#include <t3tri.h>
#ifdef _WIN32
#include <winsock2.h>
#else
#include <sys/select.h>
#include <sys/types.h>
#include <sys/socket.h>
#endif


#define SEND_BUFFER_SIZE 256
#define RECV_BUFFER_SIZE 256

#define print_protocol_error()  fprintf(stderr, TRI_PREFIX "Protocol error (%d)!\n", __LINE__)


/***********************************************************
 *
 * Simple helper functions
 *
 ***********************************************************/

void
dump_vector(FILE *dump_file, char* buf, long len)
{
#ifdef DEBUG
  long i;
  if (dump_file == NULL) return;

  for (i = 0; i < len; ++i) {
    if (isalpha(buf[i]))
      fprintf(dump_file, "%c", buf[i]);
    else
      fprintf(dump_file, "x%02X", buf[i]);
  }

  fprintf(dump_file, "\n");
#endif
}

static char *
copy_string(const char *str)
{
  char *s;

  if (str == NULL) 
    return NULL;

  s = (char*)malloc(sizeof(char) * (strlen(str) + 1));

  return strcpy(s, str);
}

static int 
lengthof_string(const char* s)
{
  if (s == NULL)
    return 0;

  return strlen(s);
}

unsigned long
bits2bytes(unsigned long bits)
{
  unsigned long bytes = 0;

  if (bits == 0)
    return 0;
    
  bytes = (bits)/8;
  if (bits % 8 != 0)
    return bytes + 1;

  return bytes;
}


BinaryString *
tut_create_BinaryString(long bytes, char bits_in_last_byte, const char* message)
{
  BinaryString *str = (BinaryString *)malloc(sizeof(BinaryString));

  str->aux = NULL;
  str->bits = bytes * 8 + bits_in_last_byte;
  str->data = (unsigned char*)malloc(sizeof(unsigned char) * bytes);

  memcpy(str->data, message, bytes);

  return str;
}


/***********************************************************
 *
 * encode_BinaryString, decode_BinaryString
 *
 * A binary string is transfered as:
 * 
 * 1 byte  - number of bytes in bitstring
 * 1 byte  - number of bits used in the last byte
 * n bytes - the byte vector with the data
 *
 ***********************************************************/

unsigned long 
encode_BinaryString(const TriMessage* msg, char buf[])
{
  unsigned long bytes = 0;

  bytes = bits2bytes(msg->bits);
  buf[0] = (char)bytes;

  if (bytes > 0) {
    buf[1] = (char)(msg->bits % 8);
    if (buf[1] > 7) {
      printf(TRI_PREFIX "Error in bitlength encoding.\n");
      exit(100);
    }
    memcpy(&(buf[2]), msg->data, bytes);

    return bytes + 2;
  } else
    return 1;
}

unsigned long 
decode_BinaryString(const char buf[], BinaryString **bstr)
{
  unsigned long bytes = 0;

  bytes = buf[0];

  if (bytes > 0) {
    *bstr = tut_create_BinaryString(bytes, buf[1], &(buf[2]));
    return bytes + 2;
  } else {
    *bstr = NULL;
    return 1;
  }
}


/***********************************************************
 *
 * encode_charstring, decode_charstring
 *
 * A binary string is transfered as:
 * 
 * 1 byte  - number of bytes in string
 * n bytes - the char vector with the string
 *
 ***********************************************************/

unsigned long
encode_charstring(char *str, char buf[])
{
  unsigned long bytes = lengthof_string(str);

  buf[0] = (char)bytes;
  if (bytes > 0)
    strcpy(&(buf[1]), str);

  return bytes + 1;
}

unsigned long
decode_charstring(const char buf[], 
                  char **str, 
                  int *len)
{
  *len = buf[0];

  if (*len > 0) {
    *str = (char*) malloc(*len + 1);

    strncpy(*str, &(buf[1]), *len);
    (*str)[*len] = '\0';
    return *len + 1; /* Bytes consumed. */
  } 

  *str = NULL;
  return 1; /* Zero length field only. */
}


/***********************************************************
 *
 * encode_parameter, decode_parameter
 *
 * A TRI parameter is transferred as 
 * 
 * 1 byte  - mode (in inout, out)
 * n bytes - the binary data (see encode_BinaryString)
 *
 ***********************************************************/

unsigned long
encode_parameter(const TriParameter *parameter,
                 char buf[])
{
  int i = 0;

  buf[0] = (char)parameter->mode;
  i = encode_BinaryString(&(parameter->par), &(buf[1]));

  return i + 1; /* Par data plus list length (1).*/
}


unsigned long
decode_parameter(const char buf[],
                 TriParameter **parameter)
{
  int i = 0;
  BinaryString *bstr;

  *parameter = (TriParameter*) malloc(sizeof(TriParameter));
  (*parameter)->mode = buf[0];

  i = decode_BinaryString(&(buf[1]), &bstr);
  
  if (bstr != NULL) {
    (*parameter)->par.bits = bstr->bits;
    (*parameter)->par.data = bstr->data;
    (*parameter)->par.aux  = bstr->aux;
  } else {
    (*parameter)->par.bits = 0;
    (*parameter)->par.data = NULL;
    (*parameter)->par.aux  = NULL;
  }

  /* The local binary string is local, only its members kept.*/
  free(bstr); bstr = NULL;

  return i + 1; /* Par data plus list length (1).*/
}


/***********************************************************
 *
 * encode_parameterlist, decode_parameterlist
 *
 * The parameter lists are encoded as:
 * 
 * 1 byte  - the number of parameters
 * n bytes - the parameter data (see encode_parameter)
 *
 ***********************************************************/

unsigned long
encode_parameterlist(const TriParameterList *parameters,
                     char buf[])
{
  int j, i = 0;

  buf[i++] = (char)parameters->length;
  for (j = 0; j < parameters->length; ++j) {
/*    buf[i++] = parameters->parList[j]->mode;
    i += encode_BinaryString(&(parameters->parList[j]->par), &(buf[i]));
    */
    i += encode_parameter(parameters->parList[j], &(buf[i]));
  }

  return i + 1; /* Par data plus list length (1).*/
}


unsigned long
decode_parameterlist(const char buf[], 
                     TriParameterList **parameters)
{
  int i = 0, j;
  *parameters = (TriParameterList*) malloc(sizeof(TriParameterList));

  (*parameters)->length = buf[i++];

  (*parameters)->parList = (TriParameter**)malloc(sizeof(TriParameter*) * 
                                                  (*parameters)->length);
  
  for (j = 0; j < (*parameters)->length; ++j)
    i += decode_parameter(&(buf[i]), &((*parameters)->parList[j]));

  return i + 1;
}


/***********************************************************
 *
 * read_header_info
 *
 * Help function to read the first heade bytes in a socket 
 * message:
 *
 * 1 byte - message type 
 * 1 byte - size of message data
 *
 ***********************************************************/

int
read_header_info(SOCKET s, 
                 int msg_type,
                 int *size)
{
  char buf[3];

  if (s == INVALID_SOCKET)
    return -1;

  /* Message type */
  if (recv(s, buf, 1, 0) != 1) {
    print_protocol_error();
    return -1;
  }

  if (buf[0] != msg_type) {
    print_protocol_error();
    return -1;
  }

  /* Total message length. */
  if (recv(s, buf, 1, 0) != 1) {
    print_protocol_error();
    return -1;
  }

  *size = buf[0];

  return 0;
}


/***********************************************************
 *
 * tut_send_message, tut_read_message
 *
 * Send and read functionality for a 'send' operation.
 *
 ***********************************************************/

void
tut_send_message(SOCKET socket,
                 const TriComponentId *componentId, 
                 const TriPortId      *tsiPortId, 
                 const TriAddress     *sutAddress, 
                 const TriMessage     *sendMessage) 
{
  char sendbuf[SEND_BUFFER_SIZE] = "";
  int i = 0;

  sendbuf[i++] = (char)TUT_SEND_MESSAGE;
  i++; /*Make room for data size. */

  i += encode_charstring(tsiPortId->portName, &(sendbuf[i]));
  
  i += encode_BinaryString(sendMessage, &(sendbuf[i]));

  sendbuf[1] = i-2; /* Amount of data. */

#ifdef DEBUG
  printf(TRI_PREFIX "Send msg (%d/%X): ", i, i); 
  dump_vector(stderr, sendbuf, i);
#endif

  (void)send(socket, sendbuf, i, 0);
}


int
tut_read_message(SOCKET          socket,
                 char **         tsi_port_name, 
                 BinaryString**  data) 
{
  char readbuf[RECV_BUFFER_SIZE] = "";
  int i = 0;
  int datasize = 0;
  int len;

  if (read_header_info(socket, TUT_SEND_MESSAGE, &datasize) != 0)
    return -1;

  if (recv(socket, readbuf, datasize, 0) != datasize) {
    print_protocol_error();
    return -1;
  }

#ifdef DEBUG
  printf(TRI_PREFIX "Read msg (%d/%X): ", datasize+2, datasize+2);
  dump_vector(stderr, readbuf, datasize);
#endif

  i = decode_charstring(&(readbuf[0]), tsi_port_name, &len);

  i += decode_BinaryString(&(readbuf[i]), data);

  return 0;
}


/***********************************************************
 *
 * tut_send_call, tut_read_call
 *
 * Send and read functionality for a 'call' operation.
 *
 ***********************************************************/

void
tut_send_call(SOCKET socket,
              const TriComponentId   *componentId, 
              const TriPortId        *tsiPortId, 
              const TriAddress       *sutAddress, 
              const TriSignatureId   *signatureId,  
              const TriParameterList *parameterList) 
{
  char sendbuf[SEND_BUFFER_SIZE] = "";
  int i = 0;

  sendbuf[i++] = (char)TUT_CALL_MESSAGE;
  i++; /*Make room for data size. */
  
  /* TSI port name */
  i += encode_charstring(tsiPortId->portName, &(sendbuf[i]));

  /* Signature ID */
  i += encode_charstring(signatureId->moduleName, &(sendbuf[i]));
  i += encode_charstring(signatureId->objectName, &(sendbuf[i]));

  /* Actual parameters */
  i += encode_parameterlist(parameterList, &(sendbuf[i]));

  if (i > 256)
    fprintf(stderr, "Error! Data size too large to encode as 'char'!\n");

  sendbuf[1] = i-2; /* Amount of data. */

#ifdef DEBUG
  printf(TRI_PREFIX "Send call (%d/%X): ", i, i); 
  dump_vector(stderr, sendbuf, i);
#endif

  (void)send(socket, sendbuf, i, 0);
}


int
tut_read_call(SOCKET socket,
              char             **tsi_port_name,
              TriSignatureId   **signatureId,  
              TriParameterList **parameterList) 
{
  char readbuf[RECV_BUFFER_SIZE] = "";
  int i = 0;
  int datasize = 0;
  int len = 0;

  if (read_header_info(socket, TUT_CALL_MESSAGE, &datasize) != 0)
    return -1;

  if (recv(socket, readbuf, datasize, 0) != datasize) {
    print_protocol_error();
    return -1;
  }

#ifdef DEBUG
  printf(TRI_PREFIX "Read call (%d/%X): ", datasize+2, datasize+2);
  dump_vector(stderr, readbuf, datasize);
#endif

  i += decode_charstring(&(readbuf[i]), tsi_port_name, &len);

  *signatureId = (TriSignatureId*)malloc(sizeof(TriSignatureId));
  i += decode_charstring(&(readbuf[i]), &((*signatureId)->moduleName), &len);
  i += decode_charstring(&(readbuf[i]), &((*signatureId)->objectName), &len);
  (*signatureId)->aux = NULL;

  i += decode_parameterlist(&(readbuf[i]), parameterList);

  return 0;
}


/***********************************************************
 *
 * tut_send_reply, tut_read_reply
 *
 * Send and read functionality for a 'reply' operation.
 *
 ***********************************************************/

void
tut_send_reply(SOCKET socket,
              const TriComponentId   *componentId, 
              const TriPortId        *tsiPortId, 
              const TriAddress       *sutAddress,                 
              const TriSignatureId   *signatureId, 
              const TriParameterList *parameterList, 
              const TriParameter     *returnValue)
{
  char sendbuf[SEND_BUFFER_SIZE] = "";
  int i = 0;

  sendbuf[i++] = (char)TUT_REPLY_MESSAGE;
  i++; /*Make room for data size. */
  
  /* TSI port name */
  i += encode_charstring(tsiPortId->portName, &(sendbuf[i]));

  /* Signature ID */
  i += encode_charstring(signatureId->moduleName, &(sendbuf[i]));
  i += encode_charstring(signatureId->objectName, &(sendbuf[i]));

  /* Actual parameters */
  i += encode_parameterlist(parameterList, &(sendbuf[i]));

  /* Return value */
  i += encode_parameter(returnValue, &(sendbuf[i]));

  if (i > 256)
    fprintf(stderr, "Error! Data size too large to encode as 'char'!\n");

  sendbuf[1] = i-2; /* Amount of data. */

#ifdef DEBUG
  printf(TRI_PREFIX "Send reply (%d/%X): ", i, i); 
  dump_vector(stderr, sendbuf, i);
#endif

  (void)send(socket, sendbuf, i, 0);
}


int
tut_read_reply(SOCKET socket,
              char             **tsi_port_name,
              TriSignatureId   **signatureId,  
              TriParameterList **parameterList,
              TriParameter     **returnValue) 
{
  char readbuf[RECV_BUFFER_SIZE] = "";
  int i = 0;
  int datasize = 0;
  int len = 0;

  if (read_header_info(socket, TUT_REPLY_MESSAGE, &datasize) != 0)
    return -1;

  if (recv(socket, readbuf, datasize, 0) != datasize) {
    print_protocol_error();
    return -1;
  }

#ifdef DEBUG
  printf(TRI_PREFIX "Read reply (%d/%X): ", datasize+2, datasize+2);
  dump_vector(stderr, readbuf, datasize);
#endif

  i += decode_charstring(&(readbuf[i]), tsi_port_name, &len);

  *signatureId = (TriSignatureId*)malloc(sizeof(TriSignatureId));
  i += decode_charstring(&(readbuf[i]), &((*signatureId)->moduleName), &len);
  i += decode_charstring(&(readbuf[i]), &((*signatureId)->objectName), &len);
  (*signatureId)->aux = NULL;

  i += decode_parameterlist(&(readbuf[i]), parameterList);

  /* Return value */
  i += decode_parameter(&(readbuf[i]), returnValue);

  return 0;
}


/***********************************************************
 *
 * tut_socket_reader
 *
 * This thread reads from the socket and dispatches the incoming 
 * data to the intended ports and their queues.
 *
 * If a port is being communicated on and it can not be found 
 * in the mapping information (portlist), the thread will sleep 
 * and wait for the map statement to come. This means that no 
 * more data will be read in the meantime. 
 * This solution was the simplest and acceptable for the 
 * purpose of this tutorial. 
 *
 ***********************************************************/

#ifdef _WIN32
#define THREAD_RETURN 0
unsigned tut_socket_reader(void *reader_struct)
#else
#define THREAD_RETURN NULL
void * tut_socket_reader(void *reader_struct)
#endif
{
  reader_struct_t r_struct;
  fd_set fd;
  SOCKET sock;

#ifdef DEBUG
  printf(TRI_PREFIX "Reader thread started.\n");
#endif

  if (reader_struct == NULL) {
    exit(200);
  }

  r_struct = (reader_struct_t)reader_struct;

  sock  = r_struct->socket;

  FD_ZERO(&fd);
  FD_SET(sock, &fd);

  while (1) {
    int i=select(sock+1, &fd, NULL, NULL, NULL);

    if (i == SOCKET_ERROR) {
      printf(TRI_PREFIX "Socket communication ended, terminating.\n");
      exit(42);
    }
    
    if (i == 1) { /* Incoming data on one (1) socket, i.e. sock. */
      int bytes_recv = SOCKET_ERROR;
      char recvbuf[256] = "";
      port_mapping_t port = NULL;

      bytes_recv = recv(sock, recvbuf, 1, MSG_PEEK ); /* Message type. */ 
      if (bytes_recv == SOCKET_ERROR || bytes_recv == 0) {
        printf(TRI_PREFIX "Socket communication ended, terminating.\n");
        exit(42);
      }

#ifdef DEBUG
      printf(TRI_PREFIX "Message %d detected.\n", recvbuf[0]);
#endif

      switch(recvbuf[0]) {
      case TUT_SEND_MESSAGE: 
        {
          char *tsi_port_name;
          BinaryString *data;

          if (tut_read_message(sock, &tsi_port_name, &data) != 0) {
            fprintf(stderr, "Terminating reader thread.\n");
            return THREAD_RETURN;
          }

          /*port = find_or_wait_for_port_mapping(r_struct->portlist, NULL, tsi_port_name);*/

          while ((port = tut_find_port_mapping(r_struct->portlist, 
                                               NULL, 
                                               tsi_port_name)) == NULL) {
#ifdef DEBUG
            printf(TRI_PREFIX "Waiting for mapping.\n");
#endif
            Sleep(500);
          }

          triEnqueueMsg(port->tsi_port, 
                        NULL, /* sutAddress */
                        &(port->comp_port->compInst), /* componentId */ 
                        data); /* receivedMessage */

          free(tsi_port_name);
          free(data);
        }
        break;
      case TUT_CALL_MESSAGE: 
        {
          char             *tsi_port_name;
          TriParameterList *plist;
          TriSignatureId   *signatureId;

          if (tut_read_call(sock, &tsi_port_name, &signatureId, &plist) != 0) {
            fprintf(stderr, "Terminating reader thread.\n");
            return THREAD_RETURN;
          }

          while ((port = tut_find_port_mapping(r_struct->portlist, 
                                                      NULL, 
                                                      tsi_port_name)) == NULL) {
#ifdef DEBUG
            printf(TRI_PREFIX "Waiting for mapping.\n");
#endif
            Sleep(500);
          }

          triEnqueueCall(port->tsi_port,
                         NULL, /* sutAddress */
                         &(port->comp_port->compInst), /* componentId */
                         signatureId,
                         plist);

          free(tsi_port_name);
          for (i = 0; i < plist->length; ++i) {
            free(plist->parList[i]->par.data);
            free(plist->parList[i]);
          }
          free(plist->parList);
          free(plist);
          free(signatureId->moduleName);
          free(signatureId->objectName);
          free(signatureId);
        }
        break;
      case TUT_REPLY_MESSAGE: 
        {
          char             *tsi_port_name;
          TriParameterList *plist;
          TriSignatureId   *signatureId;
          TriParameter     *retval;

          if (tut_read_reply(sock, &tsi_port_name, &signatureId, &plist, &retval) != 0) {
            fprintf(stderr, "Terminating reader thread.\n");
            return THREAD_RETURN;
          }

          while ((port = tut_find_port_mapping(r_struct->portlist, 
                                                      NULL, 
                                                      tsi_port_name)) == NULL) {
#ifdef DEBUG
            printf(TRI_PREFIX "Waiting for mapping.\n");
#endif
            Sleep(500);
          }

          triEnqueueReply(port->tsi_port,
                         NULL, /* sutAddress */
                         &(port->comp_port->compInst), /* componentId */
                         signatureId,
                         plist,
                         retval);

          free(tsi_port_name);
          for (i = 0; i < plist->length; ++i) {
            free(plist->parList[i]->par.data);
            free(plist->parList[i]);
          }
          free(plist->parList);
          free(plist);
          free(signatureId->moduleName);
          free(signatureId->objectName);
          free(signatureId);
        }
        break;
      default:
        fprintf(stderr, TRI_PREFIX "Unknown message (%X).\n", recvbuf[0]);
        return THREAD_RETURN;
      }
    }
  }
}

#ifndef _WIN32
#include <sys/poll.h>

void Sleep(unsigned long milliseconds)
{
  poll(NULL, 0, milliseconds);
}
#endif
