/*************************************************************************
 * Tutorial logging.
 *
 * 2010 IBM Corporation
 *
 * The logging for the tutorial is made in HTML. Each component logs to a 
 * separate file which is named "tut-<timestamp>-<i>.html". 
 * The <timestamp> is the same for all files created in the same ETS 
 * execution. The <i> is an incremented number starting with 1.
 * This means that the CPC (control part component) will have number one
 * and the MTC and PTC's will follow.
 *
 *************************************************************************/

#include "t3rts.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>


/* Part of the filename which all components' logfile will have in common.*/
static char filenamebase[64];
static unsigned int inc = 1;

/* Log object definition */
struct log_object 
{
  int header_generated;
  FILE* file;
};

/* Mapping table from event kind identifier enum to string form. */
static const char *scope_kind_name[] = 
{
  "function",
  "external_function",
  "testcase",
  "teststep",
  "control_part",
  "undefined"
};


/***********************************************************
 *
 * tut_log_event
 *
 * This function is called for each log event generated by
 * the RTS. The event kind identifiers can be found in the 
 * 't3rts_event_logging.h' file. 
 * The following event kinds are logged specially while 
 * the other events are logged in a common way (in the default
 * clause of the switch statement):
 *
 *  t3rt_log_kind_event_testcase_started_c
 *  t3rt_log_kind_event_testcase_ended_c
 *  t3rt_log_kind_event_component_created_c
 *
 ***********************************************************/

void
tut_log_event(t3rt_log_t log,
              t3rt_log_event_kind_t event,
              t3rt_source_location_t source_location,
              t3rt_value_t params[],
              t3rt_binary_string_t origin,
              const char *origin_name,
              bool forwarded,
              t3rt_binary_string_t aux,
              t3rt_context_t ctx)
{
  struct log_object *l = (struct log_object *)t3rt_log_get_auxiliary(log, ctx);
  FILE* f = l->file;

  /* First time here, output which component we are logging for. */
  if (l->header_generated == 0) {
    fprintf(f, "<h1>%s</h1><hr/>\n", origin_name);
    l->header_generated = 1;
  }

  if (strcmp(origin_name, "CPC") != 0) {
    int i =0;
 
    /* To minimize the amount of logging made when the test is 
       going in and out of named alt statements, those events 
       are ignored, as well as the alt_default and alt_wait 
       events. 
    */
    if (event == t3rt_log_kind_event_scope_entered_c) {
      t3rt_scope_kind_t kind;
      t3rt_log_extract_scope_entered(params, NULL, &kind, ctx);
      if (kind == t3rt_scope_teststep_c)
        return;
    }
    if (event == t3rt_log_kind_event_scope_left_c) {
      t3rt_scope_kind_t kind;
      t3rt_log_extract_scope_left(params, NULL, &kind, ctx);
      if (kind == t3rt_scope_teststep_c)
        return;
    }
    if (event == t3rt_log_event_alt_defaults_c ||
        event == t3rt_log_event_alt_wait_c)
      return;

    /* Log event */
    fprintf(f, "<tr valign=\"top\">"
               "<td>%s</td><td>",t3rt_log_event_kind_string(event));
    while(params[i] != NULL) {
      fprintf(f, "%s<br/>", t3rt_value_to_string(params[i], ctx));
      i++;
    }
    fprintf(f, "</td>\n</tr>\n");
  } else {

    /* For CPC only these events are logged: */
    if (event != t3rt_log_kind_event_testcase_started_c &&
        event != t3rt_log_kind_event_testcase_verdict_c &&
        event != t3rt_log_kind_event_component_created_c)
      return;
    
    /* CPC logging only: */
    fprintf(f, "<tr valign=\"top\">");
        
    switch(event) {
    case t3rt_log_kind_event_testcase_started_c: {
      const char *module_name, *testcase_name;
      double timeout;
      char timeout_str[64];
      t3rt_log_extract_testcase_started(params,
        &module_name, 
        &testcase_name, 
        &timeout,
        ctx);
      if (timeout > 0)
        sprintf(timeout_str, "%.0f sec", timeout);
      else
        strcpy(timeout_str, "&lt;no timeout&gt;");

      fprintf(f, 
        "<td>%s</td><td>%s.%s (timeout=%s)</td>", 
        t3rt_log_event_kind_string(event), module_name, testcase_name, timeout_str);
                                                 }
      break;
      
    case t3rt_log_kind_event_testcase_verdict_c: {
      t3rt_verdict_t  verdict;
      t3rt_log_extract_test_case_verdict(params,
        NULL, 
        &verdict, 
        NULL,
        ctx);
      
      fprintf(f, "<td>%s</td><td><em><strong>%s</strong></em></td>",
        t3rt_log_event_kind_string(event), t3rt_verdict_string(verdict));
                                                 }
      break;
      
    case t3rt_log_kind_event_component_created_c: {
      const char *component_name;
      char comp_file[256] = "";
      sprintf(comp_file, "tut-%s-%d.html", filenamebase, inc-1);
      t3rt_log_extract_component_created(params, &component_name, NULL, ctx);
      fprintf(f, "<td>%s<br/><a href=\"%s\">%s</a></td><td>%s</td>",
        t3rt_log_event_kind_string(event), comp_file, comp_file, component_name);
                                                  }
      break;
      
    default:
      break;  
    }

    fprintf(f, "</tr>\n");
  }  
}


/***********************************************************
 *
 * tut_log_init
 *
 * This function is called one time only, in the end of the 
 * ETS execution. The common filename timestamp is formed to
 * be used later in the 'tut_log_open' function. 
 *
 ***********************************************************/

void
tut_log_init(t3rt_log_mechanism_t log_mechanism,
         t3rt_context_t context)
{
  time_t t;
  struct tm *time_struct;
  
  time(&t);
  time_struct = localtime(&t);

  sprintf(filenamebase, "%02d%02d%02d", 
          time_struct->tm_hour, 
          time_struct->tm_min, 
          time_struct->tm_sec);
}


/***********************************************************
 *
 * tut_log_finalize
 *
 * This function is called one time only, in the beginning 
 * of the ETS execution. Nothing needs to be done. 
 *
 ***********************************************************/

void 
tut_log_finalize(t3rt_log_mechanism_t log_mechanism)
{
  char index_filename[256];
  FILE *f;
  unsigned int i;

  sprintf(index_filename, "tut-%s-index.html", filenamebase);

  f = fopen(index_filename, "w");

  fprintf(f, 
    "<html><head></head><body>\n"
    "<p>Generated files for Tutorial test execution (%s):</p>\n"
    "<ul>\n",
    filenamebase);
  
  for (i = 1; i < inc; ++i) {
    fprintf(f, 
      "<li><a href=\"tut-%s-%d.html\">tut-%s-%d.html</a></li>\n", 
      filenamebase, i, filenamebase, i);
  }

  fprintf(f, "</ul>\n</body></html>\n");
}


/***********************************************************
 *
 * tut_log_open
 *
 * This function is responsible for storing the information 
 * necessary for the logmechanism into the log object.
 * The file for this log instance is opened. 
 * This function is called once for each component instance. 
 *
 ***********************************************************/

void
tut_log_open(t3rt_log_t log,
             t3rt_context_t ctx)
{
  char filename[128];
  time_t t;
  struct log_object *l = (struct log_object *)malloc(sizeof(struct log_object));
  
  sprintf(filename, "tut-%s-%d.html", filenamebase, inc++);
  
  l->file = fopen(filename, "w");
  l->header_generated = 0;

  t3rt_log_set_auxiliary(log, l, ctx);

  time(&t);
    
  fprintf(l->file, 
    "<html>\n<head><title></title></head>\n<body>\n"
    "<p>Test started at <strong>%s</strong></p>\n"
    "<table border=\"1\">\n<tr><th>Log event</th><th>Parameters</th>\n", ctime(&t));
}


/***********************************************************
 *
 * tut_log_close
 *
 * The opened file for this log instance is closed. 
 * This function is called once for each component instance. 
 *
 ***********************************************************/

void
tut_log_close(t3rt_log_t log,
              t3rt_context_t ctx)
{
  struct log_object *l = (struct log_object *)t3rt_log_get_auxiliary(log, ctx);

  fprintf(l->file, "</table>\n</body>\n</html>\n");

  fclose(l->file);
}


/***********************************************************
 *
 * t3ud_register_log_mechanisms
 *
 * The tutorial log functions are registered into the RTS. 
 *
 ***********************************************************/

void
t3ud_register_log_mechanisms (int argc, char * argv [])
{
  t3rt_log_register_listener("Tutorial Logging",
                             t3rt_log_mechanism_version_1_c,
                             tut_log_init, 
                             tut_log_finalize,
                             tut_log_open,
                             tut_log_close,
                             tut_log_event);
}

