/*************************************************************************
 * Tutorial port representation and handling (tut_ports.c)
 *
 * 2010 IBM Corporation
 *
 * This module contains the functions used for handling port mappings.
 * See tut_ports.h for definitions. 
 *
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "tut_ports.h"

/***********************************************************
 *
 * tut_compare_binary
 *
 * Compares two TRI-defined binary strings. 
 * Zero (0) is returned if they are equal, non-zero if they 
 * are not.
 *
 ***********************************************************/

int 
tut_compare_binary(const BinaryString *bs1, const BinaryString *bs2)
{
  int bytes, i;
  assert(bs1 != NULL);
  assert(bs2 != NULL);

  if (bs1->bits != bs2->bits)
    return -1;

  bytes = bs1->bits / 8;
  for (i = 0; i < bytes; ++i) {
    if (bs1->data[i] != bs2->data[i])
      return -1;
  }

  if ((bs1->bits % 8) != 0)
    printf(TRI_PREFIX 
     "Warning! The binary string comparision only support full bytes.\n");

  return 0;
}


/***********************************************************
 *
 * compare_port
 *
 * Compares two TRI-defined port identifiers.
 * Zero (0) is returned if they are equal, non-zero if they 
 * are not.
 *
 ***********************************************************/

static int 
compare_port(const TriPortId *p1, const TriPortId *p2)
{
  assert(p1 != NULL);
  assert(p2 != NULL);
  assert(p1->portName != NULL);
  assert(p2->portName != NULL);

  if (strcmp(p1->portName, p2->portName) != 0)
    return -1;

  if (p1->portIndex != p2->portIndex)
    return -1;

  if (strcmp(p1->compInst.compType.moduleName, 
             p2->compInst.compType.moduleName) != 0)
    return -1;

  if (strcmp(p1->compInst.compType.objectName, 
             p2->compInst.compType.objectName) != 0)
    return -1;

  return 0;
}


/***********************************************************
 *
 * delete_port_id
 *
 * Frees up memory for a TRI port identifer. The pointer
 * it also NULLed.
 *
 ***********************************************************/

static void
delete_port_id(TriPortId **pid)
{
  TriPortId *p;

  if (pid == NULL || *pid == NULL) 
    return;

  p = *pid;

  free(p->portType.moduleName);
  free(p->portType.objectName);
  free(p->portName);
  
  free(p->compInst.compType.moduleName);
  free(p->compInst.compType.objectName);

  *pid = NULL;
}


/***********************************************************
 *
 * copy_string
 *
 * Makes a char vector copy of a string. It handles NULL 
 * source strings and adds a null termination character
 * ('\0').
 *
 ***********************************************************/

static char *
copy_string(const char *str)
{
  char *s;
  int len;

  if (str == NULL) 
    return NULL;

  len = strlen(str);
  s = (char*)malloc(sizeof(char) * len + 1);
  s[len] = '\0';

  return strcpy(s, str);
}


/***********************************************************
 *
 * tut_copy_binary
 *
 * Copies a TRI binary string from 'src' into 'dest'. 
 *
 ***********************************************************/

int 
tut_copy_binary(BinaryString *dest, const BinaryString *src)
{
  if (dest == NULL || src == NULL) {
    printf(TRI_PREFIX "BinaryString copy failed.\n");
    return TUT_FAILED;
  }

  if (src->bits > 0) { 
    long bytes = (src->bits)/8;
    if (src->bits % 8 != 0)
      bytes += 1;

    dest->bits = src->bits;
    dest->data = (unsigned char*)malloc(sizeof(unsigned char) * bytes);
    memcpy(dest->data, src->data, bytes);
  }

  return TUT_OK;
}


/***********************************************************
 *
 * tut_find_port_mapping_noprot, tut_find_port_mapping
 *
 * Lookup function for port mappings given the component 
 * port id and/or the mapped TSI port name. 
 *
 * It returns the first occurrence only, NULL if non is found.
 * 
 * MT-level: The '..._noprot' function is not protected from 
 * multiple thread access to the port mapping list. The 
 * tut_find_port_mapping uses a mutex for this purpose.
 *
 * This is done because some other functions requires this 
 * lookup functionality. There, the mutex has already been 
 * taken. A second attempt to take the mutex will block, 
 * even if it is the same thread that already has the mutex 
 * that makes this attempt. These functions uses the non-
 * protected version.
 *
 ***********************************************************/

port_mapping_t
tut_find_port_mapping_noprot(port_list_t ports,
                             const TriPortId *compPortId,
                             const char *tsi_port_name)
{
  if (ports == NULL || ports->list == NULL)
    return NULL;

  if (compPortId == NULL && tsi_port_name == NULL)
    return NULL;

  if (compPortId != NULL) {
    port_mapping_t pm = ports->list;
    while (pm != NULL) {
      if (pm->comp_port != NULL &&
          compare_port(compPortId, pm->comp_port) == 0) {
        return pm;
      }
      pm = pm->next;
    }
    return NULL;
  }

  if (tsi_port_name != NULL) {
    port_mapping_t pm = ports->list;
    while (pm != NULL) {
      if (pm->tsi_port != NULL &&
          strcmp(pm->tsi_port->portName, tsi_port_name) == 0) {
        return pm;
      }
      pm = pm->next;
    }
  }

  return NULL;
}

port_mapping_t
tut_find_port_mapping(port_list_t ports,
                      const TriPortId *compPortId,
                      const char *tsi_port_name)
{
  port_mapping_t pm;

#ifdef _WIN32
  WaitForSingleObject(ports->mutex, INFINITE);
#else
  pthread_mutex_lock(&ports->mutex);
#endif
  
  pm = tut_find_port_mapping_noprot(ports, compPortId, tsi_port_name);

#ifdef _WIN32
  ReleaseMutex(ports->mutex);
#else
  pthread_mutex_unlock(&ports->mutex);
#endif

  return pm;
}

/***********************************************************
 *
 * tut_add_port_mapping
 *
 * Adds a pointer to a port mapping object and returns the 
 * index where it was inserted. If there is no room, -1 is 
 * returned.
 *
 ***********************************************************/

void
tut_add_port_mapping(port_mapping_t pm, port_list_t ports)
{
#ifdef _WIN32
  WaitForSingleObject(ports->mutex, INFINITE);
#else
  pthread_mutex_lock(&ports->mutex);
#endif

  pm->next = ports->list;

  if (ports->list != NULL) {
    ports->list->prev = pm;
    pm->next = ports->list;
  }

  ports->list = pm;
  pm->prev = NULL;

#ifdef _WIN32
  ReleaseMutex(ports->mutex);
#else
  pthread_mutex_unlock(&ports->mutex);
#endif
}


/***********************************************************
 *
 * delete_port_mapping_noprot
 *
 * Remove the port mapping representation from the list of 
 * ports.
 *
 ***********************************************************/

void 
delete_port_mapping_noprot(port_list_t ports, port_mapping_t *portmap)
{
  port_mapping_t pm = *portmap;

  if (ports->list == pm) {
    ports->list = pm->next;
  }

  if (pm->prev != NULL) pm->prev->next = pm->next;
  if (pm->next != NULL) pm->next->prev = pm->prev;

  pm->prev = NULL; 
  pm->next = NULL;

  delete_port_id(&(pm->comp_port));
  delete_port_id(&(pm->tsi_port));

  free(*portmap);
  *portmap = NULL;
}


/***********************************************************
 *
 * tut_delete_port_mapping
 *
 * Remove the port mapping representation from the list of 
 * ports given the component and TSI port identifiers.
 *
 ***********************************************************/

void
tut_delete_port_mapping(port_list_t ports, 
                        const TriPortId *compPortId,
                        const TriPortId *tsiPortId)
{
  port_mapping_t pm;

  if (ports == NULL)
    return;

#ifdef _WIN32
  WaitForSingleObject(ports->mutex, INFINITE);
#else
  pthread_mutex_lock(&ports->mutex);
#endif

  pm = tut_find_port_mapping_noprot(ports, compPortId, tsiPortId->portName);
  
  delete_port_mapping_noprot(ports, &pm);

#ifdef _WIN32
  ReleaseMutex(ports->mutex);
#else
  pthread_mutex_unlock(&ports->mutex);
#endif
}


/***********************************************************
 *
 * tut_clear_port_mappings
 *
 * Remove all the port mappings from the list of ports. 
 *
 ***********************************************************/

void
tut_clear_port_mappings(port_list_t ports)
{
  if (ports == NULL)
    return;

#ifdef _WIN32
  WaitForSingleObject(ports->mutex, INFINITE);
#else
  pthread_mutex_lock(&ports->mutex);
#endif

  while (ports->list != NULL) {
    delete_port_mapping_noprot(ports, &(ports->list));
  }

#ifdef _WIN32
  ReleaseMutex(ports->mutex);
#else
  pthread_mutex_unlock(&ports->mutex);
#endif
}


/***********************************************************
 *
 * tut_copy_TriPortId
 *
 * Returns a copy of the port identifer.
 *
 ***********************************************************/

TriPortId *
tut_copy_TriPortId(const TriPortId *portid)
{
  TriPortId *p = (TriPortId *)malloc(sizeof(TriPortId));

  if (tut_copy_binary(&(p->compInst.compInst), &(portid->compInst.compInst)) != TUT_OK) {
    free(p);
    return NULL;
  }

  p->portType.moduleName = copy_string(portid->portType.moduleName);
  p->portType.objectName = copy_string(portid->portType.objectName);
  p->portName = copy_string(portid->portName);
  p->portIndex = portid->portIndex;

  p->compInst.compType.moduleName = copy_string(portid->compInst.compType.moduleName);
  p->compInst.compType.objectName = copy_string(portid->compInst.compType.objectName);

  return p;
}


