
/*
//================================================================
//
//     $Id: asntyp.typ,v 2.1 2002/07/12 08:13:44 xmso Exp $
//
//     NAME
//          asntyp.typ
//
//     DESCRIPTION
//          Type definitions for the ASN2C Run Time System.
//
//     AUTHORS
//          Pekka Ollikainen (Version 1.nn)
//          Sami Kesti       (Version 1.nn)
//          Markku Turunen
//
//          Copyright (c) Nokia Research Center
//
//     CONTENTS
//          ERR_RET
//               Type for the RTS error handling buffer.
//
//          NULLTYPE
//               C home type for ASN.1 NULL type.
//
//          sint64
//               C home type for 64 bits long integers.
//
//          BITSTRING
//               C home type for static bit string types.
//
//          OCTET
//               C home type for static octet string types.
//
//          REAL
//               C home types for ASN.1 real type.
//
//          STATSTR
//               Generic static string stub types.
//
//          DYNSTR
//               Generic dynamic string types.
//
//          ANY
//               C home type for ASN.1 NULL type.
//
//          LIST
//               Generic list type.
//
//          COMMONTIME
//               A home type for ASN.1 time types.
//
//          DATETIMEZ
//               A home type for ASN.1 time types.
//
//          asnchar
//               Character types for ASN.1 character strings.
//
//          OBJECTIDENTIFIER
//               Home type for the ASN.1 object identifier type.
//
//          OBJECTDESCRIPTOR
//               Home type for the ASN.1 object descriptor type.
//		 Changed to:
//          ASN2C_OBJECTDESCRIPTOR
//               Changed by AKU. Clashesh with visualc namespace.
//
//          EXTERNAL
//               Home type for the ASN.1 external type.
//
//          SKIPPED
//               Home type for a skipped ASN.1 type.
//
//          EXTENSIONMARKER
//               Home type for an extensionmaker component.
//
//          EXTENSIBLE_ENUMERATED
//               Home type for an extensible enumerated type.
//
//          NOBODYUSESTHISTYPE
//               Home type for a dummy sequence component.
//
//          ASNHOME
//               Generic home type for all the ASN.1 types.
//
//          ENTRYDESCR
//               Entry descriptor table element type.
//
//     NAMING CONVENTIONS
//          Type definitions for ASN.1 types have the ASN.1 type name
//          as a part of C type name. Example: OBJECTIDENTIFIER,
//          EXTENSIBLE_ENUMERATED.
//
//          If a definition is derived from the CASN 1 RTS then
//          name is not changed. Example: ERR_RET.
//
//          If a new external type definition has a counterpart
//          in the CASN 1 RTS or in the CVOPS RTS then the new type
//          definition name is derived from the counterpart name.
//          Example: sint64 and sint32.
//
//          Type specific macro definitions have a type name as 
//          a part of the macro name. Example: REAL type and
//          REAL_MINUS_INF, realIsMinusIs; LIST and listInit, 
//          firstListItem.
//
//          Low-level and internal type definitions have a prefix "Asn".
//          Example: AsnAllocType, AsnCompStatus, AsnTag.
//
//================================================================
*/

#ifndef ASNTYP_TYP
#define ASNTYP_TYP

  /*============================================================*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "env.def"
#include "options.def"
#include "general.def"

#ifndef ASNCNF_CNF
#include "asncnf.cnf"
#endif

#include "basic.typ"
#include "frame.ext"

#if ASN_HAS_TIME
#include "timers.ext"
#endif /* ASN_HAS_TIME */

#include "trace.ext"
#include "clibrary.ext"
#include <setjmp.h>

#ifndef ASNDEF_DEF
#include "asndef.def"
#endif

#if STAND_ALONE_ASN == 0

#include "vtask.ext"
#include "messages.ext"

#endif /* STAND_ALONE_ASN == 0 */

  /*============================================================*/

#if ASN_RCS_ID
static char asntyp_typ_rcsid [] =
  "$Id: asntyp.typ,v 2.1 2002/07/12 08:13:44 xmso Exp $";
#endif

  /*============================================================*/

/*
$begin-intro-type
.IP   ERR_RET
//         Type for the RTS error handling buffer.
.LP
$end-intro-type

$begin-global-type
.TH ERR_RET 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         ERR_RET - The RTS error handling buffer.
//
.SH   SYNOPSIS
.ft B
.nf
*/
                typedef struct error_return {
                     jmp_buf buf;
                } ERR_RET;
/*
.fi
.ft
.SH   DESCRIPTION
//         The ERR_RET structure contains buffer for error recovery.
//         The error recovery uses the setjmp/longjump mechanism.
//         The "buf" member must be initialized using the setjmp function.
//         After initialization a pointer to the initialized structure must 
//         be passed as an argument to a RTS initialization function, e.g.
//         setStart.
//
.SH   EXAMPLE
.nf
//         int myDecode (FRAME_PTR *toBeDecoded, Type *home)
//         {
.RS
//              ERR_RET errBuf;
//              int errCode;
//
//              if ((errCode = setjmp (errBuf.buf)) != 0){
.RS
//                   deleteFrame (toBeDecoded);
//                   return (errCode);
.RE
//              }
//
//              setStart (toBeDecoded, &errBuf);
//              decType (home);
//
//              return (0);
.RE
//         }
.fi
.SH   FILES
//         asntyp.typ
//
.SH   "SEE ALSO"
//         setjmp, longjmp, setStart
//
$end-global-type
*/

  /*============================================================*/

#if ASN_HAS_NULL

/*
$begin-intro-type
.IP   NULLTYPE
//         C home type for ASN.1 NULL type.
.LP
$end-intro-type

$begin-global-type
.TH NULLTYPE 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         NULLTYPE - C home type for ASN.1 NULL type.
//
.SH   SYNOPSIS
.ft B
.nf
*/
           typedef BYTE_TYPE NULLTYPE;
/*
.fi
.ft
.SH   DESCRIPTION
//         The type NULLTYPE is the C home type for the ASN.1 NULL type.
//
.SH   FILES
//         asntyp.typ
//
$end-global-type
*/

#endif /* ASN_HAS_NULL */

  /*============================================================*/

#if ASN_HAS_INT64

#include "sint64.h"

/*
$begin-intro-type
.IP   sint64
//         C home type for 64 bits long integers.
.LP
$end-intro-type

$begin-global-type
.TH sint64 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         sint64 - C home type for 64 bits long integers
//
.SH   SYNOPSIS
.ft B
.nf
          typedef long sint64;

          typedef struct {
               long hi;
               unsigned long lo;
          } sint64;

          typedef long long sint64;

          typedef __int64 sint64;
.fi
.ft
.SH   DESCRIPTION
//         The type sint64 is the C home type for 64 bits long integers.
//         See the sint64.h file for the actual selected implementation 
//         of sint64 type.
//
.SH   FILES
//         asntyp.typ
//
$end-global-type
*/

#endif /* ASN_HAS_INT64 */

  /*============================================================*/

#if ASN_HAS_BITSTRING

/*
$begin-intro-type
.IP   BITSTRING
//         C home type for static bit string types.
.LP
$end-intro-type

$begin-global-type
.TH BITSTRING 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         BITSTRING - C home type for static bit string types.
//
.SH   SYNOPSIS
.ft B
.nf
*/
           typedef BYTE_TYPE BITSTRING;
/*
.fi
.ft
.SH   DESCRIPTION
//         The type BITSTRING is used as the type for the "st" member
//         of static bit string types.
//
.SH   FILES
//         asntyp.typ
//
$end-global-type
*/

#endif /* ASN_HAS_BITSTRING */

  /*============================================================*/

#if ASN_HAS_OCTETSTRING

/*
$begin-intro-type
.IP   OCTET
//         C home type for static octet string types.
.LP
$end-intro-type

$begin-global-type
.TH OCTET 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         OCTET - C home type for static octet string types.
//
.SH   SYNOPSIS
.ft B
.nf
*/
           typedef BYTE_TYPE OCTET;
/*
.fi
.ft
.SH   DESCRIPTION
//         The type OCTET is used as the type for the "st" member
//         of static octet string types.
//
.SH   FILES
//         asntyp.typ
//
$end-global-type
*/

#endif /* ASN_HAS_OCTETSTRING */

  /*============================================================*/

#if ASN_HAS_REAL

/*
$begin-intro-type
.IP   REAL
//         C home types for ASN.1 real type.
.LP
$end-intro-type

$begin-global-type
.TH REAL 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         REAL - C home types for ASN.1 real types
//
.SH   SYNOPSIS
.ft B
.nf
*/
#     if ASN_HAS_REAL
           typedef double REAL;
#     endif
  
#     if ASN_HAS_REAL32
           typedef struct {
                int kind;
                float value;
           } REAL32;
#     endif
  
#     if ASN_HAS_REAL64
           typedef struct {
                int kind;
                double value;
           } REAL64;
#     endif
  
#     if ASN_HAS_REAL128
           typedef struct {
                int kind;
                long double value;
           } REAL128;
#     endif
/*
.fi
.ft
.SH   DESCRIPTION
//         The REAL and REAL<N> types are alternative C home types for
//         the ASN.1 real type. The ASN_HAS_REAL and ASN_HAS_REAL<N>
//         flags control the existence of these types.
.LP
//         The "kind" member in the REAL<N> types specifies what kind of
//         an real value a home area contains. The valid values are as 
//         follows:
.LP
.IP        REAL_PLUS_INF
//              Plus infinity value.
//
.IP        REAL_MINUS_INF
//              Minus infinity value.
//
.IP        REAL_NORMAL
//              Normal real value.
.LP
//         If the "kind" member is REAL_NORMAL then the "value" member
//         contains a floating point value.
.LP
//         The following macros can be used to manipulate the REAL<N>
//         structures:
.LP
.IP        "int realIsPlusInf (REAL<N> *home)"
//              Determine whether a home area contains a plus infinity value.
//
.IP        "int realIsMinusInf (REAL<N> *home)"
//              Determine whether a home area contains a minus infinity value.
//
.IP        "int realIsNormal (REAL<N> *home)"
//              Determine whether a home area contains a normal value.
//
.IP        "void setRealPlusInf (REAL<N> *home)"
//              Set a real value as plus infinity.
//
.IP        "void setRealMinusInf (REAL<N> *home)"
//              Set a real value as minus infinity.
//
.IP        "void setRealNormal (REAL<N> *home, <realtype> val)"
//              Set a real value as a normal floating point value "val".
//
.IP        "<realtype> realNormalValue (REAL<N> *home)"
//              Get a normal floating point value.
//
.SH   FILES
//         asntyp.typ
//
$end-global-type
*/

#define REAL_MINUS_INF (-1)
#define REAL_PLUS_INF  1
#define REAL_NORMAL    0

#define realIsMinusInf(homePtr)   ((homePtr)->kind == REAL_MINUS_INF)
#define realIsPlusInf(homePtr)    ((homePtr)->kind == REAL_PLUS_INF)
#define realIsNormal(homePtr)     ((homePtr)->kind == REAL_NORMAL)

#define setRealMinusInf(homePtr)  ((homePtr)->kind = REAL_MINUS_INF)
#define setRealPlusInf(homePtr)   ((homePtr)->kind = REAL_PLUS_INF)
#define setRealNormal(homePtr, v) \
  do { (homePtr)->kind = REAL_NORMAL; (homePtr)->value = v; } while (0)

#define realNormalValue(homePtr)  ((homePtr)->value)

#endif /* ASN_HAS_REAL */

  /*============================================================*/

#if ASN_STA_STRI

/*
$begin-intro-type
.IP   STATSTR
//         Generic static string stub types.
.LP
$end-intro-type

$begin-global-type
.TH STATSTR 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         STATSTR - Generic static string stub types.
//
.SH   SYNOPSIS
.ft B
.nf
*/
           typedef struct statstr0 {
                BYTE_TYPE st [1];
           } STATSTR0;

           typedef struct statstr8 {
                uint8 n;
                BYTE_TYPE st [1];
           } STATSTR8;

           typedef struct statstr16 {
                uint16 n;
                BYTE_TYPE st [1];
           } STATSTR16;

           typedef struct statstr {
                uint32 n;
                BYTE_TYPE st [1];
           } STATSTR;
/*
.fi
.ft
.SH   DESCRIPTION
//         The STATSTR<N> types are generic stubs for static byte arrays for
//         bit, octet and character strings. The member "n" contains the 
//         number of elements (in bits, octets or characters) stored
//         into the "st" member.
.LP
//         The STATSTR<n> types are used only in the ASNHOME structure
//         as a hook for ASN.1 type specific static string types.
//
.SH   FILES
//         asntyp.typ
//
.SH   "SEE ALSO"
//         DYNSTR, LIST, ASNHOME, stastrfnc
//
$end-global-type
*/

#endif /* ASN_STA_STRI */

  /*============================================================*/

/*
$begin-intro-type
.IP   asnchar
//         Character types for ASN.1 character strings.
.LP
$end-intro-type

$begin-global-type
.TH asnchar 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         asnchar - Character types for ASN.1 character strings.
//
.SH   SYNOPSIS
.ft B
.nf
*/
#          if ASN_HAS_CHARSTRING

           typedef char NUMERICCHAR;

           typedef char PRINTABLECHAR;

           typedef char IA5CHAR;

           typedef char VISIBLECHAR;

           typedef unsigned char T61CHAR;

           typedef unsigned char GRAPHICCHAR;

           typedef unsigned char GENERALCHAR;

           typedef unsigned char VIDEOTEXCHAR;

#          if ASN_HAS_BMPSTRING == 1
           typedef wchar_t ASN_BMPCHAR;
#          endif

#          if ASN_HAS_BMPSTRING == 2
           typedef sint16  ASN_BMPCHAR;
#          endif

#          endif /* ASN_HAS_CHARSTRING */
/*
.fi
.ft
.SH   DESCRIPTION
//         The <X>CHAR types are used as base types for the "st" arrays
//         in static character strings.
//
.SH   FILES
//         asntyp.typ
//
.SH   "SEE ALSO"
//         DYNSTR
//
$end-global-type
*/

  /*============================================================*/

#if ASN_DYN_STRI

/*
$begin-intro-type
.IP   DYNSTR
//         Generic dynamic string types.
.LP
$end-intro-type

$begin-global-type
.TH DYNSTR 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         DYNSTR - Generic dynamic string types.
//
.SH   SYNOPSIS
.ft B
.nf
*/
           typedef struct dynstr0 {
                BYTE_TYPE *st;
           } DYNSTR0;

           typedef struct dynstr8 {
                uint8 n;
                BYTE_TYPE *st;
           } DYNSTR8;

           typedef struct dynstr16 {
                uint16 n;
                BYTE_TYPE *st;
           } DYNSTR16;

           typedef struct dynstr {
                uint32 n;
                BYTE_TYPE *st;
           } DYNSTR;

#          if ASN_HAS_BMPSTRING

           typedef struct asn_BMPString0 {
                ASN_BMPCHAR *st;
           } ASN_BMPString0;

           typedef struct asn_BMPString8 {
                uint8 n;
                ASN_BMPCHAR *st;
           } ASN_BMPString8;

           typedef struct asn_BMPString16 {
                uint16 n;
                ASN_BMPCHAR *st;
           } ASN_BMPString16;

           typedef struct asn_BMPString {
                uint32 n;
                ASN_BMPCHAR *st;
           } ASN_BMPString;

#          endif
/*
.fi
.ft
.SH   DESCRIPTION
//         The DYNSTR<N> types are generic dynamic byte arrays for bit,
//         octet and character strings. The member "n" contains the number
//         of allocated elements (in bits, octets or characters) stored
//         into the "st" member.
.LP
//         The ASN_BMPString<N> types are character arrays for dynamic
//         BMPStrings.
//
.SH   FILES
//         asntyp.typ
//
.SH   "SEE ALSO"
//         LIST, STATSTR, dynstrfnc
//
$end-global-type
*/

#endif /* ASN_DYN_STRI */

  /*============================================================*/

#if ASN_HAS_ANY

/*
$begin-intro-type
.IP   ANY
//         C home type for ASN.1 NULL type.
.LP
$end-intro-type

$begin-global-type
.TH NULLTYPE 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         ANY - C home type for ASN.1 ANY type.
//
.SH   SYNOPSIS
.ft B
.nf
*/
           typedef BYTE_TYPE *ANY;
/*
.fi
.ft
.SH   DESCRIPTION
//         The type ANY is a C home type for the ASN.1 NULL type.
//
.SH   FILES
//         asntyp.typ
//
$end-global-type
*/

#endif /* ASN_HAS_ANY */

  /*============================================================*/

#if ASN_HAS_LIST || ASN_SEG_STRI

/*
$begin-intro-type
.IP   LIST
//         Generic list type.
.LP
$end-intro-type

$begin-global-type
.TH LIST 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         LIST - Generic list type.
//
.SH   SYNOPSIS
.ft B
.nf
*/
           typedef struct list {
                uint32 n;
                struct asnhome *ls;
                struct asnhome *le;
           } LIST;
/*
.fi
.ft
.SH   DESCRIPTION
//         The LIST structure is used as a home area for segmented string
//         types and segmented list types. The "n" member contains the
//         the number of string elements/list elements. The "ls" member
//         points to the first element in the list and the "le" member
//         points to the last element in the list. 
.LP
//         In case of a string type a string value is splitted into 
//         substrings and pointers to those substrings are stored into
//         list elements as values of the ASNHOME structures. The structures
//         form a liked list.
.IP
//         If case of a list type the list element values are stored
//         into ASNHOME structures. The structures form a liked list.
//         
.LP
//         The following macros can be used to manipulate the string
//         length.
.LP
.IP        "void listInit (LIST *home)"
//              Initialize a LIST structure.
//
.IP        "uint32 listLen (LIST *home)"
//              Get the list length.
//
.IP        "void setListLen (LIST *home, uint32 newLen)"
//              Set the list length.
//
.IP        "void incListLen (LIST *home)"
//              Increment the list length by one.
//
.IP        "void decListLen (LIST *home)"
//              Decrement the list length by one. The original length must be
//              greater than zero.
//
.IP        "ASNHOME *firstListItem (LIST *home)"
//              Return the first list element.
//
.IP        "ASNHOME *lastListItem (LIST *home)"
//              Return the last list element.
//
.IP        "ASNHOME *nextListItem (ASNHOME *listItem)"
//              Return the next list element.
//
.IP        "ASNHOME *prevListItem (ASNHOME *listItem)"
//              Return the previous list element.
//
.IP        "void extractFirstListItem (LIST *home, ASNHOME **listItem)"
//              Remove the first list element from a list and set
//              "*listItem" to point to the removed element. If the list
//              is empty then "*listItem" is set to NULL.
//
.SH   FILES
//         asntyp.typ
//
.SH   "SEE ALSO"
//         DYNSTR, ASNHOME, appendList, rmListItem
//
$end-global-type
*/

#define listInit(homePtr) \
  do { \
    (homePtr)->n = 0; \
    (homePtr)->ls = (homePtr)->le = NULL; \
   } while (0)

#define listLen(homePtr)             ((homePtr)->n)

#define setListLen(homePtr, newLen)  ((homePtr)->n = (newLen))

#define incListLen(homePtr)          ((homePtr)->n++)

#define decListLen(homePtr)          ((homePtr)->n--)

#define firstListItem(homePtr)       ((homePtr)->ls)

#define lastListItem(homePtr )       ((homePtr)->le)

#define nextListItem(listItem)       ((listItem)->next)

#define prevListItem(listItem)       ((listItem)->prev)

#define extractFirstListItem(homePtr, listItem) \
  do { \
    *(listItem) = (homePtr)->ls; \
    if (*listItem){ \
      rmListItem ((homePtr), *(listItem)); \
      (homePtr)->n--; \
    } \
  } while (0)

#endif /* ASN_HAS_LIST || ASN_SEG_STRI */
  
  /*============================================================*/

#if ASN_HAS_COMMONTIME

/*
$begin-intro-type
.IP   COMMONTIME
//         A home type for ASN.1 time types.
.LP
$end-intro-type

$begin-global-type
.TH COMMONTIME 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         COMMONTIME - A home type for ASN.1 time types.
//
.SH   SYNOPSIS
.ft B
.nf
*/
           typedef struct commontime {
                time_t sec;
                unsigned short msec;
                short zone;
                short dst;
           } COMMONTIME;
/*
.fi
.ft
.SH   DESCRIPTION
//         The COMMONTIME type is a home time for the ASN.1 time types
//         GeneralizedTime and UTCTime. The member "sec" contains the
//         number of seconds since the epoch (1-jan-1970, 00:00:00 GMT).
//         The "millitm" member contains milliseconds. The "zone" member
//         contains the difference between the GMT and the local time 
//         in minutes. The "dst" member is not used. If a time value is 
//         a local time without any time zone information then the "zone" 
//         member is set to LOCAL_COMMONTIME.
.LP
//         The structure is compatible with the struct timeb structure.
//         The ftime(3) function can be used to fill a struct timeb
//         structure. The ftime function does not belong to the set of
//         ANSI C standard functions.
.LP
//         The flag ASN_HAS_COMMONTIME controls existence of this type.
//
.SH   FILES
//         asntyp.typ
//
.SH   "SEE ALSO"
//         DATETIMEZ, ftime, time_to_str, str_to_time
//
$end-global-type
*/

#define LOCAL_COMMONTIME  (~((short)0))

#endif /* ASN_HAS_COMMONTIME */

  /*============================================================*/

#if ASN_HAS_TIME

/*
$begin-intro-type
.IP   DATETIMEZ
//         A home type for ASN.1 time types.
.LP
$end-intro-type

$begin-global-type
.TH DATETIMEZ 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         DATETIMEZ - A home type for ASN.1 time types.
//
.SH   SYNOPSIS
.ft B
.nf
*/
           typedef struct datetimez {
                DATETIME datetime;
                short zoneCode;
                short zone;
           } DATETIMEZ;
/*
.fi
.ft
.SH   DESCRIPTION
//         The DATETIMEZ type is a home time for the ASN.1 time types
//         GeneralizedTime and UTCTime. The member "datetime" contains the
//         date and time in the CVOPS DATETIME format. The DATETIME 
//         structure is defined as follows:
.LP
.RS
.nf
//             typedef struct dateTime {
.RS
//                  unsigned short year;
//                  unsigned short month;
//                  unsigned short day;
//                  unsigned short hour;
//                  unsigned short minute;
//                  unsigned short second;
//                  unsigned short csec;
.RE
//             } DATETIME;
.fi
.RE
.LP
//         The "zoneCode" member contains information about time zone.
//         Valid values are as follows:
.LP
.IP        COORDINATED_UNIVERSAL_TIME
//              Time is specified as coordinated universal time, e.g.
//              "199411021200Z".
//
.IP        LOCAL_TIME_WITH_ZONE
//              Time is specified as local time with time zone e.g.
//              "199411021200+0200".
//
.IP        LOCAL_TIME
//              Time is specified as local time without time zone e.g.
//              "199411021200".
.LP
//         If the "zoneCode" member is set to LOCAL_TIME_WITH_ZONE then
//         the "zone" member contains the difference between the GMT and
//         the local time in minutes.
.LP
//         The CVOPS function wsystime can be used to fill the structure.
.LP
//         The flag ASN_HAS_TIME controls existence of this type.
//
.SH   FILES
//         asntyp.typ
//
.SH   "SEE ALSO"
//         COMMONTIME, wsystime, datetimez_to_str, str_to_datetimez,
//         datetimezInit, datetimezCmp, datetimezCpy
//
$end-global-type
*/

#define COORDINATED_UNIVERSAL_TIME 0
#define LOCAL_TIME_WITH_ZONE       1
#define LOCAL_TIME                 2

#endif /* ASN_HAS_TIME */

  /*============================================================*/

#if ASN_HAS_OBJID

  /*
   * ObjIdElem, type for an element of OBJECT IDENTIFIER.
   */

#if ASN_OBJIDC_SIZE == 8
typedef uint8 ObjIdElem;
#endif

#if ASN_OBJIDC_SIZE == 16
typedef uint16 ObjIdElem;
#endif

#if ASN_OBJIDC_SIZE == 32
typedef uint32 ObjIdElem;
#endif

/*
$begin-intro-type
.IP   OBJECTIDENTIFIER
//         Home type for the ASN.1 object identifier type.
.LP
$end-intro-type

$begin-global-type
.TH OBJECTIDENTIFIER 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         OBJECTIDENTIFIER - Home type for the ASN.1 object identifier type.
//
.SH   SYNOPSIS
.ft B
.nf
*/
           typedef struct objId {
                sint16 n;
                ObjIdElem elem [MAX_OBJIDC_CNT];
           } OBJECTIDENTIFIER;
/*
.fi
.ft
.SH   DESCRIPTION
//         The OBJECTIDENTIFIER type is the home type for the ASN.1
//         object identifier type. The "n" member contains the number
//         of elements in the "elem" array.
.LP
//         The ASN_OBJIDC_SIZE flag controls the base type for the ObjIdElem
//         type. ObjIdElem can be uint8, uint16 or uint32. By default it
//         is uint32.
.LP
//         By default the MAX_OBJIDC_CNT constant is 16. It can be modified
//         if there can be object identifier values with more than 16 
//         elements.
//
.SH   FILES
//         asntyp.typ, asncnf.cnf
//
.SH   "SEE ALSO"
//         objIdCopy, objIdCmp, asnmaxcnf
//
$end-global-type
*/

#endif /* ASN_HAS_OBJID */

  /*============================================================*/

#if ASN_DYN_STRI

/*
$begin-intro-type
.IP   ASN2C_OBJECTDESCRIPTOR
//         Home type for the ASN.1 object descriptor type.
.LP
$end-intro-type

$begin-global-type
.TH ASN2C_OBJECTDESCRIPTOR 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         ASN2C_OBJECTDESCRIPTOR - Home type for the ASN.1 object descriptor type.
//
.SH   SYNOPSIS
.ft B
.nf
*/
           typedef DYNSTR ASN2C_OBJECTDESCRIPTOR;
/*
.fi
.ft
.SH   DESCRIPTION
//         The ASN2C_OBJECTDESCRIPTOR type is the home type for the ASN.1
//         object descriptor type. 
//
.SH   FILES
//         asntyp.typ
//
.SH   "SEE ALSO"
//         DYNSTR
//
$end-global-type
*/

#endif /* ASN_DYN_STRI */

  /*============================================================*/

#if ASN_HAS_EXTERNAL

/*
$begin-intro-type
.IP   EXTERNAL
//         Home type for the ASN.1 external type.
.LP
$end-intro-type

$begin-global-type
.TH EXTERNAL 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         EXTERNAL - Home type for the ASN.1 external type.
//
.SH   SYNOPSIS
.ft B
.nf
*/
#     if ASN_HAS_EXTERNAL_1990

           typedef struct external_1990 {
                sint8 compmask;
                OBJECTIDENTIFIER direct;
                sint32 indirect;
                ASN2C_OBJECTDESCRIPTOR descr;
                struct {
                     sint8 choice;
                     union {
                          FRAME_PTR single;
                          FRAME_PTR octet;
                          FRAME_PTR arbit;
                     } value;
                } encoding;
           } EXTERNAL_1990;

#     if !DX200
           typedef EXTERNAL_1990 EXTERNAL;
#     else
           typedef EXTERNAL_1990 ASN_EXTERNAL;
#     endif

#     endif /* ASN_HAS_EXTERNAL_1990 */

#     if ASN_HAS_EXTERNAL_1994

           typedef struct external_1994 {
                sint8 compmask;
                struct t_identification_EXTERNAL_1994 {
                  sint8 choice;
                  union {
                       OBJECTIDENTIFIER syntax;
                       sint32 pci;
                       struct t_ctx_neg_EXTERNAL_1994 {
                            sint32 ctx_neg_pci;
                            OBJECTIDENTIFIER ctx_neg_stx;
                       } ctx_neg;
                  } i1;
                } identification;
                ASN2C_OBJECTDESCRIPTOR descr;
                struct t_value_EXTERNAL_1994 {
                     sint8 choice;
                     union {
                          FRAME_PTR notation;
                          FRAME_PTR encoded;
                     } i1;
                } value;
           } EXTERNAL_1994;

#     endif /* ASN_HAS_EXTERNAL_1994 */
/*
.fi
.ft
.SH   DESCRIPTION
//         The EXTERNAL type is the home type for the ASN.1
//         external type. 
.LP
//         The flags ASN_HAS_EXTERNAL, ASN_HAS_EXTERNAL_1990 and
//         ASN_HAS_EXTERNAL_1994 control existence of these types.
.LP
//         See the file asntyp.typ for component access macros for
//         these types.
//
.SH   FILES
//         asntyp.typ
//
$end-global-type
*/

  /*
   * Access macros for EXTERNAL, ASN.1(1990)
   */

#define direct_EXTERNAL   0
#define indirect_EXTERNAL 1
#define descr_EXTERNAL    2
#define encoding_EXTERNAL 3

#define set_direct_EXTERNAL(home)   ((home)->compmask |= BIT(0))
#define is_direct_EXTERNAL(home)    ((home)->compmask & BIT(0))
#define rem_direct_EXTERNAL(home)   ((home)->compmask &= ~BIT(0))

#define set_indirect_EXTERNAL(home) ((home)->compmask |= BIT(1))
#define is_indirect_EXTERNAL(home)  ((home)->compmask & BIT(1))
#define rem_indirect_EXTERNAL(home) ((home)->compmask &= ~BIT(1))

#define set_descr_EXTERNAL(home)    ((home)->compmask |= BIT(2))
#define is_descr_EXTERNAL(home)     ((home)->compmask & BIT(2))
#define rem_descr_EXTERNAL(home)    ((home)->compmask &= ~BIT(2))

#define set_encoding_EXTERNAL(home) ((home)->compmask |= BIT(3))
#define is_encoding_EXTERNAL(home)  ((home)->compmask & BIT(3))
#define rem_encoding_EXTERNAL(home) ((home)->compmask &= ~BIT(3))

#define single_EXTERNAL 0
#define octet_EXTERNAL  1
#define arbit_EXTERNAL  2

#define set_single_EXTERNAL(home) ((home)->encoding.choice = 0)
#define is_single_EXTERNAL(home)  ((home)->encoding.choice == 0)
#define alt_single_EXTERNAL(home) (&((home)->encoding.value.single))

#define set_octet_EXTERNAL(home)  ((home)->encoding.choice = 1)
#define is_octet_EXTERNAL(home)   ((home)->encoding.choice == 1)
#define alt_octet_EXTERNAL(home)  (&((home)->encoding.value.octet))

#define set_arbit_EXTERNAL(home)  ((home)->encoding.choice = 2)
#define is_arbit_EXTERNAL(home)   ((home)->encoding.choice == 2)
#define alt_arbit_EXTERNAL(home)  (&((home)->encoding.value.arbit))

  /*
   * Access macros for EXTERNAL, ASN.1(1994)
   */

#define syntax_EXTERNAL_1994   0
#define pci_EXTERNAL_1994      1
#define ctx_neg_EXTERNAL_1994  2

#define descr_EXTERNAL_1994   1

#define notation_EXTERNAL_1994  0
#define encoded_EXTERNAL_1994   1

#define set_syntax_EXTERNAL_1994(home) ((home)->identification.choice = 0)
#define is_syntax_EXTERNAL_1994(home)  ((home)->identification.choice == 0)
#define alt_syntax_EXTERNAL_1994(home) (&((home)->identification.i1.syntax))

#define set_pci_EXTERNAL_1994(home) ((home)->identification.choice = 1)
#define is_pci_EXTERNAL_1994(home)  ((home)->identification.choice == 1)
#define alt_pci_EXTERNAL_1994(home) (&((home)->identification.i1.pci))

#define set_ctx_neg_EXTERNAL_1994(home) ((home)->identification.choice = 2)
#define is_ctx_neg_EXTERNAL_1994(home)  ((home)->identification.choice == 2)
#define alt_ctx_neg_EXTERNAL_1994(home) (&((home)->identification.i1.ctx_neg))

#define set_descr_EXTERNAL_1994(home)    ((home)->compmask |= BIT(1))
#define is_descr_EXTERNAL_1994(home)     ((home)->compmask & BIT(1))
#define rem_descr_EXTERNAL_1994(home)    ((home)->compmask &= ~BIT(1))

#define set_notation_EXTERNAL_1994(home)  ((home)->value.choice = 0)
#define is_notation_EXTERNAL_1994(home)   ((home)->value.choice == 0)
#define alt_notation_EXTERNAL_1994(home)  (&((home)->value.i1.notation))

#define set_encoded_EXTERNAL_1994(home)   ((home)->value.choice = 1)
#define is_encoded_EXTERNAL_1994(home)    ((home)->value.choice == 1)
#define alt_encoded_EXTERNAL_1994(home)   (&((home)->value.i1.encoded))

#endif /* ASN_HAS_EXTERNAL */

  /*============================================================*/

#if ASN_HAS_SKIP_IGNORE || ASN_HAS_SKIP_ERROR

/*
$begin-intro-type
.IP   SKIPPED
//         Home type for a skipped ASN.1 type.
.LP
$end-intro-type

$begin-global-type
.TH SKIPPED 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         SKIPPED - Home type for a skipped ASN.1 type.
//
.SH   SYNOPSIS
.ft B
.nf
*/
           typedef BYTE_TYPE SKIPPED;
/*
.fi
.ft
.SH   DESCRIPTION
//         The SKIPPED type is the home type for an ASN.1 type
//         which has been marked with the CASN option SKIP.
//
.SH   FILES
//         asntyp.typ
//
$end-global-type
*/

#endif /* ASN_HAS_SKIP_IGNORE || ASN_HAS_SKIP_ERROR */

  /*============================================================*/

#if ASN_HAS_EXTENSIONMARKER

/*
$begin-intro-type
.IP   EXTENSIONMARKER
//         Home type for an extensionmaker component.
.LP
$end-intro-type

$begin-global-type
.TH EXTENSIONMARKER 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         EXTENSIONMARKER - Home type for an extensionmarker component
//
.SH   SYNOPSIS
.ft B
.nf
*/
           typedef struct extensionMarker {
                sint16 n;
                FRAME_PTR fst;
                FRAME_PTR lst;
           } EXTENSIONMARKER;
/*
.fi
.ft
.SH   DESCRIPTION
//         The EXTENSIONMARKER type is the home type for an extensionmarker
//         component. The "n" member contains the number of unrecognized
//         components stored in the list of FRAME structures. The "fst"
//         and "lst" denote the first and last elements in the list.
//
.SH   FILES
//         asntyp.typ
//
.SH   "SEE ALSO"
//         isUnrecogComp, nextUnrecogComp
//
$end-global-type
*/

#endif /* ASN_HAS_EXTENSIONMARKER */

  /*============================================================*/

#if ASN_HAS_EXTENSIBLE_ENUMERATED

/*
$begin-intro-type
.IP   EXTENSIBLE_ENUMERATED
//         Home type for an extensible enumerated type.
.LP
$end-intro-type

$begin-global-type
.TH EXTENSIBLE_ENUMERATED 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         EXTENSIBLE_ENUMERATED - Home type for an extensionmarker component
//
.SH   SYNOPSIS
.ft B
.nf
*/
           typedef struct extensible_enumerated {
                BOOLE  extended;
                sint32 value;
           } EXTENSIBLE_ENUMERATED;
/*
.fi
.ft
.SH   DESCRIPTION
//         The EXTENSIBLE_ENUMERATED type is the home type for an extensible
//         enumerated type. The "extended" member indicates whether the
//         "value" member contains a value that belongs to the known value
//         set of an enumerated type (FALSE) or not (TRUE). In case
//         extended is TRUE the "value" member can contain something
//         else that a real defined value depending on the encoding rules
//         (e.g. enumeration index in case of PER).
//
.SH   FILES
//         asntyp.typ
//
$end-global-type
*/

#endif /* ASN_HAS_EXTENSIBLE_ENUMERATED */

  /*============================================================*/

/*
$begin-intro-type
.IP   NOBODYUSESTHISTYPE
//         Home type for a dummy sequence component.
.LP
$end-intro-type

$begin-global-type
.TH NOBODYUSESTHISTYPE 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         NOBODYUSESTHISTYPE - Home type for a dummy sequence component
//
.SH   SYNOPSIS
.ft B
.nf
*/
           typedef BYTE_TYPE NOBODYUSESTHISTYPE;
/*
.fi
.ft
.SH   DESCRIPTION
//         The NOBODYUSESTHISTYPE type is the home type for a dummy sequence
//         or set component. The dummy component is generated for sequence and
//         set types with an empty component list.
//
.SH   FILES
//         asntyp.typ
//
$end-global-type
*/

  /*============================================================*/

  /*
   * - AsnAllocType, type for pointers.
   * - AsnValueFieldType, type for the determinant member of the ASNHOME
   *   type. It indicates which union member is selected. Valid values
   *   comprise the VF_<N> constants. See ASNHOME.
   * - AsnMemBlockType, kind for an allocated memory block. Valid values
   *   comprise the MB_<N> constants. See asndef.def.
   * - AsnMainType, identifies an ASN.1 type. Valid values comprise
   *   the TYP_<N> constants. See asndef.def.
   * - AsnCompStatus, identifies the role of an type within a structure.
   *   Valid values comrise the TCS_<N> constants. See asndef.def.
   * - AsnFreeType, kind for memory blocks to be releases in freeAsn.
   *   Valid values comprise the FREE_<N> constants. See asndef.def.
   */

typedef size_t    AsnAllocSize;
typedef void      AsnAllocType;

typedef int AsnValueFieldType; /* VF_, see ASNHOME */
typedef int AsnMemBlockType;   /* MB_ */
typedef int AsnMainType;       /* TYP */
typedef int AsnCompStatus;     /* TCS_, VCS_ */
typedef int AsnFreeType;       /* FREE_ */

  /*============================================================*/

#if ASN_MEMORY

/*
$begin-internal-type
//
//    NAME
//         MEM_BLOCK
//
//    DESCRIPTION
//         An ASNHOME handle points to a list of MEM_BLOCKs that contain
//         information about allocated memory blocks that have been
//         associated with an ASNHOME handle. 
//
//         The "mb" member points to an allocated memory block.
//
//         The "mb_type" member contains the type of an allocated memory
//         block:
//
//         MB_FRM
//              A frame buffer.
//
//         MB_ASNH
//              An ASNHOME handle.
//
//         MB_STG
//              All other memory blocks.
//
//         The "next" member points to the next element in the MEM_BLOCK list/
//
$end-internal-type
*/

typedef struct mem_block {
  void             *mb; 
  struct mem_block *next;
  AsnMemBlockType   mb_type;
} MEM_BLOCK;

#define M_isHandleMemBlock(mblock)   ((mblock)->mb_type == MB_ASNH)
#define M_isFrameMemBlock(mblock)    ((mblock)->mb_type == MB_FRM)
#define M_isNormalMemBlock(mblock)   ((mblock)->mb_type == MB_STG)

#endif /* ASN_MEMORY */

  /*============================================================*/

#if ASN_HAS_ASNH || ASN_HAS_LIST || ASN_SEG_STRI || ASN_MEMORY

/*
$begin-intro-type
.IP   ASNHOME
//         Generic home type for all the ASN.1 types.
.LP
$end-intro-type

$begin-global-type
$TH ASNHOME 3
.'\" t
.\" groff -lt -man
.TH ASNHOME 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         ASNHOME - Generic home type for all the ASN.1 types
//
.SH   SYNOPSIS
.ft B
.nf
*/
           typedef struct asnhome {
                union {
#     if ASN_HAS_NULL
                     NULLTYPE                null;
#     endif
#     ifdef __cplusplus
                     BOOLE                   boolVal;
#     else
#       if ASN_HAS_BOOLEAN == 1
                     BOOLE                   bool;
#       endif
#       if ASN_HAS_BOOLEAN == 2
                     BOOLE                   boolVal;
#       endif
#     endif /* __cplusplus */
#     if ASN_HAS_INT32
                     uint8                   ui8;
                     uint16                  ui16;
                     uint32                  ui32;
                     sint8                   i8;
                     sint16                  i16;
                     sint32                  i32;
#     endif
#     if ASN_HAS_INT64
                     sint64                  i64;
#     endif
#     if ASN_HAS_EXTENSIBLE_ENUMERATED
                     EXTENSIBLE_ENUMERATED   xenum;
#     endif 
#     if ASN_HAS_REAL
                     REAL                    real;
#     endif
#     if ASN_HAS_REAL32
                     REAL32                  real32;
#     endif
#     if ASN_HAS_REAL64
                     REAL64                  real64;
#     endif
#     if ASN_HAS_REAL128
                     REAL128                 real128;
#     endif
#     if ASN_DYN_STRI
                     DYNSTR0                 dyns0;
                     DYNSTR8                 dyns8;
                     DYNSTR16                dyns16;
                     DYNSTR                  dyns;
#     endif
#     if ASN_HAS_BMPSTRING
                     ASN_BMPString0          bmpstr0;
                     ASN_BMPString8          bmpstr8;
                     ASN_BMPString16         bmpstr16;
                     ASN_BMPString           bmpstr;
#     endif
#     if ASN_HAS_ANY
                     ANY                     any;
#     endif
#     if ASN_HAS_LIST || ASN_SEG_STRI
                     LIST                    list;
#     endif
#     if ASN_HAS_FRAME
                     struct frame           *frm;
#     endif
#     if ASN_HAS_COMMONTIME
                     struct commontime      *comt;
#     endif
#     if ASN_HAS_DATETIMEZ
                     struct datetimez       *datetime;
#     endif
#     if ASN_STA_STRI
                     struct statstr0        *stas0;
                     struct statstr8        *stas8;
                     struct statstr16       *stas16;
                     struct statstr         *stas;
#     endif
#     if ASN_HAS_OBJID
                     struct objId           *obji;
#     endif
#     if ASN_HAS_EXTERNAL
                     struct external_1990   *ext;
#     endif
#     if ASN_HAS_EXTERNAL_1990
                     struct external_1990   *ext1990;
#     endif
#     if ASN_HAS_EXTERNAL_1994
                     struct external_1994   *ext1994;
#     endif
#     if ASN_HAS_EXTENSIONMARKER
                     struct extensionMarker *extensionmarker;
#     endif
                     BYTE_TYPE              *stru;
                } val;
#     if ASN_MEMORY
                MEM_BLOCK        *head;
#     endif
                struct asnhome   *prev;
                struct asnhome   *next;
                AsnValueFieldType type;
           } ASNHOME;
/*
.SH   DESCRIPTION
//         The ASNHOME type is the generic home type for all the
//         ASN.1 types. The "val" member contains a value of an ASN.1
//         type in an appropriate union member. The "head" member
//         is used in the internal memory management. The "prev" and
//         "next" members are used when an ASNHOME structure is used
//         as a list element. 
.LP
//         The "type" member indicates which "val" member has been selected.
//         The valid values for the "type" member are the VF_ constants.
.LP
.sp 0.5
.TS
.box, center;
.l | l.
.sp 0.5
//         VF_ constant        	ASNHOME.val member
.sp 0.5
._
.sp 0.5
//         VF_BOOL       	BOOLE boolVal/BOOLE bool
.sp 0.5
//         VF_UI8        	uint8 ui8
.sp 0.5
//         VF_UI16       	uint16 ui16
.sp 0.5
//         VF_UI32       	uint32 ui32
.sp 0.5
//         VF_I8         	sint8 i8
.sp 0.5
//         VF_I16       	sint16 i16
.sp 0.5
//         VF_I32        	sint32 i32
.sp 0.5
//         VF_I64       	sint64 i64
.sp 0.5
//         VF_REAL       	REAL real
.sp 0.5
//         VF_STAS0      	struct statstr8 *stas0
.sp 0.5
//         VF_STAS8      	struct statstr8 *stas8
.sp 0.5
//         VF_STAS16    	struct statstr16 *stas16
.sp 0.5
//         VF_STAS       	struct statstr *stas
.sp 0.5
//         VF_DYNS0      	DYNSTR0 dyns0
.sp 0.5
//         VF_DYNS8      	DYNSTR8 dyns8
.sp 0.5
//         VF_DYNS16      	DYNSTR16 dyns16
.sp 0.5
//         VF_DYNS       	DYNSTR dyns
.sp 0.5
//         VF_BMPSTR0      	ASN_BMPString0 bmpstr0
.sp 0.5
//         VF_BMPSTR8      	ASN_BMPString8 bmpstr8
.sp 0.5
//         VF_BMPSTR16      	ASN_BMPString16 bmpstr16
.sp 0.5
//         VF_BMPSTR       	ASN_BMPString bmpstr
.sp 0.5
//         VF_LIST       	LIST list
.sp 0.5
//         VF_FRM        	struct frame *frm
.sp 0.5
//         VF_NULL       	NULLTYPE null
.sp 0.5
//         VF_STRU       	BYTE_TYPE *stru
.sp 0.5
//         VF_ANY        	ANY any
.sp 0.5
//         VF_OBJI       	struct objId *obji
.sp 0.5
//         VF_COMT       	struct commontime *comt/struct datetimez *datetime
.sp 0.5
//         VF_EXT              	struct external_1990 *ext
.sp 0.5
//         VF_EXTERNAL_1990    	struct external_1990 *ext1990
.sp 0.5
//         VF_EXTERNAL_1994    	struct external_1994 *ext1994
.sp 0.5
.TE
.DE
//
.SH   FILES
//         asntyp.typ
//
$end-global-type
*/

#else  /* ASN_HAS_ASNH || ASN_HAS_LIST || ASN_SEG_STRI || ASN_MEMORY */

typedef void ASNHOME;

#endif /* ASN_HAS_ASNH || ASN_HAS_LIST || ASN_SEG_STRI || ASN_MEMORY */

  /*============================================================*/

#if ASN_HAS_ENTRYDESCR

/*
$begin-intro-type
.IP   ENTRYFUNC
//         Generic entry type function pointer.
.LP
$end-intro-type

$begin-global-type
.TH ENTRYFUNC 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         ENTRYFUNC - Generic entry type function pointer type.
//
.SH   SYNOPSIS
.ft B
.nf
//         typedef void (*ENTRYFUNC) (ASNG_F(void *g) void *home);
.fi
.ft
.SH   DESCRIPTION
//         The type ENTRYFUNC is the generic entry function pointer
//         type. The ASN2C generated type specific entry functions
//         can be casted to be ENTRYFUNCs. If the flag ASN_REENTRANT is 
//         set then the "g" argument is pointer to some buffer area 
//         (ASNCODG in case of encoding/decoding functions, ASNTSTG in case
//         of test functions). Otherwise the "g" argument is omitted.
//         The "home" argument points to a home area of an entry type.
//
.SH   FILES
//         asntyp.typ
//
.SH   "SEE ALSO"
//         ENTRYDESCR
//
$end-global-type
*/

typedef void (*ENTRYFUNC) (ASNG_F (void *) void *);

#endif /* ASN_HAS_ENTRYDESCR */

  /*============================================================*/

  /*
   * - AsnTag, 8 or 32 bits wide integer that can hold all the
   *   supported tag values in an internal format.
   * - AsnTag8, 8 bits wide integer that can hold tag values 
   *   when a tag number is in range 0..30.
   */

#if ASN_TAGS_8b_WIDE
typedef uint8  AsnTag;
#else
typedef uint32 AsnTag;
#endif

typedef uint8  AsnTag8;

  /*============================================================*/

  /*
   * - AsnLength, type for lengths.
   */

#if ASN_MAX_LEN_32K
typedef sint16   AsnLength;
#else
typedef FRAME_IX AsnLength;
#endif  /* ASN_MAX_LEN_254 */

  /*============================================================*/

#if ASN_HAS_ENTRYDESCR

/*
$begin-intro-type
.IP   ENTRYDESCR
//         Entry descriptor table element type.
.LP
$end-intro-type

$begin-global-type
$TH ENTRYDESCR 3
.'\" t
.\" groff -lt -man
.TH ENTRYDESCR 3 "$Date: 2002/07/12 08:13:44 $" "Nokia Research Center" "ASN2C RTS"
//
.SH   NAME
//         ENTRYDESCR - Entry descriptor table element type.
//
.SH   SYNOPSIS
.ft B
.nf
//         typedef const struct entrydescr {
!/              <members>
//         } ENTRYDESCR;
.fi
.ft
.SH   DESCRIPTION
//         The ENTRYDESCR type contains members which describe an entry
//         type and which provide hooks for handling of an instance of
//         an entry type. Configuration flags control existence of the
//         structure members.
.LP
//         The members are as follows:
.LP
.IP        "const char *name"
//              The "name" member contains the typereference of an entry type.
//              The typereference is as specified in an ASN.1 module expect
//              that hyphens (-) are translated to underscores (_). The
//              ASN_TABN flag controls existence of this member.
//
.IP        "AsnMainType etype"
//              The ASN.1 type of an entry type. One of the TYP_ constants.
//              See the table below.
//
.IP        "AsnTag tag"
//              The tag of an entry type.
//
.IP        "ENTRYFUNC enc"
.IP        "ENTRYFUNC dec"
.RS
//              The "enc" and "dec" members contain pointers to the ASN2C
//              generated type-specific encoding and decoding functions.
//              The flag ASNCODEC_MODULE controls existence of these
//              members.
.RE
.IP        "ENTRYFUNC ask"
.IP        "ENTRYFUNC print"
//              The "ask" and "print" members contain pointers to the ASN2C
//              generated type-specific ask and print functions.
//              The flag ASNTST_MODULE controls existence of these
//              members.
// 
.IP        "unsigned size"
//              The "size" member contains size of an entry home type
//              in bytes. This member ia available only in stand-alone
//              applications (flag STAND_ALONE_ASN).
//
.IP        "sint16 params"
//              The "params" member contains number of components of an
//              entry type. This member is available only in CVOPS
//              applications (flag STAND_ALONE_ASN).
//
.IP        "const sint16 *partoasn"
//              The dynamic version of "partoasn" member contains information
//              for mapping of CVOPS message parameters to virtual task
//              specific ASN.1 variables. The flag ASN_TABO controls whether
//              this version is selected. This member is available only in
//              CVOPS applications (flag STAND_ALONE_ASN).
//
.IP        "sint16 partoasn [MAX_ASN_PARAMS]"
//              The static version of "partoasn" member contains information
//              for mapping of CVOPS message parameters to virtual task
//              specific ASN.1 variables. The flag ASN_TABO controls whether
//              this version is selected. This member is available only in
//              CVOPS applications (flag STAND_ALONE_ASN).
//
.IP        "int tabRef"
//              The "tabRef" member refers to a GENTAB generated table element.
//              The ASN_TABREF flag controls existence of this member.
//
.LP
//         The TYP_ constants are as follows:
.LP
.sp 0.5
.TS
.box, center;
.l | l.
.sp 0.5
//         TYP_ constant       	ASN.1 type
.sp 0.5
._
.sp 0.5
//         TYP_NONE      	None
.sp 0.5
//         TYP_BOOLEAN    	BOOLEAN
.sp 0.5
//         TYP_INTEGER    	INTEGER
.sp 0.5
//         TYP_ENUM      	ENUMERATED
.sp 0.5
//         TYP_REAL      	REAL
.sp 0.5
//         TYP_BITSTR    	BIT STRING
.sp 0.5
//         TYP_OCTSTR    	OCTET STRING
.sp 0.5
//         TYP_NULL      	NULL
.sp 0.5
//         TYP_SEQ       	SEQUENCE
.sp 0.5
//         TYP_SEQOF    	SEQUENCE OF
.sp 0.5
//         TYP_SET          	SET
.sp 0.5
//         TYP_SETOF        	SET OF
.sp 0.5
//         TYP_CHOICE       	CHOICE
.sp 0.5
//         TYP_ANY          	ANY
.sp 0.5
//         TYP_OPEN         	Open type
.sp 0.5
//         TYP_EMB_PDV      	EMBEDDED PDV
.sp 0.5
//         TYP_OBJID        	OBJECT IDENTIFIER
.sp 0.5
//         TYP_NUMSTR       	NumericString
.sp 0.5
//         TYP_PRISTR       	PrintableString
.sp 0.5
//         TYP_TELSTR       	TeletexString
.sp 0.5
//         TYP_VIDSTR       	VideotexString
.sp 0.5
//         TYP_VISSTR       	VisibleString
.sp 0.5
//         TYP_IA5STR       	IA5String
.sp 0.5
//         TYP_GRASTR       	GraphicString
.sp 0.5
//         TYP_GENSTR       	GeneralString
.sp 0.5
//         TYP_BMPSTR       	BMPString
.sp 0.5
//         TYP_UNISTR       	UniversalString
.sp 0.5
//         TYP_CHASTR       	CHARACTER STRING
.sp 0.5
//         TYP_GENTIME      	GeneralizedTime
.sp 0.5
//         TYP_UTCTIME      	UTCTime
.sp 0.5
//         TYP_EXTERNAL        	EXTERNAL (ASN.1:1990)
.sp 0.5
//         TYP_EXTERNAL_1990   	EXTERNAL (ASN.1:1990)
.sp 0.5
//         TYP_EXTERNAL_1994   	EXTERNAL (ASN.1:1994)
.sp 0.5
//         TYP_OBJDESCR        	ObjectDescriptor
.sp 0.5
//         TYP_INSTANCE_OF     	Instance-of type
.sp 0.5
.TE
.DE
//
.SH   FILES
//         asntyp.typ
//
.SH   "SEE ALSO"
//         ENTRYFUNC
//
$end-global-type
*/

typedef const struct entrydescr {

#if ASN_TABN
  const char   *name;
#endif

  AsnMainType   etype;

#if ASN_TABTAG
  AsnTag        tag;
#endif

#if (ASNCODEC_MODULE == 1 || ASNPER_MODULE == 1 || ASNNER_MODULE == 1)
  ENTRYFUNC     enc;
  ENTRYFUNC     dec;
#endif

#if ASNTST_MODULE
  ENTRYFUNC     ask;
  ENTRYFUNC     print;
#endif

#if ASN_ALT_TAB                       /* For alternative encodings */
  ENTRYFUNC     altEnc [ASN_ALT_MAX];
  ENTRYFUNC     altDec [ASN_ALT_MAX];
#endif

#if STAND_ALONE_ASN
  size_t        size;
#else
  sint16        params;

#if ASN_TABO
  const sint16 *partoasn;
#else
  sint16        partoasn [MAX_ASN_PARAMS];
#endif
#endif

#if ASN_TABREF
  int           tabRef;               /* Refers to a GENTAB generated table */
#endif
} ENTRYDESCR;

typedef ENTRYDESCR *ENTRYDESCR_PTR;

#endif /* ASN_HAS_ENTRYDESCR */

  /*============================================================*/

#if STAND_ALONE_ASN == 0

/*
$begin-internal-type
//
//    NAME
//         C_ASN_TASK_TYPE_STR
//
//    DESCRIPTION
//         The C_ASN_TASK_TYPE_STR type contains virtual task type
//         specific ASN.1 information. The POPVAR tool generates
//         instances of this type.
//
//         The "n" member contains the number of entry types associated
//         with a given task interface (PEER, UP, DOWN etc.).
//
//         The "ed" member contains pointers to ASN2C generated entry
//         descriptor tables for each task interface.
//
$end-internal-type
*/

typedef struct c_asn_task_type_str {
  INDEX          n  [IFCOUNT];
  ENTRYDESCR_PTR ed [IFCOUNT];
} C_ASN_TASK_TYPE_STR;

  /*============================================================*/

/*
$begin-internal-type
//
//    NAME
//         C_ASN_VTASK_STR
//
//    DESCRIPTION
//         The C_ASN_VTASK_STR type contains virtual task specific ASN.1
//         variables.
//
//         The "asn" member is a pointer to POPVAR generated ASNHOME
//         pointer array/structure.
//
$end-internal-type
*/

typedef struct c_asn_vtask_str {
  ASNHOME **asn;
} C_ASN_VTASK_STR;

#endif

  /*============================================================*/

#endif /* ASNTYP_TYP */


