/*---------------------------------------------------------------------------
 * File Name: example_communication.c
 *
 * 2010 IBM Corporation
 * 
 * Description:
 * This is an example implementation of the communication part of the
 * Platform Layer.
 */

#include <stdio.h>
#include <string.h>             /* For strlen, strcmp etc */

#include "t3rts.h"

#include "example_general.h" /* For assert definition */
#include "example_communication.h"
#include "example_time.h"
#include "example_components.h"


/* 'min' is not a standard feature on all platforms: */
#define ex_min(A,B) (((A) <= (B)) ? (A) : (B))


/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
#include "t3tri.h"
#include "example_triutil.h"


#define EX_TRI_MAP_FAILED        "triMap failed!"
#define EX_TRI_UNMAP_FAILED      "triUnmap failed!"
#define EX_TRI_SEND_FAILED       "triSend failed!"
#define EX_TRI_SEND_MC_FAILED    "triSendMC failed!"
#define EX_TRI_SEND_BC_FAILED    "triSendBC failed!"
#define EX_TRI_CALL_FAILED       "triCall failed!"
#define EX_TRI_CALL_MC_FAILED    "triCallMC failed!"
#define EX_TRI_CALL_BC_FAILED    "triCallBC failed!"
#define EX_TRI_REPLY_FAILED      "triReply failed!"
#define EX_TRI_REPLY_MC_FAILED   "triReplyMC failed!"
#define EX_TRI_REPLY_BC_FAILED   "triReplyBC failed!"
#define EX_TRI_RAISE_FAILED      "triRaise failed!"
#define EX_TRI_RAISE_MC_FAILED   "triRaiseMC failed!"
#define EX_TRI_RAISE_BC_FAILED   "triRaiseBC failed!"
#define EX_TRI_SARESET_FAILED    "triSAReset failed!"
#define EX_TRI_PARESET_FAILED    "triPAReset failed!"

#define EX_TRI_ENCODE_FAILED     "Encoding of value in TRI failed!"
#define EX_TRI_ACTION_INFORMAL_FAILED     "triSUTActionInformal failed!"
#define EX_TRI_ACTION_TEMPLATE_FAILED     "triSUTActionTemplate failed!"

t3rt_context_t current_ctx;
void * current_ctx_sem;

#endif
/* T3RTS_TRI_SECTION_END */

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
# ifdef DEBUG
#  define ENABLE_DEFLECT
# endif
#else
/* T3RTS_TRI_SECTION_END */
#ifndef DISABLE_DEFLECT
# define ENABLE_DEFLECT
#endif /*DISABLE_DEFLECT*/
/* T3RTS_TRI_SECTION_START */
#endif /* T3PL_TRI */
/* T3RTS_TRI_SECTION_END */

static t3rt_type_t sys_comptype = NULL;
static t3rt_boolean_t sut_is_component = false;
static t3rt_boolean_t all_timers_active = false;



#ifdef ENABLE_DEFLECT

static void
get_reflected_port(t3rt_binary_string_t       port_address,
                   ex_component_object_id_t * dest_port_id,
                   ex_component_object_id_t * other_port_id,
                   t3rt_binary_string_t     * other_port_address,
                   unsigned long            * other_comp_address,
                   t3rt_context_t ctx);

#endif


#ifdef DEBUG
# include <stdlib.h>             /* For declaration of getenv etc */

static void
debug_print_all_ports(t3rt_context_t ctx);
#endif



static unsigned long
ex_parameter_count(t3rt_type_t type, t3rt_context_t ctx)
{
  if (t3rt_type_field_count(type, ctx) == 0)
    return 0;

  return (t3rt_type_field_count(type, ctx)
          - 1); /* Don't count with the return value */
}



void
t3pl_communication_pre_initialize(int argc, char * argv [], t3rt_context_t ctx)
{
/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  {
    t3rt_value_t val = t3rt_type_instantiate_value(t3rt_boolean_type, t3rt_temp_alloc_c, ctx);
    (void)t3rt_value_set_boolean(val, false, ctx);
    t3rt_rtconf_set_param("t3rt.codecs.builtin_as_default.enabled", val, ctx);
  }

  current_ctx = NULL;
  current_ctx_sem = t3pl_sem_create(1, ctx);
#endif
/* T3RTS_TRI_SECTION_END */
}


static void
ex_check_all_timers_active (t3rt_context_t ctx)
{
  t3rt_value_t      keyvalue;
  
  
  keyvalue = t3rt_rtconf_get_param ("t3rt.timers.assuming_all_active", ctx);

  if (keyvalue == t3rt_no_value_c || keyvalue == t3rt_illegal_value_c) {
      all_timers_active = false;
  } else {
      all_timers_active = t3rt_value_get_boolean (keyvalue, ctx);
  }
}


/*
 * Initializes the communication module
 */

void
t3pl_communication_initialize(t3rt_context_t ctx)
{
  /* Update the static variable 'all_timers_active' */
  ex_check_all_timers_active (ctx);

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  {
    TriStatus      status;
    t3rt_boolean_t sem_status;


    sem_status = t3pl_sem_wait(current_ctx_sem, ctx);
    assert(sem_status);
    assert(current_ctx == NULL);
    current_ctx = ctx;

    status = triSAReset();

    if (status != TRI_OK) {
      current_ctx = NULL;
      sem_status = t3pl_sem_post(current_ctx_sem, ctx);
      assert(sem_status);
      ex_tri_error(EX_TRI_SARESET_FAILED, ctx);
    }

    status = triPAReset();

    current_ctx = NULL;
    sem_status = t3pl_sem_post(current_ctx_sem, ctx);
    assert(sem_status);

    if (status != TRI_OK)
      ex_tri_error(EX_TRI_PARESET_FAILED, ctx);
  }
#endif
/* T3RTS_TRI_SECTION_END */
}


/*
 * Finalizes the communication module
 */

void
t3pl_communication_finalize(t3rt_context_t ctx)
{
/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  t3rt_boolean_t status;
  
  status = t3pl_sem_destroy(current_ctx_sem, ctx);
  assert(status);
#endif
/* T3RTS_TRI_SECTION_END */
}

/* 
 * ex_communication_testcase_terminated
 *
 * This function is called from t3pl_general_testcase_terminated when a 
 * test case has ended. 
 * the ex_concurrency_testcase_terminated deallocates the data storage of
 * components properly.
 */
void
ex_communication_testcase_terminated(t3rt_context_t ctx)
{
  sys_comptype = NULL;
  sut_is_component = false;
}


/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
static void
clear_TriPortId(TriPortId *triPortId, t3rt_context_t ctx)
{
  assert( triPortId != NULL );

  triPortId->aux = NULL;
  triPortId->compInst.compInst.aux = NULL;
  triPortId->compInst.compInst.bits = 0;
  triPortId->compInst.compInst.data = NULL;
  triPortId->compInst.compName = NULL;
  triPortId->compInst.compType.aux = NULL;
  triPortId->compInst.compType.moduleName = NULL;
  triPortId->compInst.compType.objectName = NULL;
  triPortId->portIndex = -1;
  triPortId->portName = NULL;
  triPortId->portType.aux = NULL;
  triPortId->portType.moduleName = NULL;
  triPortId->portType.objectName = NULL;
}

static void
create_TriPortId(t3rt_type_t port_type, const char* portname, long index, 
                 TriPortId *triPortId, t3rt_boolean_t is_system, 
                 t3rt_type_t system_comp_type, t3rt_context_t ctx)
{
  t3rt_type_t               comp_type;
  t3rt_binary_string_t      comp_addr;
  const char *              comp_name;

  clear_TriPortId(triPortId, ctx);

  triPortId->portIndex = index; 

  triPortId->portName = t3pl_memory_allocate(t3rt_perm_alloc_c, strlen(portname) + 1, ctx);
  strcpy(triPortId->portName, portname);
  
  /* Port type info */
  (void)tri__create_QualifiedName(t3rt_type_module(port_type, ctx),
                                  t3rt_type_name(port_type, ctx),
                                  NULL, t3rt_perm_alloc_c, 
                                  &(triPortId->portType), ctx);

  if (is_system) {
    comp_addr = ex_component_get_control_port(EX_SYS_COMPIDX, ctx)->address;
    comp_type = system_comp_type;
    comp_name = "SYSTEM";
  } else {
    comp_addr = t3rt_context_get_component_address(ctx);
    comp_type = t3rt_context_get_component_type(ctx);
    comp_name = t3rt_context_get_component_name(ctx);                                    
  }

  (void)tri__t3_to_tri_BinaryString(comp_addr, 
                                    t3rt_perm_alloc_c, 
                                    &(triPortId->compInst.compInst), ctx);

  triPortId->compInst.compName = (char*)t3pl_memory_allocate(t3rt_perm_alloc_c, strlen(comp_name) + 1, ctx);
  strcpy(triPortId->compInst.compName, comp_name);

  triPortId->compInst.compType.objectName = 
    (char*)t3pl_memory_allocate(t3rt_perm_alloc_c, 
                                strlen(t3rt_type_name(comp_type, ctx)) + 1,
                                ctx);
  strcpy(triPortId->compInst.compType.objectName, t3rt_type_name(comp_type, ctx));

  triPortId->compInst.compType.moduleName =  
    (char*)t3pl_memory_allocate(t3rt_perm_alloc_c, 
                                strlen(t3rt_type_module(comp_type, ctx)) + 1,
                                ctx);
  strcpy(triPortId->compInst.compType.moduleName, t3rt_type_module(comp_type, ctx));

  triPortId->compInst.compType.aux = NULL;
}

static void
destroy_TriPortId(TriPortId *triPortId,  
				  t3rt_context_t ctx)
{
  if (triPortId->compInst.compName)
    t3pl_memory_deallocate(t3rt_perm_alloc_c, triPortId->compInst.compName, ctx);
  if (triPortId->compInst.compType.objectName)
    t3pl_memory_deallocate(t3rt_perm_alloc_c, triPortId->compInst.compType.objectName, ctx);
  if (triPortId->compInst.compType.moduleName)
    t3pl_memory_deallocate(t3rt_perm_alloc_c, triPortId->compInst.compType.moduleName, ctx);

  if (triPortId->portName)
    t3pl_memory_deallocate(t3rt_perm_alloc_c, triPortId->portName, ctx);

  if (triPortId->portType.objectName)
    t3pl_memory_deallocate(t3rt_perm_alloc_c, triPortId->portType.objectName, ctx);
  if (triPortId->portType.moduleName)
    t3pl_memory_deallocate(t3rt_perm_alloc_c, triPortId->portType.moduleName, ctx);

  if (triPortId->compInst.compInst.data)
    t3pl_memory_deallocate(t3rt_perm_alloc_c, triPortId->compInst.compInst.data, ctx);
}

#endif
/* T3RTS_TRI_SECTION_END */

/*
 * Create and initialize a port. Includes any platform dependent
 * communications mechanism, address and queue initializations.
 */

void
t3pl_port_create(t3rt_value_t port_value,
                 t3rt_binary_string_t address,
                 const char* name, /* Port or port array name */
                 long index,    /* Index in port array or -1 */
                 t3rt_context_t ctx)
{
  unsigned long compidx = ex_component_get_index(NULL, ctx);
  ex_port_t port;

  port = ex_component_create_port(compidx, name, index, ctx);

  if (sys_comptype != NULL && !sut_is_component) {
    t3rt_type_t mytype = t3rt_context_get_component_type(ctx);

    if (sys_comptype == mytype) {
      sut_is_component = true;
    }
  }

  t3rt_binary_string_assign(address, port->address, ctx);

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  port->port_repr = 
    (ex_port_repr_t)t3pl_memory_allocate(t3rt_perm_alloc_c, 
                                         sizeof(ex_port_repr_struct_t), 
                                         ctx); 
  create_TriPortId(t3rt_value_type(port_value, ctx), name, 
                   index, &(port->port_repr->triPortId), false, NULL, ctx);
#endif
/* T3RTS_TRI_SECTION_END */
}


/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
#endif
/* T3RTS_TRI_SECTION_END */


void
t3pl_port_create_control_port_for_cpc(t3rt_binary_string_t address,
                                       t3rt_context_t ctx)
{
  /* Actual component and port repr created in t3pl_task_setup.
     Known upcoming CPCP address: */
  t3rt_binary_string_append_4bytes(address, 0, ctx); 
}


/*
 * Start a port
 */

void
t3pl_port_start(t3rt_binary_string_t port_address,
                t3rt_context_t ctx)
{
  ex_port_t port;
  ex_component_object_id_t objid;

  objid = ex_component_get_port_compobj(port_address, ctx);

  port = ex_component_get_port(objid, ctx);

  port->is_active = true;
}


/*
 * Stop a port
 */

void
t3pl_port_stop(t3rt_binary_string_t port_address,
               t3rt_context_t ctx)
{
  ex_port_t port;
  ex_component_object_id_t objid;

  objid = ex_component_get_port_compobj(port_address, ctx);

  port = ex_component_get_port(objid, ctx);

  port->is_active = false;
}

/*
 * Halt a port
 */

void
t3pl_port_halt(t3rt_binary_string_t port_address,
               t3rt_context_t ctx)
{
  /* NOP! */
}

/*
 * Destroy a port
 */

void
t3pl_port_destroy(t3rt_binary_string_t port_address,
                  t3rt_context_t ctx)
{
  ex_port_t port;
  ex_component_object_id_t objid;

  objid = ex_component_get_port_compobj(port_address, ctx);

  port = ex_component_get_port(objid, ctx);

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  if (port->port_repr != NULL) {
	destroy_TriPortId(&(port->port_repr->triPortId), ctx);
    port->port_repr->triPortId.portName = NULL;
    port->port_repr->triPortId.portType.moduleName = NULL;
    port->port_repr->triPortId.portType.objectName = NULL;
    t3pl_memory_deallocate(t3rt_perm_alloc_c, port->port_repr, ctx);
    port->port_repr = NULL;
  }
#endif
/* T3RTS_TRI_SECTION_END */

  ex_component_destroy_port(port, ctx);
}


/*
 * Discard any data contents of the port
 */

void
t3pl_port_clear(t3rt_binary_string_t port_address,
                t3rt_context_t ctx)
{
  /* NOP! */
}


/*
 * Map component port to system port.
 *
 *  The mapping is normally taken care of in the RTL. The function is intended
 *  to be used in the case when the integration implementation needs to be 
 *  aware of the mapping.
 *
 */

void
t3pl_port_map(t3rt_binary_string_t port_address,
              const char *tsi_port_name, /* Port or port array name */
              long index,       /* Index in port array or -1 */
              t3rt_context_t ctx)
{
  /*  Both the component port and the system port representations are assumed
      to have been created previously. 
      All we need to do here is map the PL port structure with each other. This
      is only done in the component to system port direction. The other way
      around is looked up by linear search.
   */

  ex_port_t port, sysport;
  ex_component_object_id_t objid;

  objid = ex_component_get_port_compobj(port_address, ctx);

  port = ex_component_get_port(objid, ctx);

  sysport = ex_component_find_port(EX_SYS_COMPIDX, tsi_port_name, index, ctx);

  assert(sysport != NULL);

  ex_sem_writer_start(&sysport->sem,ctx);
  port->associated_with = sysport->id;
  ex_sequence_add(sysport->sysport_associated_with, t3rt_perm_alloc_c, ctx);
  ex_sequence_last_element(sysport->sysport_associated_with) = port;
  ex_sem_writer_end(&sysport->sem,ctx);

#ifdef DEBUG_WITHOUT_THIS
  if (getenv("T3RT_DEBUG_PRINT_ALL_PORTS"))
    debug_print_all_ports(ctx);
#endif
/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI

  if (triMap(&(port->port_repr->triPortId), &(sysport->port_repr->triPortId)) != TRI_OK) {
    ex_tri_error(EX_TRI_MAP_FAILED, ctx);
  }
#endif
/* T3RTS_TRI_SECTION_END */
}



/*
 * Unmap component port from system port.
 *
 *  The mapping is normally taken care of in the RTL. The function is intended
 *  to be used in the case when the integration implementation needs to be 
 *  aware of the mapping.
 *
 */

void
t3pl_port_unmap(t3rt_binary_string_t port_address, t3rt_context_t ctx)
{
  ex_port_t port, sysport;
  ex_component_object_id_t objid;
  ex_sequence_iter(ex_port_t) iter;

  objid = ex_component_get_port_compobj(port_address, ctx);

  port = ex_component_get_port(objid, ctx);

  assert(port->associated_with != ex_component_no_object); /* Otherwise not mapped! */
  sysport = ex_component_get_port(port->associated_with, ctx);

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  {
    if (triUnmap(&(port->port_repr->triPortId), &(sysport->port_repr->triPortId))) {
      ex_tri_error(EX_TRI_UNMAP_FAILED, ctx);
    }
  }
#endif
/* T3RTS_TRI_SECTION_END */

  ex_sem_writer_start(&sysport->sem,ctx);
  port->associated_with = ex_component_no_object;
  ex_sequence_iter_start(iter,sysport->sysport_associated_with);

  while (ex_sequence_iter_next(iter)) {
    if (ex_sequence_iter_get(iter) == port) {
      if (ex_sequence_iter_remove(iter,ctx))
	break;
    }
  }
  ex_sem_writer_end(&sysport->sem,ctx);
}


/*
 * t3pl_port_component_send
 *
 * This inserts the message into the destination components event queue.
 *
 */
void
t3pl_port_component_send(t3rt_binary_string_t dest_component,
                         t3rt_binary_string_t port_address,
                         t3rt_binary_string_t data,
                         t3rt_context_t ctx)
{
  unsigned long compidx = ex_component_get_index(dest_component, ctx);

  ex_component_queue_put_message(ex_component_get_comp_queue(compidx, ctx),
                                 port_address, data, NULL, ctx);  
}


/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
static void
initialize_componentId(TriComponentId *componentId, t3rt_context_t ctx)
{
  t3rt_type_t comp_type;
  const char *comp_name;

  assert(componentId != NULL);
  
  (void)tri__t3_to_tri_BinaryString(t3rt_context_get_component_address(ctx), 
                                    t3rt_temp_alloc_c, &(componentId->compInst), 
                                    ctx);

  comp_name = t3rt_context_get_component_name(ctx);                                    
  componentId->compName = (char*)t3rt_memory_temp_allocate(strlen(comp_name) + 1, ctx);
  strcpy(componentId->compName, comp_name);
                                    
  comp_type = t3rt_context_get_component_type(ctx);

  (void)tri__create_QualifiedName((char*) t3rt_type_module(comp_type, ctx),
                                  (char*) t3rt_type_name(comp_type, ctx), NULL, 
                                  t3rt_temp_alloc_c, &(componentId->compType), ctx);

}
#endif

/* 
 *  TRI
 *  ComponentId's are created on-the-fly for each send. Allocation for the id
 *  is made in temporary memory.
 */
/* T3RTS_TRI_SECTION_END */

void
t3pl_port_sut_send(t3rt_binary_string_t port_address,
                   t3rt_binary_string_t sut_address,
                   t3rt_binary_string_t data,
                   t3rt_binary_string_t sender_port_address,
                   t3rt_context_t ctx)
{
  ex_component_object_id_t objid = ex_component_get_port_compobj(port_address, ctx);
/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  TriComponentId  componentId;
  BinaryString    sendMessage;
  BinaryString    sutAddress;
  TriStatus       triStatus;
  ex_port_t       destport;

  destport = ex_component_get_port(objid, ctx);

  initialize_componentId(&componentId, ctx);

  (void)tri__t3_to_tri_BinaryString(data, t3rt_temp_alloc_c, &sendMessage, ctx);

  if (sut_address != NULL) {
    (void)tri__t3_to_tri_BinaryString(sut_address, t3rt_temp_alloc_c, 
                                      &sutAddress, ctx);
    triStatus = triSend(&componentId, &(destport->port_repr->triPortId), 
                        &sutAddress, &sendMessage);
  } else
    triStatus = triSend(&componentId, &(destport->port_repr->triPortId), 
                        NULL, &sendMessage);
    
  if (triStatus != TRI_OK)
    ex_tri_error(EX_TRI_SEND_FAILED, ctx);

#else
/* T3RTS_TRI_SECTION_END */

  if (sut_is_component) {
    ex_component_object_id_t other_port_id, dest_port_id;
    t3rt_binary_string_t other_port_address;
    unsigned long other_compidx;

    get_reflected_port(sender_port_address, &dest_port_id, &other_port_id,
                       &other_port_address, &other_compidx, ctx);

#ifdef DEBUG
    fprintf(stderr, "*** Reflecting message on port %lx to port %lx.\n",
            dest_port_id, other_port_id);
#endif

  /* NOTE! 'sender_port_address' is used here as the SUT address from which 
   * this call will seem to originate from in the receiving end. 
   */
    ex_component_queue_put_message(ex_component_get_comp_queue(other_compidx, ctx),
                                   other_port_address, data, sender_port_address, ctx);
    return;
  }
  /* Loop-back implementation.
     Insert the message into the queue of the sending component. */
  /* NOTE! 'sender_port_address' is used here as the SUT address from which 
   * this call will seem to originate from in the receiving end. 
   */
  /* To chose the correct component queue, i.e. the senders own queue, we 
     need re-evaluate the object id for the receiving port. */
  objid = ex_component_get_port_compobj(sender_port_address, ctx);
  ex_component_queue_put_message(ex_component_get_comp_queue(ex_component_index(objid), ctx),
                                 sender_port_address, data, sender_port_address, ctx);
/* T3RTS_TRI_SECTION_START */
#endif
/* T3RTS_TRI_SECTION_END */
}

void
t3pl_port_sut_send_mc(t3rt_binary_string_t port_address,
                      t3rt_binary_string_t sut_address_list[], /* NULL-terminated list */
                      t3rt_binary_string_t data,
                      t3rt_binary_string_t sender_port_address,
                      t3rt_context_t ctx)
{
  ex_component_object_id_t objid = ex_component_get_port_compobj(port_address, ctx);
/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  TriComponentId  componentId;
  BinaryString    sendMessage;
  TriAddressList  sutAddressList;
  TriStatus       triStatus;
  ex_port_t       destport;
  unsigned long   i;
  unsigned long   size;

  destport = ex_component_get_port(objid, ctx);

  initialize_componentId(&componentId, ctx);

  (void)tri__t3_to_tri_BinaryString(data, t3rt_temp_alloc_c, &sendMessage, ctx);

  assert(sut_address_list != NULL);
  for(size = 0; sut_address_list[size]; size++) ; /* calculate list length */

  sutAddressList.length = size;
  sutAddressList.addrList = (TriAddress**)t3rt_memory_temp_allocate(size * sizeof(TriAddress*), ctx);
  for (i = 0; i < size; i++) {
    sutAddressList.addrList[i] = (TriAddress*)t3rt_memory_temp_allocate(sizeof(TriAddress), ctx);
    (void)tri__t3_to_tri_BinaryString(sut_address_list[i], t3rt_temp_alloc_c, sutAddressList.addrList[i], ctx);
  }

  triStatus = triSendMC(&componentId, &(destport->port_repr->triPortId), 
                        &sutAddressList, &sendMessage);
    
  if (triStatus != TRI_OK)
    ex_tri_error(EX_TRI_SEND_MC_FAILED, ctx);

#else
/* T3RTS_TRI_SECTION_END */

  if (sut_is_component) {
    ex_component_object_id_t other_port_id, dest_port_id;
    t3rt_binary_string_t other_port_address;
    unsigned long other_compidx;

    get_reflected_port(sender_port_address, &dest_port_id, &other_port_id,
                       &other_port_address, &other_compidx, ctx);

#ifdef DEBUG
    fprintf(stderr, "*** Reflecting message on port %lx to port %lx.\n",
            dest_port_id, other_port_id);
#endif

  /* NOTE! 'sender_port_address' is used here as the SUT address from which 
   * this call will seem to originate from in the receiving end. 
   */
    ex_component_queue_put_message(ex_component_get_comp_queue(other_compidx, ctx),
                                   other_port_address, data, sender_port_address, ctx);
    return;
  }
  /* Loop-back implementation.
     Insert the message into the queue of the sending component. */
  /* NOTE! 'sender_port_address' is used here as the SUT address from which 
   * this call will seem to originate from in the receiving end. 
   */
  /* To chose the correct component queue, i.e. the senders own queue, we 
     need re-evaluate the object id for the receiving port. */
  objid = ex_component_get_port_compobj(sender_port_address, ctx);
  ex_component_queue_put_message(ex_component_get_comp_queue(ex_component_index(objid), ctx),
                                 sender_port_address, data, sender_port_address, ctx);
/* T3RTS_TRI_SECTION_START */
#endif
/* T3RTS_TRI_SECTION_END */
}

void
t3pl_port_sut_send_bc(t3rt_binary_string_t port_address,
                      t3rt_binary_string_t data,
                      t3rt_binary_string_t sender_port_address,
                      t3rt_context_t ctx)
{
  ex_component_object_id_t objid = ex_component_get_port_compobj(port_address, ctx);
/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  TriComponentId  componentId;
  BinaryString    sendMessage;
  TriStatus       triStatus;
  ex_port_t       destport;

  destport = ex_component_get_port(objid, ctx);

  initialize_componentId(&componentId, ctx);

  (void)tri__t3_to_tri_BinaryString(data, t3rt_temp_alloc_c, &sendMessage, ctx);

  triStatus = triSendBC(&componentId, &(destport->port_repr->triPortId), &sendMessage);
    
  if (triStatus != TRI_OK)
    ex_tri_error(EX_TRI_SEND_BC_FAILED, ctx);

#else
/* T3RTS_TRI_SECTION_END */

  if (sut_is_component) {
    ex_component_object_id_t other_port_id, dest_port_id;
    t3rt_binary_string_t other_port_address;
    unsigned long other_compidx;

    get_reflected_port(sender_port_address, &dest_port_id, &other_port_id,
                       &other_port_address, &other_compidx, ctx);

#ifdef DEBUG
    fprintf(stderr, "*** Reflecting message on port %lx to port %lx.\n",
            dest_port_id, other_port_id);
#endif

  /* NOTE! 'sender_port_address' is used here as the SUT address from which 
   * this call will seem to originate from in the receiving end. 
   */
    ex_component_queue_put_message(ex_component_get_comp_queue(other_compidx, ctx),
                                   other_port_address, data, sender_port_address, ctx);
    return;
  }
  /* Loop-back implementation.
     Insert the message into the queue of the sending component. */
  /* NOTE! 'sender_port_address' is used here as the SUT address from which 
   * this call will seem to originate from in the receiving end. 
   */
  /* To chose the correct component queue, i.e. the senders own queue, we 
     need re-evaluate the object id for the receiving port. */
  objid = ex_component_get_port_compobj(sender_port_address, ctx);
  ex_component_queue_put_message(ex_component_get_comp_queue(ex_component_index(objid), ctx),
                                 sender_port_address, data, sender_port_address, ctx);
/* T3RTS_TRI_SECTION_START */
#endif
/* T3RTS_TRI_SECTION_END */
}

/*
 * Procedure based communication interface
 */


void *
t3pl_port_sut_call(t3rt_binary_string_t port_to_call,
                   t3rt_binary_string_t sut_address,
                   t3rt_binary_string_t caller_port,
                   t3rt_binary_string_t address,
                   t3rt_type_t signature_type,
                   t3rt_binary_string_t parameters[],
                   t3rt_context_t ctx)
{
  ex_component_object_id_t dest_port_id;

#ifdef ENABLE_DEFLECT
  ex_component_object_id_t other_port_id;
  t3rt_binary_string_t other_port_address;
  unsigned long other_compidx;

#endif

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
    ex_port_t         destport;
    TriComponentId    componentId;
    TriPortId        *tsiPortId;
    unsigned long     param_count;
    TriParameterList  callParams;
    unsigned long     i;
    TriSignatureId    signatureId;
    TriStatus         triStatus;
    BinaryString      sutAddress;

    dest_port_id = ex_component_get_port_compobj(port_to_call, ctx);

    destport = ex_component_get_port(dest_port_id, ctx);

    initialize_componentId(&componentId, ctx);
    tsiPortId = &(destport->port_repr->triPortId);

    /* Find out # params */
    param_count = ex_parameter_count(signature_type, ctx);

    /* Allocate suitable parameter list (temp) */
    tri__create_TriParameterList((long)param_count, t3rt_temp_alloc_c,
                                 &callParams, ctx);

    /* Fill the array with transformed data */
    for (i = 0; i < param_count; ++i) {
      t3rt_field_properties_t prop;

      prop = t3rt_type_field_properties(signature_type, i+1, ctx);
      tri__create_TriParameter(parameters[i], prop, t3rt_temp_alloc_c,
                               callParams.parList[i], ctx);
    }

    signatureId.moduleName =
      (char *)t3rt_type_definition_module(signature_type, ctx);
    signatureId.objectName =
      (char *)t3rt_type_definition_name(signature_type, ctx);
    signatureId.aux = NULL;

    if (sut_address != NULL) {
      (void)tri__t3_to_tri_BinaryString(sut_address, t3rt_temp_alloc_c, 
                                        &sutAddress, ctx);
      triStatus = triCall(&componentId, tsiPortId, &sutAddress, 
                          &signatureId, &callParams);
    } else
      triStatus = triCall(&componentId, tsiPortId, NULL, 
                          &signatureId, &callParams);

    if (triStatus != TRI_OK)
      ex_tri_error(EX_TRI_CALL_FAILED, ctx);

#endif /* T3PL_TRI */
/* T3RTS_TRI_SECTION_END */

#ifdef ENABLE_DEFLECT
  if (!sut_is_component) {
    fprintf(stderr, "PL WARNING: Call without SUT component specified!\n");
    return NULL;
  }

  get_reflected_port(caller_port, &dest_port_id, &other_port_id,
                     &other_port_address, &other_compidx, ctx);

#ifdef DEBUG
  fprintf(stderr, "*** Reflecting call on port %lx to port %lx.\n",
          dest_port_id, other_port_id);
#endif

  /* NOTE! 'caller_port' address is used here as the SUT address from which 
   * this call will seem to originate from in the receiving end. 
   */
  ex_component_queue_put_call(ex_component_get_comp_queue(other_compidx, ctx),
                              other_port_address,
                              signature_type,
                              parameters,
                              caller_port,
                              ctx);

#endif /* ENABLE_DEFLECT */

  return NULL;
}

void *
t3pl_port_sut_call_mc(t3rt_binary_string_t port_to_call,
                      t3rt_binary_string_t sut_address_list[], /* NULL-terminated list */
                      t3rt_binary_string_t caller_port,
                      t3rt_binary_string_t address,
                      t3rt_type_t          signature_type,
                      t3rt_binary_string_t parameters[],
                      t3rt_context_t ctx)
{
  ex_component_object_id_t dest_port_id;

#ifdef ENABLE_DEFLECT
  ex_component_object_id_t other_port_id;
  t3rt_binary_string_t other_port_address;
  unsigned long other_compidx;

#endif

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
    ex_port_t         destport;
    TriComponentId    componentId;
    TriPortId        *tsiPortId;
    unsigned long     param_count;
    TriParameterList  callParams;
    unsigned long     i;
    TriSignatureId    signatureId;
    TriStatus         triStatus;
    TriAddressList    sutAddressList;
    unsigned long     size;

    dest_port_id = ex_component_get_port_compobj(port_to_call, ctx);

    destport = ex_component_get_port(dest_port_id, ctx);

    initialize_componentId(&componentId, ctx);
    tsiPortId = &(destport->port_repr->triPortId);

    /* Find out # params */
    param_count = ex_parameter_count(signature_type, ctx);

    /* Allocate suitable parameter list (temp) */
    tri__create_TriParameterList((long)param_count, t3rt_temp_alloc_c,
                                 &callParams, ctx);

    /* Fill the array with transformed data */
    for (i = 0; i < param_count; ++i) {
      t3rt_field_properties_t prop;

      prop = t3rt_type_field_properties(signature_type, i+1, ctx);
      tri__create_TriParameter(parameters[i], prop, t3rt_temp_alloc_c,
                               callParams.parList[i], ctx);
    }

    signatureId.moduleName =
      (char *)t3rt_type_definition_module(signature_type, ctx);
    signatureId.objectName =
      (char *)t3rt_type_definition_name(signature_type, ctx);
    signatureId.aux = NULL;

    assert(sut_address_list != NULL);
    for(size = 0; sut_address_list[size]; size++) ; /* calculate list length */

    sutAddressList.length = size;
    sutAddressList.addrList = (TriAddress**)t3rt_memory_temp_allocate(size * sizeof(TriAddress*), ctx);
    for (i = 0; i < size; i++) {
      sutAddressList.addrList[i] = (TriAddress*)t3rt_memory_temp_allocate(sizeof(TriAddress), ctx);
      (void)tri__t3_to_tri_BinaryString(sut_address_list[i], t3rt_temp_alloc_c, sutAddressList.addrList[i], ctx);
    }

    triStatus = triCallMC(&componentId, tsiPortId, &sutAddressList, 
                          &signatureId, &callParams);

    if (triStatus != TRI_OK)
      ex_tri_error(EX_TRI_CALL_MC_FAILED, ctx);

#endif /* T3PL_TRI */
/* T3RTS_TRI_SECTION_END */

#ifdef ENABLE_DEFLECT
  if (!sut_is_component) {
    fprintf(stderr, "PL WARNING: Call without SUT component specified!\n");
    return NULL;
  }

  get_reflected_port(caller_port, &dest_port_id, &other_port_id,
                     &other_port_address, &other_compidx, ctx);

#ifdef DEBUG
  fprintf(stderr, "*** Reflecting call on port %lx to port %lx.\n",
          dest_port_id, other_port_id);
#endif

  /* NOTE! 'caller_port' address is used here as the SUT address from which 
   * this call will seem to originate from in the receiving end. 
   */
  ex_component_queue_put_call(ex_component_get_comp_queue(other_compidx, ctx),
                              other_port_address,
                              signature_type,
                              parameters,
                              caller_port,
                              ctx);

#endif /* ENABLE_DEFLECT */

  return NULL;
}

void *
t3pl_port_sut_call_bc(t3rt_binary_string_t  port_to_call,
                      t3rt_binary_string_t  caller_port,
                      t3rt_binary_string_t  address,
                      t3rt_type_t           signature_type,
                      t3rt_binary_string_t  parameters[],
                      t3rt_context_t ctx)
{
  ex_component_object_id_t dest_port_id;

#ifdef ENABLE_DEFLECT
  ex_component_object_id_t other_port_id;
  t3rt_binary_string_t other_port_address;
  unsigned long other_compidx;

#endif

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
    ex_port_t         destport;
    TriComponentId    componentId;
    TriPortId        *tsiPortId;
    unsigned long     param_count;
    TriParameterList  callParams;
    unsigned long     i;
    TriSignatureId    signatureId;
    TriStatus         triStatus;

    dest_port_id = ex_component_get_port_compobj(port_to_call, ctx);

    destport = ex_component_get_port(dest_port_id, ctx);

    initialize_componentId(&componentId, ctx);
    tsiPortId = &(destport->port_repr->triPortId);

    /* Find out # params */
    param_count = ex_parameter_count(signature_type, ctx);

    /* Allocate suitable parameter list (temp) */
    tri__create_TriParameterList((long)param_count, t3rt_temp_alloc_c,
                                 &callParams, ctx);

    /* Fill the array with transformed data */
    for (i = 0; i < param_count; ++i) {
      t3rt_field_properties_t prop;

      prop = t3rt_type_field_properties(signature_type, i+1, ctx);
      tri__create_TriParameter(parameters[i], prop, t3rt_temp_alloc_c,
                               callParams.parList[i], ctx);
    }

    signatureId.moduleName =
      (char *)t3rt_type_definition_module(signature_type, ctx);
    signatureId.objectName =
      (char *)t3rt_type_definition_name(signature_type, ctx);
    signatureId.aux = NULL;

    triStatus = triCallBC(&componentId, tsiPortId,
                          &signatureId, &callParams);

    if (triStatus != TRI_OK)
      ex_tri_error(EX_TRI_CALL_BC_FAILED, ctx);

#endif /* T3PL_TRI */
/* T3RTS_TRI_SECTION_END */

#ifdef ENABLE_DEFLECT
  if (!sut_is_component) {
    fprintf(stderr, "PL WARNING: Call without SUT component specified!\n");
    return NULL;
  }

  get_reflected_port(caller_port, &dest_port_id, &other_port_id,
                     &other_port_address, &other_compidx, ctx);

#ifdef DEBUG
  fprintf(stderr, "*** Reflecting call on port %lx to port %lx.\n",
          dest_port_id, other_port_id);
#endif

  /* NOTE! 'caller_port' address is used here as the SUT address from which 
   * this call will seem to originate from in the receiving end. 
   */
  ex_component_queue_put_call(ex_component_get_comp_queue(other_compidx, ctx),
                              other_port_address,
                              signature_type,
                              parameters,
                              caller_port,
                              ctx);

#endif /* ENABLE_DEFLECT */

  return NULL;
}

void
t3pl_port_sut_call_done(void * handle, t3rt_context_t ctx)
{
  /* Nothing implemented for default integration. */
}


void
t3pl_port_sut_call_abort(void * handle, t3rt_context_t ctx)
{
  /* Nothing implemented for default integration. */
}


void
t3pl_port_sut_reply(t3rt_binary_string_t port_to_call,
                    t3rt_binary_string_t sut_address,
                    t3rt_binary_string_t caller_port,
                    t3rt_binary_string_t address,
                    t3rt_type_t          signature_type,
                    t3rt_binary_string_t parameters[],
                    t3rt_binary_string_t return_value,
                    t3rt_context_t ctx)
{
  ex_component_object_id_t dest_port_id;

#ifdef ENABLE_DEFLECT
  ex_component_object_id_t other_port_id;
  t3rt_binary_string_t other_port_address;
  unsigned long other_compidx;
#endif

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
    unsigned long     i;
    ex_port_t         destport;
    TriComponentId    componentId;
    TriPortId        *tsiPortId;
    unsigned long     param_count;
    TriParameterList  replyParams;
    TriParameter      returnValue;
    TriSignatureId    signatureId;
    BinaryString      sutAddress;
    TriStatus         triStatus;

    dest_port_id = ex_component_get_port_compobj(port_to_call, ctx);

    destport = ex_component_get_port(dest_port_id, ctx);

    initialize_componentId(&componentId, ctx);
    tsiPortId = &(destport->port_repr->triPortId);

    /* Find out # params */
    param_count = ex_parameter_count(signature_type, ctx);

    /* Allocate suitable parameter list (temp) */
    tri__create_TriParameterList((long)param_count, t3rt_temp_alloc_c,
                                 &replyParams, ctx);

    /* Fill the array with transformed data */
    for (i = 0; i < param_count; ++i) {
      t3rt_field_properties_t prop;

      prop = t3rt_type_field_properties(signature_type, i+1, ctx);
      tri__create_TriParameter(parameters[i], prop, t3rt_temp_alloc_c,
                               replyParams.parList[i], ctx);
    }

    /* Initialization of BinaryString in returnValue.par */
    returnValue.par.aux = NULL;
    returnValue.par.bits = 0;
    returnValue.par.data = NULL;
    
    /* TRI seems to miss a property for return value, so we use 'out': */
    tri__create_TriParameter(return_value,
                             t3rt_field_property_out_value_c,
                             t3rt_temp_alloc_c,
                             &returnValue,
                             ctx);

    signatureId.moduleName =
      (char *)t3rt_type_definition_module(signature_type, ctx);
    signatureId.objectName =
      (char *)t3rt_type_definition_name(signature_type, ctx);
    signatureId.aux = NULL;

    if (sut_address != NULL) {
      (void)tri__t3_to_tri_BinaryString(sut_address, t3rt_temp_alloc_c, 
                                        &sutAddress, ctx);
      triStatus = triReply(&componentId, tsiPortId, &sutAddress,
                           &signatureId, &replyParams, &returnValue);
    } else
      triStatus = triReply(&componentId, tsiPortId, NULL,
                           &signatureId, &replyParams, &returnValue);
    
    if (triStatus != TRI_OK)
      ex_tri_error(EX_TRI_REPLY_FAILED, ctx);

#endif /* T3PL_TRI */
/* T3RTS_TRI_SECTION_END */

#ifdef ENABLE_DEFLECT
  if (!sut_is_component) {
    fprintf(stderr, "PL WARNING: Reply without SUT component specified!\n");
    return;
  }

  get_reflected_port(caller_port, &dest_port_id, &other_port_id,
                     &other_port_address, &other_compidx, ctx);

#ifdef DEBUG
  fprintf(stderr, "*** Reflecting reply on port %lx to port %lx.\n",
          dest_port_id, other_port_id);
#endif

  /* NOTE! 'caller_port' address is used here as the SUT address from which 
   * this call will seem to originate from in the receiving end. 
   */
  ex_component_queue_put_reply(ex_component_get_comp_queue(other_compidx, ctx),
                               other_port_address,
                               signature_type,
                               parameters,
                               return_value,
                               caller_port,
                               ctx);
  return;

#endif /* ENABLE_DEFLECT */
}

void
t3pl_port_sut_reply_mc(t3rt_binary_string_t port_to_call,
                       t3rt_binary_string_t sut_address_list[],
                       t3rt_binary_string_t caller_port,
                       t3rt_binary_string_t address,
                       t3rt_type_t          signature_type,
                       t3rt_binary_string_t parameters[],
                       t3rt_binary_string_t return_value,
                       t3rt_context_t ctx)
{
  ex_component_object_id_t dest_port_id;

#ifdef ENABLE_DEFLECT
  ex_component_object_id_t other_port_id;
  t3rt_binary_string_t other_port_address;
  unsigned long other_compidx;
#endif

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
    unsigned long     i;
    ex_port_t         destport;
    TriComponentId    componentId;
    TriPortId        *tsiPortId;
    unsigned long     param_count;
    TriParameterList  replyParams;
    TriParameter      returnValue;
    TriSignatureId    signatureId;
    TriStatus         triStatus;
    TriAddressList    sutAddressList;
    unsigned long     size;

    dest_port_id = ex_component_get_port_compobj(port_to_call, ctx);

    destport = ex_component_get_port(dest_port_id, ctx);

    initialize_componentId(&componentId, ctx);
    tsiPortId = &(destport->port_repr->triPortId);

    /* Find out # params */
    param_count = ex_parameter_count(signature_type, ctx);

    /* Allocate suitable parameter list (temp) */
    tri__create_TriParameterList((long)param_count, t3rt_temp_alloc_c,
                                 &replyParams, ctx);

    /* Fill the array with transformed data */
    for (i = 0; i < param_count; ++i) {
      t3rt_field_properties_t prop;

      prop = t3rt_type_field_properties(signature_type, i+1, ctx);
      tri__create_TriParameter(parameters[i], prop, t3rt_temp_alloc_c,
                               replyParams.parList[i], ctx);
    }

    /* Initialization of BinaryString in returnValue.par */
    returnValue.par.aux = NULL;
    returnValue.par.bits = 0;
    returnValue.par.data = NULL;
    
    /* TRI seems to miss a property for return value, so we use 'out': */
    tri__create_TriParameter(return_value,
                             t3rt_field_property_out_value_c,
                             t3rt_temp_alloc_c,
                             &returnValue,
                             ctx);

    signatureId.moduleName =
      (char *)t3rt_type_definition_module(signature_type, ctx);
    signatureId.objectName =
      (char *)t3rt_type_definition_name(signature_type, ctx);
    signatureId.aux = NULL;

    assert(sut_address_list != NULL);
    for(size = 0; sut_address_list[size]; size++) ; /* calculate list length */

    sutAddressList.length = size;
    sutAddressList.addrList = (TriAddress**)t3rt_memory_temp_allocate(size * sizeof(TriAddress*), ctx);
    for (i = 0; i < size; i++) {
      sutAddressList.addrList[i] = (TriAddress*)t3rt_memory_temp_allocate(sizeof(TriAddress), ctx);
      (void)tri__t3_to_tri_BinaryString(sut_address_list[i], t3rt_temp_alloc_c, sutAddressList.addrList[i], ctx);
    }

    triStatus = triReplyMC(&componentId, tsiPortId, &sutAddressList,
                           &signatureId, &replyParams, &returnValue);

    if (triStatus != TRI_OK)
      ex_tri_error(EX_TRI_REPLY_MC_FAILED, ctx);

#endif /* T3PL_TRI */
/* T3RTS_TRI_SECTION_END */

#ifdef ENABLE_DEFLECT
  if (!sut_is_component) {
    fprintf(stderr, "PL WARNING: Reply without SUT component specified!\n");
    return;
  }

  get_reflected_port(caller_port, &dest_port_id, &other_port_id,
                     &other_port_address, &other_compidx, ctx);

#ifdef DEBUG
  fprintf(stderr, "*** Reflecting reply on port %lx to port %lx.\n",
          dest_port_id, other_port_id);
#endif

  /* NOTE! 'caller_port' address is used here as the SUT address from which 
   * this call will seem to originate from in the receiving end. 
   */
  ex_component_queue_put_reply(ex_component_get_comp_queue(other_compidx, ctx),
                               other_port_address,
                               signature_type,
                               parameters,
                               return_value,
                               caller_port,
                               ctx);
  return;

#endif /* ENABLE_DEFLECT */
}

void
t3pl_port_sut_reply_bc(t3rt_binary_string_t port_to_call,
                       t3rt_binary_string_t caller_port,
                       t3rt_binary_string_t address,
                       t3rt_type_t          signature_type,
                       t3rt_binary_string_t parameters[],
                       t3rt_binary_string_t return_value,
                       t3rt_context_t ctx)
{
  ex_component_object_id_t dest_port_id;

#ifdef ENABLE_DEFLECT
  ex_component_object_id_t other_port_id;
  t3rt_binary_string_t other_port_address;
  unsigned long other_compidx;
#endif

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
    unsigned long     i;
    ex_port_t         destport;
    TriComponentId    componentId;
    TriPortId        *tsiPortId;
    unsigned long     param_count;
    TriParameterList  replyParams;
    TriParameter      returnValue;
    TriSignatureId    signatureId;
    TriStatus         triStatus;

    dest_port_id = ex_component_get_port_compobj(port_to_call, ctx);

    destport = ex_component_get_port(dest_port_id, ctx);

    initialize_componentId(&componentId, ctx);
    tsiPortId = &(destport->port_repr->triPortId);

    /* Find out # params */
    param_count = ex_parameter_count(signature_type, ctx);

    /* Allocate suitable parameter list (temp) */
    tri__create_TriParameterList((long)param_count, t3rt_temp_alloc_c,
                                 &replyParams, ctx);

    /* Fill the array with transformed data */
    for (i = 0; i < param_count; ++i) {
      t3rt_field_properties_t prop;

      prop = t3rt_type_field_properties(signature_type, i+1, ctx);
      tri__create_TriParameter(parameters[i], prop, t3rt_temp_alloc_c,
                               replyParams.parList[i], ctx);
    }

    /* Initialization of BinaryString in returnValue.par */
    returnValue.par.aux = NULL;
    returnValue.par.bits = 0;
    returnValue.par.data = NULL;
    
    /* TRI seems to miss a property for return value, so we use 'out': */
    tri__create_TriParameter(return_value,
                             t3rt_field_property_out_value_c,
                             t3rt_temp_alloc_c,
                             &returnValue,
                             ctx);

    signatureId.moduleName =
      (char *)t3rt_type_definition_module(signature_type, ctx);
    signatureId.objectName =
      (char *)t3rt_type_definition_name(signature_type, ctx);
    signatureId.aux = NULL;

    triStatus = triReplyBC(&componentId, tsiPortId,
                           &signatureId, &replyParams, &returnValue);

    if (triStatus != TRI_OK)
      ex_tri_error(EX_TRI_REPLY_BC_FAILED, ctx);

#endif /* T3PL_TRI */
/* T3RTS_TRI_SECTION_END */

#ifdef ENABLE_DEFLECT
  if (!sut_is_component) {
    fprintf(stderr, "PL WARNING: Reply without SUT component specified!\n");
    return;
  }

  get_reflected_port(caller_port, &dest_port_id, &other_port_id,
                     &other_port_address, &other_compidx, ctx);

#ifdef DEBUG
  fprintf(stderr, "*** Reflecting reply on port %lx to port %lx.\n",
          dest_port_id, other_port_id);
#endif

  /* NOTE! 'caller_port' address is used here as the SUT address from which 
   * this call will seem to originate from in the receiving end. 
   */
  ex_component_queue_put_reply(ex_component_get_comp_queue(other_compidx, ctx),
                               other_port_address,
                               signature_type,
                               parameters,
                               return_value,
                               caller_port,
                               ctx);
  return;

#endif /* ENABLE_DEFLECT */
}

void
t3pl_port_sut_raise(t3rt_binary_string_t port_to_call,
                    t3rt_binary_string_t sut_address,
                    t3rt_binary_string_t caller_port,
                    t3rt_binary_string_t address,
                    t3rt_type_t          signature_type,
                    t3rt_binary_string_t exception_value,
                    t3rt_context_t ctx)
{
  ex_component_object_id_t dest_port_id;

#ifdef ENABLE_DEFLECT
  ex_component_object_id_t other_port_id;
  t3rt_binary_string_t other_port_address;
  unsigned long other_compidx;
#endif

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
    ex_port_t       destport;
    TriComponentId  componentId;
    TriPortId      *tsiPortId;
    TriException    exceptionValue;
    TriSignatureId  signatureId;
    BinaryString    sutAddress;
    TriStatus       triStatus;

    dest_port_id = ex_component_get_port_compobj(port_to_call, ctx);

    destport = ex_component_get_port(dest_port_id, ctx);

    initialize_componentId(&componentId, ctx);
    tsiPortId = &(destport->port_repr->triPortId);

    (void)tri__t3_to_tri_BinaryString(exception_value, t3rt_temp_alloc_c,
                                      &exceptionValue, ctx);

    signatureId.moduleName =
      (char *)t3rt_type_definition_module(signature_type, ctx);
    signatureId.objectName =
      (char *)t3rt_type_definition_name(signature_type, ctx);
    signatureId.aux = NULL;

    if (sut_address != NULL) {
      (void)tri__t3_to_tri_BinaryString(sut_address, t3rt_temp_alloc_c, 
                                        &sutAddress, ctx);
      triStatus = triRaise(&componentId, tsiPortId, &sutAddress,
                           &signatureId, &exceptionValue);
    } else
      triStatus = triRaise(&componentId, tsiPortId, NULL,
                           &signatureId, &exceptionValue);
    
    if (triStatus != TRI_OK)
      ex_tri_error(EX_TRI_RAISE_FAILED, ctx);

#endif
/* T3RTS_TRI_SECTION_END */

#ifdef ENABLE_DEFLECT
  if (!sut_is_component) {
    fprintf(stderr, "PL WARNING: Raise without SUT component specified!\n");
    return;
  }

  get_reflected_port(caller_port, &dest_port_id, &other_port_id,
                     &other_port_address, &other_compidx, ctx);

#ifdef DEBUG
  fprintf(stderr, "*** Reflecting raise on port %lx to port %lx.\n",
          dest_port_id, other_port_id);
#endif

  /* NOTE! 'caller_port' address is used here as the SUT address from which 
   * this call will seem to originate from in the receiving end. 
   */
  ex_component_queue_put_exception(ex_component_get_comp_queue(other_compidx, ctx),
                                   other_port_address,
                                   signature_type,
                                   exception_value,
                                   caller_port,
                                   ctx);
  return;

#endif /* ENABLE_DEFLECT */
}

void
t3pl_port_sut_raise_mc(t3rt_binary_string_t port_to_call,
                       t3rt_binary_string_t sut_address_list[],
                       t3rt_binary_string_t caller_port,
                       t3rt_binary_string_t address,
                       t3rt_type_t          signature_type,
                       t3rt_binary_string_t exception_value,
                       t3rt_context_t ctx)
{
  ex_component_object_id_t dest_port_id;

#ifdef ENABLE_DEFLECT
  ex_component_object_id_t other_port_id;
  t3rt_binary_string_t other_port_address;
  unsigned long other_compidx;
#endif

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
    ex_port_t       destport;
    TriComponentId  componentId;
    TriPortId      *tsiPortId;
    TriException    exceptionValue;
    TriSignatureId  signatureId;
    TriStatus       triStatus;
    TriAddressList  sutAddressList;
    unsigned long   size;
    unsigned long   i;

    dest_port_id = ex_component_get_port_compobj(port_to_call, ctx);

    destport = ex_component_get_port(dest_port_id, ctx);

    initialize_componentId(&componentId, ctx);
    tsiPortId = &(destport->port_repr->triPortId);

    (void)tri__t3_to_tri_BinaryString(exception_value, t3rt_temp_alloc_c,
                                      &exceptionValue, ctx);

    signatureId.moduleName =
      (char *)t3rt_type_definition_module(signature_type, ctx);
    signatureId.objectName =
      (char *)t3rt_type_definition_name(signature_type, ctx);
    signatureId.aux = NULL;

    assert(sut_address_list != NULL);
    for(size = 0; sut_address_list[size]; size++) ; /* calculate list length */

    sutAddressList.length = size;
    sutAddressList.addrList = (TriAddress**)t3rt_memory_temp_allocate(size * sizeof(TriAddress*), ctx);
    for (i = 0; i < size; i++) {
      sutAddressList.addrList[i] = (TriAddress*)t3rt_memory_temp_allocate(sizeof(TriAddress), ctx);
      (void)tri__t3_to_tri_BinaryString(sut_address_list[i], t3rt_temp_alloc_c, sutAddressList.addrList[i], ctx);
    }

    triStatus = triRaiseMC(&componentId, tsiPortId, &sutAddressList,
                           &signatureId, &exceptionValue);
    
    if (triStatus != TRI_OK)
      ex_tri_error(EX_TRI_RAISE_MC_FAILED, ctx);

#endif
/* T3RTS_TRI_SECTION_END */

#ifdef ENABLE_DEFLECT
  if (!sut_is_component) {
    fprintf(stderr, "PL WARNING: Raise without SUT component specified!\n");
    return;
  }

  get_reflected_port(caller_port, &dest_port_id, &other_port_id,
                     &other_port_address, &other_compidx, ctx);

#ifdef DEBUG
  fprintf(stderr, "*** Reflecting raise on port %lx to port %lx.\n",
          dest_port_id, other_port_id);
#endif

  /* NOTE! 'caller_port' address is used here as the SUT address from which 
   * this call will seem to originate from in the receiving end. 
   */
  ex_component_queue_put_exception(ex_component_get_comp_queue(other_compidx, ctx),
                                   other_port_address,
                                   signature_type,
                                   exception_value,
                                   caller_port,
                                   ctx);
  return;

#endif /* ENABLE_DEFLECT */
}

void
t3pl_port_sut_raise_bc(t3rt_binary_string_t port_to_call,
                       t3rt_binary_string_t caller_port,
                       t3rt_binary_string_t address,
                       t3rt_type_t          signature_type,
                       t3rt_binary_string_t exception_value,
                       t3rt_context_t ctx)
{
  ex_component_object_id_t dest_port_id;

#ifdef ENABLE_DEFLECT
  ex_component_object_id_t other_port_id;
  t3rt_binary_string_t other_port_address;
  unsigned long other_compidx;
#endif

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
    ex_port_t       destport;
    TriComponentId  componentId;
    TriPortId      *tsiPortId;
    TriException    exceptionValue;
    TriSignatureId  signatureId;
    TriStatus       triStatus;

    dest_port_id = ex_component_get_port_compobj(port_to_call, ctx);

    destport = ex_component_get_port(dest_port_id, ctx);

    initialize_componentId(&componentId, ctx);
    tsiPortId = &(destport->port_repr->triPortId);

    (void)tri__t3_to_tri_BinaryString(exception_value, t3rt_temp_alloc_c,
                                      &exceptionValue, ctx);

    signatureId.moduleName =
      (char *)t3rt_type_definition_module(signature_type, ctx);
    signatureId.objectName =
      (char *)t3rt_type_definition_name(signature_type, ctx);
    signatureId.aux = NULL;

    triStatus = triRaiseBC(&componentId, tsiPortId,
                           &signatureId, &exceptionValue);
    
    if (triStatus != TRI_OK)
      ex_tri_error(EX_TRI_RAISE_BC_FAILED, ctx);

#endif
/* T3RTS_TRI_SECTION_END */

#ifdef ENABLE_DEFLECT
  if (!sut_is_component) {
    fprintf(stderr, "PL WARNING: Raise without SUT component specified!\n");
    return;
  }

  get_reflected_port(caller_port, &dest_port_id, &other_port_id,
                     &other_port_address, &other_compidx, ctx);

#ifdef DEBUG
  fprintf(stderr, "*** Reflecting raise on port %lx to port %lx.\n",
          dest_port_id, other_port_id);
#endif

  /* NOTE! 'caller_port' address is used here as the SUT address from which 
   * this call will seem to originate from in the receiving end. 
   */
  ex_component_queue_put_exception(ex_component_get_comp_queue(other_compidx, ctx),
                                   other_port_address,
                                   signature_type,
                                   exception_value,
                                   caller_port,
                                   ctx);
  return;

#endif /* ENABLE_DEFLECT */
}

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
static TriActionTemplate *
create_TriActionTemplate(t3rt_value_t template_value, t3rt_context_t ctx)
{
  t3rt_binary_string_t  data;
  BinaryString         *x;
  t3rt_value_t          value;

  /* error here if 'template_value' is an ambiguous template */
  value = t3rt_valueof(template_value, t3rt_temp_alloc_c, ctx);

  data = t3rt_binary_string_allocate(t3rt_temp_alloc_c, 0, ctx);
  if (t3rt_encode(value, data, ctx) != t3rt_codecs_result_succeeded_c)
  {
    ex_tri_error(EX_TRI_ENCODE_FAILED, ctx);
    return NULL;
  }

  x = (BinaryString*)t3rt_memory_temp_allocate(sizeof(BinaryString), ctx);

  return (TriActionTemplate*)tri__t3_to_tri_BinaryString(data, t3rt_temp_alloc_c, x, ctx);
}
#endif
/* T3RTS_TRI_SECTION_END */

/*
 * Perform the SUT action (implicit send in TTCN-2)
 */

void
t3pl_port_sut_action(t3rt_value_t string_or_template,
                     t3rt_context_t ctx)
{
/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  
  TriStatus status;

  if (t3rt_value_kind(string_or_template, ctx) == t3rt_type_kind_charstring)
  {
    t3rt_value_t value;

    /* this is to handle case of dynamic templates (error here is tempate is ambiguous) */
    value = t3rt_valueof(string_or_template, t3rt_temp_alloc_c, ctx); 

    status = triSUTActionInformal((char*)t3rt_value_get_string(value, ctx));

    if (status == TRI_Error)
      ex_tri_error(EX_TRI_ACTION_INFORMAL_FAILED, ctx);
  }
  else if (t3rt_value_kind(string_or_template, ctx) == t3rt_type_kind_char)
  {
    t3rt_value_t value;
    char buf[2];

    /* this is to handle case of dynamic templates (error here is template is ambiguous) */
    value = t3rt_valueof(string_or_template, t3rt_temp_alloc_c, ctx); 

    buf[0] = t3rt_value_get_char(value, ctx);
    buf[1] = '\0';

    status = triSUTActionInformal((char*)buf);

    if (status == TRI_Error)
      ex_tri_error(EX_TRI_ACTION_INFORMAL_FAILED, ctx);
  }
  else /* static templates and non-string values */
  {
    TriActionTemplate* tri_action_template = create_TriActionTemplate(string_or_template, ctx);

    if (tri_action_template == NULL)
      status = TRI_Error;
    else
      status = triSUTActionTemplate(tri_action_template);

    if (status == TRI_Error)
      ex_tri_error(EX_TRI_ACTION_TEMPLATE_FAILED, ctx);
  }

#else
/* T3RTS_TRI_SECTION_END */
  /* This function is to be left unimplemented. */
/* T3RTS_TRI_SECTION_START */
#endif
/* T3RTS_TRI_SECTION_END */

}


/*
 * Retrieves the system port address
 */

void
t3pl_port_retrieve_system_port(const char * system_port_name,
                               long index, /* Index in port array or -1 */
                               t3rt_binary_string_t system_port_address,
                               t3rt_context_t ctx)
{
  ex_port_t port = 
    ex_component_find_port(EX_SYS_COMPIDX, system_port_name, index, ctx);

  assert(port != NULL);

  t3rt_binary_string_assign(system_port_address, port->address, ctx);
}


/*
 * Releases the system port address
 */

void
t3pl_port_release_system_port(t3rt_binary_string_t system_port_address,
                              t3rt_context_t ctx)
{
  /* This function is to be left unimplemented */
}


/*
 * Returns the port for controlling the system "component".
 */

t3rt_binary_string_t
t3pl_component_get_system_control_port(t3rt_context_t ctx)
{
  ex_port_t port = 
    ex_component_get_control_port(EX_SYS_COMPIDX, ctx);

  assert(port != NULL);

#ifdef DEBUG
    fprintf(stderr, "Allocating system control port %lx.\n", port->id);
#endif

  return port->address;
}


static void
add_system_port(t3rt_type_t port_type, const char *name, 
                long index, t3rt_type_t system_comp_type,
                t3rt_context_t ctx)
{
  ex_port_t port = ex_component_create_port(EX_SYS_COMPIDX, name, index, ctx);

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  port->port_repr = 
    (ex_port_repr_t)t3pl_memory_allocate(t3rt_perm_alloc_c, 
                                         sizeof(ex_port_repr_struct_t), 
                                         ctx); 
  create_TriPortId(port_type, name, index, &(port->port_repr->triPortId), 
                   true, system_comp_type, ctx);
#endif
/* T3RTS_TRI_SECTION_END */
}



/* For all ports and arrays of ports in the system component type, create a
   port struct entry in the all_port sequence. */
static void
create_system_ports(t3rt_type_t component_type, t3rt_context_t ctx)
{
  unsigned long i;

  assert (t3rt_type_kind(component_type, ctx) == t3rt_type_kind_component);

  for (i = 0; i < t3rt_type_field_count(component_type, ctx); ++i) {
    t3rt_type_t port_type;
    const char* port_name;

    port_type = t3rt_type_field_type(component_type, i, ctx);
    port_name = t3rt_type_field_name(component_type, i, ctx);

    if (t3rt_type_kind(port_type, ctx) == t3rt_type_kind_port) {
      add_system_port(port_type, port_name, -1, 
                      component_type, ctx);
    }
    else if (t3rt_type_kind(port_type, ctx) == t3rt_type_kind_array) {
      t3rt_type_t port_element_type =
        t3rt_type_array_contained_type(port_type, ctx);

      if (t3rt_type_kind(port_element_type, ctx) == t3rt_type_kind_port) {
        long j;
        unsigned long count = t3rt_type_array_size(port_type, ctx);

        /* Loop through all ports and create system port instances for them: */
        for (j = 0; j < (long)count; ++j) {
          add_system_port(port_element_type, port_name, j, component_type,ctx);
        }
      }
    }
  }
}


/*
 * Set the component type of the system component
 */

void
t3pl_component_set_system_component_type(t3rt_type_t component_type,
                                         t3rt_context_t ctx)
{
  unsigned long compidx = ex_component_create(1, -1, ctx);

  assert(compidx == EX_SYS_COMPIDX);

  (void)t3pl_component_get_system_control_port(ctx); /* Make sure that it exists */

  /* Here the regular ports of the system are to be created */
  create_system_ports(component_type, ctx);

  assert(sys_comptype == NULL);
  sys_comptype = component_type;
}

void 
ex_dispatch_queue_element(ex_eventqueue_element_t qelem, 
                             t3rt_context_t ctx)
{   
  switch (qelem->kind) {
    case ex_eventqueue_element_message: 
      t3rt_port_insert_message(qelem->port_address, qelem->sender /* sut_address */,
                               qelem->data, ctx);
      break;
    case ex_eventqueue_element_call:
      t3rt_port_insert_call(qelem->port_address, qelem->sender /* sut_address */,
                            qelem->signature_type, qelem->data_vector, 
                            NULL /* sender, obsolete parameter */, ctx);
      break;
    case ex_eventqueue_element_reply:
      t3rt_port_insert_reply(qelem->port_address, qelem->sender /* sut_address */,
                             qelem->signature_type, qelem->data_vector, 
                             qelem->data, NULL /* sender, obsolete parameter */, ctx);
      break;
    case ex_eventqueue_element_exception:
      t3rt_port_insert_exception(qelem->port_address, qelem->sender /* sut_address */,
                                 qelem->signature_type, qelem->data,
                                 NULL /* sender, obsolete parameter */, ctx);
      break;
    case ex_eventqueue_element_timeout:
      t3rt_timer_timed_out(qelem->timer_handle, ctx);
      break;
    default: 
      assert( false ); /* Never reached! */
      break;
  };
}


t3rt_boolean_t
queue_exists(ex_eventqueue_t queue)
{
  if (queue == NULL) {
    fprintf(stderr, "Component queue missing! Discarding message\n");
    return false;
  }
  return true;
}


/* 
 * Blocks the current component waiting for some external stimuli
 * (i.e. some message received or timer timing out).  This function
 * will return information if there was data received and/or a timeout
 * occurred.  In either case the time_to_soonest_timeout is updated
 * according to how long it really took.
 */

t3rt_snapshot_return_t
t3pl_component_wait (double * real_time_wait,
                     double * time_to_soonest_timeout, 
                     t3rt_context_t ctx)
{
  ex_port_t port;
  double    timeout;
  double    prev_timeout;
  ex_eventqueue_t queue;         /* Test! */
  ex_eventqueue_element_t qelem; /* Test! */
  t3rt_boolean_t status;
  unsigned long compidx;

  if (all_timers_active) {
      *time_to_soonest_timeout = t3rt_duration_forever_c;
  }

  compidx = ex_component_get_index(NULL, ctx);

  port = ex_component_get_control_port(compidx, ctx);

  if (!port->is_active) {
    while(true) ex_thread_sleep(100, ctx); /* Wait for thread to be killed. CONFIGURATION: Time value */
  }

  queue = ex_component_get_comp_queue(compidx, ctx)->events;
  
  status = queue_exists(queue);
  assert(status);

  /* Process any events already present in the component queue. */
  if (status && ((qelem = ex_event_tryget(queue, NULL, NULL, ctx)) != NULL)) { /*lint !e774 */
    do {
      ex_dispatch_queue_element(qelem, ctx);
      ex_eventqueue_element_delete(&qelem, ctx);
    } while ((qelem = ex_event_tryget(queue, NULL, NULL, ctx)) != NULL);

    if (*time_to_soonest_timeout == t3rt_duration_nowait_c || /*lint !e777 */
        *real_time_wait == t3rt_duration_nowait_c) /*lint !e777 */
      return t3rt_snapshot_return_timeout_and_data_received;
    return t3rt_snapshot_return_data_received;
  }
  
  /* Since no events are in the component queue already, we have to wait 
     until any (number of) event arrives. The minimum of
     (real_time_wait, time_to_soonest_timeout) is the longest we wait. */

  /*lint -e777 Testing floats for equality - intended */
  if (*time_to_soonest_timeout == t3rt_duration_nowait_c ||
      *real_time_wait == t3rt_duration_nowait_c) {
    timeout = 0.0;          /* Just poll */
  }
  else if (*time_to_soonest_timeout == t3rt_duration_forever_c) {
    if (*real_time_wait == t3rt_duration_forever_c) {
      timeout = t3rt_duration_forever_c;
    } else {
      assert(*real_time_wait >= 0.0);
      timeout = *real_time_wait;
    }
  }
  else {
    assert(*time_to_soonest_timeout >= 0.0);

    if (*real_time_wait == t3rt_duration_forever_c) {
      timeout = *time_to_soonest_timeout;
    } else {
      timeout = ex_min(*time_to_soonest_timeout, *real_time_wait);
    }
  }

  /* ex_thread_sleep(0, ctx); STo: not necessary and a real performance killer*/ /* Substitute for 'thread yield' */

  prev_timeout = timeout;
  qelem = ex_event_tryget(queue, NULL, &timeout, ctx); /*lint !e774 */

  if (*time_to_soonest_timeout != t3rt_duration_forever_c &&
      *time_to_soonest_timeout != t3rt_duration_nowait_c) {
    *time_to_soonest_timeout -= prev_timeout - timeout;
    if (*time_to_soonest_timeout <= 0.0) *time_to_soonest_timeout = 0.0;
  }

  if (*real_time_wait != t3rt_duration_forever_c &&
      *real_time_wait != t3rt_duration_nowait_c) {
    *real_time_wait -= prev_timeout - timeout;
    if (*real_time_wait <= 0.0) *real_time_wait = 0.0;
  }

  if (qelem != NULL) {
    do {
      ex_dispatch_queue_element(qelem, ctx);
      ex_eventqueue_element_delete(&qelem, ctx);
    } while ((qelem = ex_event_tryget(queue, NULL, NULL, ctx)) != NULL);

    if (*time_to_soonest_timeout == 0.0 || *real_time_wait == 0.0) {
      return t3rt_snapshot_return_timeout_and_data_received;
    }

    return t3rt_snapshot_return_data_received;
  }

  return t3rt_snapshot_return_timeout;
}


/* 
 * Extract control message (probably not on the top of event queue)
 * and adds it into control port
 */

void 
t3pl_component_control (t3rt_context_t ctx)
{
  ex_port_t port;
  ex_eventqueue_t queue;         /* Test! */
  ex_eventqueue_element_t qelem; /* Test! */
  t3rt_boolean_t status;
  unsigned long compidx;

  compidx = ex_component_get_index(NULL, ctx);

  port = ex_component_get_control_port(compidx, ctx);

  if (!port->is_active) {
    while(true) ex_thread_sleep(100, ctx); /* Wait for thread to be killed. CONFIGURATION: Time value */
  }

  queue = ex_component_get_comp_queue(compidx, ctx)->events;
  
  status = queue_exists(queue);
  assert(status);

  /* Process any control events already present in the component queue. */
  if (status && ((qelem = ex_event_tryget(queue, port->address, NULL, ctx)) != NULL)) { /*lint !e774 */
    do {
      ex_dispatch_queue_element(qelem, ctx);
      assert(t3rt_binary_string_is_equal(qelem->port_address, t3rt_context_get_component_address(ctx), ctx));
      ex_eventqueue_element_delete(&qelem, ctx);
    } while ((qelem = ex_event_tryget(queue, port->address, NULL, ctx)) != NULL);
  }
}

#ifdef ENABLE_DEFLECT


static void
get_reflected_port(t3rt_binary_string_t       port_address,
                   ex_component_object_id_t * dest_port_id,
                   ex_component_object_id_t * other_port_id,
                   t3rt_binary_string_t     * other_port_address,
                   unsigned long            * other_compidx,
                   t3rt_context_t ctx)
{
  ex_port_t other_port;

  *dest_port_id = ex_component_get_port_compobj(port_address, ctx);

  other_port = ex_find_reflected_port(*dest_port_id, ctx);

  assert(other_port != NULL);

  *other_port_id = other_port->id;

  *other_compidx = ex_component_index(*other_port_id);
  
  *other_port_address = other_port->address;
}


#endif /* ENABLE_DEFLECT */


void
ex_component_queue_put_call(ex_component_queue_t  comp_queue,
                            t3rt_binary_string_t  port_addr,
                            t3rt_type_t           sig_type,
                            t3rt_binary_string_t *parameters,
                            t3rt_binary_string_t  sender,
                            t3rt_context_t        ctx)
{
  ex_eventqueue_element_t elem;
  ex_eventqueue_t queue;

  assert(comp_queue != NULL);
    
  queue = comp_queue->events;

  if (queue_exists(queue)) {
    elem = ex_eventqueue_element_create(ex_eventqueue_element_call,
                                        ctx);

    ex_eventqueue_element_set_port_address(elem, port_addr, ctx);
    ex_eventqueue_element_set_signature_type(elem, sig_type, ctx);
    ex_eventqueue_element_set_data_vector(elem, parameters,
                                          ex_parameter_count(sig_type, ctx),
                                          ctx);
    ex_eventqueue_element_set_sender(elem, sender, ctx);

    ex_event_put(queue, elem, ctx);
  }
}


void
ex_component_queue_put_reply(ex_component_queue_t  comp_queue,
                             t3rt_binary_string_t  port_addr,
                             t3rt_type_t           sig_type,
                             t3rt_binary_string_t *reply_data,
                             t3rt_binary_string_t  return_data,
                             t3rt_binary_string_t  sender,
                             t3rt_context_t        ctx)
{
  ex_eventqueue_element_t elem;
  ex_eventqueue_t queue;

  assert(comp_queue != NULL);
    
  queue = comp_queue->events;

  if (queue_exists(queue)) {
      elem = ex_eventqueue_element_create(ex_eventqueue_element_reply,
                                          ctx);

      ex_eventqueue_element_set_port_address(elem, port_addr, ctx);
      ex_eventqueue_element_set_signature_type(elem, sig_type, ctx);
      ex_eventqueue_element_set_data(elem, return_data, ctx);
      ex_eventqueue_element_set_data_vector(elem, reply_data,
                                            ex_parameter_count(sig_type, ctx),
                                            ctx);
      ex_eventqueue_element_set_sender(elem, sender, ctx);

      ex_event_put(queue, elem, ctx);
  }
}


void
ex_component_queue_put_exception(ex_component_queue_t comp_queue,
                                 t3rt_binary_string_t port_addr,
                                 t3rt_type_t sig_type,
                                 t3rt_binary_string_t exception_data,
                                 t3rt_binary_string_t sender,
                                 t3rt_context_t ctx)
{
  ex_eventqueue_element_t elem;
  ex_eventqueue_t queue;

  assert(comp_queue != NULL);
    
  queue = comp_queue->events;

  if (queue_exists(queue)) {
      elem = ex_eventqueue_element_create(ex_eventqueue_element_exception,
                                          ctx);

      ex_eventqueue_element_set_port_address(elem, port_addr, ctx);
      ex_eventqueue_element_set_signature_type(elem, sig_type, ctx);
      ex_eventqueue_element_set_data(elem, exception_data, ctx);
      ex_eventqueue_element_set_sender(elem, sender, ctx);

      ex_event_put(queue, elem, ctx);
  }
}


void
ex_component_queue_put_message(ex_component_queue_t comp_queue,
                               t3rt_binary_string_t port_address,
                               t3rt_binary_string_t data,
                               t3rt_binary_string_t sender,
                               t3rt_context_t ctx)
{
  ex_eventqueue_element_t elem;
  ex_eventqueue_t queue;

  assert(comp_queue != NULL);
    
  queue = comp_queue->events;

  if (queue_exists(queue)) {
      elem = ex_eventqueue_element_create(ex_eventqueue_element_message,
                                          ctx);

      ex_eventqueue_element_set_port_address(elem, port_address, ctx);
      ex_eventqueue_element_set_data(elem, data, ctx);
      ex_eventqueue_element_set_sender(elem, sender, ctx);

      ex_event_put(queue, elem, ctx);
  }
}


void
ex_component_queue_put_timeout(ex_component_queue_t comp_queue,
                               t3rt_timer_handle_t  timer_handle,
                               t3rt_context_t ctx)
{
  ex_eventqueue_element_t elem;
  ex_eventqueue_t queue;

  assert(comp_queue != NULL);
    
  queue = comp_queue->events;

  if (queue_exists(queue)) {
      elem = ex_eventqueue_element_create(ex_eventqueue_element_timeout,
                                          ctx);

      ex_eventqueue_element_set_timer_handle(elem, timer_handle, ctx);

      ex_event_put(queue, elem, ctx);
  }
}


#if 0 /*  was ifdef DEBUG */
static void
debug_print_all_ports(t3rt_context_t ctx)
{
  port_struct_t *port;
  port_iter_t iter;

  t3pl_sem_wait(port_sem, ctx);

  /* Iterate through the ports, find the system port and map them.*/
  ex_sequence_iter_start(iter, all_ports);

  while (ex_sequence_iter_next(iter)) {
    port = &ex_sequence_iter_get(iter);

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
    fprintf(stderr, "Port %2ld %s %sactive %s %sport TRI=(%s of %s.%s) %s[%ld] mapped to %ld.\n",
                    ex_sequence_iter_get_index(iter),
                    (port->port_address != NULL ? 
                      t3rt__binary_string_to_hexstring(port->port_address, false, ctx) : ""),
                    (port->is_active ? "  " : "in"), /* active or inactive */
                    (port->is_system_port ? "   system" : "component"),
                    (port->is_control_port ? "control " : ""),
                    (port->triPortId.portName != NULL ? port->triPortId.portName : "<null>"),
                    (port->triPortId.portType.moduleName != NULL ? port->triPortId.portType.moduleName : "<null>"),
                    (port->triPortId.portType.objectName != NULL ? port->triPortId.portType.objectName : "<null>"),
                    (port->port_name != NULL ? port->port_name : "<?>"),
                    port->port_index,
                    port->mapped_to_tsi_port);
#else
/* T3RTS_TRI_SECTION_END */
    fprintf(stderr,
            "Port %2ld %sactive %s %sport %s[%ld] mapped to %ld.\n", 
            ex_sequence_iter_get_index(iter), 
            (port->is_active ? "  " : "in"), /* active or inactive */
            (port->is_system_port ? "   system" : "component"),
            (port->is_control_port ? "control " : ""),
            (port->port_name != NULL ? port->port_name : "<?>"),
            port->port_index,
            port->mapped_to_tsi_port);
/* T3RTS_TRI_SECTION_START */
#endif
/* T3RTS_TRI_SECTION_END */
  }

  t3pl_sem_post(port_sem, ctx);
}
#endif /* DEBUG */


/*
 * This function assumes that for every (non-control) system port, there
 * exist exactly two component (non-control) ports mapped to it.
 * This function finds, given one of these component ports, the other
 * component port, and at the same time verifies that the above assumption
 * holds.
 */

#define ASSOC_PORTS_LENGTH 5

ex_port_t 
ex_find_reflected_port(ex_component_object_id_t pid, 
                       t3rt_context_t ctx)
{
  ex_port_t                 port;
  ex_port_t                 assoc_ports[ASSOC_PORTS_LENGTH];
  unsigned long             found;

  port = ex_component_get_port(pid, ctx);

  found = ex_components_find_associated_ports(port->associated_with, 
                                              assoc_ports, 
                                              ASSOC_PORTS_LENGTH, 
                                              ctx);

  assert(found == 2); /* "self" and the other */
  
  if (assoc_ports[0] == port) 
    return assoc_ports[1];

  return assoc_ports[0];
}


/* End of File */

