/*---------------------------------------------------------------------------
 * File Name: example_components.c
 *
 * 2010 IBM Corporation
 *
 * Description:
 * XXX
 */

/*
$Id: example_components.c,v 2.8 2005/10/10 07:52:30 x-rupaya Exp $
*/

#include <stdio.h>
#include <string.h>             /* For strlen, strcmp etc */
#include <stdlib.h>

#include "example_general.h" /* For assert definition */
#include "example_components.h"
#include "example_concurrency.h"


static ex_rw_semaphore_struct_t                 components_sem;
static ex_component_object_id_t                 writer;
static ex_sequence_head(ex_component_struct_t)  components;

typedef ex_sequence_iter(ex_port_struct_t)      ex_component_port_iter_t;

typedef ex_sequence_iter(ex_component_struct_t) ex_component_iter_t;



void
ex_component_pre_initialize(t3rt_context_t ctx)
{
  ex_sequence_init(components);
  ex_sem_alloc(&components_sem, ctx);
}

void
ex_component_initialize(t3rt_context_t ctx)
{
}

void
ex_component_finalize(t3rt_context_t ctx)
{
  ex_sem_dealloc(&components_sem, ctx);
  ex_sequence_deallocate(components, t3rt_perm_alloc_c, ctx);
}


/*
 * Removes the component elements from the global list. All except the 
 * CPC component which is always the first element.
 *
 * Intended to be called by the main thread only when *all* components but 
 * the CPC has has been killed! 
 *
 * The assumption is that the component data for all other components has 
 * already been removed one way or the other (t3pl_task_kill or 
 * t3pl_task_exit). 
 */

void
ex_component_release_components(t3rt_context_t ctx)
{
  unsigned long size;

  ex_component_writer_start(ctx);

  size = ex_sequence_size(components);

  while (size-- > 1) { /* Leave CPC */
    ex_component_delete_MT_unsafe(size, ctx);
    ex_sequence_remove(components, (long)size, ctx);
  }

  ex_component_writer_end(ctx);
}

void
ex_component_release_cpc(t3rt_context_t ctx)
{
  unsigned long size;

  ex_component_writer_start(ctx);

  size = ex_sequence_size(components);

  assert(size == 1);
  
  ex_component_delete_MT_unsafe(EX_CPC_COMPIDX, ctx);
  ex_sequence_remove(components, EX_CPC_COMPIDX, ctx);
  
  ex_component_writer_end(ctx);
}

/*
 *
 * NOTE!
 * This function is only intended to be called from the 
 * t3pl_port_create_control_port_for_cpc since that thread is definately 
 * alone when executing.
 *
 */
unsigned long
ex_component_create(long number_of_ports, /* -1 for 'unknown' */
                    long number_of_timers, /* -1 for 'unknown' */
                    t3rt_context_t ctx)
{
  unsigned long size;
  ex_component_t component;
  ex_port_t ctrl_port;

  ex_component_writer_start(ctx);

  size = ex_sequence_size(components);

  ex_sequence_add(components, t3rt_perm_alloc_c, ctx);

  component = &ex_sequence_last_element(components);

  component->ctx = NULL;        /* Filled in later */
  ex_sequence_init(component->component_ports.ports);
  ex_sequence_init(component->component_timers.timers);
  component->component_queue =
    t3pl_memory_allocate(t3rt_perm_alloc_c,
                         sizeof(ex_component_queue_struct_t),
                         ctx);
  component->task_info = NULL;
  component->component_queue->id = ex_component_object_id(size, 0);
  component->component_queue->events = ex_eventqueue_create(ctx);

  /* Allocate room for requested number of ports: */
  if (number_of_ports > 0) {
    long i = number_of_ports;
    while (i-- > 0)
      ex_sequence_add(component->component_ports.ports, t3rt_perm_alloc_c, ctx);

    i = number_of_ports;
    while (i-- > 0) 
      ex_sequence_remove(component->component_ports.ports, i, ctx);
  }

  /* Create control port. (always at index zero) */
  ctrl_port = ex_component_create_port(size, "control port", -1, ctx);
  assert(ctrl_port != NULL && ex_object_index(ctrl_port->id) == 0);

  /* Allocate room for requested number of timers: */
  if (number_of_timers > 0) {
    long i = number_of_timers;
    while (i-- > 0)
      ex_sequence_add(component->component_timers.timers, t3rt_perm_alloc_c, ctx);

    i = number_of_timers;
    while (i-- > 0)
      ex_sequence_remove(component->component_timers.timers, i, ctx);
  }

  ex_component_writer_end(ctx);

  return size;
}

static void
ex_component_destroy_component_queue(ex_component_queue_t *queue, 
                                     t3rt_context_t ctx)
{
  assert(queue != NULL && *queue != NULL);

  (*queue)->id = ex_component_no_object;
  ex_eventqueue_delete(&((*queue)->events), ctx);
  t3pl_memory_deallocate(t3rt_perm_alloc_c, (*queue), ctx);
  *queue = NULL;
}


void
ex_component_delete_MT_unsafe(unsigned long compidx, t3rt_context_t ctx)
{
  ex_component_t component;
  unsigned long count;
  assert(compidx < ex_sequence_size(components));

#ifdef DEBUG
  assert(writer == ex_component_get_index(NULL, ctx));
#endif

  component = &ex_sequence_element(components, compidx);

  count = ex_sequence_size(component->component_timers.timers);
  /* Destroy the timers: */
  while (count-- > 0) {
    ex_timer_t timer =
      &ex_sequence_element(component->component_timers.timers, count);

    /* The timer might have been deleted from the RTS prior to the 
       component being terminated. If so, the handle has already been 
       nullified.*/ 
    if (timer->handle != NULL)
      t3pl_timer_delete(&(timer->handle), ctx);

	ex_sequence_remove(component->component_timers.timers, (long)count, ctx);
  }

  ex_sequence_deallocate(component->component_timers.timers, t3rt_perm_alloc_c, ctx);

  count = ex_sequence_size(component->component_ports.ports);
  while (count-- > 0) {
    ex_port_t port =
      &ex_sequence_element(component->component_ports.ports, count);

    t3pl_port_destroy(port->address, ctx);
  }
  count = ex_sequence_size(component->component_ports.ports);
  /* Destroy the ports: */
  while (count-- > 0) {
    ex_sequence_remove(component->component_ports.ports, (long)count, ctx);
  }

  ex_sequence_deallocate(component->component_ports.ports, t3rt_perm_alloc_c, ctx);

  /* TODO: The context should be cleaned up by calling a function passed from
   * t3pl_task_kill. That would change the prototype for t3pl_task_kill. 
   */
  component->ctx = NULL;

  ex_component_destroy_component_queue(&(component->component_queue), ctx);

  ex_task_destroy_task_info(&(component->task_info), ctx);
}


/*
 * ex_component_terminate_MT_unsafe
 *
 * This is called from t3pl_task_exit when a component terminates.
 * Its job is only to delete its own timers (but not the list of timers), 
 * its task info but nothing else. The rest is done when terminating the 
 * test case (in ex_component_delete_MT_unsafe).
 *
 * A special case for component destruction is when an error occurs during 
 * MTC startup. In that case we try to kill the system component (at index 
 * EX_SYS_COMPIDX) but it might not have been created. So we have to be more
 * relaxed.
 *
 */
void
ex_component_terminate_MT_unsafe(unsigned long compidx, t3rt_context_t ctx)
{
  ex_component_t component;
  unsigned long count;

  /* The special case of MTC startup error */
  if (compidx == EX_SYS_COMPIDX && 
      ex_sequence_size(components) == 2) /* I.e. only CPC and MTC currently created. */
    return;

  assert(compidx < ex_sequence_size(components));

#ifdef DEBUG
  assert(writer == ex_component_get_index(NULL, ctx));
#endif

  component = &ex_sequence_element(components, compidx);

  count = ex_sequence_size(component->component_timers.timers);
  /* Destroy the timers: */
  while (count-- > 0) {
    ex_timer_t timer =
      &ex_sequence_element(component->component_timers.timers, count);

    /* The timer might have been deleted from the RTS prior to the 
       component being terminated. If so, the handle has already been 
       nullified. We can not delete the timer here because it would leave 
       dangling pointers in the RT timer values (for declared timers in the 
       component). */ 
    if (timer->handle != NULL)
      t3pl_timer_stop(timer->handle, ctx);
  }

  /* TODO: The context should be cleaned up by calling a function passed from
   * t3pl_task_kill. That would change the prototype for t3pl_task_kill. 
   */
  component->ctx = NULL;
}



void
ex_component_set_context(unsigned long compidx,
                         t3rt_context_t ctx)
{
  ex_component_writer_start(ctx);

  assert(ex_sequence_element(components, compidx).ctx == NULL);

  ex_sequence_element(components, compidx).ctx = ctx;

  ex_component_writer_end(ctx);
}


t3rt_context_t
ex_component_get_context(unsigned long compidx, t3rt_context_t ctx)
{
  t3rt_context_t context;

  ex_component_reader_start(ctx);

  context = ex_sequence_element(components, compidx).ctx;

  ex_component_reader_end(ctx);

  return context;
}


t3rt_context_t
ex_component_get_context_MT_unsafe(unsigned long compidx)
{
  return ex_sequence_element(components, compidx).ctx;
}


ex_component_port_list_t
ex_component_get_ports(unsigned long compidx, 
                       t3rt_context_t ctx)
{
  //We should never get here since this function is a source of race conditions!!!
  //DO NOT USE IT!!!
  assert(false);

  return NULL;

/*
  ex_component_port_list_t ports;

  ex_component_reader_start(ctx);

  ports = &(ex_sequence_element(components, compidx).component_ports);

  ex_component_reader_end(ctx);

  return ports;
*/
}


ex_port_t
ex_component_get_port(ex_component_object_id_t port_object_id,
                      t3rt_context_t ctx)
{
  unsigned long compidx = ex_component_index(port_object_id);
  unsigned long objidx = ex_object_index(port_object_id);
  ex_component_port_list_t ports;
  ex_port_t port;

  ex_component_reader_start(ctx);

  ports = &(ex_sequence_element(components, compidx).component_ports);

  assert(objidx < ex_sequence_size(ports->ports));

  port = &ex_sequence_element(ports->ports, objidx);

  ex_component_reader_end(ctx);

  return port;
}


ex_port_t 
ex_component_get_control_port(unsigned long compidx, t3rt_context_t ctx)
{
  ex_component_port_list_t ports;
  ex_port_t port;

  ex_component_reader_start(ctx);

  ports = &(ex_sequence_element(components, compidx).component_ports);

  port = &ex_sequence_element(ports->ports, 0);

  ex_component_reader_end(ctx);

  return port;
}


ex_component_timer_list_t
ex_component_get_timers(unsigned long compidx, t3rt_context_t ctx)
{
  //We should never get here since this function is a source of race conditions!!!
  //DO NOT USE IT!!!
  assert(false);

  return NULL;

/*
  ex_component_timer_list_t timers;

  ex_component_reader_start(ctx);

  timers = &(ex_sequence_element(components, compidx).component_timers);

  ex_component_reader_end(ctx);

  return timers;
*/
}


ex_timer_t
ex_component_get_timer(ex_component_object_id_t timer_object_id,
                       t3rt_context_t ctx)
{
  unsigned long compidx = ex_component_index(timer_object_id);
  unsigned long objidx = ex_object_index(timer_object_id);
  ex_component_timer_list_t timers;
  ex_timer_t timer;

  ex_component_reader_start(ctx);

  timers = &(ex_sequence_element(components, compidx).component_timers);

  if (objidx >= ex_sequence_size(timers->timers))
    timer = NULL;
  else
    timer = &ex_sequence_element(timers->timers, objidx);

  ex_component_reader_end(ctx);

  return timer;
}


ex_component_queue_t
ex_component_get_comp_queue(unsigned long compidx, t3rt_context_t ctx)
{
  ex_component_queue_t queue;

  ex_component_reader_start(ctx);

  queue = ex_sequence_element(components, compidx).component_queue;

  ex_component_reader_end(ctx);

  return queue;
}


ex_task_info_t
ex_component_get_task_info(unsigned long compidx, t3rt_context_t ctx)
{
  ex_task_info_t ti;

  ex_component_reader_start(ctx);

  ti = ex_sequence_element(components, compidx).task_info;

  ex_component_reader_end(ctx);

  return ti;
}


void
ex_component_set_task_info(unsigned long compidx, 
                           ex_task_info_t task_info, 
                           t3rt_context_t ctx)
{
  ex_component_writer_start(ctx);

  ex_sequence_element(components, compidx).task_info = task_info;

  ex_component_writer_end(ctx);
}


/* 
 * ex_component_create_port
 *
 * Called from t3pl_task_create...
 */
ex_port_t
ex_component_create_port(unsigned long compidx,
                         const char * port_name,
                         long port_index, /* -1 for unused */
                         t3rt_context_t ctx)
{
  ex_component_port_list_t ports;
  unsigned long size;
  ex_port_t port;

  ex_component_reader_start(ctx);

  ports = &(ex_sequence_element(components, compidx).component_ports);

  size = ex_sequence_size(ports->ports);

  ex_sequence_add(ports->ports, t3rt_perm_alloc_c, ctx);

  port = &ex_sequence_last_element(ports->ports);

  port->id = ex_component_object_id(compidx, size);
  port->port_name = port_name;
  port->port_index = port_index;
  port->is_active = true;
  ex_sem_alloc(&port->sem,ctx);
  port->associated_with = ex_component_no_object;
  ex_sequence_init(port->sysport_associated_with);
  port->port_repr = NULL;
  port->address = ex_component_get_port_address(port->id, ctx);

  ex_component_reader_end(ctx);

  return port;
}


ex_timer_t
ex_component_create_timer(unsigned long compidx, t3rt_context_t ctx)
{
  ex_component_timer_list_t timers;
  unsigned long size;
  ex_timer_t timer;
  ex_component_object_id_t * timer_handle;

  ex_component_reader_start(ctx);

  timers = &(ex_sequence_element(components, compidx).component_timers);

  size = ex_sequence_size(timers->timers);

  ex_sequence_add(timers->timers, t3rt_perm_alloc_c, ctx);

  timer = &ex_sequence_last_element(timers->timers);

  timer->id = ex_component_object_id(compidx, size);

  timer_handle = (ex_component_object_id_t *)
    t3pl_memory_allocate(t3rt_perm_alloc_c,
                         sizeof(ex_component_object_id_t),
                         ctx);

  *timer_handle = timer->id;

  timer->handle = (t3rt_timer_handle_t)timer_handle;
  timer->timer_repr = NULL;

  ex_component_reader_end(ctx);

  return timer;
}


void
ex_component_destroy_port(ex_port_t port, t3rt_context_t ctx)
{
  port->id = ex_component_no_object;
  port->port_name = NULL;
  port->port_index = -1;
  port->is_active = false;
  ex_sem_dealloc(&port->sem, ctx);
  port->associated_with = ex_component_no_object;
  ex_sequence_deallocate(port->sysport_associated_with, t3rt_perm_alloc_c, ctx);

  t3rt_binary_string_deallocate_all(&(port->address), ctx);
  port->address = NULL;

  /* NOTE! port_repr must be deallocated *before* calling this function. */

  port->port_repr = NULL;
}




void
ex_component_destroy_timer(ex_timer_t timer, t3rt_context_t ctx)
{  

  t3pl_memory_deallocate(t3rt_perm_alloc_c, timer->handle, ctx);
  timer->handle = NULL;
  
  timer->id = ex_component_no_object;
  
  /* NOTE! timer_repr must be deallocated *before* calling this function. */
  timer->timer_repr = NULL;  

}

void
ex_component_writer_start(t3rt_context_t ctx) 
{
  ex_sem_writer_start(&components_sem, ctx);

  writer = ex_component_get_index(NULL, ctx);
}

void
ex_component_writer_end(t3rt_context_t ctx) 
{
  writer = ex_component_no_object;

  ex_sem_writer_end(&components_sem, ctx);
}

void
ex_component_reader_start(t3rt_context_t ctx) 
{
  if (writer != ex_component_no_object &&
      writer == ex_component_get_index(NULL, ctx))
    return;

  ex_sem_reader_start(&components_sem, ctx);
}

void
ex_component_reader_end(t3rt_context_t ctx) 
{
  if (writer != ex_component_no_object &&
      writer == ex_component_get_index(NULL, ctx))
    return;

  ex_sem_reader_end(&components_sem, ctx);
}


/* Move the following three functions to components.c? */
unsigned long
ex_component_get_index(t3rt_binary_string_t address, t3rt_context_t ctx)
{
  t3rt_binary_string_t addr = address;
  ex_component_object_id_t objid;

  if (addr == NULL)
    addr = t3rt_context_get_component_address(ctx);

  /* Message sent from non-active context */
  if (t3rt_binary_string_length(addr, ctx) == 0)
    return ex_component_no_object;

  objid = ex_component_get_port_compobj(addr, ctx);

  return ex_component_index(objid);
}


ex_component_object_id_t
ex_component_get_port_compobj(t3rt_binary_string_t port_address,
                              t3rt_context_t ctx)
{
  unsigned long portidx;
  t3rt_binary_string_iter_t iter;

  assert(t3rt_binary_string_length(port_address, ctx) == 32);

  t3rt_binary_string_start(port_address, &iter, ctx);

  portidx = t3rt_bstring_iter_get_4bytes(&iter, ctx);

  /* TODO: Replace with proper assert?: assert(portidx < ex_sequence_size(all_ports)); */

  return portidx;
}


t3rt_binary_string_t
ex_component_get_port_address(ex_component_object_id_t objid,
                              t3rt_context_t ctx)
{
  t3rt_binary_string_t retval;

  retval = t3rt_binary_string_allocate(t3rt_perm_alloc_c, 32, ctx);

  t3rt_binary_string_append_4bytes(retval, (unsigned long)objid, ctx);

  return retval;
}


ex_port_t
ex_component_find_port(unsigned long compidx,
                       const char* name, 
                       long array_index, 
                       t3rt_context_t ctx)
{
  ex_component_port_list_t ports;
  ex_port_t found_port = NULL;
  ex_component_port_iter_t iter;

  ex_component_reader_start(ctx);

  ports = &(ex_sequence_element(components, compidx).component_ports);

  assert(ports != NULL);
  
  ex_sequence_iter_start(iter, ports->ports);

  while (ex_sequence_iter_next(iter)) {
    ex_port_t port = &ex_sequence_iter_get(iter);
    if (port->port_index == array_index   &&
        port->port_name != NULL     && 
        strcmp(port->port_name, name) == 0) 
    {
      found_port = port;
      break;
    }   
  }
  
  ex_component_reader_end(ctx);

  return found_port;
}


unsigned long
ex_components_find_associated_ports(ex_component_object_id_t pid, 
                                    ex_port_t assoc_ports[], 
                                    unsigned long assoc_ports_length, 
                                    t3rt_context_t ctx)
{
  ex_component_iter_t comp_iter;
  unsigned long count = 0;

  ex_component_reader_start(ctx);

  ex_sequence_iter_start(comp_iter, components);

  while (ex_sequence_iter_next(comp_iter)) {
    ex_component_t           comp = &ex_sequence_iter_get(comp_iter);
    ex_component_port_iter_t iter;
    ex_component_port_list_t ports = &(comp->component_ports);
  
    assert(ports != NULL);
  
    ex_sequence_iter_start(iter, ports->ports); /*lint !e413 Likely use of null pointer. */

    (void)ex_sequence_iter_next(iter); /* Skip control port.*/
    while (ex_sequence_iter_next(iter)) {
      ex_port_t port = &ex_sequence_iter_get(iter);
      if (port->associated_with == pid) {
        if (count < assoc_ports_length)
          assoc_ports[count] = port;
        count++;
      }   
    }
  }

  ex_component_reader_end(ctx);

  return count;
}



/* End of File */
