/*---------------------------------------------------------------------------
 * File Name: t3pl_concurrency.c
 *
 * 2010 IBM Corporation
 * 
 * Description:
 *     This is an example implementation of the concurrency part of the
 * Platform Layer.
 */


#include <stdio.h>
#include <stdlib.h>

#include "t3rts.h"

#include "example_general.h" /* For assert definition */
#include "example_concurrency.h"
#include "example_components.h"


/***************************************************************************/
#if defined(_WIN32) && ! defined(WIN32_GCC)   /* Win32 platform */
/***************************************************************************/

#include <windows.h>
#include <process.h>

#define THREAD_MAIN_CALLING_CONVENTION  __stdcall 
#define THREAD_MAIN_RETURN_TYPE         unsigned
#define THREAD_RETURN                   return 0

typedef HANDLE              ex_thread_handle_t;
typedef unsigned            ex_thread_id_t;

#define ex_create_thread(status, handle, id, function, args)            \
    handle = (HANDLE)_beginthreadex (NULL /* security */,               \
                                     0 /* stack size*/,                 \
                                     function,                          \
                                     args,                              \
                                     0 /* create as running */,         \
                                     &(id));                            \
    status = (handle != (HANDLE)0)

#define ex_exit_thread()                   _endthreadex(0)
#define ex_terminate_thread(status, handle)                             \
    status = (TerminateThread (handle, 1UL) == TRUE)

#define ex_wait_for_thread_termination(status, handle)                  \
    status = (WaitForSingleObject (handle, INFINITE) == TRUE)

#define ex_get_current_thread_id(id)                                    \
    id = (unsigned)GetCurrentThreadId ();

#define ex_close_thread_handle(handle)      (CloseHandle (handle) == TRUE)


    /* Windows specific thread init function */
static void
ex_init_thread (void)
{
}

/***************************************************************************/
#else           /* Unix platform */
/***************************************************************************/

#include <pthread.h>
#include <semaphore.h>
#include <time.h>
#include <errno.h>      /* for errno and ETIMEDOUT etc. */
#include <unistd.h>             /* For fork, execvp etc */


#define THREAD_MAIN_CALLING_CONVENTION
#define THREAD_MAIN_RETURN_TYPE         void *
#define THREAD_RETURN                   return NULL

typedef pthread_t           ex_thread_handle_t;
typedef pthread_t           ex_thread_id_t;

#define ex_create_thread(status, handle, id, function, args)                         \
    status = (pthread_create (&(handle), NULL, function, args) == 0) ? true : false; \
    id = handle;

#define ex_exit_thread()                   pthread_exit(NULL)
#define ex_terminate_thread(status, handle)                              \
  status = ex_terminate_thread_aux(handle)

#define ex_wait_for_thread_termination(status, handle)                   \
  status = (pthread_join(handle, NULL) != 0) ? true : false


static t3rt_boolean_t
ex_terminate_thread_aux(ex_thread_handle_t handle)
{
  int retval;

  retval = pthread_cancel(handle);

  if (retval == ESRCH)        /* Thread already terminated... */
    return 1;

  assert(retval == 0);

  return(retval == 0);
}


#define ex_get_current_thread_id(id)                                    \
    id = pthread_self ();

#define ex_close_thread_handle(handle)      true


    /* UNIX specific thread init function */
static void
ex_init_thread (void)
{
    int status;

    status = pthread_setcancelstate (PTHREAD_CANCEL_ENABLE, NULL);
    assert (status == 0);
    status = pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
    assert (status == 0);
}

/***************************************************************************/
#endif              
/***************************************************************************/



/***************************************************************************/
                           /*  Common part */
/***************************************************************************/



typedef struct ex_task_info_internal_t {
  ex_thread_id_t      thread_id;
  ex_thread_handle_t  thread_handle;
  void *              ready_sem;  /* Used in t3pl_task_create */
} ex_task_info_struct_t;



/* 
 * Main thread function 
 */

THREAD_MAIN_RETURN_TYPE     THREAD_MAIN_CALLING_CONVENTION
ex_thread_main (void * control_port_address_ref)
{
    t3rt_binary_string_t    control_port_address;

    assert(control_port_address_ref != NULL);

    ex_init_thread();

    control_port_address = (t3rt_binary_string_t)control_port_address_ref;

    t3rt_component_main(control_port_address, NULL);

    THREAD_RETURN;
}



void
ex_thread_sleep (unsigned long millisec_delay, t3rt_context_t context)
{
#if defined(_WIN32) && ! defined(WIN32_GCC)   /* Win32 platform */

    Sleep (millisec_delay);

#else           /* UNIX platform */

    struct timespec     ts;
    struct timespec     timeleft;

    int                 status;

    ts.tv_sec  = (millisec_delay / 1000L);
    ts.tv_nsec = (millisec_delay % 1000L) * 1000000L;

    do {
      status = nanosleep (&ts, &timeleft);
      ts.tv_sec  = timeleft.tv_sec;
      ts.tv_nsec = timeleft.tv_nsec;
    } while (status != 0 && errno == EINTR);

    assert(status == 0 || errno == EINTR);

#endif
}


/*****************************************/


void
t3pl_concurrency_pre_initialize (int argc, char * argv [], t3rt_context_t ctx)
{
  ex_component_pre_initialize(ctx);
}


void
t3pl_concurrency_initialize (t3rt_context_t ctx)
{
  ex_component_initialize(ctx);
}


/*
 * This function is executing alone (i.e. no other threads are running) and no 
 * protective mechanisms are necessary.
 */
void
t3pl_concurrency_finalize (t3rt_context_t ctx)
{
  ex_component_finalize(ctx);
}


void
ex_concurrency_testcase_terminated (t3rt_context_t ctx)
{
  ex_component_release_components(ctx);
}


/*
 * Start the first (non-CPC) component of the current process.
 * This function is supposed to end by calling the t3rt_component_main
 * function with the newly created control port address of this
 * component along with the provided context.
 */

void
t3pl_concurrency_start_separate_component(int argc, const char* argv[],
                                          t3rt_context_t ctx)
{
  t3rt_binary_string_t address;
  unsigned long compidx;
  ex_port_t port;

  /* Warning:
   * Observe that control port addresses are supposed to be globally
   * unique.  This means that this process need to somhow arrange for
   * creating component indices that are separate from all other
   * processes.
   *
   */

  compidx = ex_component_create(1, -1, ctx);
  port = ex_component_get_port(ex_component_object_id(compidx, 0), ctx);
  address = t3rt_binary_string_copy(port->address, t3rt_perm_alloc_c, ctx);

  /* OBSERVE!
   * Here is where the control port address is supposed to be
   * transferred back to the component that has started this
   * component.  The arguments from the command line may be
   * helpful in this procedure.
   */

  t3rt_component_main(address, ctx);
}


void
t3pl_task_register_context(t3rt_context_t ctx)
{
  ex_task_info_t   task_info;
  unsigned long compidx = ex_component_get_index(NULL, ctx);

  task_info = ex_component_get_task_info(compidx, ctx);

  assert(ex_component_get_context(compidx, ctx) == NULL);

  ex_component_set_context(compidx, ctx);

  /* Inform t3pl_task_create that we are ready to preceed: */
  if (compidx != EX_CPC_COMPIDX) {
    t3rt_boolean_t status = t3pl_sem_post(task_info->ready_sem, ctx);
    assert(status);
  }
}



/*
 *  This function is called from the t3rt_component_main function in an 
 *  attempt to setup whatever is necessary for the component to 
 *  communicate through its ports.
 *  It is called after the control port has been created but before any
 *  other port is created and before any communication between components 
 *  is made.
 *
 */

void
t3pl_task_setup(t3rt_binary_string_t compaddr, t3rt_context_t ctx)
{
  if (ex_component_get_index(compaddr, ctx) == EX_CPC_COMPIDX) {
    unsigned long compidx = ex_component_create(1, -1, ctx);

    assert(compidx == 0); /* CPC must be first component! */
  }
}


static void
create_task_in_new_process(int argc,
                           const char * argv[],
                           t3rt_value_t component_value,
                           t3rt_binary_string_t address,
                           t3rt_context_t context)
{
#if defined(_WIN32) && ! defined(WIN32_GCC)   /* Win32 platform */
  assert(false);                /* FUTURE_EXTENSION: NYI for windows */
#else           /* UNIX platform */

  pid_t childpid;

  childpid = fork();

  if (childpid == -1) {
    perror("fork failed");
    exit(1);
  }

  if (childpid != 0) {
    /* In parent, succeeded starting process. */

    /* TODO: Here we should retrieve the control port of the new
     * component and return it.
     */
    ex_thread_sleep(30000, context); /* CONFIGURATION: 30s */

    return;
  }

  /* In child, succeeded starting process. */

  /* This will transfer execution to (eventually) the
     t3pl_concurrency_start_separate_component function: */
  execvp(argv[0], (char * const *)argv);

  perror("execvp failed");
  exit(1);
#endif
}


static ex_task_info_t
ex_task_create_task_info(t3rt_context_t ctx)
{
  ex_task_info_t info = 
    t3pl_memory_allocate(t3rt_perm_alloc_c, 
                         sizeof(ex_task_info_struct_t),
                         ctx);
  info->ready_sem = NULL;
  return info;
}

void 
ex_task_destroy_task_info(ex_task_info_t *info, t3rt_context_t ctx)
{
  if (*info == NULL)
    return;

  /* The ready_sem is dealloctad immediately after use in task_create. */
  assert((*info)->ready_sem == NULL);

  (void)ex_close_thread_handle((*info)->thread_handle);

  t3pl_memory_deallocate(t3rt_perm_alloc_c, *info, ctx);
  
  *info = NULL;
}


/* 
 * Create a port to control the task and a thread of execution executing 
 * the 't3rt_component_main' function. 
 * 
 * The binary string 'address' is set to the address of the control port 
 * of the new task.
 */

void
t3pl_task_create(int argc,
                 const char * argv[],
                 t3rt_value_t component_value,
                 t3rt_binary_string_t address,
                 t3rt_context_t ctx)
{
  ex_task_info_t   task_info;
  t3rt_boolean_t   op_status = false;
  t3rt_value_t     is_separate;
  unsigned long    compidx;
  ex_port_t        port;
  t3rt_binary_string_t control_port;
  t3rt_boolean_t status;

  is_separate = t3rt_rtconf_get_param ("separate_process", ctx);

  if (is_separate != t3rt_illegal_value_c && 
      is_separate != t3rt_no_value_c &&
      t3rt_value_get_boolean (is_separate, ctx)) {
    create_task_in_new_process(argc, argv, component_value, address, ctx);
    return;
  }

  compidx = ex_component_create(1, -1, ctx);

  task_info = ex_task_create_task_info(ctx);
  ex_component_set_task_info(compidx, task_info, ctx);

  task_info->ready_sem = t3pl_sem_create(0, ctx);

  port = ex_component_get_port(ex_component_object_id(compidx, 0), ctx);
  
  t3rt_binary_string_assign(address, port->address, ctx);

  control_port = t3rt_binary_string_copy(port->address, t3rt_perm_alloc_c, ctx);

  ex_create_thread(op_status,                /* status of operation */
                   task_info->thread_handle, /* storage for thread handle */
                   task_info->thread_id,     /* storage for thread id */
                   ex_thread_main,           /* main thread function */
                   (void *)control_port);    /* thread argument */
  assert(op_status);
  if (!op_status) {
    /* OS failed to create new thread */
    t3rt_error_description_t      err;

    err.error_code = 1;
    err.error_message = "Operating system failed to create new thread!";
    err.detailed_description = NULL;

    t3rt_report_fatal_system_error(err, ctx);
  }

  /* Wait here until 't3pl_task_register_context' in the new thread
   * declares itself ready in order to prevent race conditions: */
  status = t3pl_sem_wait(task_info->ready_sem, ctx);
  assert(status);

  assert(ex_component_get_context(compidx, ctx) != NULL);

  /* Deallocate the ready_sem here, as we will not need it any more: */
  status = t3pl_sem_destroy(task_info->ready_sem, ctx);
  assert(status);

  task_info->ready_sem = NULL;
}


/*
 * Force task to stop executing. If the component did shutdown according 
 * to normal shutdown procedures the 'shutdown_acknowledged' flag will be 
 * true.
 */

void
t3pl_task_kill(t3rt_binary_string_t address, 
               const t3rt_boolean_t shutdown_acknowledged,
               t3rt_context_t ctx)
{
  t3rt_boolean_t  status;
  unsigned long   compidx;
  ex_task_info_t  task_info;

  compidx = ex_component_get_index(address, ctx);
  task_info = ex_component_get_task_info(compidx, ctx);

  if (shutdown_acknowledged) { 
    ex_thread_id_t  thread_id;
    /* Assuming the thread will (or already has shut down itself) */
    /* If task_info is absent, the thread has already killed itself,
       otherwise we wait. */
    ex_get_current_thread_id(thread_id);

    if (task_info != NULL) {
      assert(thread_id != task_info->thread_id);
      ex_wait_for_thread_termination(status, task_info->thread_handle);
    }
  } else {
    assert(task_info != NULL);

    ex_terminate_thread(status, task_info->thread_handle);
    assert(status);

    ex_thread_sleep(100, ctx);

    /* Critical section start */
    ex_component_writer_start(ctx);
    ex_component_terminate_MT_unsafe(compidx, ctx);
    ex_component_writer_end(ctx);
    /* Critical section end */

    /* A call to context cleanup to deallocate the killed threads context
       should be made here. That would require an API change so it is not 
       presently done (in 2.1). 
       
       When context is passed to t3pl_task_kill like in t3pl_task_exit, 
       the function ex_component_delete_MT_unsafe the t3pl_timer_delete call 
       must be replaced by a call to t3pl_timer_stop (as in 
       ex_component_terminate_MT_unsafe).
       
       Without call to this function there will be memory leaks! */
    /*ctx_cleanup_f(ctx);*/

  }
}


/*
 * Called to terminate this task
 */

void
t3pl_task_exit (t3rt_context_cleanup_function_t ctx_cleanup_f,
                t3rt_context_t ctx)
{
  unsigned long compidx = ex_component_get_index(NULL, ctx);

  ex_component_writer_start(ctx);

  /* The allocated system component must be deallocated somewhere. 
     This is a good spot. */
  if (compidx == EX_MTC_COMPIDX)
    ex_component_terminate_MT_unsafe(EX_SYS_COMPIDX, ctx);

  ex_component_terminate_MT_unsafe(compidx, ctx);

  /* This functions should be called prior to termination to ensure deallocation
     of ctx structures. Without call to this function there will be
     a memory leak! */
  
  ex_component_writer_end(ctx);

  /* Context cleanup function accesses writer semaphor so
   * it should be called outside write_start -> witer_end sequence
   * to prevent locks 
   */

  ctx_cleanup_f (ctx);

  ex_exit_thread();
}


/*
 * This function returns globally unique Id number for current task
 * This may be process id or thread id or etc.
 * Note: this function does not receive context through parameters
 */

unsigned long
t3pl_task_id()
{
  ex_thread_id_t id;
  ex_get_current_thread_id(id);
  return (unsigned long)id;
}

/* End of File */

