/*---------------------------------------------------------------------------
 * File Name: example_eventqueue.c
 *
 * 2010 IBM Corporation
 *
 * Description:
 * This is an example implementation of the eventqueue part of the
 * Platform Layer.
 */


#include <stdio.h>
#include <stdlib.h>

#include "t3rts.h"

#include "example_general.h" /* For assert definition */
#include "example_eventqueue.h"
#include "example_time.h"         /* ex_get_current_relative_time */ 
#include "example_concurrency.h"  /* ex_thread_sleep */



/*
 *  This is the head of the queue that keeps track of head and tail.
 *
 *  The queue has no max size and thus require no write semaphore.
 *  It has a read semaphore to protect/block on reading from the
 *  empty queue and another semaphore to protect modification of the queue.
 *
 */

typedef struct ex_eventqueue_struct_internal_t
{
  ex_eventqueue_element_t    head, tail;
  void *                     queue_protection;
  void *                     read_sem;
  unsigned long              size;
  t3rt_boolean_t             inactive;
} ex_eventqueue_struct_t;



/*************************************************************************
 *
 *  This support contains a FIFO queue, definition of the queue elements
 *  and synchronization mechanisms (semaphores).
 *
 */


ex_eventqueue_t
ex_eventqueue_create(t3rt_context_t ctx)
{
  ex_eventqueue_t queue = (ex_eventqueue_t)
    t3pl_memory_allocate(t3rt_perm_alloc_c,
                         sizeof(ex_eventqueue_struct_t),
                         ctx);
  queue->head             = NULL;
  queue->tail             = NULL;
  queue->size             = 0;
  queue->queue_protection = t3pl_sem_create(1, ctx);
  queue->read_sem         = t3pl_sem_create(0, ctx); /* 0 = empty queue */
  queue->inactive         = false;
  return queue;
}


void
ex_eventqueue_clear(ex_eventqueue_t queue, t3rt_context_t ctx)
{
  ex_eventqueue_element_t       elem;
  ex_eventqueue_element_t       next_elem;

  elem = queue->head;
  while (elem != NULL) {
    next_elem = elem->next;
    ex_eventqueue_element_delete (&elem, ctx);
    elem = next_elem;
  }

  queue->head = NULL;
  queue->tail = NULL;
  queue->size = 0;

  /* TODO: Issue warnings on discarded elements?! */
}


void
ex_eventqueue_inactivate(ex_eventqueue_t queue, t3rt_context_t ctx)
{
  ex_eventqueue_clear(queue, ctx);

  queue->inactive = true;
}


void
ex_eventqueue_delete(ex_eventqueue_t * queue, t3rt_context_t ctx)
{
  t3rt_boolean_t status;

  ex_eventqueue_clear(*queue, ctx);

  status = t3pl_sem_destroy((*queue)->queue_protection, ctx);
  assert (status);
  status = t3pl_sem_destroy((*queue)->read_sem, ctx);
  assert (status);

  t3pl_memory_deallocate (t3rt_perm_alloc_c, (*queue), ctx);

  *queue = NULL;
}


unsigned long
ex_event_size(ex_eventqueue_t queue, t3rt_context_t ctx)
{
  unsigned long size;
  t3rt_boolean_t status;

  assert( queue != NULL );

  status = t3pl_sem_wait(queue->queue_protection, ctx);
  assert (status);

  size = queue->size;

  status = t3pl_sem_post(queue->queue_protection, ctx);
  assert (status);

  return size;
}



ex_eventqueue_element_t
ex_eventqueue_element_create(ex_eventqueue_element_kind_t kind,
                             t3rt_context_t ctx)
{
  ex_eventqueue_element_t elem = (ex_eventqueue_element_t)
    t3pl_memory_allocate(t3rt_perm_alloc_c,
                         sizeof(ex_eventqueue_element_struct_t),
                         ctx);

  elem->kind                        = kind;
  elem->port_address                = NULL;
  elem->signature_type              = t3rt_undefined_type;
  elem->data                        = NULL;
  elem->data_vector                 = NULL;
  elem->data_vector_length          = 0;
  elem->sender                      = NULL;
  elem->timer_handle                = NULL;
  elem->next                        = NULL;

  return elem;
}


void
ex_eventqueue_element_delete(ex_eventqueue_element_t * elem,
                             t3rt_context_t ctx)
{
  if (elem == NULL) return;

  if ((*elem)->port_address != NULL)
    t3rt_binary_string_deallocate_all(&((*elem)->port_address), ctx);

  if ((*elem)->data != NULL)
    t3rt_binary_string_deallocate_all(&((*elem)->data), ctx);

  if ((*elem)->sender!= NULL)
    t3rt_binary_string_deallocate_all(&((*elem)->sender), ctx);

  if ((*elem)->data_vector != NULL) {
    unsigned long i;

    for (i = 0; i < (*elem)->data_vector_length; ++i)
      t3rt_binary_string_deallocate_all(&((*elem)->data_vector[i]), ctx);

    t3pl_memory_deallocate(t3rt_perm_alloc_c, (*elem)->data_vector, ctx);
  }

  (*elem)->port_address                = NULL;
  (*elem)->signature_type              = t3rt_undefined_type;
  (*elem)->data                        = NULL;
  (*elem)->data_vector                 = NULL;
  (*elem)->data_vector_length          = 0;
  (*elem)->sender                      = NULL;
  (*elem)->timer_handle                = NULL;
  (*elem)->next                        = NULL;

  t3pl_memory_deallocate(t3rt_perm_alloc_c, *elem, ctx);

  *elem = NULL;
}


void
ex_event_put(ex_eventqueue_t         queue,
             ex_eventqueue_element_t element,
             t3rt_context_t ctx)
{
  t3rt_boolean_t status;

  assert( queue != NULL );
  assert( element != NULL );

  status = t3pl_sem_wait(queue->queue_protection, ctx);
  assert (status);

  if (queue->tail == NULL) { /* Empty queue */
    queue->head = queue->tail = element;
  } else {
    queue->tail->next = element;
    queue->tail       = element;
  }

  queue->size++;

  status = t3pl_sem_post(queue->queue_protection, ctx);
  assert (status);

  (void)t3pl_sem_post(queue->read_sem, ctx);
}

/* msg->port_address is NULL in timeout event */
#define is_message_to_address(msg, addr, ctx) \
  ((msg) && (msg)->port_address && (addr) && t3rt_binary_string_is_equal((msg)->port_address, addr, ctx))

/* helper function that extracts element from event queue */  
static ex_eventqueue_element_t
extract_eventqueue_element(ex_eventqueue_t queue, t3rt_binary_string_t dest_port_address, t3rt_context_t ctx)
{
  ex_eventqueue_element_t result = NULL;
  t3rt_boolean_t status;

  status = t3pl_sem_wait(queue->queue_protection, ctx);
  assert (status);

  if (queue->head != NULL) {
    ex_eventqueue_element_t prev = NULL;
    
    result = queue->head;
    
    /* search for matching element */
    if (dest_port_address) {
      while (result && !is_message_to_address(result, dest_port_address, ctx))
      {
        prev = result;
        result = result->next;
      }
    }

    /* found matching element in the queue */
    if (result) {
      if (queue->head == result)
        queue->head = result->next;
      else
        prev->next = result->next;

      if (queue->tail == result)
        queue->tail = prev;

      queue->size--;
    }
  }

  status = t3pl_sem_post(queue->queue_protection, ctx);
  assert (status);

  return result;
}
  
/*
 * Even if this function is allowed access to the queue, the queue can
 * have been deactivated and in that case NULL is returned.
 * If dest_port != NULL function returns only message if it's sent to dest_port
 */
ex_eventqueue_element_t
ex_event_get(ex_eventqueue_t queue, t3rt_binary_string_t dest_port_address,
             t3rt_context_t  ctx)
{
  ex_eventqueue_element_t result = NULL;

  assert( queue != NULL );

  if (queue->inactive) {
    fprintf(stderr, "PL WARNING: Get on inactive component queue!\n");
    return NULL;
  }

  (void)t3pl_sem_wait(queue->read_sem, ctx);

  result = extract_eventqueue_element(queue, dest_port_address, ctx);
  if (!result)
    /* We must increment semaphore since we haven't extracted data from non-empty(!) queue */
    (void)t3pl_sem_post(queue->read_sem, ctx);
  
  return result;
}


static ex_eventqueue_element_t
ex_event_tryget_support(ex_eventqueue_t queue,
                        t3rt_binary_string_t dest_port_address,
                        double * timeout,
                        t3rt_context_t  ctx)
{
  ex_eventqueue_element_t result = NULL;
  double save_time;

  assert(timeout != NULL);
  assert(*timeout >= 0.0);

#ifdef DEBUG_CONTROL_PROTOCOL
  fflush(stdout);
  fprintf(stderr, "tryget with timeout %f %08x<%d>.\n", *timeout,
          (int)ctx, (int)ctx->index);
#endif

  save_time = ex_get_current_relative_time();

  if (t3pl_sem_timedwait(queue->read_sem, *timeout, ctx)) {
    *timeout -= ex_get_current_relative_time() - save_time;
    if (*timeout <= 0.0)  *timeout = 0.0;

    result = extract_eventqueue_element(queue, dest_port_address, ctx);
    if (result)
      return result;
    else
      /* We must increment semaphore since we haven't extracted data from non-empty(!) queue */
      (void)t3pl_sem_post(queue->read_sem, ctx);
  }

  *timeout = 0.0;
  
#ifdef DEBUG
  fprintf(stderr, "tryget timed out %f %08x<%d>.\n", *timeout,
          (int)ctx, (int)ctx->index);
#endif

  return NULL;
}  

/*
 * The timeout may be specified as pointer to t3rt_duration_nowait_c
 * (or as NULL) to signify that no waiting should be done.  This
 * function returns NULL if a timeout occured or if the queue is inactive.
 */
ex_eventqueue_element_t
ex_event_tryget(ex_eventqueue_t queue,
                t3rt_binary_string_t dest_port_address,
                double * timeout,
                t3rt_context_t  ctx)
{
  ex_eventqueue_element_t result = NULL;

  assert( queue != NULL );

  if (queue->inactive) {
    fprintf(stderr, "PL WARNING: Tryget on inactive component queue!\n");
    return NULL;
  }

  if (timeout != NULL && *timeout == t3rt_duration_forever_c) /*lint !e777 */
    return ex_event_get(queue, dest_port_address, ctx);

  /* Use a helper function for the tricky case: */
  if (timeout != NULL && *timeout != t3rt_duration_nowait_c) /*lint !e777 */
    return ex_event_tryget_support(queue, dest_port_address, timeout, ctx);

  if (t3pl_sem_trywait(queue->read_sem, ctx))
  {
    result = extract_eventqueue_element(queue, dest_port_address, ctx);
    if (!result)
      /* We must increment semaphore since we haven't extracted data from non-empty(!) queue */
      (void)t3pl_sem_post(queue->read_sem, ctx);
  }

  return result;
}


void
ex_eventqueue_element_set_port_address(ex_eventqueue_element_t element,
                                       t3rt_binary_string_t    port_addr,
                                       t3rt_context_t          ctx)
{
  element->port_address =
    t3rt_binary_string_copy(port_addr, t3rt_perm_alloc_c, ctx);
}


void
ex_eventqueue_element_set_signature_type(ex_eventqueue_element_t element,
                                         t3rt_type_t             type,
                                         t3rt_context_t          ctx)
{
  element->signature_type = type;
}


void
ex_eventqueue_element_set_data(ex_eventqueue_element_t element,
                               t3rt_binary_string_t    data,
                               t3rt_context_t          ctx)
{
  if (data != NULL)
    element->data =
      t3rt_binary_string_copy(data, t3rt_perm_alloc_c, ctx);
  else
    element->data = NULL;
}


void
ex_eventqueue_element_set_data_vector(ex_eventqueue_element_t  element,
                                      t3rt_binary_string_t *   data_vector,
                                      unsigned long            size,
                                      t3rt_context_t           ctx)
{
  unsigned long i;

  element->data_vector =
    t3pl_memory_allocate(t3rt_perm_alloc_c,
                         size * sizeof(t3rt_binary_string_t*),
                         ctx);

  for (i = 0; i < size; ++i)
    if (data_vector[i] != NULL)
      element->data_vector[i] = t3rt_binary_string_copy(data_vector[i],
                                                        t3rt_perm_alloc_c,
                                                        ctx);
    else
      element->data_vector[i] = NULL;

  element->data_vector_length = size;
}


void
ex_eventqueue_element_set_sender(ex_eventqueue_element_t element,
                                 t3rt_binary_string_t    sender,
                                 t3rt_context_t          ctx)
{
  if (sender != NULL)
    element->sender = t3rt_binary_string_copy(sender, t3rt_perm_alloc_c, ctx);
  else
    element->sender = NULL;
}


void
ex_eventqueue_element_set_timer_handle(ex_eventqueue_element_t element,
                                       t3rt_timer_handle_t     handle,
                                       t3rt_context_t          ctx)
{
  element->timer_handle = handle;
}



/* End of File */
