/*---------------------------------------------------------------------------
 * File Name: example_general.c
 *
 * 2010 IBM Corporation
 *
 * Description:
 * This is an example implementation of the general part of the
 * Platform Layer.
 */

#include <string.h>             /* For memcpy, memmove etc */

#include "t3rts.h"
#include "example_general.h" /* For assert definition */
#include "example_communication.h"



/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
#include "t3tri.h"
#include "example_triutil.h"

#define EX_TRI_EXTERNAL_FUNCTION_FAILED "triExternalFunction failed!"
#define EX_TRI_EXECUTE_TESTCASE_FAILED  "triExecuteTestcase failed!"
#define EX_TRI_END_TESTCASE_FAILED      "triEndTestcase failed!"
#define EX_TRI_DECODE_FAILED            "Decoding of value in TRI failed!"
#define EX_TRI_ENCODE_FAILED            "Encoding of value in TRI failed!"


extern t3rt_context_t current_ctx;
extern void *         current_ctx_sem;

unsigned long
component_port_count(t3rt_type_t type, t3rt_context_t ctx)
{
  unsigned long i, count = 0;

  /* System port id list */
  for(i = 0; i < t3rt_type_field_count(type, ctx); ++i) {
    t3rt_type_t type_elem = t3rt_type_field_type(type, i, ctx);
    if (t3rt_type_kind(type_elem, ctx) == t3rt_type_kind_port)
      count++;

    if (t3rt_type_kind(type_elem, ctx) == t3rt_type_kind_array &&
        t3rt_type_kind(t3rt_type_array_contained_type(type_elem, ctx), ctx) == 
          t3rt_type_kind_port)
          count += t3rt_type_array_size(type_elem, ctx);

  }

  return count;
}

#endif /* T3PL_TRI */
/* T3RTS_TRI_SECTION_END */

#ifdef DEBUG
void example_assert(t3rt_boolean_t cond, const char *cond_str, int line, const char* file)
{
  if (cond)
    return;

  fflush(stdout);
  fprintf(stderr, "%s(%d): Assert '%s' failed!\n", file, line, cond_str);

  exit(1); 
}
#endif

void
t3pl_general_prepare_testcase(const char* module, 
                              const char* testcase, 
                              t3rt_type_t mtc_type, 
                              t3rt_type_t system_type,
                              t3rt_context_t ctx)
{
/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  TriTestCaseId qname;
  TriPortIdList tsiPortIdList;
  unsigned long i, ports_count = 0;
  TriStatus status;
  t3rt_boolean_t sem_status;

  /* Test case name */
  (void)tri__create_QualifiedName(module, testcase, NULL, 
                                  t3rt_temp_alloc_c, &qname, ctx);

  /* Go from the component reference type to the component "record" type: */
  system_type = t3rt_type_field_type(system_type, 0, ctx);

  tsiPortIdList.length = (long) component_port_count(system_type, ctx);
  tsiPortIdList.portIdList = (TriPortId**)
    t3rt_memory_temp_allocate(((unsigned long)sizeof(TriPortId*) * 
                               (unsigned long)tsiPortIdList.length
                              ), 
                              ctx);


  /* System port id list. It can be ports or arrays of ports. */
  for(i = 0; i < t3rt_type_field_count(system_type, ctx); ++i) {
    ex_port_t   port;
    const char* name = t3rt_type_field_name(system_type, i, ctx);
    t3rt_type_t type_elem = t3rt_type_field_type(system_type, i, ctx);

    if (t3rt_type_kind(type_elem, ctx) == t3rt_type_kind_port) {
      port = ex_component_find_port(EX_SYS_COMPIDX, name, -1, ctx);
      tsiPortIdList.portIdList[ports_count++] = &(port->port_repr->triPortId);
    } else if (t3rt_type_kind(type_elem, ctx) == t3rt_type_kind_array &&
               t3rt_type_kind(t3rt_type_array_contained_type(type_elem, ctx), ctx) == 
               t3rt_type_kind_port) {
      unsigned long j;

      for(j = 0; j < t3rt_type_array_size(type_elem, ctx); ++j) {
        port = ex_component_find_port(EX_SYS_COMPIDX, name, (long)j, ctx);
        tsiPortIdList.portIdList[ports_count++] = &(port->port_repr->triPortId);
      }
    }
  }

  tsiPortIdList.length = (long)ports_count;

  sem_status = t3pl_sem_wait(current_ctx_sem, ctx);
  assert(sem_status);
  assert(current_ctx == NULL);
  current_ctx = ctx;

  status = triExecuteTestcase(&qname, &tsiPortIdList);

  current_ctx = NULL;
  sem_status = t3pl_sem_post(current_ctx_sem, ctx);
  assert(sem_status);

  if (status != TRI_OK)
    ex_tri_error(EX_TRI_EXECUTE_TESTCASE_FAILED, ctx);

#else
/* T3RTS_TRI_SECTION_END */
  /* Go from the component reference type to the component "record" type: */
  mtc_type = t3rt_type_field_type(mtc_type, 0, ctx);
  system_type = t3rt_type_field_type(system_type, 0, ctx);
/* T3RTS_TRI_SECTION_START */
#endif
/* T3RTS_TRI_SECTION_END */
}

void
t3pl_general_postprocess_testcase(const char* module, 
                                  const char* testcase, 
                                  t3rt_type_t mtc_type, 
                                  t3rt_type_t system_type,
                                  t3rt_context_t ctx)
{
/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  TriStatus status;

  status = triEndTestcase();

  if (status != TRI_OK)
    ex_tri_error(EX_TRI_END_TESTCASE_FAILED, ctx);
#endif
/* T3RTS_TRI_SECTION_END */
}

void
t3pl_general_testcase_terminated(const char* module, 
                                 const char* testcase, 
                                 t3rt_type_t mtc_type, 
                                 t3rt_type_t system_type,
                                 t3rt_context_t ctx)
{
/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  ex_communication_testcase_terminated(ctx);
  ex_concurrency_testcase_terminated(ctx);
#endif
/* T3RTS_TRI_SECTION_END */
}

void
t3pl_general_control_terminated(const char* module, 
                                t3rt_context_t ctx)
{
  ex_component_release_cpc(ctx);
}

/*
 *  t3pl_call_external_function
 *
 *  The return value should be assign with the return value received from the
 *  actual function call. The return_value parameter can be 't3rt_no_value_c' 
 *  in which case no return values expected. 
 */
void
t3pl_call_external_function(t3rt_type_t signature_type, 
                            t3rt_value_t parameters[], 
                            t3rt_value_t return_value, 
                            t3rt_context_t ctx)
{
/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  TriFunctionId    functionId;  /* QualifiedName */
  TriParameterList parameterList;
  TriParameter     returnValue;
  unsigned long i;

  (void)tri__create_QualifiedName(t3rt_type_module(signature_type, ctx), 
                                  t3rt_type_name(signature_type, ctx),
                                  NULL, t3rt_temp_alloc_c, &functionId, ctx);

    /* Create parameter list. */
  if (t3rt_type_field_count(signature_type, ctx) == 0)
    parameterList.length = 0;
  else /* Don't count with the return value */
    parameterList.length = (long int)(t3rt_type_field_count(signature_type, ctx) - 1); 

  if (parameterList.length == 0) {
    parameterList.length = 0;
    parameterList.parList = NULL;
  } else {
    parameterList.parList = 
      (TriParameter **)t3rt_memory_temp_allocate((unsigned long)parameterList.length * 
                                                  sizeof(TriParameter*),
				                                 ctx);

    for(i = 0; i < (unsigned long)parameterList.length; ++i) {
      t3rt_binary_string_t enc_par = NULL;

      if (t3rt_type_field_properties(signature_type, i + 1, ctx) != 
            t3rt_field_property_out_value_c) {

        enc_par = t3rt_binary_string_allocate(t3rt_temp_alloc_c, 0, ctx);

        if (t3rt_encode(parameters[i], enc_par, ctx) != t3rt_codecs_result_succeeded_c) {
          ex_tri_error(EX_TRI_ENCODE_FAILED, ctx);
        }
      }

      parameterList.parList[i] = (TriParameter*)
        t3rt_memory_temp_allocate(sizeof(TriParameter), ctx);

      /* Initialization of BinaryString in parameterList.parList[i]->par */
      parameterList.parList[i]->par.aux = NULL;
      parameterList.parList[i]->par.bits = 0;
      parameterList.parList[i]->par.data = NULL;
      
      tri__create_TriParameter(enc_par,
                               t3rt_type_field_properties(signature_type, i + 1, ctx),
                               t3rt_temp_alloc_c,
                               parameterList.parList[i], ctx);
    }
  }

  /* Call the TRI implementation and handle the result. */
  if (triExternalFunction(&functionId, &parameterList, &returnValue) == TRI_OK) {
    t3rt_value_t              temp_value;
    t3rt_binary_string_t      enc_par;
    t3rt_binary_string_iter_t iter;

    if (return_value != t3rt_no_value_c) { /* Return value expected. */
      enc_par = tri__tri_to_t3_binary_string(&(returnValue.par), t3rt_temp_alloc_c, ctx);
      t3rt_binary_string_start(enc_par, &iter, ctx);

      if (t3rt_decode(&iter,
                      t3rt_type_field_type(signature_type, 0, ctx),
                      t3rt_temp_alloc_c, &temp_value, ctx) !=
        t3rt_codecs_result_succeeded_c) {
        assert(false);
      }

      t3rt_value_assign(return_value, temp_value, ctx);
    }

    /* "Unpack" the inout and out parameters. */
    for(i = 0; i < (unsigned long)parameterList.length; ++i) {
      TriParameter *tri_par = parameterList.parList[i];

      if (tri_par->mode == TRI_IN) {
        assert(t3rt_type_field_properties(signature_type, i + 1, ctx) ==
                 t3rt_field_property_in_value_c); 
        continue;
      }

      if (tri_par->mode == TRI_INOUT) 
        assert(t3rt_type_field_properties(signature_type, i + 1, ctx) == 
                 t3rt_field_property_inout_value_c);

      if (tri_par->mode == TRI_OUT)        
        assert(t3rt_type_field_properties(signature_type, i + 1, ctx) == 
                 t3rt_field_property_out_value_c);

      enc_par = tri__tri_to_t3_binary_string(&(tri_par->par), t3rt_temp_alloc_c, ctx);
      t3rt_binary_string_start(enc_par, &iter, ctx);

      if (t3rt_decode(&iter, t3rt_type_field_type(signature_type, i + 1, ctx), 
                      t3rt_temp_alloc_c, &temp_value, ctx) != 
          t3rt_codecs_result_succeeded_c) {
        ex_tri_error(EX_TRI_DECODE_FAILED, ctx);
      }

      assert( parameters[i] != t3rt_no_value_c );

      t3rt_value_assign(parameters[i], temp_value, ctx);
    }
  } else {
    ex_tri_error(EX_TRI_EXTERNAL_FUNCTION_FAILED, ctx);
  }

#endif
/* T3RTS_TRI_SECTION_END */
}


/* End of File */

