/*---------------------------------------------------------------------------
 * File Name: example_semaphore.h
 *
 * 2010 IBM Corporation
 *
 * Description:
 * This is an example implementation of a semaphore utility abstraction.
 *
 * A semaphore is a data object used for managing access to a shared
 * resource between multiple threads.
 *
 * There is a fundamental semaphore functionality available via the
 * t3pl_sem_* functions provided by this module, that this module also
 * uses to build upon to provide so called reader-writer semaphores.
 * This enhanced type of semaphore is used to manage access to a
 * shared resource that can have multiple readers accessing the
 * resource simultaneously, but only a single writer at a time.
 *
 * The readers-writers semaphores are used by first allocating an
 * ex_rw_semaphore_struct_t structure and initialize it by calling the
 * ex_sem_alloc function and later using the result in other
 * ex_sem_*{start,end} function calls.
 *
 */


#include "example_general.h" /* For assert definition */
#include "example_semaphore.h"


/***************************************************************************/
#ifdef _WIN32   /* Win32 platform */
/***************************************************************************/

#include <windows.h>
#include <limits.h>

typedef HANDLE              ex_sem_t;

#define ex_create_sem(status, sem_ptr, available_resources)                             \
        *(sem_ptr) = CreateSemaphore (NULL, (long)available_resources, LONG_MAX, NULL); \
        status = (*(sem_ptr) != NULL);

#define ex_wait_sem(sem_ptr)                   WaitForSingleObject (*(sem_ptr), INFINITE)
#define ex_try_wait_sem(sem_ptr)               WaitForSingleObject (*(sem_ptr), 0)
#define ex_timed_wait_sem(sem_ptr, wait_sec)   WaitForSingleObject(*(sem_ptr), (unsigned long)(wait_sec*1000));
#define ex_wait_sem_success                    WAIT_OBJECT_0
#define ex_wait_sem_success2                   WAIT_ABANDONED_0

#define ex_post_sem(sem_ptr)                   ReleaseSemaphore (*(sem_ptr), 1, NULL)
#define ex_post_sem_success                    TRUE

#define ex_destroy_sem(sem_ptr)                CloseHandle (*(sem_ptr))
#define ex_destroy_sem_success                 TRUE

/***************************************************************************/
#else           /* Unix platform */
/***************************************************************************/

#include <pthread.h>
#include <sys/time.h>
#include <errno.h>      /* for errno and ETIMEDOUT etc. */

typedef struct {
          unsigned long n_resources;
          pthread_mutex_t mutex; 
          pthread_cond_t  cond;
        } ex_sem_t;

#define sem_resources(sem_ptr) (sem_ptr->n_resources)
#define mutex_ptr(sem_ptr) (&(sem_ptr->mutex))
#define cond_ptr(sem_ptr) (&(sem_ptr->cond))

#define ex_create_sem(status, sem_ptr, available_resources) \
        { \
          sem_resources(sem_ptr) = available_resources; \
          status = (pthread_mutex_init (mutex_ptr(sem_ptr), NULL) == 0 && \
                    pthread_cond_init (cond_ptr(sem_ptr), NULL) == 0); \
        }

#define ex_destroy_sem(sem_ptr)            !(pthread_mutex_destroy (mutex_ptr(sem_ptr)) == 0 && \
                                             pthread_cond_destroy (cond_ptr(sem_ptr)) == 0)

#define ex_destroy_sem_success             0
        
static int 
ex_wait_sem(ex_sem_t *sem_ptr) 
{
  int status;
  
  status = pthread_mutex_lock(mutex_ptr(sem_ptr));
  if (status == 0)
  {
    while (sem_resources(sem_ptr) == 0)
    {
	  pthread_cond_wait(cond_ptr(sem_ptr), mutex_ptr(sem_ptr));
    }
    assert(sem_resources(sem_ptr) > 0);
    sem_resources(sem_ptr) -= 1;
  
    status = pthread_mutex_unlock(mutex_ptr(sem_ptr));
    assert(status == 0);
    
    return 0;
  }
  
  return -1;
}

static int 
ex_try_wait_sem(ex_sem_t *sem_ptr) 
{
  int status;  

  status = pthread_mutex_lock(mutex_ptr(sem_ptr));
  if (status == 0)
  {
    int retcode = 0;
    
    assert(sem_resources(sem_ptr) >= 0);
    if (sem_resources(sem_ptr) > 0)
      sem_resources(sem_ptr) -= 1;
    else
      retcode = -1;

    status = pthread_mutex_unlock(mutex_ptr(sem_ptr));
    assert(status == 0);
    
    return retcode;
  } 
  
  return -1;
}

#define NANOSEC 1000000000

static int 
ex_timed_wait_sem(ex_sem_t *sem_ptr, double wait_time) 
{
  int status;
  struct timespec timeout;
  struct timeval now;
  unsigned long wait_seconds = (unsigned long)wait_time;
  unsigned long wait_nseconds = (unsigned long)((wait_time - wait_seconds)*NANOSEC);

  gettimeofday(&now, NULL);
  timeout.tv_sec = now.tv_sec + wait_seconds;
  timeout.tv_nsec = now.tv_usec*1000 + wait_nseconds;
  if (timeout.tv_nsec >= NANOSEC) 
  {
    timeout.tv_sec ++ ;
    timeout.tv_nsec -= NANOSEC;
  }

  status = pthread_mutex_lock(mutex_ptr(sem_ptr));
  if (status == 0) 
  {
    int retcode = 0;
  
    while (sem_resources(sem_ptr) == 0 && retcode != ETIMEDOUT) 
    {
      retcode = pthread_cond_timedwait(cond_ptr(sem_ptr), mutex_ptr(sem_ptr), &timeout);
    }  

    if (retcode != ETIMEDOUT)
    {
      assert(sem_resources(sem_ptr) > 0);
      sem_resources(sem_ptr) -= 1;
    }

    status = pthread_mutex_unlock(mutex_ptr(sem_ptr)); 
    assert(status == 0);
    
    return retcode == ETIMEDOUT ? -1 : 0;
  }    
  
  return -1;
}

#define ex_wait_sem_success                0
#define ex_wait_sem_success2               0

static int 
ex_post_sem(ex_sem_t *sem_ptr) 
{
  int status;
  
  status = pthread_mutex_lock(mutex_ptr(sem_ptr));
  if (status == 0) 
  {
    assert(sem_resources(sem_ptr) >= 0);
    sem_resources(sem_ptr) += 1;

    pthread_cond_signal(cond_ptr(sem_ptr));
    
    status = pthread_mutex_unlock(mutex_ptr(sem_ptr));
    assert(status == 0);
    
    return 0;
  }
  
  return -1;
}

#define ex_post_sem_success                0

/***************************************************************************/
#endif
/***************************************************************************/

#ifdef _WIN32
static void
report_windows_error(DWORD err, t3rt_context_t ctx)
{
  LPVOID lpMsgBuf;


  (void)FormatMessage( 
    FORMAT_MESSAGE_ALLOCATE_BUFFER | 
    FORMAT_MESSAGE_FROM_SYSTEM | 
    FORMAT_MESSAGE_IGNORE_INSERTS,
    NULL,
    err,
    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), /* Default language */
    (LPTSTR) &lpMsgBuf,
    0,
    NULL 
    );

  /* Log the string. */
  t3rt_log_string_to_all(t3rt_log_kind_msg_error_c,
                         (LPCTSTR)lpMsgBuf,
                         ctx);

  /* Free the buffer. */
  LocalFree (lpMsgBuf);
}
#endif


/* =========================================================================
 * Simple semaphores
 * =========================================================================
 */


/*
 * Initialize semaphore
 */

void *
t3pl_sem_create(unsigned int available_resources, t3rt_context_t ctx)
{
  ex_sem_t *     sem_ptr;
  t3rt_boolean_t status;


  sem_ptr = (ex_sem_t *)t3pl_memory_allocate(t3rt_perm_alloc_c,
                                             sizeof(ex_sem_t),
                                             ctx);

  ex_create_sem(status, sem_ptr, available_resources);
  assert(status);

  return sem_ptr;
}


/*
 * Wait on semaphore
 */

t3rt_boolean_t
t3pl_sem_wait(void *sem, t3rt_context_t ctx)
{
  ex_sem_t *         sem_ptr = (ex_sem_t *)sem;
  unsigned long      status = (unsigned long)-1;
  t3rt_boolean_t     result;


  assert(sem_ptr != NULL);
  status = (unsigned long)ex_wait_sem(sem_ptr);

  result = (status == ex_wait_sem_success || status == ex_wait_sem_success2);

#ifdef _WIN32
  if (!result) report_windows_error(GetLastError(), ctx);
#endif

  return result;
}


/*
 * Trywait semaphore
 */

t3rt_boolean_t
t3pl_sem_trywait(void * sem, t3rt_context_t ctx)
{
  ex_sem_t *         sem_ptr = (ex_sem_t *)sem;
  unsigned long      status = (unsigned long)-1;
  t3rt_boolean_t     result;

  assert(sem_ptr != NULL);
  status = (unsigned long)ex_try_wait_sem(sem_ptr);

  result = (status == ex_wait_sem_success || status == ex_wait_sem_success2);

  return result;
}

/*
 * Timout semaphore
 */
 
t3rt_boolean_t
t3pl_sem_timedwait(void * sem, double wait_seconds, t3rt_context_t ctx)
{
  ex_sem_t *         sem_ptr = (ex_sem_t *)sem;
  unsigned long      status = (unsigned long)-1;
  t3rt_boolean_t     result;

  assert(sem_ptr != NULL);

  /*Waiting with timeout (NOTE: timout is provided in seconds)
    Conversion to other time values is done insed ex_*_sem functions
  */
  status = (unsigned long)ex_timed_wait_sem(sem_ptr, wait_seconds);

  result = (status == ex_wait_sem_success || status == ex_wait_sem_success2);

  return result;
}

/*
 * Release semaphore
 */

t3rt_boolean_t
t3pl_sem_post(void * sem, t3rt_context_t ctx)
{
  ex_sem_t *          sem_ptr = (ex_sem_t *)sem;
  unsigned long       status = (unsigned long)-1;
  t3rt_boolean_t      result;

  assert(sem_ptr != NULL);

  status = (unsigned long)ex_post_sem(sem_ptr);

  result = (status == ex_post_sem_success);

#ifdef _WIN32
  if (!result) report_windows_error(GetLastError(), ctx);
#endif

  return result;
}


/*
 * Destroy semaphore
 */

t3rt_boolean_t
t3pl_sem_destroy(void * sem, t3rt_context_t ctx)
{
  ex_sem_t *      sem_ptr = (ex_sem_t *)sem;
  int             status = -1;
  t3rt_boolean_t  result;

  assert(sem_ptr != NULL);

  status = ex_destroy_sem(sem_ptr);
  result = (status == ex_destroy_sem_success);

  t3pl_memory_deallocate(t3rt_perm_alloc_c, sem, ctx);

  return result;
}


/* =========================================================================
 * Reader-writer semaphores
 * =========================================================================
 */

/*
 * This uses three simple semaphores and a counter to implement a
 * semaphore that allows multiple 'readers' simultaneous access to the
 * critical resource that is managed by it, but only a single 'writer'
 * - and that situation may also only occur when no 'reader' is
 * accessing the critical resource.
 *
 * To prevent a situation called 'live-lock', where the 'writer' is
 * indefinitely blocked from access by multiple 'readers' taking turns
 * holding the semaphore, the implementation uses an extra simple
 * semaphore (called req_sem) to "serialize" the writer among the
 * 'readers'.
 *
 * The other two simple semaphores are used to signify if a(ny number
 * of) reader(s) and a writer is accessing the critical resource at
 * the moment.  The counter is used to count the number of readers
 * currently accessing the critical resource (i.e. are inside the
 * critical section).
 *
 * The write semaphore is taken by the writer at start and released at
 * the end of the critical section.  The read semaphore, however, is
 * taken by the first reader to enter the critical section, but it is
 * not released until the last reader leaves the critical section.
 *
 */

void
ex_sem_alloc(ex_rw_semaphore_t rw_sem, t3rt_context_t ctx)
{
  assert(rw_sem != NULL);

  rw_sem->req_sem = t3pl_sem_create(1, ctx);
  rw_sem->write_sem = t3pl_sem_create(1, ctx);
  rw_sem->read_sem = t3pl_sem_create(1, ctx);
  rw_sem->num_readers = 0;
}


void
ex_sem_dealloc(ex_rw_semaphore_t rw_sem, t3rt_context_t ctx)
{
  assert(rw_sem != NULL);

  (void)t3pl_sem_destroy(rw_sem->req_sem, ctx);
  (void)t3pl_sem_destroy(rw_sem->write_sem, ctx);
  (void)t3pl_sem_destroy(rw_sem->read_sem, ctx);
  rw_sem->req_sem = NULL;
  rw_sem->write_sem = NULL;
  rw_sem->read_sem = NULL;
  rw_sem->num_readers = 0;
}


void
ex_sem_reader_start(ex_rw_semaphore_t rw_sem, t3rt_context_t ctx)
{
  void * write_request_semaphore;
  void * write_semaphore;
  void * read_semaphore;
  unsigned long * number_of_readers;
  t3rt_boolean_t status;

  assert(rw_sem != NULL);

  write_request_semaphore = rw_sem->req_sem;
  write_semaphore = rw_sem->write_sem;
  read_semaphore = rw_sem->read_sem;
  number_of_readers = &rw_sem->num_readers;

  /* Touching write-request semaphore to prevent new readers
   * before writer request completed
   */
  status = t3pl_sem_wait(write_request_semaphore, ctx);
  assert(status); /* ERRORHANDLING! */
  status = t3pl_sem_post(write_request_semaphore, ctx);
  assert(status); /* ERRORHANDLING! */


  status = t3pl_sem_wait(read_semaphore, ctx);
  assert(status); /* ERRORHANDLING! */

  if (*number_of_readers == 0) {
    status = t3pl_sem_wait(write_semaphore, ctx);
    assert(status); /* ERRORHANDLING! */
  }

  (*number_of_readers)++;

  status = t3pl_sem_post(read_semaphore, ctx);
  assert(status); /* ERRORHANDLING! */
}


void
ex_sem_reader_end(ex_rw_semaphore_t rw_sem, t3rt_context_t ctx)
{
  void * write_semaphore = rw_sem->write_sem;
  void * read_semaphore = rw_sem->read_sem;
  unsigned long * number_of_readers = &rw_sem->num_readers;
  t3rt_boolean_t status;


  status = t3pl_sem_wait(read_semaphore, ctx);
  assert(status); /* ERRORHANDLING! */

  --(*number_of_readers);

  if (*number_of_readers == 0) {
    status = t3pl_sem_post(write_semaphore, ctx);
    assert(status); /* ERRORHANDLING! */
  }

  status = t3pl_sem_post(read_semaphore, ctx);
  assert(status); /* ERRORHANDLING! */
}


void
ex_sem_writer_start(ex_rw_semaphore_t rw_sem, t3rt_context_t ctx)
{
  void * write_request_semaphore = rw_sem->req_sem;
  void * write_semaphore = rw_sem->write_sem;
  t3rt_boolean_t status;

  status = t3pl_sem_wait(write_request_semaphore, ctx);
  assert(status); /* ERRORHANDLING! */

  status = t3pl_sem_wait(write_semaphore, ctx);
  assert(status); /* ERRORHANDLING! */
}


void
ex_sem_writer_end(ex_rw_semaphore_t rw_sem, t3rt_context_t ctx)
{
  void * write_request_semaphore = rw_sem->req_sem;
  void * write_semaphore = rw_sem->write_sem;
  t3rt_boolean_t status;

  status = t3pl_sem_post(write_semaphore, ctx);
  assert(status); /* ERRORHANDLING! */

  status = t3pl_sem_post(write_request_semaphore, ctx);
  assert(status); /* ERRORHANDLING! */
}



/* End of File */
