/*---------------------------------------------------------------------------
 * File Name: example_sequence.h
 *
 * 2010 IBM Corporation
 *
 * Description:
 *  This is an example implementation of a sequence container utility.
 */


#ifndef __example_sequence_h
#define __example_sequence_h

#ifdef __cplusplus
extern "C" {
#endif

#include "t3rts.h"



/*
 * A sequence is a data structure used to hold a varying number of elements.
 * This particular implementation of sequences is used by first declaring
 * a named sequence head, for example in a struct, like this:
 *
 * typedef struct XXX_internal_t {
 * ...
 * ex_sequence_head(ex_element_struct_type_t) my_element_q;
 * ...
 * } XXX_t;
 *
 * (where ex_element_struct_type_t is the type of each element
 *  of the sequence).
 *
 * Access and manipulation of the sequence is then accomplished by using
 * the ex_sequence_{init,insert,remove,add,size,element,...} macros.
 *
 * Examples:
 * XXX_t xxx;
 * ex_sequence_init(xxx.my_element_q);
 * ex_sequence_add(xxx.my_element_q, t3pl_temp_alloc_c, ctx);
 * ex_sequence_last_element(xxx.my_element_q).elem_attr = 42;
 *
 */

#define ex_sequence_head(ELTYPE) struct {                    \
                                        unsigned long alloc;    \
                                        unsigned long size;     \
                                        ELTYPE* seq;            \
                                    }

/* See comment block below for example of use: */
#define ex_sequence_iter(ELTYPE) struct {                                \
                                      ex_sequence_head(ELTYPE) *  seq;   \
                                      ELTYPE *                       pos;   \
                                      long                           index; \
                                    }

#define ex_sequence_init(SEQ) \
  ((SEQ).seq = NULL, (SEQ).alloc = 0, (SEQ).size = 0)

#define ex_sequence_deallocate(SEQ,AS,CTX) \
  ex_sequence_deallocate_aux(&(SEQ),AS,CTX)

#define ex_sequence_insert(SEQ,IDX,AS,CTX) \
  ex_sequence_insert_aux(&(SEQ),IDX,sizeof((SEQ).seq[0]),AS,CTX)

#define ex_sequence_remove(SEQ,IDX,CTX) \
  ex_sequence_remove_aux(&(SEQ),IDX,sizeof((SEQ).seq[0]),CTX)

#define ex_sequence_add(SEQ,AS,CTX) \
  ex_sequence_insert_aux(&(SEQ),(long)(SEQ).size,sizeof((SEQ).seq[0]),AS,CTX)

#define ex_sequence_size(SEQ) (SEQ).size
#define ex_sequence_element(SEQ,IDX) (SEQ).seq[IDX]
#define ex_sequence_first_element(SEQ) (SEQ).seq[0]
#define ex_sequence_last_element(SEQ) (SEQ).seq[(SEQ).size - 1]


typedef struct ex_sequence_head_struct_internal_t {
  unsigned long     alloc;
  unsigned long     size;
  void*             seq;
} ex_sequence_head_struct_t;

void
ex_sequence_insert_aux(void *arg, long idx, unsigned long element_size,
                          t3rt_alloc_strategy_t strategy, t3rt_context_t ctx);

void
ex_sequence_remove_aux(void *arg, long idx, unsigned long element_size,
                          t3rt_context_t ctx);

void
ex_sequence_deallocate_aux(void *arg, t3rt_alloc_strategy_t strategy,
                              t3rt_context_t ctx);

/*
 * Sequence iterator
 *
 * This is a pure forward iterator for the above sequence mechanism.
 * 
 * The iterator will be corrupted if the underlying sequence is enlarged 
 * beyond its allocated size (i.e. reallocated). 
 * 
 * Example:
 *
 * ex_sequence_iter(ex_element_struct_type_t) iter;
 * ex_sequence_iter_start(iter, xxx.my_element_q);
 * ...
 * while (ex_sequence_iter_next(iter)) {
 *   ...
 *   ex_element_struct_type_t var = ex_sequence_iter_get(iter);
 *   ...
 *   ex_element_struct_type_t *var2 = &(ex_sequence_iter_get(iter));
 *   ...
 *   while (should_be_removed(ex_sequence_iter_get(iter))) {
 *     if (ex_sequence_iter_remove(iter, ctx))
 *       break;
 *   }
 * }
 *
 */

#define ex_sequence_iter_start(ITER, SEQ) \
  ((ITER).seq = (void*) &(SEQ), \
   (ITER).pos = NULL, (ITER).index = -1)

#define ex_sequence_iter_reset(ITER) \
  ((ITER).pos = NULL, (SEQ).index = -1)

#define ex_sequence_iter_next(ITER) \
  ((ITER).index >= 0 ? \
    ((ITER).pos += 1, ++((ITER).index) < (long)(ITER).seq->size) : \
    ((ITER).pos = (ITER).seq->seq, (ITER).index = 0, (ITER).seq->size > 0))

#define ex_sequence_iter_get(ITER) \
  (*((ITER).pos))

#define ex_sequence_iter_get_index(ITER) \
  ((ITER).index)

#define ex_sequence_iter_remove(ITER,CTX) \
  (ex_sequence_remove(*(ITER).seq,(ITER).index,CTX), \
   (((ITER).index) >= (long)(ITER).seq->size))


#ifdef __cplusplus
}
#endif
   
#endif

/* End of File */

