/*---------------------------------------------------------------------------
 * File Name: t3pl_time.c
 *
 * 2010 IBM Corporation
 *
 * Description:
 *     This is an example implementation of the time part of the
 * Platform Layer.
 */


#include <sys/types.h>
#include <sys/timeb.h>
#include <time.h>

#if defined(_WIN32) || defined(WIN32_GCC)   /* Win32 platform and Cygwin! */
#ifndef _WIN32_WINNT
/* This is necessary to include Platform SDK headers */
#define _WIN32_WINNT 0x0501
#endif
#include <windows.h>
#else
#include <sys/time.h>
#endif

#include "t3rts.h"

#include "example_general.h" /* For assert definition */
#include "example_time.h"
#include "example_components.h"

static double ex_start_time;

double
ex_get_current_time (void)
{
#if defined(_WIN32) || defined(WIN32_GCC)   /* Win32 platform and Cygwin! */
    static t3rt_boolean_t first_chance = true;
    static LARGE_INTEGER freq;
    static LARGE_INTEGER tick_start;
    static double time_start;

    LARGE_INTEGER now;
    double elapsed;

    if (first_chance) 
    {
       struct timeb tmb;
       struct tm *tms;

       ftime(&tmb);
       /* 'tms' buffer is thread unsafe, but it's Ok here since
        * first call to this function is from t3pl_time_pre_initialize().
        */
       tms = localtime(&tmb.time);
       time_start = (tms->tm_hour * 3600 + tms->tm_min * 60 + tms->tm_sec) + (tmb.millitm * 0.001);

       (void)QueryPerformanceFrequency(&freq);
       (void)QueryPerformanceCounter(&tick_start);
       first_chance = false;
    }

    (void)QueryPerformanceCounter(&now);

    elapsed = (double)(now.QuadPart - tick_start.QuadPart) / freq.QuadPart;

    return time_start + elapsed;
#else           /* Unix platform */
    static t3rt_boolean_t first_chance = true;
    static double time_start;
    static struct timeval tick_start;

    struct timeval now;
    double elapsed;

    if (first_chance) 
    {
       struct timeb tmb;
       struct tm *tms;

       ftime(&tmb);
       /* 'tms' buffer is thread unsafe, but it's Ok here since
        * first call to this function is from t3pl_time_pre_initialize().
        */
       tms = localtime(&tmb.time);
       time_start = (tms->tm_hour * 3600 + tms->tm_min * 60 + tms->tm_sec) + (tmb.millitm * 0.001);

       (void)gettimeofday(&tick_start, NULL);

       first_chance = false;
    }

    (void)gettimeofday(&now, NULL);

    elapsed = (double)(now.tv_sec - tick_start.tv_sec) + ((double)(now.tv_usec - tick_start.tv_usec))*0.000001;

    return time_start + elapsed;
#endif
}

double
ex_get_current_relative_time (void)
{
    return ex_get_current_time() - ex_start_time;
}


static ex_timer_t
get_timer_from_handle(const t3rt_timer_handle_t handle,
                      t3rt_context_t ctx)
{
  assert(handle != NULL);

  return ex_component_get_timer(*(ex_component_object_id_t*)handle, ctx);
}


/*
 * ex_timer_update
 *
 * Updates the timer representation to the current state of the timer.
 *
 */

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
static void
ex_timer_update(ex_timer_repr_t timer,
                t3rt_context_t ctx)
{
  unsigned char is_running;

  if (triTimerRunning(&(timer->tri_TimerId), &is_running) == TRI_Error)
    ex_tri_error(EX_TRI_TIMER_RUNNING_FAILED, ctx);

  switch (timer->state) {
  case t3rt_timer_state_running: {
    if (is_running == 0) {
      timer->elapsed_time = 0.0;
      timer->state = t3rt_timer_state_timedout;
    } else {
      if (triReadTimer(&(timer->tri_TimerId), &(timer->elapsed_time)) == TRI_Error)
        ex_tri_error(EX_TRI_READ_TIMER_FAILED, ctx);

      if (timer->elapsed_time >= timer->duration) /* Timer timedout */
        timer->state = t3rt_timer_state_timedout;
    }
                           }
    break;
  case t3rt_timer_state_timedout:
  case t3rt_timer_state_stopped:
    break;
  default: {
    assert (false); /* Should never happen! */
           }
    break;
  }
}
#else /* T3PL_TRI */
/* T3RTS_TRI_SECTION_END */
static void
ex_timer_update (ex_timer_repr_t timer,
                 t3rt_context_t ctx)
{
  switch (timer->state) {
  case t3rt_timer_state_running: {
    if (timer->stop_time <= ex_get_current_relative_time()) {
      timer->state = t3rt_timer_state_timedout;
    }
    break;
                           }
  case t3rt_timer_state_timedout:
  case t3rt_timer_state_stopped:
    break;
  default: {
    assert (false); /* Should never happen! */
           }
    break;
  }
}
/* T3RTS_TRI_SECTION_START */
#endif
/* T3RTS_TRI_SECTION_END */



/*****************************************/


/*
 * This function is presently considered to be a no-op since its provided
 * functionality is not required for that phase
 */

void
t3pl_time_pre_initialize (int argc, char * argv [], t3rt_context_t ctx)
{
    ex_start_time = ex_get_current_time();
}


/*
 * Initialize the timer module
 */

void
t3pl_time_initialize (t3rt_context_t ctx)
{
}


/*
 * Finalize the timer module
 */

void
t3pl_time_finalize (t3rt_context_t ctx)
{
}

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI

/*
 * The timerid is an encoded unsigned long (ex_component_object_id_t)
 */

static void
encode_timer_handle(ex_component_object_id_t objid,
                    TriTimerId *timerId,
                    t3rt_context_t ctx)
{
  assert( sizeof(long) >= 4 );

  timerId->data =
    (unsigned char*)t3pl_memory_allocate(t3rt_perm_alloc_c,
                                         sizeof(unsigned long),
                                         ctx);
  timerId->data[0] = (unsigned char)(objid >> 24);
  timerId->data[1] = (unsigned char)((objid & 0x00FF0000) >> 16);
  timerId->data[2] = (unsigned char)((objid & 0x0000FF00) >> 8);
  timerId->data[3] = (unsigned char)(objid & 0x000000FF);
  timerId->bits = 32;
  timerId->aux = NULL;
}

/*
 * The timerid is an encoded unsigned long (ex_component_object_id_t)
 *
 */

ex_component_object_id_t
ex_decode_timer_handle(const TriTimerId *timerId)
{
  unsigned long first = 0, second = 0, third = 0, fourth= 0;

  assert(timerId->data != NULL &&
         timerId->bits == 32);

  first  = ((unsigned long) timerId->data[0]) << 24;
  second = ((unsigned long) timerId->data[1]) << 16;
  third  = ((unsigned long) timerId->data[2]) <<  8;
  fourth =  (unsigned long) timerId->data[3];

  return (first | second | third | fourth);
}

#endif /* T3PL_TRI */
/* T3RTS_TRI_SECTION_END */


/*
 * decodes RTS timer handle to obtain TRI timer handle 
 * Note: return type is t3rt_binary_string_t, not TriBinaryString 
 *
 * TRI: This implementation is also valid for the TRI integration.
 *
 */

t3rt_binary_string_t 
t3pl_timer_decode(t3rt_timer_handle_t handle,
                  t3rt_alloc_strategy_t strategy,  
                  t3rt_context_t ctx)
{
/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  ex_timer_t timer = get_timer_from_handle(handle, ctx);

  assert(timer != NULL && timer->timer_repr != NULL);

  return tri__tri_to_t3_binary_string (&timer->timer_repr->tri_TimerId, 
                                       strategy, 
                                       ctx);
#else
/* T3RTS_TRI_SECTION_END */
  return t3rt_binary_string_allocate(strategy, 0, ctx);
/* T3RTS_TRI_SECTION_START */
#endif
/* T3RTS_TRI_SECTION_END */
}

/*
 * Create a timer instance into the "stopped" state
 *
 *
 * TRI: This implementation is also valid for the TRI integration.
 *
 */
void
t3pl_timer_create (t3rt_timer_handle_t *handle,
                   t3rt_context_t ctx)
{
  ex_timer_t      timer;
  unsigned long compidx;

  assert(handle != NULL);

  compidx = ex_component_get_index(NULL, ctx);

  timer = ex_component_create_timer(compidx, ctx); /* This also creates the handle. */

  timer->timer_repr =
    (ex_timer_repr_t)t3pl_memory_allocate(t3rt_perm_alloc_c,
                                          sizeof(ex_timer_repr_struct_t),
                                          ctx);

  assert(timer->timer_repr != NULL);

  *handle = timer->handle;

  timer->timer_repr->state = t3rt_timer_state_stopped;
/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  timer->timer_repr->duration   = 0.0;
  timer->timer_repr->elapsed_time = 0.0;
  encode_timer_handle(timer->id, &(timer->timer_repr->tri_TimerId), ctx);
#else
/* T3RTS_TRI_SECTION_END */
  timer->timer_repr->start_time = 0.0;
  timer->timer_repr->stop_time  = 0.0;
/* T3RTS_TRI_SECTION_START */
#endif
/* T3RTS_TRI_SECTION_END */
}


/*
 * Delete the timer instance. It will never be used again
 */

void
t3pl_timer_delete(t3rt_timer_handle_t *handle,
                  t3rt_context_t ctx)
{
  ex_timer_t timer = get_timer_from_handle(*handle, ctx);

  assert(timer != NULL && timer->timer_repr != NULL);

  /* Stop the timer here, if applicable. */
  if (timer->timer_repr->state == t3rt_timer_state_running)
    t3pl_timer_stop(*handle, ctx);
  /* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  clear_BinaryString(&(timer->timer_repr->tri_TimerId), t3rt_perm_alloc_c, ctx);
#endif
  /* T3RTS_TRI_SECTION_END */
  
  t3rt__memory_deallocate( t3rt_perm_alloc_c,
						  timer->timer_repr,
						  ctx);

  ex_component_destroy_timer(timer, ctx);

  *handle = NULL; /* Handle deallocated in ex_component_destroy_timer. */ 
}


/*
 * Start a created timer instance. Also used to restart the timer.
 * This should set the timer into the "running" state
 */

void
t3pl_timer_start(const t3rt_timer_handle_t handle,
                 double duration,
                 t3rt_context_t ctx)
{
  ex_timer_t timer = get_timer_from_handle(handle, ctx);

  assert(timer != NULL && timer->timer_repr != NULL);

  timer->timer_repr->state = t3rt_timer_state_running;
  /* T3RTS_TRI_SECTION_START */
#ifndef T3PL_TRI
  /* T3RTS_TRI_SECTION_END */
  timer->timer_repr->start_time = ex_get_current_relative_time();
  timer->timer_repr->stop_time = timer->timer_repr->start_time + duration;
  /* T3RTS_TRI_SECTION_START */
#else /* TRI */
  timer->timer_repr->duration = duration;
  timer->timer_repr->elapsed_time = 0.0;

  if (triStartTimer(&(timer->timer_repr->tri_TimerId), duration) == TRI_Error)
    ex_tri_error(EX_TRI_START_TIMER_FAILED, ctx);
#endif /* TRI */
  /* T3RTS_TRI_SECTION_END */
}


/*
 * Stop a timer instance. This should set the timer into the "stopped" state
 */

void
t3pl_timer_stop(const t3rt_timer_handle_t handle,
                t3rt_context_t ctx)
{
  ex_timer_t timer = get_timer_from_handle(handle, ctx);

  assert(timer != NULL && timer->timer_repr != NULL);

  if (timer->timer_repr->state == t3rt_timer_state_stopped)
    return;

  timer->timer_repr->state = t3rt_timer_state_stopped;

/* T3RTS_TRI_SECTION_START */
#ifdef T3PL_TRI
  if (triStopTimer(&(timer->timer_repr->tri_TimerId)) == TRI_Error)
    ex_tri_error(EX_TRI_STOP_TIMER_FAILED, ctx);
#endif
/* T3RTS_TRI_SECTION_END */
}


/*
 * Read the current value of a timer
 */

t3rt_timer_state_t
t3pl_timer_read(const t3rt_timer_handle_t handle,
                double * elapsed_time,
                t3rt_context_t ctx)
{
  double             timer_time = 0.0;
  ex_timer_repr_t    timer;

  timer = get_timer_from_handle(handle, ctx)->timer_repr;

  ex_timer_update (timer, ctx);

  if (timer->state == t3rt_timer_state_running) {
/* T3RTS_TRI_SECTION_START */
#ifndef T3PL_TRI
/* T3RTS_TRI_SECTION_END */
    timer_time = ex_get_current_relative_time() - timer->start_time;
/* T3RTS_TRI_SECTION_START */
#else /* T3PL_TRI */
    timer_time = timer->elapsed_time;
#endif
/* T3RTS_TRI_SECTION_END */
    /* Here we check that the elapsed time is not zero. */
    if (timer_time <= 0.0)
      timer_time = 0.0000001; /* Preventing zero time. */
  }

  if (elapsed_time != NULL)
    *elapsed_time = timer_time;

  return timer->state;
}


/* End of File */

