/**************************************************************************
 * File Name: t3tri.c
 *
 * 2010 IBM Corporation
 *
 * Description:
 *     TRI VERSION 4.2
 *
 */

#ifdef T3PL_TRI


#include <string.h>

#include "t3rts.h"

#include "t3tri.h"

#include "example_general.h" /* For assert definition */
#include "example_communication.h"
#include "example_concurrency.h"
#include "example_time.h"
#include "example_triutil.h"

extern t3rt_context_t current_ctx;


#define MAPPED_PORTS_LEN 50


static ex_port_t
find_destination_port(const TriPortId   *tsiPortId,
                      const TriComponentId *componentId,
                      t3rt_context_t *ctx_out)
{
  ex_port_t       sysport, current_port;
  ex_sequence_iter(ex_port_t) iter;


  /* This is an unsafe context since the CPC is active but it is most likely 
   * waiting for final verdict in an 'execute' statement so it should work... 
   */
  *ctx_out = ex_component_get_context_MT_unsafe(EX_CPC_COMPIDX);

  sysport = ex_component_find_port(EX_SYS_COMPIDX, 
                                   tsiPortId->portName, 
                                   tsiPortId->portIndex, 
                                   *ctx_out);

  if (NULL == sysport) {
    fprintf(stderr, "ERROR: No system component - no testcase running?\n");
    fflush(stderr);
    return NULL;
  }

  ex_sem_reader_start(&sysport->sem,*ctx_out);

  if (1 == ex_sequence_size(sysport->sysport_associated_with)) {
    ex_sem_reader_end(&sysport->sem,*ctx_out);
    return ex_sequence_first_element(sysport->sysport_associated_with);
  }

  ex_sequence_iter_start(iter,sysport->sysport_associated_with);

  while (ex_sequence_iter_next(iter)) {
    current_port = ex_sequence_iter_get(iter);
    if (NULL != current_port) {
      if (t3tri_componentId_is_equal(&(current_port->port_repr->triPortId.compInst), componentId)) {
	ex_sem_reader_end(&sysport->sem,*ctx_out);
	return current_port;
      }
    }
  }

  ex_sem_reader_end(&sysport->sem,*ctx_out);

  fprintf(stderr, "ERROR: Port not found.\n");
  fflush(stderr);
  return NULL;
}


static ex_component_queue_t
object_id_to_component_queue(ex_component_object_id_t objid, t3rt_context_t ctx)
{
  return ex_component_get_comp_queue(ex_component_index(objid), ctx);
}

/**************************************************************************
 *
 * SA -> TE
 *
 * Functions that are called from the SUT Adapter.
 *
 */
void
triEnqueueMsg(const TriPortId      *tsiPortId,
              const TriAddress     *sutAddress,
              const TriComponentId *componentId,
              const TriMessage     *receivedMessage)
{
  t3rt_binary_string_t  data = NULL;
  t3rt_binary_string_t  sut_address = NULL;
  ex_port_t             dest_port = NULL;
  t3rt_context_t        ctx = NULL;

  dest_port = find_destination_port(tsiPortId, componentId, &ctx);

  assert(ctx != NULL);

  if (NULL == dest_port) {
    fprintf(stderr, "ERROR (in triEnqueueMessage): could not locate destination port - message discarded!\n");
    fflush(stderr);
    return;
  }

  if (receivedMessage == NULL) {
    fprintf(stderr, "ERROR (in triEnqueueMessage): message not provided - message discarded!\n");
    fflush(stderr);
    return;
  }

  data = tri__tri_to_t3_binary_string(receivedMessage, t3rt_perm_alloc_c, ctx);

  if (sutAddress != NULL && sutAddress->bits >= 0)
    sut_address = tri__tri_to_t3_binary_string(sutAddress, t3rt_perm_alloc_c, ctx);

  ex_component_queue_put_message(object_id_to_component_queue(dest_port->id, ctx), 
                                 dest_port->address,
                                 data,
                                 sut_address,
                                 ctx);

  /* free memory */ 
  if (data) t3rt_binary_string_deallocate_all(&data, ctx);
  if (sut_address) t3rt_binary_string_deallocate_all(&sut_address, ctx);
}

/*----------------------------------------------------------*/
/* Common function looking for signature */
typedef enum { TRI_ENQUEUE_CALL = 0, TRI_ENQUEUE_REPLY, TRI_ENQUEUE_EXCEPTION } TRI_ENQUEUE_TYPE;
static t3rt_type_t triGetSignatureType(const TriSignatureId   *signatureId,
                                const TriParameterList        *parameterList,
                                t3rt_context_t                 ctx,
                                TRI_ENQUEUE_TYPE               enqueueType)
{
  static char *strCallFNames[] = { "triEnqueueCall", "triEnqueueReply", "triEnqueueException" };
  static char *strActionNames[] = { "call", "reply", "exception" };
  t3rt_symbol_entry_t   symbol;
  char                 *moduleName;
  char                 *objectName;
  t3rt_type_t           signatureType;
  unsigned long         signatureParCount, actualParCount;

  moduleName = signatureId->moduleName;
  objectName = signatureId->objectName;

  if (!moduleName || (strlen(moduleName) == 0)) {
    fprintf(stderr, "WARNING (in %s): signature module name not specified, directing search in modules!\n", strCallFNames[enqueueType]);
    fflush(stderr);
    moduleName = NULL;
  }

  if (!objectName || (strlen(objectName) == 0)) {
    fprintf(stderr, "ERROR (in %s): signature name not specified - %s discarded!\n", strCallFNames[enqueueType], strActionNames[enqueueType]);
    fflush(stderr);
    return t3rt_illegal_type;
  }

  actualParCount = parameterList ? parameterList->length : 0;

  if(moduleName)
    symbol = t3rt_find_element(moduleName, objectName, ctx);
  else
  {
    t3rt_symbol_entry_t   *symbols;
    unsigned long i, number_symbols = t3rt_find_element_in_all_modules(&symbols, objectName, ctx);

    symbol = NULL;
    for(i = 0; i < number_symbols; i++)
      if(t3rt_type_kind(signatureType = t3rt_symbol_table_entry_type(symbols[i], ctx), ctx) == t3rt_type_kind_signature) {
        if(enqueueType == TRI_ENQUEUE_EXCEPTION)
          signatureParCount = 0;
        else {
          signatureParCount = t3rt_type_field_count(signatureType, ctx);
          if(signatureParCount > 0)
            signatureParCount--;
        }
        if(signatureParCount == actualParCount) {
          if((symbol == NULL) /* first occurence found */ ||
            (t3rt_symbol_table_entry_kind(symbol, ctx) != t3rt_symbol_entry_kind_signature_type) /* it was imported symbol */)
            symbol = symbols[i];
          else if(t3rt_symbol_table_entry_kind(symbols[i], ctx) == t3rt_symbol_entry_kind_signature_type) {
            fprintf(stderr, "ERROR (in %s): ambiguity: signature '%s' found in several modules - %s discarded!\n",
            strCallFNames[enqueueType], objectName, strActionNames[enqueueType]);
            fflush(stderr);
            return t3rt_illegal_type;
          }
        }
      }
  }

  if (!symbol || t3rt_type_kind(signatureType = t3rt_symbol_table_entry_type(symbol, ctx), ctx) != t3rt_type_kind_signature) {
    fprintf(stderr, "ERROR (in %s): object '%s' not found in symbol table or not a signature type - %s discarded!\n",
      strCallFNames[enqueueType], objectName, strActionNames[enqueueType]);
    fflush(stderr);
    return t3rt_illegal_type;
  }

  if(enqueueType != TRI_ENQUEUE_EXCEPTION) {
    signatureParCount = t3rt_type_field_count(signatureType, ctx);
    if(signatureParCount > 0)
      signatureParCount--;
    if (actualParCount != signatureParCount)
    {
      fprintf(stderr, "ERROR (in %s): parameter list length mismatch for procedure '%s' - %s discarded!\n",
        strCallFNames[enqueueType], objectName, strActionNames[enqueueType]);
      fflush(stderr);
      return t3rt_illegal_type;
    }
  }

  return signatureType;
}

void
triEnqueueCall (const TriPortId        *tsiPortId,
                const TriAddress       *sutAddress,
                const TriComponentId   *componentId,
                const TriSignatureId   *signatureId,
                const TriParameterList *parameterList)
{
  t3rt_binary_string_t  sut_address = NULL;
  t3rt_binary_string_t *parameters = NULL;
  long                  i;
  ex_port_t             dest_port = NULL;
  t3rt_context_t        ctx = NULL;
  t3rt_type_t           signatureType;

  dest_port = find_destination_port(tsiPortId, componentId, &ctx);

  assert(ctx != NULL);

  if (NULL == dest_port) {
    fprintf(stderr, "ERROR (in triEnqueueCall): could not locate destination port - call discarded!\n");
    fflush(stderr);
    return;
  }

  signatureType = triGetSignatureType(signatureId, parameterList, ctx, TRI_ENQUEUE_CALL);
  if(signatureType == t3rt_illegal_type)
    return;

  if (t3rt_type_field_count(signatureType, ctx) > 1) {
    /* signature has parameters */
    parameters = (t3rt_binary_string_t *)
      t3pl_memory_allocate(t3rt_perm_alloc_c,
                           (unsigned long)(parameterList->length * sizeof(t3rt_binary_string_t*)), /*lint !e737 */
                           ctx);

    for(i = 0; i < parameterList->length; ++i) {
      TriParameter *tpar = parameterList->parList[i];
      t3rt_field_properties_t field_props = t3rt_type_field_properties(signatureType, i+1, ctx);

      if (!tpar || tpar->mode == TRI_OUT) {
        parameters[i] = NULL;
      } else {
        parameters[i] = tri__tri_to_t3_binary_string(&(tpar->par), t3rt_perm_alloc_c, ctx);
      }

      if (parameters[i] == NULL && 
          (field_props == t3rt_field_property_in_value_c || field_props == t3rt_field_property_inout_value_c)) {

        fprintf(stderr, "ERROR (in triEnqueueCall): some of the 'in' and 'inout' parameters for procedure '%s' are not provided - call discarded!\n", signatureId->objectName);
        fflush(stderr);
        return;
      }
    }
  }

  if (sutAddress != NULL && sutAddress->bits >= 0)
    sut_address = tri__tri_to_t3_binary_string(sutAddress, t3rt_perm_alloc_c, ctx);

  ex_component_queue_put_call(object_id_to_component_queue(dest_port->id, ctx), 
                              dest_port->address,
                              signatureType,
                              parameters,
                              sut_address,
                              ctx);
                              
  /* free memory */ 
  for(i = 0; parameterList && i < parameterList->length; ++i) {
    if (parameters && parameters[i]) t3rt_binary_string_deallocate_all(&parameters[i], ctx);
  }
  if (parameters) t3pl_memory_deallocate(t3rt_perm_alloc_c, parameters, ctx);
  if (sut_address) t3rt_binary_string_deallocate_all(&sut_address, ctx);
}



void
triEnqueueReply (const TriPortId        *tsiPortId,
                 const TriAddress       *sutAddress,
                 const TriComponentId   *componentId,
                 const TriSignatureId   *signatureId,
                 const TriParameterList *parameterList,
                 const TriParameter     *returnValue)
{
  t3rt_binary_string_t  sut_address = NULL;
  t3rt_binary_string_t *parameters = NULL;
  t3rt_binary_string_t  retval = NULL;
  long                  i;
  ex_port_t             dest_port = NULL;
  t3rt_context_t        ctx = NULL;
  t3rt_type_t           signatureType;

  dest_port = find_destination_port(tsiPortId, componentId, &ctx);

  assert(ctx != NULL);

  if (NULL == dest_port) {
    fprintf(stderr, "ERROR (in triEnqueueReply): could not locate destination port - reply discarded!\n");
    fflush(stderr);
    return;
  }

  signatureType = triGetSignatureType(signatureId, parameterList, ctx, TRI_ENQUEUE_REPLY);
  if(signatureType == t3rt_illegal_type)
    return;

  if (t3rt_type_field_count(signatureType, ctx) > 1) {
    /* signature has parameters */
    parameters = (t3rt_binary_string_t *)
      t3pl_memory_allocate(t3rt_perm_alloc_c,
                           (unsigned long)(parameterList->length * sizeof(t3rt_binary_string_t*)), /*lint !e737 */
                           ctx);

    for(i = 0; i < parameterList->length; ++i) {
      TriParameter *tpar = parameterList->parList[i];
      t3rt_field_properties_t field_props = t3rt_type_field_properties(signatureType, i+1, ctx);

      if (!tpar || tpar->mode == TRI_IN) {
        parameters[i] = NULL;
      } else {
        parameters[i] = tri__tri_to_t3_binary_string(&(tpar->par), t3rt_perm_alloc_c, ctx);
      }

      if (parameters[i] == NULL && 
          (field_props == t3rt_field_property_out_value_c || field_props == t3rt_field_property_inout_value_c)) {

        fprintf(stderr, "ERROR (in triEnqueueReply): some of the 'out' and 'inout' parameters for procedure '%s' are not provided - reply discarded!\n", signatureId->objectName);
        fflush(stderr);
        return;
      }
    }
  }

  if (returnValue != NULL)
    retval = tri__tri_to_t3_binary_string(&(returnValue->par), t3rt_perm_alloc_c, ctx);

  if (retval == NULL && t3rt_type_field_type(signatureType, 0, ctx) != t3rt_undefined_type) {
    fprintf(stderr, "ERROR (in triEnqueueReply): return value for procedure '%s' is not provided - reply discarded!\n", signatureId->objectName);
    fflush(stderr);
    return;
  }

  if (sutAddress != NULL && sutAddress->bits >= 0)
    sut_address = tri__tri_to_t3_binary_string(sutAddress, t3rt_perm_alloc_c, ctx);

  ex_component_queue_put_reply(object_id_to_component_queue(dest_port->id, ctx), 
                               dest_port->address,
                               signatureType,
                               parameters,
                               retval,
                               sut_address,
                               ctx);
                               
  /* free memory */ 
  for(i = 0; parameterList && i < parameterList->length; ++i) {
    if (parameters && parameters[i]) t3rt_binary_string_deallocate_all(&parameters[i], ctx);
  }
  if (parameters) t3pl_memory_deallocate(t3rt_perm_alloc_c, parameters, ctx);
  if (retval) t3rt_binary_string_deallocate_all(&retval, ctx);
  if (sut_address) t3rt_binary_string_deallocate_all(&sut_address, ctx);
}



void
triEnqueueException (const TriPortId      *tsiPortId,
                     const TriAddress     *sutAddress,
                     const TriComponentId *componentId,
                     const TriSignatureId *signatureId,
                     const TriException   *exception)
{
  t3rt_binary_string_t sut_address = NULL;
  t3rt_binary_string_t exc = NULL;
  ex_port_t            dest_port = NULL;
  t3rt_context_t       ctx = NULL;
  t3rt_type_t           signatureType;

  dest_port = find_destination_port(tsiPortId, componentId, &ctx);

  assert(ctx != NULL);

  if (NULL == dest_port) {
    fprintf(stderr, "ERROR (in triEnqueueException): could not locate destination port - exception discarded!\n");
    fflush(stderr);
    return;
  }

  signatureType = triGetSignatureType(signatureId, NULL, ctx, TRI_ENQUEUE_EXCEPTION);
  if(signatureType == t3rt_illegal_type)
    return;

  if (exception != NULL)
    exc = tri__tri_to_t3_binary_string(exception, t3rt_perm_alloc_c, ctx);

  if (exc == NULL) {
    fprintf(stderr, "ERROR (in triEnqueueException): exception value for procedure '%s' is not provided - exception discarded!\n", signatureId->objectName);
    fflush(stderr);
    return;
  }

  if (sutAddress != NULL && sutAddress->bits >= 0)
    sut_address = tri__tri_to_t3_binary_string(sutAddress, t3rt_perm_alloc_c, ctx);

  ex_component_queue_put_exception(object_id_to_component_queue(dest_port->id, ctx), 
                                   dest_port->address,
                                   signatureType,
                                   exc,
                                   sut_address,
                                   ctx);
                                   
  /* free memory */ 
  if (exc) t3rt_binary_string_deallocate_all(&exc, ctx);
  if (sut_address) t3rt_binary_string_deallocate_all(&sut_address, ctx);
}


/**************************************************************************
 *
 * PA -> TE
 *
 * Functions that are called from the Platform Adaptater.
 *
 */

void
triTimeout (const TriTimerId *timerId)
{
  t3rt_context_t ctx = NULL;
  ex_component_object_id_t objid;
  ex_timer_t timer;

  ctx = ex_component_get_context_MT_unsafe(EX_CPC_COMPIDX);

  objid = ex_decode_timer_handle(timerId);

  timer = ex_component_get_timer(objid, ctx);

  if (timer == NULL)
    return;

  ex_component_queue_put_timeout(object_id_to_component_queue(objid, ctx),
                                 timer->handle,
                                 ctx);
}


/**************************************************************************
 *
 * TRI Support functions
 *
 *
 */

void
clear_BinaryString(BinaryString *str,
                   t3rt_alloc_strategy_t strategy,
                   t3rt_context_t context)
{
  str->aux = NULL;
  str->bits = 0;
  t3pl_memory_deallocate(strategy, str->data, context);
  str->data = NULL;
}


QualifiedName *
tri__create_QualifiedName(const char* moduleName, const char* objectName,
                          void* aux, t3rt_alloc_strategy_t strategy,
                          QualifiedName *name, t3rt_context_t ctx)
{
  assert(name != NULL);

  name->moduleName = (char*)t3rt__memory_allocate(strategy, strlen(moduleName) + 1, ctx);
  strcpy(name->moduleName, moduleName);

  name->objectName = (char*)t3rt__memory_allocate(strategy, strlen(objectName) + 1, ctx);
  strcpy(name->objectName, objectName);

  name->aux = aux;

  return name;
}


void
tri__create_TriParameterList(long param_count,
                             t3rt_alloc_strategy_t strategy,
                             TriParameterList *parlist,
                             t3rt_context_t ctx)
{
  parlist->length = param_count;
  parlist->parList = (TriParameter**)
    t3rt__memory_allocate(strategy, 
                          (unsigned long)(param_count * sizeof(*parlist)), /*lint !e737 */
                          ctx);

  while (--param_count >= 0) {
    TriParameter * par = (TriParameter*)
      t3rt__memory_allocate(strategy, sizeof(parlist->parList[0][0]),
                            ctx);

    parlist->parList[param_count] = par;

    par->mode = TRI_IN;         /* Just anything valid will do right now */

    /* Initialization of BinaryString in par->par: */
    par->par.aux = NULL;
    par->par.bits = 0;
    par->par.data = NULL;
  }
}


void
tri__create_TriParameter(t3rt_binary_string_t enc_par,
                         t3rt_field_properties_t props,
                         t3rt_alloc_strategy_t strategy,
                         TriParameter *par,
                         t3rt_context_t ctx)
{
  switch (props) {
  default:
    assert(false); break;
  case t3rt_field_property_in_value_c:    par->mode = TRI_IN; break;
  case t3rt_field_property_out_value_c:   par->mode = TRI_OUT; break;
  case t3rt_field_property_inout_value_c: par->mode = TRI_INOUT; break;
  }

  if (enc_par != NULL)
    (void)tri__t3_to_tri_BinaryString(enc_par, strategy, &(par->par), ctx);
  else
    clear_BinaryString(&(par->par), strategy, ctx);
}



BinaryString *
tri__t3_to_tri_BinaryString(t3rt_binary_string_t t3string,
                            t3rt_alloc_strategy_t strategy,
                            BinaryString *bstring,
                            t3rt_context_t ctx)
{
  t3rt_binary_string_iter_t bstr_iter;
  unsigned long             allocated_bytesize;

  assert(bstring != NULL);

  allocated_bytesize = t3rt_binary_string_length (t3string, ctx);
  allocated_bytesize = (allocated_bytesize / 8) + (((allocated_bytesize % 8) > 0) ? 1 : 0);

  if (strategy == t3rt_temp_alloc_c)
    bstring->data = (unsigned char*)t3rt_memory_temp_allocate(allocated_bytesize, ctx);
  else
    bstring->data = (unsigned char*)t3pl_memory_allocate(strategy, allocated_bytesize, ctx);

  t3rt_binary_string_start (t3string, &bstr_iter, ctx);
  t3rt_bstring_iter_get_nbits (&bstr_iter,
                               t3rt_binary_string_length (t3string, ctx),
                               bstring->data,
                               ctx);

  bstring->bits = (long)t3rt_binary_string_length(t3string, ctx);
  bstring->aux  = NULL;

  return bstring;
}

t3rt_binary_string_t
tri__tri_to_t3_binary_string(const BinaryString *tri_str,
                             t3rt_alloc_strategy_t strategy,
                             t3rt_context_t ctx)
{
  t3rt_binary_string_t t3str =
    t3rt_binary_string_construct(strategy,
                                 (unsigned long)tri_str->bits,
                                 tri_str->data,
                                 ctx);
  return t3str;
}


/*
 *  Primitives
 */


/* TriComponentId */

bool
t3tri_componentId_is_equal(const TriComponentId *comp1,
                        const TriComponentId *comp2)
{
  if (comp1 == comp2) return true;

  return (t3tri_BinaryString_is_equal(&(comp1->compInst), &(comp2->compInst)) &&
          (strcmp(comp1->compType.moduleName, comp2->compType.moduleName) == 0) &&
          (strcmp(comp1->compType.objectName, comp2->compType.objectName) == 0));
}


bool
t3tri_BinaryString_is_equal(const BinaryString *str1,
                          const BinaryString *str2)
{
  unsigned int      bytesize;
  if (str1 == str2) return true;

  if (str1->bits != str2->bits) return false;

  bytesize = (unsigned int)str1->bits / 8;

  if (str1->bits % 8 > 0) bytesize++;

  return (memcmp(str1->data, str2->data, bytesize) == 0);
}


/*
 * This function is only supposed to be called from within
 * the triPAReset, triSAReset or triExecuteTestcase functions.
 */
t3rt_context_t
t3tri_get_context()
{
  return current_ctx;
}


#endif /* T3PL_TRI */

/* End of File */

