/**************************************************************************
*
* File Name: t3tri_saexample.c
*
* 2010 IBM Corporation
*
* Description:
*
* Example implementation of a SUT Adapter (SA).
*
* This SA-implementation acts as an underlying switch between 2 
* TTCN-3 test components. Each system port at the SA have exactly 2 
* component ports that can be mapped explicitly in the TTCN-3 
* Executable (TE). 
* This means that the SA does not know which of the mapped components
* that acts as the SUT - this has to be explicitly defined in the TE and
* is therefore controlled by the user.
*
* Requirements for the SA-implementation to work:
*  - There shall be exactly 2 mapped component ports for each system port.
* 
* Example on how a configuration might look like when test component A 
* (TCA) and test component B (TCB) are both mapped through their port P 
* to the system port *TSI* in the SA. This way either TCA or TCB can act 
* as the SUT when the TE is executing and sending events to P will be 
* switched to the corresponding other test component.
*
* ===============================================
*      ---------------     ---------------
* TE   |     TCA     |     |     TCB     |
*      |             |     |             |
*      -----------P---     ---P-----------
*                  \         /     
* ==================\=======/====================
*                    \     /
*              -------*TSI*--------
*              |                  |
* TRI          |       SA         |
*              |                  |
*              --------------------
* ===============================================
*
*/

#include <stdio.h>

#include "example_triutil.h"
#include "example_general.h"
#include "example_concurrency.h"
#include "example_sequence.h"

#define ex_sequence_static_init(TYPE,VAR) \
static TYPE VAR = { 0, 0, (void*)NULL}

typedef struct
{  
  TriPortId  systemPort;
  TriPortId  compPorts[2];
} TriSystemMappedPorts_t;

/* Create sequence structure */
typedef ex_sequence_head(TriSystemMappedPorts_t) TriSystemMappedPorts;

/* Global structure of mapped ports.*/
ex_sequence_static_init(TriSystemMappedPorts,sysPorts);

/* Semaphore to protect the datastructures. */
void *tri_sem;

/* shorthand for iterator */
#define get_seq()  ex_sequence_iter_get(iterSeq)
/* shorthand for iterator index */
#define get_seq_idx()  ex_sequence_iter_get_index(iterSeq)

/* Internal help function that takes the sending componentId together with 
* system-portId and tries to find a matching two-way mapping. If found, 
* receiving componentId is returned - otherwise NULL.  
*/
static TriComponentId *
getMappedComp(const TriComponentId *componentId, 
              const TriPortId *tsiPortId)
{
  /* Set up new datastructure for iteration */
  ex_sequence_iter(TriSystemMappedPorts_t) iterSeq;
  
  ex_sequence_iter_start(iterSeq, sysPorts);
  
  while(ex_sequence_iter_next(iterSeq)){
    /* First of all, find the correct system port */
    if(strcmp(get_seq().systemPort.portName, tsiPortId->portName) == 0){
      /*  Then, find the correct component port */
      if(t3tri_componentId_is_equal(&(get_seq().compPorts[0].compInst), 
                                    componentId)) {
        /* OK, check if we have a two-way mapping */
        if(get_seq().compPorts[1].portName != NULL) {
          return &(ex_sequence_element(sysPorts,
                                       get_seq_idx()).compPorts[1].compInst);
        }
        /* No two-way mapping was found */
        printf("No two-way mapping found!\n");
        return NULL;
      }

      if (t3tri_componentId_is_equal(&(get_seq().compPorts[1].compInst), 
                                     componentId)) { 
        /* OK, check if we have a two-way mapping */
        if (get_seq().compPorts[0].portName != NULL) {
          return &(ex_sequence_element(sysPorts,
                                       get_seq_idx()).compPorts[0].compInst);
        }
        /* No two-way mapping was found */
        printf("No two-way mapping found!\n");
        return NULL;
      }
      /* Sending component was not found - should never end up here! */
      printf("Sending component is not mapped!\n");
      return NULL; 
    }
  }
  /* No correct system port found - should never end up here! */
  printf("No correct system port found!\n");
  return NULL; 
}



/**************************************************************************
*
* TE -> SA
*
* Function to be implemented by the System Adaptor
*
*/

TriStatus 
triSAReset() 
{ 
  printf("\n***>> triSAReset called\n");
  return TRI_OK; 
}

TriStatus 
triExecuteTestcase(const TriTestCaseId *testCaseId,  
                   const TriPortIdList *tsiPortList) 
{ 
  long int i;
  t3rt_boolean_t status;
  t3rt_context_t cur_ctx = t3tri_get_context();
  printf("\n***>> triExecuteTestcase called\n");
  
  if(ex_sequence_size(sysPorts) != 0){
    /* Free/clear earlier structure of mapped ports */
    ex_sequence_deallocate(sysPorts, t3rt_perm_alloc_c, cur_ctx);
    if (tri_sem != NULL) {
      status = t3pl_sem_destroy(tri_sem, cur_ctx);
      assert(status);
      tri_sem = NULL;
    }
  }
  
  tri_sem = t3pl_sem_create(1, cur_ctx);
  
  status = t3pl_sem_wait(tri_sem, cur_ctx);
  assert(status);
  
  /* Create data structure */
  ex_sequence_init(sysPorts);
  
  /* Copy system ports to structure */
  for (i = 0; i < tsiPortList->length; i++){
    ex_sequence_add(sysPorts, t3rt_perm_alloc_c, cur_ctx);
    ex_sequence_last_element(sysPorts).systemPort = 
      *(tsiPortList->portIdList[i]);
    
    /* Initiation -> set portNames to NULL to control mappings */
    ex_sequence_last_element(sysPorts).compPorts[0].portName = NULL;
    ex_sequence_last_element(sysPorts).compPorts[1].portName = NULL;
  }
  
  status = t3pl_sem_post(tri_sem, cur_ctx);
  assert(status);
  return TRI_OK; 
}

TriStatus 
triEndTestcase() 
{ 
  t3rt_boolean_t status;
  t3rt_context_t cur_ctx = t3tri_get_context();
  printf("\n***>> triEndTestcase called\n");

  if(ex_sequence_size(sysPorts) != 0){
    /* Free/clear earlier structure of mapped ports */
    ex_sequence_deallocate(sysPorts, t3rt_perm_alloc_c, cur_ctx);
    if (tri_sem != NULL) {
      status = t3pl_sem_destroy(tri_sem, cur_ctx);
      assert(status);
      tri_sem = NULL;
    }
  }

  return TRI_OK; 
}

TriStatus 
triMap(const TriPortId *compPortId,  
       const TriPortId *tsiPortId) 
{ 
  t3rt_context_t cur_ctx = NULL;
  t3rt_boolean_t status;
  
  /* Set up new datastructure for iteration */
  ex_sequence_iter(TriSystemMappedPorts_t) iterSeq;
  
  status = t3pl_sem_wait(tri_sem, cur_ctx);
  assert(status);
  
  ex_sequence_iter_start(iterSeq, sysPorts);
  
  printf("\n***>> triMap called\n");  
  
  while (ex_sequence_iter_next(iterSeq)){
    if (t3tri_componentId_is_equal(&(get_seq().systemPort.compInst), 
                                   &(tsiPortId->compInst))          &&
        get_seq().systemPort.portIndex == tsiPortId->portIndex      &&
        strcmp(get_seq().systemPort.portName, tsiPortId->portName) == 0) {
      /* OK, we have found the correct system port, try to map. */
      if(get_seq().compPorts[0].portName == NULL) {
        get_seq().compPorts[0] = *compPortId;
        status = t3pl_sem_post(tri_sem, cur_ctx);
        assert(status);
        return TRI_OK; 
      }
      if (get_seq().compPorts[1].portName == NULL){
        get_seq().compPorts[1] = *compPortId;   
        status = t3pl_sem_post(tri_sem, cur_ctx);
        assert(status);
        return TRI_OK; 
      }
      /* No free port found. */
      status = t3pl_sem_post(tri_sem, cur_ctx);
      assert(status);
      printf("triMap: No free port found - both ports are already mapped!\n");
      return TRI_Error;
    }
  }

  status = t3pl_sem_post(tri_sem, cur_ctx);
  assert(status);
  /*  No correct system port found - should never end up here! */
  printf("triMap: No correct system port found!\n");
  return TRI_Error; 
}


TriStatus 
triUnmap(const TriPortId *compPortId, 
         const TriPortId *tsiPortId) 
{ 
  t3rt_context_t cur_ctx = NULL;
  t3rt_boolean_t status;
  
  /* Set up new datastructure for iteration */
  ex_sequence_iter(TriSystemMappedPorts_t) iterSeq;
  
  status = t3pl_sem_wait(tri_sem, cur_ctx);
  assert(status);
  
  ex_sequence_iter_start(iterSeq, sysPorts);
  
  printf("\n***>> triUnmap called\n");
  
  while(ex_sequence_iter_next(iterSeq)){
    if(t3tri_componentId_is_equal(&(get_seq().systemPort.compInst), 
      &(tsiPortId->compInst)) &&
      get_seq().systemPort.portIndex == tsiPortId->portIndex &&
      strcmp(get_seq().systemPort.portName, tsiPortId->portName) == 0) 
    {
      /* OK, we have found the correct system port, 
         find the mapped component port and unmap */
      if (t3tri_componentId_is_equal(&(get_seq().compPorts[0].compInst), 
                                     &(compPortId->compInst)) &&
          get_seq().compPorts[0].portIndex == compPortId->portIndex &&
          strcmp(get_seq().compPorts[0].portName, compPortId->portName) == 0) 
      {
        get_seq().compPorts[0].portName = NULL;
        status = t3pl_sem_post(tri_sem, cur_ctx);
        assert(status);
        return TRI_OK; 
      }

      if (t3tri_componentId_is_equal(&(get_seq().compPorts[1].compInst), 
                                     &(compPortId->compInst)) &&
          get_seq().compPorts[1].portIndex == compPortId->portIndex &&
          strcmp(get_seq().compPorts[1].portName, compPortId->portName) == 0) 
      {
        get_seq().compPorts[1].portName = NULL;
        status = t3pl_sem_post(tri_sem, cur_ctx);
        assert(status);
        return TRI_OK; 
      }

      /* Component port not found. */
      status = t3pl_sem_post(tri_sem, cur_ctx);
      assert(status);
      printf("triUnmap: Component port not found\n");
      return TRI_Error;
    }
  }

  /* No correct system port found - should never end up here! */
  status = t3pl_sem_post(tri_sem, cur_ctx);
  printf("triUnmap: No correct system port found!\n");
  return TRI_Error; 
  
}


TriStatus 
triSend(const TriComponentId *componentId, 
        const TriPortId      *tsiPortId, 
        const TriAddress     *sutAddress, 
        const TriMessage     *sendMessage) 
{
  t3rt_context_t cur_ctx = NULL;
  t3rt_boolean_t status;
  TriComponentId *dstCompId;
  
  printf("\n***>> triSend called\n");
  status = t3pl_sem_wait(tri_sem, cur_ctx);
  assert(status);
  if((dstCompId = getMappedComp(componentId, tsiPortId)) == NULL){
    status = t3pl_sem_post(tri_sem, cur_ctx);
    assert(status);
    /* No explicit mapping - use regular "loopback functionality"! */
    triEnqueueMsg(tsiPortId, sutAddress, componentId, sendMessage);
    return TRI_OK;
  }
  
  triEnqueueMsg(tsiPortId, sutAddress, dstCompId, sendMessage);
  status = t3pl_sem_post(tri_sem, cur_ctx);
  assert(status);
  return TRI_OK;
}


TriStatus 
triSendMC (const TriComponentId *componentId,
           const TriPortId      *tsiPortId,
           const TriAddressList *sutAddresses,
           const TriMessage     *sendMessage)
{
  printf("\n***>> triSendMC called\n");
  return TRI_Error;
}           


TriStatus
triSendBC (const TriComponentId *componentId,
           const TriPortId      *tsiPortId,
           const TriMessage     *sendMessage)
{
  printf("\n***>> triSendBC called\n");
  return TRI_Error;
}           


TriStatus 
triCall(const TriComponentId   *componentId, 
        const TriPortId        *tsiPortId,  
        const TriAddress       *sutAddress,  
        const TriSignatureId   *signatureId,  
        const TriParameterList *parameterList) 
{
  t3rt_context_t cur_ctx = NULL;
  TriComponentId *dstCompId;
  t3rt_boolean_t status;

  printf("\n***>> triCall called\n");
  
  status = t3pl_sem_wait(tri_sem, cur_ctx);
  assert(status);
  
  if ((dstCompId = getMappedComp(componentId, tsiPortId)) == NULL) {
    status = t3pl_sem_post(tri_sem, cur_ctx);
    assert(status);
    return TRI_Error;
  }
  
  triEnqueueCall(tsiPortId, sutAddress, dstCompId, 
                 signatureId, parameterList);
  status = t3pl_sem_post(tri_sem, cur_ctx);
  assert(status);
  return TRI_OK;
}


TriStatus
triCallMC (const TriComponentId   *componentId,
           const TriPortId        *tsiPortId, 
           const TriAddressList   *sutAddresses, 
           const TriSignatureId   *signatureId, 
           const TriParameterList *parameterList)
{
  printf("\n***>> triCallMC called\n");
  return TRI_Error; 
}           


TriStatus
triCallBC (const TriComponentId   *componentId,
           const TriPortId        *tsiPortId,
           const TriSignatureId   *signatureId,
           const TriParameterList *parameterList)
{
  printf("\n***>> triCallBC called\n");
  return TRI_Error; 
}           


TriStatus 
triReply(const TriComponentId    *componentId, 
         const TriPortId         *tsiPortId,  
         const TriAddress        *sutAddress,  
         const TriSignatureId    *signatureId,  
         const TriParameterList  *parameterList,  
         const TriParameter      *returnValue) 
{ 
  t3rt_context_t cur_ctx = NULL;
  t3rt_boolean_t status;
  TriComponentId *dstCompId;

  printf("\n***>> triReply called\n");
  
  status = t3pl_sem_wait(tri_sem, cur_ctx);
  assert(status);
  
  if ((dstCompId = getMappedComp(componentId, tsiPortId)) == NULL) {
    status = t3pl_sem_post(tri_sem, cur_ctx);
    assert(status);
    return TRI_Error;
  }
  
  triEnqueueReply(tsiPortId, sutAddress, dstCompId, 
                  signatureId, parameterList, returnValue);
  status = t3pl_sem_post(tri_sem, cur_ctx);
  assert(status);
  return TRI_OK;
}


TriStatus 
triReplyMC (const TriComponentId   *componentId,
            const TriPortId        *tsiPortId,
            const TriAddressList   *sutAddresses,
            const TriSignatureId   *signatureId,
            const TriParameterList *parameterList,
            const TriParameter     *returnValue)
{
  printf("\n***>> triReplyMC called\n");
  return TRI_Error; 
}
            

TriStatus
triReplyBC (const TriComponentId   *componentId,
            const TriPortId        *tsiPortId,
            const TriSignatureId   *signatureId,
            const TriParameterList *parameterList,
            const TriParameter     *returnValue)
{
  printf("\n***>> triReplyBC called\n");
  return TRI_Error; 
}


TriStatus
triRaise(const TriComponentId  *componentId,
         const TriPortId       *tsiPortId,
         const TriAddress      *sutAddress,
         const TriSignatureId  *signatureId,
         const TriException    *exception) 
{ 
  t3rt_context_t cur_ctx = NULL;
  t3rt_boolean_t status;
  TriComponentId *dstCompId;

  printf("\n***>> triRaise called\n");
  
  status = t3pl_sem_wait(tri_sem, cur_ctx);
  assert(status);
  
  if ((dstCompId = getMappedComp(componentId, tsiPortId)) == NULL) {
    status = t3pl_sem_post(tri_sem, cur_ctx);
    assert(status);
    return TRI_Error;
  }
  
  triEnqueueException(tsiPortId, sutAddress, dstCompId, 
                      signatureId, exception);
  status = t3pl_sem_post(tri_sem, cur_ctx);
  assert(status);
  return TRI_OK;
}


TriStatus
triRaiseMC (const TriComponentId *componentId,
            const TriPortId      *tsiPortId,
            const TriAddressList *sutAddresses,
            const TriSignatureId *signatureId,
            const TriException   *exception)
{
  printf("\n***>> triRaiseMC called\n");
  return TRI_Error; 
}
            

TriStatus
triRaiseBC (const TriComponentId *componentId,
            const TriPortId      *tsiPortId,
            const TriSignatureId *signatureId,
            const TriException   *exception)
{
  printf("\n***>> triRaiseBC called\n");
  return TRI_Error; 
}            


TriStatus 
triSUTActionInformal(char* description) 
{ 
  printf("\n***>> triSUTActionInformal called\n");
  return TRI_Error; 
}


TriStatus 
triSUTActionTemplate(const TriActionTemplate* templateValue) 
{ 
  printf("\n***>> triSUTActionTemplate called\n");
  return TRI_Error; 
}
