# --------------------------------------------------------------------
# File : ObjectReport.tcl
# 
# Desc : This file contains a sample TCL script used to generate a
#        HTML report of some classes of objects inside a TTCN
#        project. The project is traversed and the different objects
#        are sorted into different categories.
# --------------------------------------------------------------------



# --------------------------------------------------------------------
# Proc : GetTempDir
# Desc : This procedure is a help procedure used to locate a
#        temporary directory in the user's drive.
# --------------------------------------------------------------------

proc GetTempDir {} {
    global env
    if [info exist env(TEMP)] {
	return $env(TEMP)
    }
    if [info exist env(TMP)] {
	return $env(TMP)
    }
    if [info exist env(TMPDIR)] {
	return $env(TMPDIR)
    }
    foreach dir {"C:/Temp" "C:/Tmp" "C:/Windows/Temp" "C:/Windows/Tmp" "/tmp"} {
	if [file isdirectory "$dir"] {
	    return $dir
	}
    }
    output "No temporary directory available.\n"
    return ""
}



# --------------------------------------------------------------------
# Proc : SetHtmlFile
# Desc : This procedure is a help procedure used to set the name
#        of the generated HTML file.
# --------------------------------------------------------------------

proc SetHtmlFile {identObj} {
    global fHtml HtmlName
    global DerivedFiles
    set DerivedFiles {}
    set tmpdir [GetTempDir]
    set HtmlName "[file join $tmpdir ObjectReport[pid].htm]"
    set fHtml [open $HtmlName w]
}



# --------------------------------------------------------------------
# Proc : HtmlHeader
# Desc : This procedure is a help procedure used to generate the
#        reusable header of the HTML file. A customer can select to
#        totally change this for his/her own purposes.
# --------------------------------------------------------------------

proc HtmlHeader {} {
    global fHtml Application
    global ScriptsDirectory

    puts $fHtml "
<html>
<head>
<title>TTCN-3 Object Report</title>
</head>

<body>
<IMG SRC=\"$ScriptsDirectory/Images/logo.jpg\" valign=center>   
<H1 align=center>TTCN-3 Object Report</H1>
<p> This document contains information concerning the TTCN Objects of the 
project.

<p>The following information is available:
<UL>
<LI> <A HREF=#Modules> Modules List</A>
<LI> <A HREF=#Imports> Imports List</A>
<LI> <A HREF=#Constants> Constants List</A>
<LI> <A HREF=#Ports> Ports List</A>
<LI> <A HREF=#Components> Components List</A>
<LI> <A HREF=#Groups> Groups List </A>
<LI> <A HREF=#SimpleTypes> SimpleTypes List </A>
<LI> <A HREF=#StructuredTypes> StructuredTypes List </A>
<LI> <A HREF=#Templates> Templates List</A>
<LI> <A HREF=#Signatures> Signatures List</A>
<LI> <A HREF=#TestCases> TestCases List</A>
</UL>
"

}



# --------------------------------------------------------------------
# Proc : HtmlCloser
# Desc : This procedure is a help procedure used to to generate the
#        reusable footer of the HTML file. A customer can select to
#        totally change this for his/her own purposes.
# --------------------------------------------------------------------

proc HtmlCloser {} {
    global fHtml
	puts $fHtml "
</body>

</html>
"
close $fHtml
}



# --------------------------------------------------------------------
# Proc : GetPath
# Desc : This procedure is used to get the path of a given TTCN
#        object in order to be able to print this out for the user.
# --------------------------------------------------------------------

proc GetPath {identObj {car /}} {
    set name [Get $identObj name]
    if {[Class $identObj] == "Module"} {
    	# No more path element when we are on the top
    	return $name
    }

    set composite [GetRole $identObj composite]

    if { [Class $composite] == "Group" } {
        # We got a group owner
	set name [GetPath $composite $car]$car$name
    } elseif { [Class $composite] == "Module" } {
        # We got a module owner
	set name [GetPath $composite $car]$car$name
    }
    return $name
}



# --------------------------------------------------------------------
# Proc : IsKindOf
# Desc : This procedure is a help procedure used to check if a given
#        object is of a member of a given class. This procedure is
#        used as a filter when looking for objects of a given kind.
# --------------------------------------------------------------------

proc IsKindOf {class identObj} {
  expr ![string compare [Class $identObj] $class]
}



# --------------------------------------------------------------------
# Proc : BuildModule
# Desc : This procedure is used generate the report for the modules
#        in the project. As the module report is slightly different
#        than the report of other objects, this is done separately.
# --------------------------------------------------------------------

proc BuildModule {item} {
  global fHtml
	puts $fHtml "
<H3><BR> <A NAME=Modules>Modules of the project</A></H2>
<P> In the following array you will find for each module of the project:
<UL>
<LI>The name of the module.
<LI>The line number of the module in its owner file
<LI>The name of the file in which this module can be found
</UL>
<P>To locate the module in the file just follow the link.
<P>
<div align=left>
  <table border=1 >
    <tr >
      <td width=150 align=center>Module Name</td>
      <td width=50 align=center>Line Number</td>
      <td width=150 align=center>File Name</td>
	</tr>"

  MapRole $item module ForEach_Module
	puts $fHtml "
  </table>
</div>
"
}



# --------------------------------------------------------------------
# Proc : ForEach_Module
# Desc : This procedure is used to traverse through the different
#        module in the project.
# --------------------------------------------------------------------

proc ForEach_Module {identObj} {
     global fHtml
     set pathName [Get [GetRole $identObj file] pathName]
     set path [GetPath $identObj]
     puts $fHtml "
<tr>
<td><A HREF=\"tlog://<$pathName>:TTCN3:PATHDCD,$pathName,$path\">[Get $identObj name]</td>
<td align=right>[Get $identObj lineNumber]</td>
<td>[file tail $pathName]</td>
</tr>
"
  return 1
}



# --------------------------------------------------------------------
# Proc : BuildOther
# Desc : This procedure is used to build reports for objects of 
#        different types when traversing the project.
# --------------------------------------------------------------------

proc BuildOther {kind class item} {
  global fHtml
  global typeFilter 
  set kinds ${kind}s
	puts $fHtml "
<BR><BR><H3> <A NAME=$kinds>$kinds of the project</A></H2>
<P> In the following array you will find for each $kind of the project:
<UL>
<LI>The name of the $kind,
<LI>The path of the $kind,
<LI>The line number of the $kind in its owner file,
<LI>The name of the file in which this $kind can be found.
</UL>
<P>To locate a $kind in the file just follow the link.
<P>
<div align=left>
  <table border=1 >
    <tr >
      <td width=150 align=center>$kind Name</td>
      <td width=150 align=center>Path</td>
      <td width=50 align=center>Line Number</td>
      <td width=150 align=center>File Name</td>
      </tr>"

      set typeFilter $class
      MapRole $item module ForEachModule

      puts $fHtml "
  </table>
</div>
"
}



# --------------------------------------------------------------------
# Proc : ForEachModuleElement
# Desc : This procedure is the one that is called for each module
#        element in order to print out its information. If the module
#        element is a group, that group is traversed recursivelly.
# --------------------------------------------------------------------

proc ForEachModuleElement {moduleElement} {
  if { [Class $moduleElement] == "Group" } {
    foreach item $moduleElement {
      MapRole $item moduleElement ForEachModuleElement
    }
  }

  global typeFilter

  if { ! [IsKindOf $typeFilter $moduleElement] } {
  return 1
  }

  if { [Class $moduleElement] != "Comment" } {
     global fHtml
     set pathName [Get [GetRole $moduleElement definedIn] pathName]
     set path ""
     set pathGui ""     
     set path [GetPath $moduleElement]
     set pathGui [GetPath $moduleElement]
     puts $fHtml "
    <tr>
    <td><A HREF=\"tlog://<$pathName>:TTCN3:PATHDCD,$pathName,$path\">[Get $moduleElement name]</td>
    <td>$pathGui</td>
    <td align=right>[Get $moduleElement lineNumber]</td>
    <td>[file tail $pathName]</td>
    </tr>
    "
    return 1
  }
}



# --------------------------------------------------------------------
# Proc : ForEachModule
# Desc : This procedure is used to traverse through the different
#        module in the project.
# --------------------------------------------------------------------    

proc ForEachModule {module} {
  MapRole $module moduleElement ForEachModuleElement
  return 1
}

# --------------------------------------------------------------------
# Proc : Main
# Desc : The script starts of by setting the HTML file we are going
#        to use. It then creates the header for the file and then
#        the whole project is traversed to generate reports for 
#        different types of objects. These categories are here for
#        example imports, constants, ports, etc. Finally the HTML
#        file is closed and ready for display.
# --------------------------------------------------------------------

SetHtmlFile {}
HtmlHeader

foreach item $TTCN3 {
  BuildModule $item
  BuildOther Import Import $item
  BuildOther Constant Constant $item
  BuildOther Port Port $item
  BuildOther Component Component $item
  BuildOther Group Group $item
  BuildOther SimpleType PredefinedType $item
  BuildOther StructuredType StructuredType $item
  BuildOther Template Template $item
  BuildOther Function Function $item
  BuildOther Signature Signature $item
  BuildOther TestCase TestCase $item
}

HtmlCloser

set cmd "HtmlReport -delete -use ObjectReport \"$HtmlName\""
set cmd "$cmd $DerivedFiles"
eval $cmd
